; fesslix.nsi

;--------------------------------

!include "MUI2.nsh"
!include "logiclib.nsh"

; The version of Fesslix
!define VERSION "552"
!define FLXPYTH "Python"

; The name of the installer
Name "Fesslix"

; The file to write
OutFile "fesslix-installer-win32-${VERSION}_${FLXPYTH}.exe"

;--------------------------------
Var welcome_title
Var welcome_text

;--------------------------------

; The default installation directory
InstallDir $PROGRAMFILES\Fesslix

; Registry key to check for directory (so if you install again, it will 
; overwrite the old one automatically)
InstallDirRegKey HKLM "Software\Fesslix" "Install_Dir"

; Request application privileges for Windows Vista (user,admin)
RequestExecutionLevel admin
  ;InstallButtonText $install_button_text

;--------------------------------

;Interface Settings

  !define MUI_ABORTWARNING
  !define MUI_ICON "fesslix.ico"

;--------------------------------

; Pages

!define MUI_PAGE_CUSTOMFUNCTION_PRE WelcomePagePre
!define MUI_WELCOMEPAGE_TITLE '$welcome_title'
!define MUI_WELCOMEPAGE_TEXT '$welcome_text'
!insertmacro MUI_PAGE_WELCOME

!insertmacro MUI_PAGE_LICENSE "COPYING"
!insertmacro MUI_PAGE_COMPONENTS
!insertmacro MUI_PAGE_DIRECTORY
!insertmacro MUI_PAGE_INSTFILES
  
!insertmacro MUI_UNPAGE_CONFIRM
!insertmacro MUI_UNPAGE_INSTFILES

!insertmacro MUI_PAGE_FINISH

;--------------------------------
;Languages
 
  !insertmacro MUI_LANGUAGE "English"

;--------------------------------
;Version Information

  VIProductVersion "0.0.${VERSION}.0"
  VIAddVersionKey /LANG=${LANG_ENGLISH} "ProductName" "Fesslix"
  VIAddVersionKey /LANG=${LANG_ENGLISH} "Comments" "Stochastic Analysis"
  ;VIAddVersionKey /LANG=${LANG_ENGLISH} "CompanyName" "Fake company"
  ;VIAddVersionKey /LANG=${LANG_ENGLISH} "LegalTrademarks" "Test Application is a trademark of Fake company"
  VIAddVersionKey /LANG=${LANG_ENGLISH} "LegalCopyright" "2010-2012 Wolfgang Betz"
  VIAddVersionKey /LANG=${LANG_ENGLISH} "FileDescription" "Fesslix"
  VIAddVersionKey /LANG=${LANG_ENGLISH} "FileVersion" "trunk.${VERSION}.${FLXPYTH}"

;--------------------------------

; The stuff to install
Section "Fesslix ${VERSION}.${FLXPYTH}"

  SectionIn RO
  
  ; Set output path to the installation directory.
  SetOutPath $INSTDIR
  
  ; Put file there
  File "fesslix.exe"
  File "fesslix.ini"
  File "gausspoints.dat"
  ;File "libgsl-0.dll"
  ;File "libgslcblas-0.dll"
  File "COPYING"
  File "fesslix_file.ico"
  File "fesslix.ico"
  
  ; Write the installation path into the registry
  WriteRegStr HKLM SOFTWARE\Fesslix "Install_Dir" "$INSTDIR"
  
  ; Write the uninstall keys for Windows
  WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\Fesslix" "DisplayName" "Fesslix"
  WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\Fesslix" "UninstallString" '"$INSTDIR\uninstall.exe"'
  WriteRegDWORD HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\Fesslix" "NoModify" 1
  WriteRegDWORD HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\Fesslix" "NoRepair" 1
  WriteUninstaller "uninstall.exe"
  
SectionEnd

; Optional section (can be disabled by the user)
Section "Start Menu Shortcuts"
  CreateDirectory "$SMPROGRAMS\Fesslix"
  CreateShortCut "$SMPROGRAMS\Fesslix\Uninstall.lnk" "$INSTDIR\uninstall.exe" "" "$INSTDIR\fesslix_file.ico" 0
  CreateShortCut "$SMPROGRAMS\Fesslix\Fesslix.lnk" "$INSTDIR\fesslix.exe" "-i" "$INSTDIR\fesslix.ico" 0
SectionEnd

Section "Register File Type"
  WriteRegStr HKCR ".flx" "" "Fesslix.ParaFile"
  WriteRegStr HKCR "Fesslix.ParaFile" "" "Fesslix Parameter File"
  WriteRegStr HKCR "Fesslix.ParaFile\DefaultIcon" "" "$INSTDIR\fesslix.ico"
  WriteRegExpandStr HKCR "Fesslix.ParaFile\shell\open\command" "" '"%SystemRoot%\system32\NOTEPAD.EXE" "%1"'
  WriteRegExpandStr HKCR "Fesslix.ParaFile\shell\run\command" "" '"$INSTDIR\fesslix.exe" "%1" -i --pause=yes'
  WriteRegExpandStr HKCR "Fesslix.ParaFile\shell\edit\command" "" '"%SystemRoot%\system32\NOTEPAD.EXE" "%1"'
SectionEnd

Section "Install Tests"
  SetOutPath $INSTDIR\test
  File /r "test\*.*"
SectionEnd

Section "Install Examples"
  SetOutPath $INSTDIR\examples
  File /r "examples\*.*"
SectionEnd

;--------------------------------

; Uninstaller

Section "Uninstall"
  
  ; Remove registry keys
  DeleteRegKey HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\Fesslix"
  DeleteRegKey HKLM SOFTWARE\Fesslix
  DeleteRegKey HKCR .flx
  DeleteRegKey HKCR Fesslix.ParaFile

  ; Remove files and uninstaller
  Delete $INSTDIR\fesslix.exe
  Delete $INSTDIR\uninstall.exe
  Delete $INSTDIR\COPYING
  Delete "$INSTDIR\*.*"

  ; Remove shortcuts, if any
  Delete "$SMPROGRAMS\Fesslix\*.*"

  ; Remove directories used
  RMDir "$SMPROGRAMS\Fesslix"
  RMDir "$INSTDIR"

SectionEnd

  Function .onMouseOverSection
    FindWindow $R0 "#32770" "" $HWNDPARENT
    GetDlgItem $R0 $R0 1043 ; description item (must be added to the UI)

    StrCmp $0 0 "" +2
      SendMessage $R0 ${WM_SETTEXT} 0 "STR:Installs the files needed to run Fesslix (Version: ${VERSION}.${FLXPYTH})"

    StrCmp $0 1 "" +2
      SendMessage $R0 ${WM_SETTEXT} 0 "STR:Adds an entry to the Windows Start Menu"

    StrCmp $0 2 "" +2
      SendMessage $R0 ${WM_SETTEXT} 0 "STR:Register the file-type .flx"

    StrCmp $0 3 "" +2
      SendMessage $R0 ${WM_SETTEXT} 0 "STR:Install a directory containing the Fesslix tests"

    StrCmp $0 4 "" +2
      SendMessage $R0 ${WM_SETTEXT} 0 "STR:Install a directory containing the Fesslix examples"
  FunctionEnd


!define MUI_WELCOMEPAGE_TITLE 'Welcome to the Fesslix Setup Wizard'
!define MUI_WELCOMEPAGE_TEXT ''

Function WelcomePagePre
  ;strcpy $install_button_text "Install"
  strcpy $welcome_title "Welcome to the Fesslix Setup Wizard"
  strcpy $welcome_text "This wizard will guide you through the installation of Fesslix.$\n$\nFesslix is a Free Software program for stochastic analysis. The source code is available free of charge under the Free Software GPL license.$\nNOTE: At this stage, Fesslix is first and foremost a research code and not adequate for practical application. $\n$\nClick next to continue."
FunctionEnd


