/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxvi.h"
#include "flxdefault.h"

#include <iostream>

#include "po_part0.h"
  


int main(int argc, char** argv)
{
  std::ostringstream flxOutputStream;
  std::ostringstream& flxErrorStream = flxOutputStream;
// load the Fesslix-stuff
  ReadStream *reader = NULL;
  GlobalVar.set_stdcout(&flxOutputStream);
  GlobalVar.set_slog(&flxOutputStream);
  GlobalVar.allowInteractiveMode = false;
  
  #include "po_part1.txt"
  
  try {
    #include "po_part2.txt"
  
  // Deal with the log-files
    create_log(logFile, logTrunc);
    
  // try to open the file with the Gauss points
    ReadStream* gaussRS = NULL;
    open_GaussFile(gaussFile, gaussRS);
    
  // create output-logFile
    if (!GlobalVar.check_logNOTcout()) logOutput=false;
    if (logOutput) {
      flxcout = new flxStreamAlloc(flxOutputStream, GlobalVar.get_log());
    } else {
      flxcout = &flxOutputStream;
    }
    
  // produce some logging-infos
    if (do_logInfo) {
      FesslixMain::logInfo(GlobalVar.slog(4),false);
    }
    
  // create main object
    FesslixMain fesslix(*flxcout, flxErrorStream, gaussRS);
    FlxViObj_Flx::set_fesslix_ptr(&fesslix);
  // actually start the flxvi-GUI
    FlxViManager viManager(flxOutputStream, flxErrorStream);	// load the Fesslix-vi-Manager
    
  try {
    #include "po_part3.txt"
  }
  catch ( FlxEndE const &e) {
    GlobalVar.logLevel_strong_reset();
    GlobalVar.slog(4) << "end: 'end' was called" << std::endl;
    if (reader != NULL) delete reader;
  }
  catch ( FlxException_NeglectInInteractive &e) {
    flxErrorStream << std::endl << e.what() << std::endl;
    GlobalVar.logLevel_strong_reset();
    if (GlobalVar.check_logNOTcout()) {
      GlobalVar.slog(1) << std::endl << e.what() << std::endl;
    }
    if (reader != NULL) delete reader;
  }
  
  viManager.do_log();
  flxOutputStream << "Fesslix: starting user-input-mode" << std::endl;
  viManager.do_log();
  
  // start the command-input mode
    viManager.execute();

    
    
  // return success
    print_RETURN_SUCCESS();
    delete_FLXcout(logOutput,flxcout);
    return RETURN_SUCCESS;
  }
  #include "po_part4.txt"
  
}




