/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "fesslix.h"

#include "flxpython.h"

#include <limits>

using namespace std;


FesslixMain::FesslixMain ( ostream& coutV, ostream& cerrV, ReadStream* gaussRS ) {

  // initialise output-streams
    GlobalVar.alert.set_err_stream(cerrV);
    dataBox.OstreamBox.insert("cout", &coutV);
    dataBox.OstreamBox.insert("cerr", &cerrV);
    dataBox.OstreamBox.insert("log", &(GlobalVar.slog(0)) );

  // Read Gauss-Point Information
    dataBox.GaussInt.gaussRS = gaussRS;
    dataBox.GaussInt.ReadGP();  
  // create readers ...
    funReader = new FlxFunctionReader();
      FlxReadManager::set_funReader(funReader);
    mtxReader = new FlxMatrixReader();
      FlxReadManager::set_mtxReader(mtxReader);

  // create the evaluation class
    EvaluateCmd = new FlxEvaluateCmd();

  // create ObjReaders
    FlxCreateObjReaders* cor;
    // The LoadLib-Objects
      EvaluateCmd->get_ObjReadBox()->insert("loadlib", new FlxObjReadLoadLib(EvaluateCmd->get_ObjReadBox()));
    // Common Objects
      cor = new FlxCreateObjReaders_Common();
      cor->createFunReaders( &dataBox);
      cor->createObjReaders( EvaluateCmd->get_ObjReadBox() );
      delete cor;
    // FEM Objects   
      cor = new FlxCreateObjReaders_FEM();
      cor->createFunReaders( &dataBox);    
      cor->createObjReaders( EvaluateCmd->get_ObjReadBox() );
      delete cor; 
    // Python Objects
      #if FLX_USE_PYTHON
	cor = new FlxCreateObjReaders_Python();
	cor->createFunReaders( &dataBox);
	cor->createObjReaders( EvaluateCmd->get_ObjReadBox() );
	delete cor;
      #endif
}

void FesslixMain::logInfo(ostream& lout, bool pversion)
{
  lout << "Fesslix: compiled with the options ..." << std::endl;
  // FLXDEBUG
    lout << "  FLX_DEBUG                     ";
    #if FLX_DEBUG
      lout << "ON";
    #else 
      lout << "OFF";
    #endif
    lout << std::endl;
  // FLXDEBUG_COUT
    lout << "  FLX_DEBUG_COUT                ";
    #if FLX_DEBUG_COUT
      lout << "ON";
    #else 
      lout << "OFF";
    #endif
    lout << std::endl;
  // FLX_EXPNUM_1D_QUAD_APPROX
    lout << "  FLX_EXPNUM_1D_QUAD_APPROX     ";
    #if FLX_EXPNUM_1D_QUAD_APPROX
      lout << "ON";
    #else 
      lout << "OFF";
    #endif
    lout << std::endl;
  // FLX_USE_ARPACK
    lout << "  FLX_USE_ARPACK                ";
    #if FLX_USE_ARPACK
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_USE_GSL
    lout << "  FLX_USE_GSL                   ";
    #if FLX_USE_GSL 
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_BOOST_PO
    lout << "  FLX_BOOST_PO                  ";
    #if FLX_BOOST_PO
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_MCI
    lout << "  FLX_KAHAN_MCI                 ";
    #if FLX_KAHAN_MCI
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_KDE
    lout << "  FLX_KAHAN_KDE                 ";
    #if FLX_KAHAN_KDE
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_DOT
    lout << "  FLX_KAHAN_DOT                 ";
    #if FLX_KAHAN_DOT
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_2NORM
    lout << "  FLX_KAHAN_2NORM               ";
    #if FLX_KAHAN_2NORM
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_MTX_FULL
    lout << "  FLX_KAHAN_MTX_FULL            ";
    #if FLX_KAHAN_MTX_FULL
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_MTX_LTRI
    lout << "  FLX_KAHAN_MTX_LTRI            ";
    #if FLX_KAHAN_MTX_LTRI
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_MTX_SLTRI
    lout << "  FLX_KAHAN_MTX_SLTRI           ";
    #if FLX_KAHAN_MTX_SLTRI
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_MTX_SSILU
    lout << "  FLX_KAHAN_MTX_SSILU           ";
    #if FLX_KAHAN_MTX_SSILU
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_MTX_SYM
    lout << "  FLX_KAHAN_MTX_SYM             ";
    #if FLX_KAHAN_MTX_SYM
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_MTX_SSYM
    lout << "  FLX_KAHAN_MTX_SSYM            ";
    #if FLX_KAHAN_MTX_SSYM
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_MTX_SSFEM
    lout << "  FLX_KAHAN_MTX_SSFEM           ";
    #if FLX_KAHAN_MTX_SSFEM
      lout << "ON";
    #else 
      lout << "OFF";
    #endif  
    lout << std::endl;
  // FLX_KAHAN_MTX_PFEM
    lout << "  FLX_KAHAN_MTX_PFEM            ";
    #if FLX_KAHAN_MTX_PFEM
      lout << "ON";
    #else 
      lout << "OFF";
    #endif
    lout << std::endl;
  // FLX_KAHAN_CG_STAGE
    lout << "  FLX_KAHAN_CG_STAGE            ";
    lout << FLX_KAHAN_CG_STAGE << std::endl;    
    
  // Operating System
    lout << "  OS of compilation:            ";
    #ifdef __unix__
      lout << "UNIX ";
      #define FLX_OS_DET
    #endif
    #ifdef __WIN32__
      lout << "WIN32 ";
      #define FLX_OS_DET
    #endif
    #ifdef __WINDOWS__
      lout << "Windows ";
      #define FLX_OS_DET
    #endif
    #ifdef __linux__
      lout << "Linux ";
      #define FLX_OS_DET
    #endif
    #ifndef FLX_OS_DET
      lout << "OTHER";
    #endif
    lout << std::endl;
  // Time of compilation
    lout << "  Compiled on " << __DATE__ << " at " << __TIME__ << std::endl;
  
  lout << "Fesslix: numeric limits ..." << std::endl;
  // tdouble
    lout << "  tdouble                       " << std::endl;
    lout << "        digits:                 " << numeric_limits<tdouble>::digits10 << std::endl;
    lout << "        epsilon:                " << numeric_limits<tdouble>::epsilon() << std::endl;
    lout << "        min:                    " << numeric_limits<tdouble>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<tdouble>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<tdouble>::round_style << std::endl;
  // tfloat
    lout << "  tfloat                        " << std::endl;
    lout << "        digits:                 " << numeric_limits<tfloat>::digits10 << std::endl;
    lout << "        epsilon:                " << numeric_limits<tfloat>::epsilon() << std::endl;
    lout << "        min:                    " << numeric_limits<tfloat>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<tfloat>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<tfloat>::round_style << std::endl;
  // double
    lout << "  double                        " << std::endl;
    lout << "        digits:                 " << numeric_limits<double>::digits10 << std::endl;
    lout << "        epsilon:                " << numeric_limits<double>::epsilon() << std::endl;
    lout << "        min:                    " << numeric_limits<double>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<double>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<double>::round_style << std::endl;
  // long double
    lout << "  long double                   " << std::endl;
    lout << "        digits:                 " << numeric_limits<long double>::digits10 << std::endl;
    lout << "        epsilon:                " << numeric_limits<long double>::epsilon() << std::endl;
    lout << "        min:                    " << numeric_limits<long double>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<long double>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<long double>::round_style << std::endl;
  // float
    lout << "  float                         " << std::endl;
    lout << "        digits:                 " << numeric_limits<float>::digits10 << std::endl;
    lout << "        epsilon:                " << numeric_limits<float>::epsilon() << std::endl;
    lout << "        min:                    " << numeric_limits<float>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<float>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<float>::round_style << std::endl;
  // int
    lout << "  int                           " << std::endl;
    lout << "        digits:                 " << numeric_limits<int>::digits10 << std::endl;
    lout << "        min:                    " << numeric_limits<int>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<int>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<int>::round_style << std::endl;
  // long
    lout << "  long                          " << std::endl;
    lout << "        digits:                 " << numeric_limits<long>::digits10 << std::endl;
    lout << "        min:                    " << numeric_limits<long>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<long>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<long>::round_style << std::endl;
  // unsigned int
    lout << "  unsigned int                  " << std::endl;
    lout << "        digits:                 " << numeric_limits<unsigned int>::digits10 << std::endl;
    lout << "        min:                    " << numeric_limits<unsigned int>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<unsigned int>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<unsigned int>::round_style << std::endl;
  // unsigned long
    lout << "  unsigned long                 " << std::endl;
    lout << "        digits:                 " << numeric_limits<unsigned long>::digits10 << std::endl;
    lout << "        min:                    " << numeric_limits<unsigned long>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<unsigned long>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<unsigned long>::round_style << std::endl;
  // tuint
    lout << "  tuint                         " << std::endl;
    lout << "        digits:                 " << numeric_limits<tuint>::digits10 << std::endl;
    lout << "        min:                    " << numeric_limits<tuint>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<tuint>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<tuint>::round_style << std::endl;
  // tnlong
    lout << "  tnlong                        " << std::endl;
    lout << "        digits:                 " << numeric_limits<tnlong>::digits10 << std::endl;
    lout << "        min:                    " << numeric_limits<tnlong>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<tnlong>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<tnlong>::round_style << std::endl;
  // tulong
    lout << "  tulong                        " << std::endl;
    lout << "        digits:                 " << numeric_limits<tulong>::digits10 << std::endl;
    lout << "        min:                    " << numeric_limits<tulong>::min() << std::endl;
    lout << "        max:                    " << numeric_limits<tulong>::max() << std::endl;
    lout << "        round_style:            " << numeric_limits<tulong>::round_style << std::endl;
    
  if (pversion) return;
  lout << "Fesslix: started with the configuration ..." << std::endl;
  // gauss.maxnumb
    lout << "  gauss.maxnumb                 ";
    lout << GaussIntegration::GaussPointMaxArraySize;
    lout << std::endl;
  // legendre.numb
    lout << "  legendre.numb                 ";
    lout << GlobalVar.LegendrePolyH_init_numb;
    lout << std::endl;
  // expnum.1D.intervals
    lout << "  expnum.1D.intervals           ";
    lout << GlobalVar.Exp_Num_Intervals_1D;
    lout << std::endl;
  // expnum.1D.samples
    lout << "  expnum.1D.samples             ";
    lout << GlobalVar.Exp_Num_Samples_1D;
    lout << std::endl;
  // expnum.1D.redist.diff
    lout << "  expnum.1D.redist.diff         ";
    lout << GlobalVar.Exp_Num_Diff_Redistribute_1D;
    lout << std::endl;
  // expnum.1D.lbound
    lout << "  expnum.1D.lbound              ";
    lout << GlobalVar.Exp_Num_Lbound_1D;
    lout << std::endl;
  // expnum.1D.ubound
    lout << "  expnum.1D.ubound              ";
    lout << GlobalVar.Exp_Num_Ubound_1D;
    lout << std::endl;
  // expnum.mci.bound
    lout << "  expnum.mci.bound              ";
    lout << GlobalVar.Exp_Num_Bound_MCI;
    lout << std::endl;
  // expnum.mci.intervals
    lout << "  expnum.mci.intervals          ";
    lout << GlobalVar.Exp_Num_Intervals_MCI;
    lout << std::endl;
  // expnum.mci.samplesperinterval
    lout << "  expnum.mci.samplesperinterval ";
    lout << GlobalVar.Exp_Num_SmpPerInterval_MCI;
    lout << std::endl;
  // expnum.mci.redist.smp
    lout << "  expnum.mci.redist.smp         ";
    lout << GlobalVar.Exp_Num_Smp_Redistribute_MCI;
    lout << std::endl;
  // expnum.mci.redist.diff
    lout << "  expnum.mci.redist.diff        ";
    lout << GlobalVar.Exp_Num_Diff_Redistribute_MCI;
    lout << std::endl;
  // expnum.rnd.moments
    lout << "  expnum.rnd.moments            ";
    lout << GlobalVar.Exp_Num_Rnd_Moments;
    lout << std::endl;
  // expnum.rnd.moments.central
    lout << "  expnum.rnd.moments.central    ";
    lout << GlobalVar.Exp_Num_Rnd_Moments_Central;
    lout << std::endl;
  // TOL
    lout << "  TOL                           " << GlobalVar.TOL() << std::endl;
  // floating point conversion
    GlobalVar.Double2String_log();
  
}

string FesslixMain::get_version()
{
  std::ostringstream ssV;
  ssV << "trunk." << VERSION_REV << " (" << VERSION_DATE << ")";
  return ssV.str();
}

FesslixMain::~FesslixMain () {
  delete EvaluateCmd;
  delete funReader;
  delete mtxReader;
}

void FesslixMain::evaluate () {
  FlxObjBase* ob=NULL;
  if (reader == NULL) {	// Check if a reader is available
    ostringstream ssV_2;
    ssV_2 << "No reader available.";
    throw FlxException("FesslixMain::evaluate_1", ssV_2.str() ); 
  }
  try {
    while ( reader->getNextType() != ReadStream::ENDOFFILE ) {
      // print the prompt for output of input to log
	GlobalVar.prelog_append(GlobalVar.get_flxPrompt());
      ob = EvaluateCmd->evaluateCmd();
      ob->exec();
      GlobalVar.logLevel_strong_reset();
      GlobalVar.slog_flush();
      delete ob; ob=NULL;
    }
  } catch ( FlxEndE &e) {
    GlobalVar.logLevel_strong_reset();
    if (ob) delete ob;
    throw;
  } catch ( FlxExitE &e) {
    GlobalVar.logLevel_strong_reset();
    GlobalVar.slog(4) << "exit: 'exit;' was called" << std::endl;
    if (ob) delete ob;
  } catch ( FlxException &e ) {
    GlobalVar.logLevel_strong_reset();
    if (ob) delete ob;
    throw;
  }
}

void FesslixMain::evaluate(ReadStream* reader)
{
  dataBox.ReadManager.push(reader);
  try {
    evaluate();
    dataBox.ReadManager.pop();
  } catch ( FlxEndE &e) {
    dataBox.ReadManager.pop();
    throw;
  } catch ( FlxException &e ) {
    dataBox.ReadManager.pop();
    throw;
  }
}



#if FLX_DEBUG
void FesslixMain::test()
{  
  GlobalVar.slog(5) << "Fesslix: test-section ------------ START -----------------------" << std::endl;
  
//   FlxRnd_Beta bd(10,1e7);
//   const tdouble p1 = bd.y2x(.1);
//   const tdouble p2 = bd.x2y(p1);
//   std::cout << p1 << ", " << p2 << std::endl;
  
  GlobalVar.slog(5) << "Fesslix: test-section ------------ END -------------------------" << std::endl;
}
#endif

