/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_fesslix_H
#define fesslix_fesslix_H

#include "flxobjcommon.h"
#include "flxlibs.h"
#include "flxversion.h"

/**
* @brief The Main-Class
*/
class FesslixMain : public FlxReaderBase {
  private:
    FlxData dataBox;
    FlxFunctionReader *funReader;
    FlxMatrixReader *mtxReader;
    FlxEvaluateCmd *EvaluateCmd;
    FlxDefParaBox AllDefParaBox;
  public:
    /**
    * @brief Create a FesslixMain Class
    * @param coutV (cout)
    * @param cerrV (cerr)
    * @param gaussI (the input file stream to read in the gauss points - will be autom. deleted after input!)
    */
    FesslixMain ( std::ostream& coutV, std::ostream& cerrV, ReadStream* gaussRS );
    ~FesslixMain ();
    
    /**
    * @brief logs some informations
    */
    static void logInfo (std::ostream& lout, bool pversion);
    static std::string get_version();
    
    /**
    * @brief Evaluate the File (this will take some time - depending on the size of the file).
    */
    void evaluate ();
    void evaluate (ReadStream *reader); 
    
    #if FLX_DEBUG
      void test ();
    #endif
};

#endif // fesslix_fesslix_H
