/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxMtx.h"

// void MtxZero(tdoubleBoostMtx& mtx)
// {
//   for (tnlong i = 0; i<mtx.size1(); i++) {
//     for (tnlong j=0;j<mtx.size2();j++) {
//       mtx(i,j)=0.0;
//     }
//   }
// }
// 
// void MtxZero(tdoubleBoostSymMtx& mtx)
// {
//   for (tnlong i = 0; i<mtx.size1(); i++) {
//     for (tnlong j=0;j<=i;j++) {
//       mtx(i,j)=0.0;
//     }
//   }
// }


void MtxProd_BTKB(const FlxMtx& B, const FlxMtxSym& E, FlxMtxSym& K)
{
  #if FLX_DEBUG
    if (K.ncols() != B.ncols() || B.nrows() != E.nrows() ) {
      std::ostringstream ssV;
      ssV << "Matrices can not be multiplied.";
      throw FlxException("MtxProd_BTKB_1", ssV.str() );
    }
  #endif
  const tdouble* Bp = &B.mtx[0];
  const tdouble* Ep = &E.mtx[0];
  tdouble* Kp = &K.mtx[0];
  const tnlong a = K.nrows();
  const tnlong b = E.nrows();
  tVec h(b); 
  tdouble* hp = &h[0];
  #if FLX_KAHAN_MTX_FULL
    pdouble t;
  #else
    tdouble t;
  #endif
  tnlong n, m, k, j;
  for (tnlong i = 0; i < a; ++i) {
    h=0.0;
    for (n=0; n < b; ++n) {
      t = 0.0;
      for (m=0; m < b; ++m) {
	t+=Bp[m*a+i]*((m<=n)?(Ep[(n*n+n)/2+m]):(Ep[(m*m+m)/2+n]));
      }
      #if FLX_KAHAN_MTX_FULL
	hp[n]=t.cast2double();
      #else
	hp[n]=t;
      #endif
    }
    for (j=i; j<a; ++j) {
      k=(j*j+j)/2+i;
      t = 0.0;
      for (n=0; n<b; ++n) {
	t+=hp[n]*Bp[n*a+j];
      }
      #if FLX_KAHAN_MTX_FULL
	Kp[k]=t.cast2double();
      #else
	Kp[k]=t;
      #endif
    }
  }
}

FlxMtxSym MtxProd_BTKB(const FlxMtx& B, const FlxMtxSym& E)
{
  FlxMtxSym K(B.ncols());
  MtxProd_BTKB(B,E,K);
  return K;
}

void MtxProd_BTKB(const FlxMtxTransformation& B, const FlxMtxSym& E, FlxMtxSym& K)
{
  #if FLX_DEBUG
    if (K.ncols() != B.ncols() || B.nrows() != E.nrows() ) {
      std::ostringstream ssV;
      ssV << "Matrices can not be multiplied.";
      throw FlxException("MtxProd_BTKB_2", ssV.str() );
    }
  #endif
  const size_t S = B.Ttm.size();
  tnlong c = 0; tnlong r;
  const FlxMtx *M, *N;
  size_t Mrows,Mcols,Nrows;
  tVec& Km = K.mtx;
  register tnlong l,j,k,i;	// loop variables
  tnlong m,n; 
  tVec h(K.nrows());
  // search maximum dim of g and f
    m=0;
    for (i=0;i<S;++i) {
      if (B.Ttm[i]->nrows() > m) m = B.Ttm[i]->nrows();
    }
    tVec g(m);
    tVec f(m);
  for (i=0; i<S;++i) {
    M = B.Ttm[i];
    const tVec& Mm = M->mtx;
    Mrows = M->nrows(); Mcols = M->ncols();
    for (j=0;j<Mcols;++j) {
      g = Mm[std::slice(j,Mrows,Mrows)];
      E.MultMv_slice(g,h,c,c+Mcols-1);
      r = 0;
      for (k=0;k<S;++k) {
	N = B.Ttm[k];
	Nrows = N->nrows();
	if (r+Nrows-1>=c+j) { 
	  N->TransposeMmultVec(h[std::slice(r,Nrows,1)],f);
	  n = (r>c+j)?(r):(c+j);
	  for (l=r+Nrows;l>n;--l) {
	    m = l - 1;
	    Km[(m*m+m)/2+c+j]=f[l-r-1];
	  }
	}
	r+=Nrows;
      }
    }
    c += Mcols;
  }
}

FlxMtxSym MtxProd_BTKB(const FlxMtxTransformation& B, const FlxMtxSym& E)
{
  FlxMtxSym K(B.ncols());
  MtxProd_BTKB(B,E,K);
  return K;
}

void MtxProd_BTKB(const FlxMtxLTri& B, const FlxMtxSym& E, FlxMtxSym& K)
{
  #if FLX_DEBUG
    if (K.ncols() != B.ncols() || B.nrows() != E.nrows() ) {
      std::ostringstream ssV;
      ssV << "Matrices can not be multiplied.";
      throw FlxException("MtxProd_BTKB_3", ssV.str() );
    }
  #endif
  const tdouble* Bp = &B.mtx[0];
  const tdouble* Ep = &E.mtx[0];
  tdouble* Kp = &K.mtx[0];
  const tnlong a = K.nrows();
  const tnlong b = E.nrows();
  #if FLX_KAHAN_MTX_LTRI
    pdouble t;
  #else
    tdouble t;
  #endif
  tVec h(b);
  tdouble* hp = &h[0];
  register tnlong i,n,j,k,m;
  for (i = 0; i < a; ++i) {
    for (n=0; n < b; ++n) {
      t = 0.0;
      for (m=i; m < b; ++m) {
	if (n>=m) {
	  t+=Bp[(m*m+m)/2+i]*Ep[(n*n+n)/2+m];
	} else {
	  t+=Bp[(m*m+m)/2+i]*Ep[(m*m+m)/2+n];
	}
      }
      #if FLX_KAHAN_MTX_LTRI
	hp[n]=t.cast2double();
      #else
	hp[n]=t;
      #endif
    }
    for (j=i; j<a; ++j) {
      k=(j*j+j)/2+i;
      t = 0.0;
      for (n=j; n< b; ++n) {
	t+=hp[n]*Bp[(n*n+n)/2+j];
      }
      #if FLX_KAHAN_MTX_LTRI
	Kp[k]=t.cast2double();
      #else
	Kp[k]=t;
      #endif
    }
  } 
}

FlxMtxSym MtxProd_BTKB(const FlxMtxLTri& B, const FlxMtxSym& E)
{
  FlxMtxSym K(B.ncols());
  MtxProd_BTKB(B,E,K);
  return K;
}


void MtxProd_BTKB(const FlxMtx_base& B, const FlxMtx_baseS& E, FlxMtxSym& K)
{
  #if FLX_DEBUG
    if (K.ncols() != B.ncols() || B.nrows() != E.nrows() ) {
      std::ostringstream ssV;
      ssV << "Matrices can not be multiplied.";
      throw FlxException("MtxProd_BTKB_4", ssV.str() );
    }
  #endif
  tdouble* Kp = &K.mtx[0];
  const tnlong a = K.nrows();
  const tnlong b = E.nrows();
  #if FLX_KAHAN_MTX_SYM
    pdouble t;
  #else
    tdouble t;
  #endif
  tVec h(b);
  tdouble* hp = &h[0];
  register tnlong i,n,m,j,k;
  for (i = 0; i < a; i++) {
    for (n=0; n < b; n++) {
      t = 0.0;
      for (m=0; m < b; m++) {
	t+=B(m,i)*E(n,m);
      }
      #if FLX_KAHAN_MTX_SYM
	hp[n] = t.cast2double();
      #else
	hp[n] = t;
      #endif
    }
    for (j=i; j<a; j++) {
      k=(j*j+j)/2+i;
      t = 0.0;
      for (n=0; n<b; n++) {
	t+=hp[n]*B(n,j);
      }
      #if FLX_KAHAN_MTX_SYM
	Kp[k] = t.cast2double();
      #else
	Kp[k] = t;
      #endif
    }
  }
}

FlxMtxSym MtxProd_BTKB(const FlxMtx_base& B, const FlxMtx_baseS& E)
{
  FlxMtxSym K(B.ncols());
  MtxProd_BTKB(B,E,K);
  return K;
}

void MtxProd_BTKB_1D(const FlxMtx& B, const FlxMtxSym& E, FlxMtxSym& K)
{
  #if FLX_DEBUG
    if (K.ncols()!=6 || B.nrows()!=3 || B.ncols()!=3 || E.nrows()!=2 ) {
      std::ostringstream ssV;
      ssV << "Matrices can not be multiplied.";
      throw FlxException("MtxProd_BTKB_1D_1", ssV.str() );
    }
  #endif
  tdouble t1,t2,t3;
  const tVec& Emtx = E.get_mtx_tVec();
  const tVec& Bmtx = B.get_mtx_tVec();
  tdoublePtr Kmtx = K.get_VecPointer();
  for (tuint i=0; i<3; ++i) {
    t1 = Emtx[0] * Bmtx[i];
    t2 = Emtx[1] * Bmtx[i];
    t3 = Emtx[2] * Bmtx[i];
    for (tuint j=i; j<3; ++j) {
      Kmtx[(j*j+j)/2+i] = t1 * Bmtx[j];
    }
    for (tuint j=0; j<3; ++j) {
      Kmtx[(j*j+7*j)/2+i+6] = t2 * Bmtx[j];
    }
    for (tuint j=i; j<3; ++j) {
      Kmtx[(j*j+7*j)/2+i+9] = t3 * Bmtx[j];
    }
  }
}

void MtxProd_BTKB_1D_part(const FlxMtx& B, const FlxMtxSym& E, FlxMtxSym& K, bool tranformFirst)
{
  #if FLX_DEBUG
    if (K.ncols()!=4 || B.nrows()!=3 || B.ncols()!=3 || E.nrows()!=2 ) {
      std::ostringstream ssV;
      ssV << "Matrices can not be multiplied.";
      throw FlxException("MtxProd_BTKB_1D_part_1", ssV.str() );
    }
  #endif
  tdouble t1,t2,t3;
  const tVec& Emtx = E.get_mtx_tVec();	// the element stiffness matrix (local)
  const tVec& Bmtx = B.get_mtx_tVec();	// the transformation matrix
  tdoublePtr Kmtx = K.get_VecPointer();	// the element stiffness matrix (global)
  if (tranformFirst) {
    Kmtx[9] = Emtx[2];
    for (tuint i=0; i<3; ++i) {
      t1 = Emtx[0] * Bmtx[i];
      t2 = Emtx[1] * Bmtx[i];
      for (tuint j=i; j<3; ++j) {
	Kmtx[(j*j+j)/2+i] = t1 * Bmtx[j];
      }
      Kmtx[6+i] = t2;
    }
  } else {
    Kmtx[0] = Emtx[0];
    for (tuint i=0; i<3; ++i) {
      t2 = Emtx[1] * Bmtx[i];
      t3 = Emtx[2] * Bmtx[i];
      for (tuint j=i; j<3; ++j) {
	Kmtx[(j*(j+3))/2+i+2] = t3 * Bmtx[j];
      }
      Kmtx[(i*(i+3))/2+1] = t2;
    }
  }
}

FlxMtxSym VecDyadProd(const tVec& v)
{
  const size_t N = v.size();
  FlxMtxSym B(N);
  tdoublePtr Bmtx = B.get_VecPointer();
  for (size_t i = 0; i < N; ++i) {
    for (size_t j = 0; j <= i; ++j) {
      Bmtx[(i*i+i)/2+j]=v[i]*v[j];
    }
  }
  return B;
}

void tVec_simple_plot(std::ostream& os, const tVec& V)
{
  for (size_t i = 0; i < V.size(); i++) {
    if (i!=0) os << "\t";
    os << GlobalVar.Double2String( V[i], true );
  }
}

std::ostream& operator<<(std::ostream& os, const tVec& V)
{
  os << "( ";
  for (size_t i = 0; i < V.size(); i++) {
    if (i!=0) os << ", ";
    os << GlobalVar.Double2String( V[i] );
  }
  os << " )";
  return os;
}

std::ostream& operator<<(std::ostream& os, const bVec& V)
{
  os << "(";
  for (size_t i = 0; i < V.size(); i++) {
    if (i!=0) os << ", ";
    os << ((V[i]==true)?("1"):("0"));
  }
  os << " )";
  return os;
}

std::ostream& operator<<(std::ostream& os, const FlxMtx_base& M)
{
  M.output_Mtx(os);
  return os;
}

std::ostream& operator<<(std::ostream& os, const std::vector< tuint >& V)
{
  os << "(";
  for (size_t i = 0; i < V.size(); i++) {
    if (i!=0) os << ",";
    os << V[i];
  }
  os << ")";
  return os;
}

FlxMtx FlxMtx_base::operator*(const FlxMtx_base& rhs) const
{
  const FlxMtx_base& lhs = *this;
  #if FLX_DEBUG
    if (lhs.ncols() != rhs.nrows()) {
      std::ostringstream ssV;
      ssV << "Matrices can not be multiplied.";
      throw FlxException("FlxMtxSparsSym::operator*_1", ssV.str() );
    }
  #endif
  const tnlong a = lhs.nrows();
  const tnlong b = rhs.ncols();
  const tnlong c = lhs.ncols();
  FlxMtx M(a,b);
  register tnlong k; register tnlong j; register tnlong i;
  for (i = 0; i<a;i++) {
    for (j = 0; j < b; j++) {
      for (k = 0; k < c; k++) {
	M(i,j)+=lhs(i,k)*rhs(k,j);
      }
    }
  }
  return M;
}

void tVec_cast2tuintNo0(const tVec& V, iVec& R)
{
  #if FLX_DEBUG
    if (V.size() != R.size()) {
      std::ostringstream ssV;
      ssV << "Vector sizes do not match.";
      throw FlxException("tVec_cast2tuintNo0_1", ssV.str() );
    }
  #endif
  const size_t S = V.size();
  tuint* Rp = &R[0];
  const tdouble* Vp = &V[0];
  for (tuint i=0;i<S;++i) {
    if ( Vp[i] <= 0.0) { 
      std::ostringstream ssV; ssV << "Number must not be negative or zero ([" << i+1 << "]=" << Vp[i] << ").";
      FlxException("tVec_cast2tuintNo0_2", "Expected unsigned integer!", ssV.str() );     
    } else {
      Rp[i] = static_cast<tuint>(Vp[i]);
    }
  }
}

const tVec FlxMtx_base::operator*(const tVec& v ) const
{
  #if FLX_DEBUG
    if (v.size() != ncols()) {
      std::ostringstream ssV;
      ssV << "Vector/Matrix sizes do not match.";
      throw FlxException("FlxMtxSparsSym::operator*_2", ssV.str() );
    }
  #endif
  tVec w(nrows());
  MultMv(v,w);
  return w;
}

const pVec FlxMtx_base::operator*(const pVec& v) const
{
  #if FLX_DEBUG
    if (v.size() != ncols()) {
      std::ostringstream ssV;
      ssV << "Vector/Matrix sizes do not match.";
      throw FlxException("FlxMtxSparsSym::operator*_3", ssV.str() );
    }
  #endif
  pVec w(nrows());
  MultMv(v,w);
  return w;
}

void FlxMtx_base::MultMv(const pVec& v, pVec& w) const
{
  throw FlxException_NotImplemented("FlxMtx_base::MultMv");
}

FlxMtx_base* FlxMtx_base::copy()
{
  throw FlxException_NotImplemented("FlxMtx_base::copy");
}

void FlxMtx_base::output_Mtx(std::ostream& sout) const
{
  tnlong N = nrows();
  tnlong M = ncols();
  for (tnlong i = 0; i<N;i++) {
    for (tnlong j = 0; j<M; j++) {
      sout << GlobalVar.Double2String( operator()(i,j) ) << "\t";
    }
    sout << std::endl;
  }
}

void FlxMtx_base::output_OctaveMtx(std::ostream& sout, bool checkTOLv, bool doendl) const
{
  tnlong N = nrows();
  tnlong M = ncols();
  sout << "[ ";
  for (tnlong i = 0; i<N;i++) {
    for (tnlong j = 0; j<M; j++) {
      sout << GlobalVar.Double2String(operator()(i,j),checkTOLv);
      if (j < M-1) sout << " ";
    }
    if (i < N-1) sout << "; ";
    else sout << "]";
    if (doendl) sout << std::endl;
  }
}

const tdouble FlxMtx_base::max() const
{
  const tnlong N = nrows();
  const tnlong M = ncols();
  tdouble d = operator()(0,0);
  tdouble dt;
  for (tnlong i = 0; i<N;i++) {
    for (tnlong j = 0; j<M; j++) {
      dt = operator()(i,j);
      if ( dt > d) {
	d = dt;
      }
    }
  }
  return d;
}

const tdouble FlxMtx_base::min() const
{
  const tnlong N = nrows();
  const tnlong M = ncols();
  tdouble d = operator()(0,0);
  tdouble dt;
  for (tnlong i = 0; i<N;i++) {
    for (tnlong j = 0; j<M; j++) {
      dt = operator()(i,j);
      if (dt < d) {
	d = dt;
      }
    }
  }
  return d;
}

void FlxMtx_baseS::assembleMinv(int i)
{
  if (Minv!=NULL) {
    delete Minv;
    Minv = NULL;
  }
  if (i == 0) {
    Minv = new FlxMtxIdentity(this->nrows());
  } else if ( i == 1) {
    FlxMtxDiag* d = new FlxMtxDiag(*this);
    Minv = d->get_Inverse();
    delete d;
  } else if ( i == 3) {
    FlxMtxSparsSym* S = dynamic_cast<FlxMtxSparsSym*>(this);
    if (S != NULL) {
      FlxMtxLTri L(S->nrows());
      L.CholeskyDec(*S);
      L.Invert();
      Minv = new FlxMtxSym(L);
      return;
    } 
    FlxMtxSym* SF = dynamic_cast<FlxMtxSym*>(this);
    if (SF != NULL) {
      FlxMtxLTri L(SF->nrows());
      L.CholeskyDec(*SF);
      L.Invert();
      Minv = new FlxMtxSym(L);
      return;
    } 
    std::ostringstream ssV;
    ssV << "Matrix is not a symmetric sparse matrix nor an symmetric dense matrix.";
    throw FlxException("FlxMtx_baseS::assembleMinv_1", ssV.str() );
  } else if ( i==4 || i == 5 || i == 6 ) {
    FlxMtxSparsSym* S = dynamic_cast<FlxMtxSparsSym*>(this);
    if (S != NULL) {
      if (i==4) {
	Minv = new FlxMtxSparsSymLU(*S);
      } else if (i==5) {
	Minv = new FlxMtxSparsSymILU(*S);
      } else {
	Minv = new FlxMtxSparsSymILU(*S,true);
      }
      return;
    }
    std::ostringstream ssV;
    ssV << "Matrix is not a symmetric sparse matrix.";
    throw FlxException("FlxMtx_baseS::assembleMinv_2", ssV.str() );
  } else {
    Minv = NULL;
  }
}

FlxMtx_baseS::~FlxMtx_baseS()
{
  if (Minv!=NULL) delete Minv;
}

void FlxMtx_baseS::preconding(const tVec& r, tVec& z, int i) const
{
  throw FlxException_NotImplemented("FlxMtx_baseS::preconding_1");
}

void FlxMtx_baseS::preconding(const pVec& r, pVec& z, int i) const
{
  throw FlxException_NotImplemented("FlxMtx_baseS::preconding_2");
}

const bool FlxMtx_baseS::solve_CG(tVec& x, tVec& f, tdouble& eps, tnlong& iter, tuint pcn, bool startZero)
{
  if (ncols() != nrows() || nrows() != f.size() || f.size() != x.size() ) {
    std::ostringstream ssV;
    ssV << "ncols=" << ncols() << "; nrows=" << nrows() << "; f_size=" << f.size() << "; x_size=" << x.size() << ";";
    throw FlxException("FlxMtx_base::solve_CG_1", "Vector/Matrix sizes do not match.", ssV.str() );
  }
  
  // check for trivial solution
  if (NormMax(f)==0.0) {
    x *= 0.0;
    iter = 0;
    eps = 0.0;
    return true;
  }
  
  assembleMinv(pcn);
  
  if (startZero) { x*=0.0; }
  
  const tnlong maxiter = iter;
  
  #if FLX_KAHAN_CG_STAGE
    pVec X(f.size());
      pVec_assemble(X,x);
  #endif
  #if FLX_KAHAN_CG_STAGE_4
    pVec r(f.size());
      pVec_assemble(r,f - (*this)*x);
    pVec z(f.size());
  #else
    tVec r = f - (*this)*x;
    #if FLX_KAHAN_CG_STAGE_3 || FLX_KAHAN_CG_STAGE_2
      pVec R(f.size());
	pVec_assemble(R,r);
    #endif
    tVec z(f.size());
  #endif
  if (Minv==NULL) {
    preconding(r, z, pcn);
  } else {
    Minv->MultMv(r,z);
  }
  #if FLX_KAHAN_CG_STAGE_4
    pVec p = z;
    pdouble zr = dot(r,z);
  #else
    tVec p = z;
    #if FLX_KAHAN_CG_STAGE_3
      pVec P(f.size());
	pVec_assemble(P,p);
    #endif
    tdouble zr = dot(r,z);
  #endif
  const double stp = eps*Norm2(f);
 
  #if FLX_KAHAN_CG_STAGE_4
    tVec rr(f.size());
    tVec_assemble(rr,r);
  #else
    tVec& rr = r;
  #endif
  if (NormMax(rr) <= GlobalVar.TOL() ) {
    eps = 0.0; 
    iter = 0;
    return true;
  }

  #if FLX_KAHAN_CG_STAGE_4
    pdouble pap,alpha,zrold,beta;
    pVec mp(p.size());
  #else
    tdouble pap,alpha,zrold,beta;
    tVec mp(p.size());
  #endif
  for (iter = 0; iter < maxiter; iter++) {
    MultMv(p,mp);	//     mp = (*this)*p;
    pap = dot(p,mp);
    alpha = zr/pap;
    #if FLX_KAHAN_CG_STAGE_4
      X+=alpha*p;
      r-=alpha*mp;
	tVec_assemble(rr,r);
    #else
      #if FLX_KAHAN_CG_STAGE_3 || FLX_KAHAN_CG_STAGE_2
	pVec_add(X,alpha*p);
	pVec_add(R,(-1*alpha)*mp);
	  tVec_assemble(r,R);
      #else
	#if FLX_KAHAN_CG_STAGE_1
	  pVec_add(X,alpha*p);
	#else
	  x+=alpha*p;
	#endif
	r-=alpha*mp;
      #endif
    #endif
    if (Norm2(rr) <= stp) {
      break;
    }
    if (Minv==NULL) {
      preconding(r, z, pcn);
    } else {
      Minv->MultMv(r,z);
    }
    zrold = zr;
    zr = dot(r,z);
    beta = zr/zrold;
    #if FLX_KAHAN_CG_STAGE_4
      p*=beta;
      p+=z;
    #else
      #if FLX_KAHAN_CG_STAGE_3
	pVec_mult(P,beta);
	pVec_add(P,z);
	tVec_assemble(p,P);
      #else
	p*=beta;
	p+=z;
      #endif
    #endif
  }

  #if FLX_KAHAN_CG_STAGE_4
    tVec_assemble(rr,r);
  #endif
  eps = Norm2(rr);
  #if FLX_KAHAN_CG_STAGE
    tVec_assemble(x,X);
    tVec_assemble(rr,(*this)*X);
    rr = f - rr;
  #else
    rr = f - (*this)*x;
  #endif
  
  // check for difference ...
    tdouble de = Norm2(rr);
    if (de > stp && eps > 0.0) {
      std::ostringstream ssV;
      ssV << "Drifting error (" << GlobalVar.Double2String(fabs(de-eps)/eps) << "). ";
      ssV << "Residuum (" << Norm2(rr) << ") not within bound (" << stp << "). (iter=" << iter << "; unknowns=" << x.size() << ")";
      GlobalVar.alert.alert("FlxMtx_base::solve_CG_3",ssV.str());
    }
    #if FLX_DEBUG
      if (eps>0.0) {
	GlobalVar.slog(4) << "  Drifting Error: " << GlobalVar.Double2String(fabs(de-eps)/eps) << std::endl;
      }
    #endif
    eps = de;
  checkTOL(x);
  if (iter==maxiter) return false;
  else return true;
}

FlxMtxDiag* FlxMtx_baseS::get_diag() const
{
  return new FlxMtxDiag(*this);
}

FlxMtx_baseS* FlxMtx_baseS::get_Inverse()
{
  throw FlxException_NotImplemented("FlxMtx_baseS::get_Inverse");
}

void FlxMtx_baseS::output_Mtx(std::ostream& sout) const
{
  for (tnlong i =0;i<nrows();i++) {
    for (tnlong j = 0; j <= i; j++) {
      sout << GlobalVar.Double2String( operator()(i,j) ) << "\t";
    }
    sout << std::endl;
  }
}

FlxMtx& FlxMtx::operator+=(const FlxMtx& m)
{
  #if FLX_DEBUG
    if (m.ncols() != ncols() || m.nrows() != nrows() ) {
      std::ostringstream ssV;
      ssV << "Matrices can not be added.";
      throw FlxException("FlxMtxSym::operator+=", ssV.str() );
    }
  #endif
  mtx+=m.mtx;
  return *this;
}

// FlxMtx FlxMtx::operator*(const FlxMtx_base& m) const
// {
//   #if FLX_DEBUG
//     if (m.nrows() != ncols()) {
//       std::ostringstream ssV;
//       ssV << "Matrices can not be multiplied.";
//       throw FlxException("FlxMtx::operator*", ssV.str() );
//     }
//   #endif
//   FlxMtx r(nrows(),m.ncols());
//   for (tnlong i = 0; i < nrows(); i++) { // rows
//     for (tnlong j = 0; j < r.ncols(); j++) { // colums
//       r(i,j)=0.0;
//       for (tnlong k = 0; k < ncols(); k++) {  // Scalar-Product
// 	r(i,j)+=operator()(i,k)*r(k,j);
//       }
//     }
//   }
// }

tdouble& FlxMtx::operator()(const tnlong& i, const tnlong& j)
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= ncols()) {
      std::ostringstream ssV;
      ssV << "Index (" << i << ", " << j << ") not within matrix.";
      throw FlxException("FlxMtx::operator()_1", ssV.str() );
    }
  #endif
  return mtx[i*csize+j];
}

const tdouble FlxMtx::operator()(const tnlong& i, const tnlong& j) const
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= ncols()) {
      std::ostringstream ssV;
      ssV << "Index (" << i << "," << j << ") not within matrix (" << nrows() << "," << ncols() << ").";
      throw FlxException("FlxMtx::operator()_2", ssV.str() );
    }
  #endif
  return mtx[i*csize+j];
}

void FlxMtx::MultMv(const tVec& v, tVec& w) const
{
  #if FLX_KAHAN_MTX_FULL
    pdouble t;
  #else
    tdouble t;
  #endif
  size_t n = 0;
  const tdouble* mtxp = &mtx[0];
  const tdouble* vp = &v[0];
  tdouble* wp = &w[0];
  for (tnlong i = 0; i<rsize;++i) {
    t = 0.0;
    for (tnlong j=0;j<csize;++j) {
      t+=mtxp[n]*vp[j];
      ++n;
    }
    #if FLX_KAHAN_MTX_FULL
      wp[i] = t.cast2double();
    #else
      wp[i] = t;
    #endif
  }
}

FlxMtx* FlxMtx::copy()
{
  return new FlxMtx(*this);
}

void FlxMtx::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= ncols()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtx::add_value", ssV.str() );
    }
  #endif
  mtx[i*ncols()+j]+=v;
}

void FlxMtx::TransposeMmultM(FlxMtxSym& res) const
{
  #if FLX_DEBUG
    if (res.nrows() != ncols()) {
      std::ostringstream ssV;
      ssV << "ERROR.";
      throw FlxException("FlxMtx::TransposeMmultM", ssV.str() );
    }
  #endif
  const tnlong c = ncols();
  const tnlong r = nrows();
  #if FLX_KAHAN_MTX_FULL
    pdouble d;
  #else
    tdouble d;
  #endif
  const tdouble* mtxp = &mtx[0];
  register tnlong k; register tnlong j; register tnlong i;
  for (i = 0; i < c; ++i) {
    for (j = 0; j <= i; ++j) {
      d = 0.0;
      for (k = 0; k < r; ++k) {
	d+= mtxp[k*ncols()+i]*mtxp[k*c+j];
      }
      #if FLX_KAHAN_MTX_FULL
	res.operator()(i,j) = d.cast2double();
      #else
	res.operator()(i,j) = d;
      #endif
    }
  }
}

void FlxMtx::TransposeMmultVec(const tVec& v, tVec& w) const
{
  #if FLX_DEBUG
    if (v.size() != nrows() || w.size() != ncols() ) {
      std::ostringstream ssV;
      ssV << "ERROR.";
      throw FlxException("FlxMtx::TransposeMmultM", ssV.str() );
    }
  #endif
  #if FLX_KAHAN_MTX_FULL
    pdouble d;
  #else
    register tdouble d;
  #endif
  const tdouble* mtxp = &mtx[0];
  const tdouble* vp = &v[0];
  tdouble* wp = &w[0];
  register tnlong k,i;
  for (i = 0; i < csize; ++i) {
    d = 0.0;
    for (k = 0; k < rsize; ++k) {
      d+= mtxp[k*csize+i]*vp[k];
    }
    #if FLX_KAHAN_MTX_FULL
      wp[i] = d.cast2double();
    #else
      wp[i] = d;
    #endif
  }
}

FlxMtxTransformation::FlxMtxTransformation(const std::vector< FlxMtx* >& Ttm): Ttm(Ttm)
{
  rsize=0;
  for (size_t i=0; i < Ttm.size(); ++i) {
    #if FLX_DEBUG
      if (Ttm[i]->ncols() != Ttm[i]->nrows() ) {
	std::ostringstream ssV;
	ssV << "ERROR.";
	throw FlxException("FlxMtxTransformation::FlxMtxTransformation_1", ssV.str() );
      }
    #endif
    rsize+= Ttm[i]->ncols();
  }
}

FlxMtxTransformation::~FlxMtxTransformation()
{
  for (size_t i=0; i<Ttm.size(); ++i) {
    if (Ttm[i]!=NULL) {
      for (size_t j=i+1;j<Ttm.size();++j) {
	if (Ttm[i]==Ttm[j]) {
	  Ttm[j] = NULL;
	}
      }
      delete Ttm[i];
      Ttm[i] = NULL;
    }
  }
}

FlxMtxTransformation& FlxMtxTransformation::operator*=(const tdouble& s)
{
  throw FlxException_NotImplemented("FlxMtxTransformation::operator*=");
}

void FlxMtxTransformation::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  throw FlxException_NotImplemented("FlxMtxTransformation::add_value");
}

const tdouble FlxMtxTransformation::operator()(const tnlong& n, const tnlong& m) const
{
  tnlong i = n;
  tnlong j = m;
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= ncols()) {
      std::ostringstream ssV;
      ssV << "Index (" << i << "," << j << ") not within matrix (" << nrows() << "," << ncols() << ").";
      throw FlxException("FlxMtxTransformation::operator()_1", ssV.str() );
    }
  #endif
  tnlong c=0;
  for (size_t ii=0;ii<Ttm.size();++ii) {
    if (i<c+Ttm[ii]->nrows()) {
      if (j<c+Ttm[ii]->nrows() && j>=c) {
	i-=c;
	j-=c;
	return Ttm[ii]->operator()(i,j);
      } else {
	return 0.0;
      }
    }
    c+=Ttm[ii]->nrows();
  }
  std::ostringstream ssV;
  ssV << "ERROR";
  throw FlxException("FlxMtxTransformation::operator()_2", ssV.str() );
}

void FlxMtxTransformation::MultMv(const tVec& v, tVec& w) const
{
  #if FLX_DEBUG
    if (v.size()!=w.size() || v.size()!=rsize ) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxMtxTransformation::MultMv_1", ssV.str() );
    }
  #endif
  tnlong c = 0;
  tnlong s;
  const size_t Ttms = Ttm.size();
  for (size_t i = 0; i < Ttms; ++i) {
    s = Ttm[i]->nrows();
    w[std::slice(c,s,1)] = (*Ttm[i])*v[std::slice(c,s,1)];
    c+=s;
  }
}

inline void FlxMtxSym::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= ncols()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxSym::add_value", ssV.str() );
    }
  #endif
  tnlong iV = i; tnlong jV=j;
  if (jV>iV) std::swap(iV,jV);
  mtx[(iV*iV+iV)/2+jV]+=v;
}

FlxMtxSym& FlxMtxSym::operator+=(const FlxMtxSym& m)
{
  #if FLX_DEBUG
    if (m.ncols() != ncols() || m.nrows() != nrows() ) {
      std::ostringstream ssV;
      ssV << "Matrices can not be added.";
      throw FlxException("FlxMtxSym::operator+=", ssV.str() );
    }
  #endif
  mtx+=m.mtx;
  return *this;
}

tdouble& FlxMtxSym::operator()(const tnlong& i, const tnlong& j)
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= ncols()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxSym::operator()_1", ssV.str() );
    }
  #endif
  tnlong iV = i; tnlong jV=j;
  if (jV>iV) std::swap(iV,jV);
  return mtx[(iV*iV+iV)/2+jV];
}

void FlxMtxSym::get_VecPointer_full(tdouble* vec) const
{
  const tdouble* mtxp = &mtx[0];
  register tnlong i,j;
  for (i=0;i<msize;++i) {
    for (j=0;j<i;++j) {
      vec[j*msize+i] = vec[i*msize+j] = mtxp[(i*i+i)/2+j];
    }
    vec[i*msize+i] = mtxp[(i*i+i)/2+i];
  }
}

FlxMtxSym::FlxMtxSym(const FlxMtxLTri& LTL): msize(LTL.msize),mtx((msize*msize+msize)/2)
{
  const tnlong N = nrows();
  register tdouble sum;
  #if FLX_KAHAN_MTX_SYM
    pdouble t;
  #else
    tdouble& t = sum;
  #endif
  const tdouble* LTLmtxp = &LTL.mtx[0];
  tdouble* mtxp = &mtx[0];
  register tnlong i, j, k,m;
  for (i = 0; i < N; ++i) {
    for (j = 0; j <= i; ++j) {
      t = 0.0;
      for (k = i; k < N; ++k) {
	m = (k*k+k)/2;
	t += LTLmtxp[m+j]*LTLmtxp[m+i];
      }
      #if FLX_KAHAN_MTX_SYM
	sum = t.cast2double();
      #endif
      mtxp[(i*i+i)/2+j] = sum;
    }
  }
}

FlxMtxSym::FlxMtxSym(const FlxMtx_baseS& S)
: msize(S.nrows()), mtx((msize*msize+msize)/2)
{
  register tnlong i,j;
  for (i = 0; i < msize; ++i) {
    for ( j = 0; j <= i; ++j) {
      mtx[(i*i+i)/2+j] = S.operator()(i,j);
      #if FLX_DEBUG
	if ( fabs(mtx[(i*i+i)/2+j]) <= GlobalVar.TOL() ) mtx[(i*i+i)/2+j] = 0.0;
      #endif
    }
  }
}

const tdouble FlxMtxSym::operator()(const tnlong& i, const tnlong& j) const
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= ncols()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxSym::operator()_2", ssV.str() );
    }
  #endif
  tnlong iV = i; tnlong jV=j;
  if (jV>iV) std::swap(iV,jV);
  return mtx[(iV*iV+iV)/2+jV];
}

const bool FlxMtxSym::isPositiveDefinite(tuint& r, const bool fixIt)
{
  return isPositiveDefinite_ext(mtx,msize,r,fixIt);
}

const bool FlxMtxSym::isPositiveDefinite_ext(tVec& mtxe, tuint rows, tuint& r, const bool fixIt)
{
  const tdouble* mtxp = &mtxe[0];
  const tdouble GVT = GlobalVar.TOL();
  for (register size_t i = 0; i < rows; ++i) {
    if ( mtxp[(i*i+i)/2+i] <= GVT ) {
      bool stop = false;
      if (fixIt) {
	stop = true;
	for (tuint j=0;j<i && stop;++j) {
	  if (mtxp[(i*i+i)/2+j] > GVT) {
	    stop = false;
	  }
	}
	for (tuint j=i+1;j<rows;++j) {
	  if (mtxp[(j*j+j)/2+i] > GVT) {
	    stop = false;
	  }
	}
	if (stop) {
	  mtxe[(i*i+i)/2+i] = 1.0;
	  std::ostringstream ssV;
	  ssV << "Fixed row " << i+1 << ".";
	  GlobalVar.alert.alert("FlxMtxSym::isPositiveDefinite",ssV.str());
	}
      }
      if (!stop) {
	r = i+1;
	return false;
      }
    }
  }
  return true;
}

inline void FlxMtxSym::preconding(const tVec& r, tVec& z, int precn) const
{
  if (precn == 2) {
    const tdouble omega=1.2;
    tdouble sum;
    #if FLX_KAHAN_MTX_SYM
      pdouble t;
    #else
      tdouble& t = sum;
    #endif
    for (size_t i=0;i<z.size();++i) z[i]=0.0;
    for (size_t i=0;i<msize;++i) {
      t = 0.0;
      for (size_t j = 0;j<i;++j) t += operator()(i,j)*z[j];
      #if FLX_KAHAN_MTX_SYM
	sum = t.cast2double();
      #endif
      z[i] = (r[i]-omega*sum)/operator()(i,i);
    }
    for (size_t i = msize;i>0;--i) {
      t = 0.0;
      for (size_t j = i; j<msize;++j) {
	t+=operator()(i-1,j)*z[j];
      }
      #if FLX_KAHAN_MTX_SYM
	sum = t.cast2double();
      #endif
      z[i-1]-=omega*sum/operator()(i-1,i-1);
    }
  } else {
    FlxMtx_baseS::preconding(r,z,precn);
  }
}

void FlxMtxSym::MultMv(const tVec& v, tVec& w) const
{
  #if FLX_KAHAN_MTX_SYM
    pVec W(nrows());
    pdouble* Wp = &W[0];
  #else
    tVec& W = w;
    tdouble* Wp = &W[0];
  #endif
  const tdouble* vp = &v[0];
  const tdouble* mtxp = &mtx[0];
  register size_t n = 0;
  register size_t i, j;
  for (i = 0; i<msize;++i) {
    Wp[i] = 0.0;
    for (j=0;j<i;++j) {
      Wp[i]+=mtxp[n]*vp[j];
      Wp[j]+=mtxp[n]*vp[i];
      ++n;
    }
    Wp[i]+=mtxp[n]*vp[i];
    ++n;
  }
  #if FLX_KAHAN_MTX_SYM
    tVec_assemble(w,W);
  #endif
}

void FlxMtxSym::MultMv_slice(const tVec& v, tVec& w, const size_t &li, const size_t &ui) const
{
  #if FLX_DEBUG
    if (w.size() != nrows() || li > ui || ui >= nrows() || v.size() != ui-li+1 ) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxMtxSym::MultMv_slice", ssV.str() );
    }
  #endif
  register size_t j, i;
  register size_t n, m;
  #if FLX_KAHAN_MTX_SYM
    pVec W(nrows());
    pdouble* Wp = &W[0];
  #else
    tVec& W = w;
    tdouble* Wp = &W[0];
  #endif
  const tdouble* vp = &v[0];
  const tdouble* mtxp = &mtx[0];
  for (i = 0; i<msize;++i) {
    W[i] = 0.0;
    // lower part of the matrix
    n = (i*i+i)/2+li;
    m = ((ui<i)?ui:i);
    for (j=li;j<=m;++j) {
      Wp[i]+=mtxp[n]*vp[j-li];
      ++n;
    }
    // uper part of the matrix
    m = ((li<=i)?i+1:li);
    for (j=m;j<=ui;++j) {
      Wp[i]+=mtxp[(j*j+j)/2+i]*vp[j-li];
    }
  }
  #if FLX_KAHAN_MTX_SYM
    tVec_assemble(w,W);
  #endif
}

void FlxMtxSym::MultMv(const pVec& v, pVec& w) const
{
  register size_t n = 0;
  register size_t i, j;
  const pdouble* vp = &v[0];
  pdouble* wp = &w[0];
  const tdouble* mtxp = &mtx[0];
  for (i = 0; i<msize;++i) {
    wp[i] = 0.0;
    for (j=0;j<i;++j) {
      wp[i]+=vp[j]*mtxp[n];
      wp[j]+=vp[i]*mtxp[n];
      ++n;
    }
    wp[i]+=vp[i]*mtxp[n];
    ++n;
  }
}

FlxMtxSym* FlxMtxSym::copy()
{
  return new FlxMtxSym(*this);
}

void FlxMtxDiag::add_value(const tnlong& ii, const tnlong& jj, const tdouble& v)
{
  if (v==0.0) return;
  tnlong i=ii;
  tnlong j=jj;
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= nrows()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxDiag::add_value_1", ssV.str() );
    }
  #endif
  if (i == j) {
    mtx[i]+=v;
  } else {
    std::ostringstream ssV;
    ssV << "Not allowed to add value '" << v << "' at this position (" << i << ", " << j << ") - diagonal matrix.";
    throw FlxException("FlxMtxDiag::add_value_2", ssV.str() );
  }
}

const bool FlxMtxDiag::isPositiveDefinite(tuint& r, const bool fixIt)
{
  if (mtx.min() <= GlobalVar.TOL() ) return false;
  else return true;
}

FlxMtx_baseS* FlxMtxDiag::get_Inverse()
{
  FlxMtxDiag* Inv = new FlxMtxDiag(msize);
  tdouble* Invmtxp = &Inv->mtx[0];
  const tdouble* mtxp = &mtx[0];
  for (register tnlong i = 0; i < msize; ++i) {
    Invmtxp[i] = 1/mtxp[i];
  }
  return Inv;
}

FlxMtxDiag::FlxMtxDiag(const FlxMtx_baseS& mb)
: msize(mb.ncols()),mtx(mb.ncols())
{
  for (tnlong i = 0; i < msize; ++i) {
    mtx[i] = mb.operator()(i,i);
  }
}

FlxMtx_baseS* FlxMtxIdentity::get_Inverse()
{
  return new FlxMtxIdentity(msize);
}

FlxMtxIdentity& FlxMtxIdentity::operator*=(const tdouble& s)
{
  std::ostringstream ssV;
  ssV << "Invalid operation.";
  throw FlxException("FlxMtxIdentity::operator*=", ssV.str() );
}

FlxMtxIdentity& FlxMtxIdentity::operator+=(const FlxMtx_base& m)
{
  std::ostringstream ssV;
  ssV << "Invalid operation.";
  throw FlxException("FlxMtxIdentity::operator+=", ssV.str() );
}

void FlxMtxIdentity::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  std::ostringstream ssV;
  ssV << "Invalid operation.";
  throw FlxException("FlxMtxIdentity::add_value", ssV.str() );
}

void FlxMtxSparsSym::set_value(const tnlong& ii, const tnlong& jj, const tdouble& v)
{
  tnlong i=ii;
  tnlong j=jj;
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= nrows()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxSparsSym::set_value_1", ssV.str() );
    }
  #endif
  if (i == j) {
    sa[i]=v;
    return;
  }
  if (j > i) std::swap(i, j);
  for (register tnlong k = ija[i];k<ija[i+1];k++) {
    if (ija[k]==j) {
      sa[k]=v;
      return;
    }
    else if (ija[k]>j) {
      break;
    }
  }
  std::ostringstream ssV;
  ssV << "Not allowed to set value '" << v << "' at this position (" << i << ", " << j << ") - sparse matrix.";
  throw FlxException("FlxMtxSparsSym::set_value_2", ssV.str() );
}

void FlxMtxSparsSym::add_value(const tnlong& ii, const tnlong& jj, const tdouble& v)
{
  if (v==0.0) return;
  tnlong i=ii;
  tnlong j=jj;
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= nrows()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxSparsSym::add_value_1", ssV.str() );
    }
  #endif
  if (i == j) {
    sa[i]+=v;
    return;
  }
  if (j > i) std::swap(i, j);
  for (register tnlong k = ija[i];k<ija[i+1];k++) {
    if (ija[k]==j) {
      sa[k]+=v;
      return;
    }
    else if (ija[k]>j) {
      break;
    }
  }
  std::ostringstream ssV;
  ssV << "Not allowed to add value '" << v << "' at this position (" << i << ", " << j << ") - sparse matrix.";
  throw FlxException("FlxMtxSparsSym::add_value_2", ssV.str() );
}

void FlxMtxDiag::MultMv(const tVec& v, tVec& w) const
{
  const tdouble* mtxp = &mtx[0];
  const tdouble* vp = &v[0];
  tdouble* wp = &w[0];
  for (register tnlong i = 0; i < msize; ++i) {
    wp[i] = vp[i]*mtxp[i];
  }
}

void FlxMtxDiag::MultMv(const pVec& v, pVec& w) const
{
  const tdouble* mtxp = &mtx[0];
  const pdouble* vp = &v[0];
  pdouble* wp = &w[0];
  for (register tnlong i = 0; i < msize; ++i) {
    wp[i] = vp[i]*mtxp[i];
  }
}

void FlxMtxSparsSym::MultMv(const tVec& v, tVec& w) const
{
  const tnlong N = nrows();
  #if FLX_KAHAN_MTX_SSYM
    pVec W(N);
    pdouble* Wp = &W[0];
  #else
    tVec& W = w;
    tdouble* Wp = &W[0];
  #endif
  const tdouble* vp = &v[0];
  register tnlong i, k;
  for (i=0;i<N;++i) {
    Wp[i] = sa[i]*vp[i];
  }
  for (i=0;i<N;++i) {
    for (k=ija[i];k<ija[i+1];++k) {
      Wp[i] += sa[k]*vp[ija[k]];
      Wp[ija[k]]+=sa[k]*vp[i];
    }
  }
  #if FLX_KAHAN_MTX_SSYM
    tVec_assemble(w,W);
  #endif
}

void FlxMtxSparsSym::MultMv(const pVec& v, pVec& w) const
{
  pdouble* wp = &w[0];
  const pdouble* vp = &v[0];
  const tnlong N = nrows();
  register tnlong k,i;
  for (i=0;i<N;++i) {
    wp[i] = vp[i]*sa[i];
  }
  for (i=0;i<N;++i) {
    for (k=ija[i];k<ija[i+1];++k) {
      wp[i] += vp[ija[k]]*sa[k];
      wp[ija[k]]+=vp[i]*sa[k];
    }
  }
}

FlxMtx_base& FlxMtxSparsSym::operator*=(const tdouble& s)
{
  const tnlong N=ija[nrows()];
  for (register tnlong i=0;i<N;++i) sa[i]*=s;
  return *this;
}

FlxMtxSparsSym::FlxMtxSparsSym(const FlxMtxSym& Mtx)
{
  const tnlong n = Mtx.nrows();
  const tVec& mtx = Mtx.get_mtx_tVec();
  const tdouble* mtxp = &mtx[0];
  const tnlong mtxs = mtx.size();
  tnlong nmax = n+1;
  const tdouble TOL = GlobalVar.TOL();
  register tnlong i, j;
  for (i = 0; i < mtxs; ++i) {
    if (fabs(mtxp[i])>TOL ) ++nmax;
  }
  for (i=1; i<=n; ++i) {
    if (fabs(mtxp[(i*i+i)/2-1])>TOL ) --nmax;
  }
  
//   if (nmax==n+1) {
//     sa = new tdouble[nmax+1];
//     ija = new tnlong[nmax+1];
//     ija[nmax]=nmax;
//   } else {
    sa = new tdouble[nmax];
    ija = new tnlong[nmax];
//   }
  ija[0] = n+1;
  tnlong k=n;
  tnlong c=0;
  for (i=0;i<n;++i) {
    for (j=0;j<i;++j) {
      if (fabs(mtxp[c])>TOL ) {
	++k;
	sa[k]=mtxp[c];
	ija[k]=j;
      }
      ++c;
    }
    ija[i+1]=k+1;
    sa[i] = mtxp[c++];
  }
  
}

FlxMtxSparsSym::~FlxMtxSparsSym()
{
  delete[] sa;
  delete[] ija;
}

const tdouble FlxMtxSparsSym::operator()(const tnlong& ii, const tnlong& jj) const
{
  tnlong i=ii;
  tnlong j=jj;
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= nrows()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxSparsSym::operator()_2", ssV.str() );
    }
  #endif
  if (i == j) return sa[i];
  if (j > i) std::swap(i, j);
  for (register tnlong k = ija[i];k<ija[i+1];k++) {
    if (ija[k]==j) {
      return sa[k];
    }
    else if (ija[k]>j) {
      return 0.0;
    }
  }
  return 0.0;
}

const bool FlxMtxSparsSym::isPositiveDefinite(tuint& r, const bool fixIt)
{
  const tdouble GVT = GlobalVar.TOL();
  for (register tnlong i = nrows(); i > 0; --i) {
    if ( sa[i-1] <= GVT ) { 
      bool stop = false;
      if (fixIt) {
	stop = true;
	if (ija[i-1]!=ija[i]) stop = false;
	for (tuint j=i;j<nrows();++j) {
	  if (operator()(j,i-1) > GVT) {
	    stop = false;
	  }
	}
	if (stop) {
	  sa[i-1] = 1.0;
	  std::ostringstream ssV;
	  ssV << "Fixed row " << i+1 << ".";
	  GlobalVar.alert.alert("FlxMtxSparsSym::isPositiveDefinite",ssV.str());
	}
      }
      if (!stop) {
	r = i;
	return false;
      }
    }
  }
  return true;
}

void FlxMtxSparsSym::output_Mtx(std::ostream& sout) const
{
  for (tnlong i =0;i<nrows();i++) {
    tnlong k=0;
    for (tnlong j = ija[i]; j<ija[i+1];j++) {
      while (k<ija[j]) {
	sout << "0" << "\t";
	k++;
      }
      sout << GlobalVar.Double2String(sa[j]) << "\t";
      k++;
    }
    while (k<i) {
      sout << "0" << "\t";
      k++;
    }
    sout << GlobalVar.Double2String(sa[i]);
    sout << std::endl;
  }
}

void FlxMtxSparsSym::preconding(const tVec& r, tVec& z, const int precn) const
{
  tdouble sum = 0;
  #if FLX_KAHAN_MTX_SYM
    pdouble t;
  #else
    tdouble& t = sum;
  #endif
  if (precn == 2) {
    const tnlong N = nrows();
    register tnlong i, j;
    const tdouble omega=1.2;
    for (i=0;i<N;++i) {
      t = 0;
      for (j = ija[i];j<ija[i+1];++j) {
	t += sa[j]*z[ija[j]];
      }
      #if FLX_KAHAN_MTX_SYM
	sum = t.cast2double();
      #endif
      z[i] = (r[i]-omega*sum)/sa[i];
    }
    for (i = N;i>=1;i--) {
      t = 0;
      for (j = i; j<N;++j) {
	t+=operator()(j,i-1)*z[j];
      }
      #if FLX_KAHAN_MTX_SYM
	sum = t.cast2double();
      #endif
      z[i-1]-=omega*sum/sa[i-1];
    }
  } else {
    FlxMtx_baseS::preconding(r,z,precn);
  }
}

FlxMtxLTri* FlxMtxLTri::copy()
{
  return new FlxMtxLTri(*this);
}

void FlxMtxLTri::MultMv(const tVec& v, tVec& w) const
{
  size_t n = mtx.size();
  #if FLX_KAHAN_MTX_LTRI
    pdouble t;
  #else
    tdouble t;
  #endif
  const tdouble* vp = &v[0];
  tdouble* wp = &w[0];
  const tdouble* mtxp = &mtx[0];
  register tnlong j,i;
  for (i = msize; i>0;--i) {
    t = 0.0;
    for (j=i;j>0;--j) {
      t+=mtxp[n-1]*vp[j-1];
      --n;
    }
    #if FLX_KAHAN_MTX_LTRI
      wp[i-1]=t.cast2double();
    #else
      wp[i-1]=t;
    #endif
  }
}

const tdouble FlxMtxLTri::operator()(const tnlong& i, const tnlong& j) const
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= ncols()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxLTri::operator()_1", ssV.str() );
    }
  #endif
  if (j>i) return 0.0;
  else return mtx[(i*i+i)/2+j];
}

void FlxMtxLTri::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= ncols()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxLTri::add_value_1", ssV.str() );
    }
  #endif
  if (j<=i) mtx[(i*i+i)/2+j]+=v;
  else {
    std::ostringstream ssV;
      ssV << "Not allowed to add value at this position. (" << i << ", " << j << ")";
      throw FlxException("FlxMtxLTri::add_value_2", ssV.str() );
  }
}

FlxMtxLTri& FlxMtxLTri::CholeskyDec(FlxMtxSym& Sm)
{
  mtx = Sm.get_mtx_tVec();
  tdouble* mtxp = &mtx[0];
  const tnlong N = nrows();
  tdouble sum, d;
  #if FLX_KAHAN_MTX_LTRI
    pdouble t;
  #else
    tdouble& t=sum;
  #endif
  register tnlong i, j, k;
  for (i = 0; i < N; ++i) {
    for (j = 0; j < i; ++j) {
      t = mtxp[(i*i+i)/2+j];
      for (k=0;k<j;k++) {
	t-=mtxp[(i*i+i)/2+k]*mtxp[(j*j+j)/2+k];
      }
      #if FLX_KAHAN_MTX_LTRI
	sum = t.cast2double();
      #endif
      mtxp[(i*i+i)/2+j] = sum/mtxp[(j*j+j)/2+j];
    }
    t = mtxp[(i*i+i)/2+i];
    for (k=0;k<i;++k) {
      d = mtxp[(i*i+i)/2+k];
      t-=pow2(d);
    }
    #if FLX_KAHAN_MTX_LTRI
      sum = t.cast2double();
    #endif
    if (sum<=0.0) {
      std::ostringstream ssV;
      ssV << "Matrix is not positiv definite. (sum = " << sum << ") [line: " << i+1 << "]";
      #if FLX_DEBUG
	std::ostringstream ssV2;
	ssV2 << "Correlation matrix:" << std::endl;
	this->output_Mtx(ssV2);
	throw FlxException("FlxMtxLTri::CholeskyDec_1dbg", ssV.str(), ssV2.str() );
      #else
	throw FlxException("FlxMtxLTri::CholeskyDec_1", ssV.str() );
      #endif
    } else {
      mtxp[(i*i+i)/2+i] = sqrt(sum);
    }
  }
  return *this;
}

FlxMtxLTri& FlxMtxLTri::CholeskyDec(FlxMtxSparsSym& Sm)
{
  const tnlong N = nrows();
  register tdouble sum, d, d1;
  #if FLX_KAHAN_MTX_LTRI
    pdouble t;
    pdouble sum2;
  #else
    tdouble& t=sum;
    tdouble sum2;
  #endif
  tdouble* mtxp = &mtx[0];
  register tnlong j, k, l, ii, jj;
  for (register tnlong i = 0; i < N; ++i) {
    l = Sm.ija[i];
    ii = (i*i+i)/2;
    sum2=0.0;
    for (j = 0; j < i; ++j) {
      jj = (j*j+j)/2;
      if (l<Sm.ija[i+1] && j==Sm.ija[l] ) {
	t = Sm.sa[l];
	++l;
      } else {
	t = 0.0;
      }
      for (k=0;k<j;++k) {
	d=mtxp[ii+k];
	d1=mtxp[jj+k];
	t-=d*d1;
      }
      #if FLX_KAHAN_MTX_LTRI
	sum = t.cast2double();
      #endif
      sum/=mtxp[jj+j];
      mtxp[ii+j] = sum;
      sum2+=pow2(sum);
    }
    // calculate diagonal entries
    #if FLX_KAHAN_MTX_LTRI
      sum = Sm.sa[i] - sum2.cast2double();
    #else
      sum = Sm.sa[i] - sum2;
    #endif
    if (sum<=0.0) {
      std::ostringstream ssV;
      ssV << "Matrix is not positiv definite. (sum = " << sum << ")";
      #if FLX_DEBUG
	std::ostringstream ssV2;
	ssV2 << "Correlation matrix:" << std::endl;
	this->output_Mtx(ssV2);
	throw FlxException("FlxMtxLTri::CholeskyDec_2dbg", ssV.str(), ssV2.str() );
      #else
	throw FlxException("FlxMtxLTri::CholeskyDec_2", ssV.str() );
      #endif
    } else {
      mtxp[ii+i] = sqrt(sum);
    }
  }
  return *this;
}


FlxMtxLTri& FlxMtxLTri::Invert()
{
  const tnlong n = this->ncols();
  tdouble* mtxp = &mtx[0];
  register tnlong i,j, m;
  register size_t k = 0;
  register tdouble a;
  #if FLX_KAHAN_MTX_LTRI
    pdouble t;
  #else 
    tdouble& t=a;
  #endif
  for (i = 0; i < n; ++i) {		// rows
    for (j = 0; j < i; ++j) {		// columns
      t = 0.0;
      for (m = j; m < i; ++m) {		// summation
	t+=mtxp[(i*i+i)/2+m]*mtxp[(m*m+m)/2+j];
      }
      #if FLX_KAHAN_MTX_LTRI
	a = t.cast2double();
      #endif
      mtxp[k]=-a/mtxp[(i*i+i)/2+i];
      ++k;
    }
    mtxp[k]=1/(mtxp[k]);
    ++k;
  }
  
  return *this;  
}

void FlxMtxLTri::TransMultVec(tVec& v)
{
  #if FLX_KAHAN_MTX_LTRI
    pdouble a;
  #else
    tdouble a;
  #endif
  const size_t vs = v.size();
  const tdouble* mtxp = &mtx[0];
  tdouble* vp = &v[0];
  for (size_t i = 0; i < vs; i++) {
    a = 0.0;
    for (size_t j = i; j < vs; j++) {
      a+=mtxp[(j*j+j)/2+i]*vp[j];
    }
    #if FLX_KAHAN_MTX_LTRI
      vp[i]=a.cast2double();
    #else
      vp[i]=a;
    #endif
  }
}

FlxMtxSpars::FlxMtxSpars(const tnlong nmax)
:sa(new tdouble[nmax]),ija(new tnlong[nmax])
{

}

FlxMtxSpars::FlxMtxSpars(const FlxMtxSpars& mtx): FlxMtx_base(mtx)
{
  const tnlong nmax = mtx.ija[mtx.ija[0]-1];
  sa = new tdouble[nmax];
  ija = new tnlong[nmax];
  const tdouble* mtxsap = &mtx.sa[0];
  const tnlong* mtxijap = &mtx.ija[0];
  for (tnlong i=0;i<nmax;++i) {
    sa[i] = mtxsap[i];
    ija[i] = mtxijap[i];
  }
}

FlxMtxSpars::FlxMtxSpars(const FlxMtx& mtx): FlxMtx_base(mtx)
{
  const tnlong n = mtx.nrows();
  const tnlong m = mtx.ncols();
  tnlong nmax = n+1;
  const tdouble* mtxmtxp = &mtx.mtx[0];
  const tdouble GVT = GlobalVar.TOL();
  register tnlong j,i;
  for (i = 0; i<n;++i) {
    for (j = 0; j<m;++j) {
      if (fabs(mtxmtxp[i*m+j])>GVT ) nmax++;
    }
  }  
  sa = new tdouble[nmax];
  ija = new tnlong[nmax];
  ija[0] = n+1;
  tnlong k=n;
  tdouble dt;
  for (i=0;i<n;++i) {
    for (j=0;j<m;++j) {
      dt = mtxmtxp[i*m+j];
      if (fabs(dt)>GVT ) {
	++k;
	sa[k]=dt;
	ija[k]=j;
      }
    }
    ija[i+1]=k+1;
  }
}

FlxMtxSpars::~FlxMtxSpars()
{
  delete[] sa;
  delete[] ija;
}

FlxMtxSpars* FlxMtxSpars::copy()
{
  return new FlxMtxSpars(*this);
}

FlxMtxSpars& FlxMtxSpars::operator*=(const tdouble& s)
{
  const tnlong N=ija[nrows()];
  for (register tnlong i=0;i<N;i++) sa[i]*=s;
  return *this;
}

const tdouble FlxMtxSpars::operator()(const tnlong& i, const tnlong& j) const
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= nrows()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxSpars::operator()_2", ssV.str() );
    }
  #endif
  for (register tnlong k = ija[i];k<ija[i+1];++k) {
    if (ija[k]==j) {
      return sa[k];
    }
    else if (ija[k]>j) {
      return 0.0;
    }
  }
  return 0.0;
}

void FlxMtxSpars::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  if (v==0.0) return;
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= nrows()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxSpars::add_value_1", ssV.str() );
    }
    if (j > i) {
      std::ostringstream ssV;
      ssV << "Not allowed to add value '" << v << "' at this position (" << i << ", " << j << ") - sparse matrix.";
      throw FlxException("FlxMtxSpars::add_value_2", ssV.str() );
    }
  #endif
  
  for (register tnlong k = ija[i];k<ija[i+1];++k) {
    if (ija[k]==j) {
      sa[k]+=v;
      return;
    }
    else if (ija[k]>j) {
      break;
    }
  }
  std::ostringstream ssV;
  ssV << "Not allowed to add value '" << v << "' at this position (" << i << ", " << j << ") - sparse matrix.";
  throw FlxException("FlxMtxSpars::add_value_3", ssV.str() );
}

void FlxMtxSpars::MultMv(const tVec& v, tVec& w) const
{
  const tnlong N = nrows();
  #if FLX_KAHAN_MTX_SLTRI
    pdouble t;
  #else
    tdouble t;
  #endif
  const tdouble* vp = &v[0];
  tdouble* wp = &w[0];
  register tnlong i, k;
  for (i=0;i<N;++i) {
    t=0.0;
    for (k=ija[i];k<ija[i+1];++k) {
      t+= sa[k]*vp[ija[k]];
    }
    #if FLX_KAHAN_MTX_SLTRI
      wp[i]=t.cast2double();
    #else
      wp[i]=t;
    #endif
  }
}

const tdouble FlxMtxSpars::MultRowV(const tVec& v, const tnlong rownumber) const
{
  #if FLX_KAHAN_MTX_SLTRI
    pdouble t = 0.0;
  #else
    tdouble t = 0.0;
  #endif
  const tdouble* vp = &v[0];
  for (register tnlong k=ija[rownumber];k<ija[rownumber+1];++k) {
    t+= sa[k]*vp[ija[k]];
  }
  #if FLX_KAHAN_MTX_SLTRI
    return t.cast2double();
  #else
    return t;
  #endif
}

void FlxMtxSpars::TransMultVec(const tVec& v, tVec& w)
{
  const tnlong N = nrows();
  w = 0.0;
  tdouble* wp = &w[0];
  const tdouble* vp = &v[0];
  register tnlong i, k;
  for (i=0;i<N;++i) {
    for (k=ija[i];k<ija[i+1];++k) {
      wp[ija[k]]+= sa[k]*vp[i];
    }
  }
}

FlxMtxSparsLTri::FlxMtxSparsLTri(FlxMtxDiag& mtx)
{
  const tuint N = mtx.nrows();
  sa = new tdouble[N+1];
  ija = new tnlong[N+1];
  
  register tnlong i;
  for (i=0;i<=N;++i) {
    ija[i] = N+1;
  }
  for (i=0;i<N;++i) {
    sa[i] = mtx.operator()(i,i);
  }
}

FlxMtxSparsLTri::FlxMtxSparsLTri(const FlxMtxLTri& mtx)
{
  const tnlong n = mtx.nrows();
  tnlong nmax = n+1;
  const tdouble* mtxmtxp = &mtx.mtx[0];
  const tdouble GVT = GlobalVar.TOL();
  register tnlong j,i;
  for (i = 1; i<n;i++) {
    for (j = 0; j<i;j++) {
      if (fabs(mtxmtxp[(i*i+i)/2+j])>GVT ) nmax++;
    }
  }  
  sa = new tdouble[nmax];
  ija = new tnlong[nmax];
  for (i = 0;i<n;i++) sa[i] = mtx(i,i);
  ija[0] = n+1;
  tnlong k=n;
  tdouble dt;
  for (i=0;i<n;i++) {
    for (j=0;j<i;j++) {
      dt = mtx(i,j);
      if (fabs(dt)>GVT ) {
	k++;
	sa[k]=dt;
	ija[k]=j;
      }
    }
    ija[i+1]=k+1;
  }
}

FlxMtxSparsLTri::FlxMtxSparsLTri(const FlxMtxSparsLTri& mtx)
:FlxMtxSpars(mtx)
{

}

void FlxMtxSparsLTri::CholeskyDec(tdoublePtr& saP, tnlong*& ijaP, FlxMtxSparsSym& mtx)
{
  register tnlong N = mtx.nrows();
  // reserve memory in advance (on the safe side)
    // TODO: maybe a smarter solution is possible???
    tVec saT(N*N+1);
    tdouble* saTp = &saT[0];
    std::valarray<tnlong>ijaT(saT.size());
    tnlong* ijaTp = &ijaT[0];
  const tdouble* mtxsap = &mtx.sa[0];
  const tnlong* mtxijap = &mtx.ija[0];
  register tnlong i,j,k,k2,l,pos;
  register tdouble sum;
  const tnlong& lmax = mtx.ija[N];
  
  // copy initial setting
    for (i=0;i<N;++i) {
      saTp[i] = mtxsap[i];
    }
    pos=N+1;
    ijaTp[0] = pos;

  #if FLX_KAHAN_MTX_SLTRI
    pdouble t;
    pdouble sum2;
  #else
    tdouble& t = sum;
    tdouble sum2;
  #endif
  for (tnlong i = 0; i < N; ++i) {
    sum2=0.0;
    l = mtxijap[i];
    for (j = 0; j < i; ++j) {
      if (l<lmax && j==mtxijap[l] && l<mtxijap[i+1]) {
	t = mtxsap[l];
	++l;
      } else {
	t = 0.0;
      }
      k2 = ijaTp[j];
      for (k=ijaTp[i];k<pos;++k) {
	while (ijaTp[k2]<ijaTp[k] && k2 < ijaTp[j+1]) ++k2;
	if (k2>=ijaTp[j+1]) break;
	if (ijaTp[k]==ijaTp[k2]) {
	  t-=saTp[k]*saTp[k2];
	}
      }
      #if FLX_KAHAN_MTX_SLTRI
	sum = t.cast2double();
      #endif
      if (sum!=0.0) {
	sum/=saTp[j];
	saTp[pos] = sum;
	sum2+=pow2(sum);
	ijaTp[pos]=j;
	++pos;
      }
    }
    // calculate diagonal entry
      #if FLX_KAHAN_MTX_SLTRI
	saTp[i] = saTp[i]-sum2.cast2double();
      #else
	saTp[i] = saTp[i]-sum2;
      #endif
      if (saTp[i]<=0.0) {
	std::ostringstream ssV;
	ssV << "Matrix is not positiv definite. (sum = " << saT[i] << " in row " << i+1 << ")";
	throw FlxException("FlxMtxSparsLTri::CholeskyDec", ssV.str() );
      }
      saTp[i] = sqrt(saTp[i]);
    ijaTp[i+1] = pos;
  }
  
  saP = new tdouble[pos];
  ijaP = new tnlong[pos];
  for (i=0;i<pos;++i) {
    saP[i] = saTp[i];
    ijaP[i] = ijaTp[i];
  }
  #if FLX_DEBUG
    if (N>1) {
      GlobalVar.slog(4) << "  Sparseness: ";
      GlobalVar.slog(4) << GlobalVar.Double2String((pos-N-1)*2.0/(N*N-N)) << " (original: ";
      GlobalVar.slog(4) << GlobalVar.Double2String((mtx.ija[mtx.ija[0]-1]-mtx.ija[0]-1)*2.0/(mtx.ija[0]*mtx.ija[0]-mtx.ija[0])) << ")";
      GlobalVar.slog(4) << std::endl;
    }
  #endif
}

FlxMtxSparsLTri::FlxMtxSparsLTri(FlxMtxSparsSym& Sm)
{
  CholeskyDec(sa,ija,Sm);
}

FlxMtxSparsLTri* FlxMtxSparsLTri::copy()
{
  return new FlxMtxSparsLTri(*this);
}

void FlxMtxSparsLTri::MultInv(const tVec& v, tVec& w)
{
  const tnlong n = nrows();
  tdouble sum;
  #if FLX_KAHAN_MTX_SLTRI
    pdouble t;
  #else
    tdouble& t=sum;
  #endif
  const tdouble* vp = &v[0];
  tdouble* wp = &w[0];
  register tnlong i, j;
  for (i=0;i<n;++i) {
    t=0.0;
    for (j=ija[i];j<ija[i+1];++j) {
      t+=sa[j]*wp[ija[j]];
    }
    #if FLX_KAHAN_MTX_SLTRI
      sum = t.cast2double();
    #endif
    wp[i]=(vp[i]-sum)/sa[i];
  }
}

void FlxMtxSparsLTri::MultMv(const tVec& v, tVec& w) const
{
  const tnlong N = nrows();
  register tnlong i, k;
  #if FLX_KAHAN_MTX_SLTRI
    pdouble t;
  #else
    tdouble t;
  #endif
  const tdouble* vp = &v[0];
  tdouble* wp = &w[0];
  for (i=N;i>0;--i) {
    t=sa[i-1]*vp[i-1];
    for (k=ija[i];k>ija[i-1];--k) {
      t+= sa[k-1]*vp[ija[k-1]];
    }
    #if FLX_KAHAN_MTX_SLTRI
      wp[i-1]=t.cast2double();
    #else
      wp[i-1]=t;
    #endif
  }
}

const tdouble FlxMtxSparsLTri::MultRowV(const tVec& v, const tnlong rownumber) const
{
  throw FlxException_NotImplemented("FlxMtxSparsLTri::MultRowV_1");
}

void FlxMtxSparsLTri::TransMultVec(tVec& v)
{
  register tnlong N = nrows();
  register tnlong i, k;
  #if FLX_KAHAN_MTX_SLTRI
    pVec w(N);
    pVec_assemble(w,v);
    pdouble* wp = &w[0];
  #else
    tVec& w = v;
    tdouble* wp = &w[0];
  #endif
  for (i=0;i<N;++i) {
    for (k=ija[i];k<ija[i+1];++k) {
      wp[ija[k]] += wp[i]*sa[k];
    }
    wp[i]=wp[i]*sa[i];
  }
  #if FLX_KAHAN_MTX_SLTRI
    tVec_assemble(v,w);
  #endif
}

void FlxMtxSparsLTri::TransMultVec(const tVec& v, tVec& w)
{
  throw FlxException_NotImplemented("FlxMtxSparsLTri::TransMultVec");
}

FlxMtxSparsLTri* FlxMtxSparsLTri::MultM_Mdiag(const FlxMtxDiag& D) const
{
  #if FLX_DEBUG
    if (nrows()!=D.nrows()) {
      throw FlxException("FlxMtxSparsLTri::MultM_Mdiag", "ERROR" );
    }
  #endif
  const tnlong nmax = ija[ija[0]-1];
  FlxMtxSparsLTri* res = new FlxMtxSparsLTri(nmax);
  
  for (register tnlong i=0;i<nmax;++i) {
    res->ija[i] = ija[i];
  }
  const tnlong N = nrows();
  register tnlong i, k;
  register tdouble t;
  for (i=0;i<N;++i) {
    t = D.operator()(i,i);
    res->sa[i] = sa[i]*t;
    for (k=ija[i];k<ija[i+1];++k) {
      res->sa[k] = sa[k]*t;
    }
  }
  
  return res;
}

const tdouble FlxMtxSparsLTri::operator()(const tnlong& i, const tnlong& j) const
{
  if (i == j) return sa[i];
  if (j > i) return 0.0;
  return FlxMtxSpars::operator()(i,j);
}

void FlxMtxSparsLTri::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  if (v==0.0) return;
  if (i == j) {
    sa[i]+=v;
    return;
  }
  FlxMtxSpars::add_value(i,j,v);
}

FlxMtxSparsSymLU::FlxMtxSparsSymLU(FlxMtxSparsSym& mtx)
{
  sa = NULL; 
  ija = NULL;
  FlxMtxSparsLTri::CholeskyDec(sa,ija,mtx);
}

FlxMtxSparsSymILU::FlxMtxSparsSymILU(const FlxMtxSparsSym& mtx, bool mod0diagentry)
{
  const tnlong nmax = mtx.ija[mtx.nrows()];
  sa = new tdouble[nmax];
  ija = new tnlong[nmax];
  const tdouble* mtxsap = &mtx.sa[0];
  const tnlong* mtxijap = &mtx.ija[0];
  for (tnlong i = 0; i < nmax; ++i) {
    sa[i] = mtxsap[i];
    ija[i] = mtxijap[i];
  }
  
  const tnlong N = nrows();
  tdouble sum;
  register tnlong i, j, k;
  
  #if FLX_KAHAN_MTX_SSILU
    pdouble t;
  #else
    tdouble& t = sum;
  #endif
  
  for (i = 0; i < N; ++i) {
    for (j=ija[i];j<ija[i+1];++j) {
      t = sa[j];
      for (k=ija[i];k<j;++k) {
	t-=sa[k]*operator()(ija[j],ija[k]);  // TODO more efficient version possible ???
      }
      #if FLX_KAHAN_MTX_SSILU
	sum = t.cast2double();
      #endif
      sa[j] = sum/(sa[ija[j]]);
    }
    t = sa[i];
    for (k=ija[i];k<ija[i+1];++k) {
      t-=pow2(sa[k]);
    }
    #if FLX_KAHAN_MTX_SSILU
      sum = t.cast2double();
    #endif
    if (sum<=0.0) {
      if (!mod0diagentry) {
	std::ostringstream ssV;
	ssV << "Matrix is not positiv definite. (sum=" << sum << " in i=" << i << ")"; // <<  std::endl << *this;
	throw FlxException("FlxMtxSparsSymILU::FlxMtxSparsSymILU_1", ssV.str() );
      } else {
	sa[i] = sqrt(sa[i]);
	GlobalVar.alert.alert("FlxMtxSparsSymILU::FlxMtxSparsSymILU_2","ILU: modified diagonal entry.");
      }
    } else {
      sa[i] = sqrt(sum);
    }
  }
}

void FlxMtxSparsSymILU::MultMv(const tVec& v, tVec& w) const
{
  // calculate L*y=f
  const tnlong N = nrows();
  
  #if FLX_KAHAN_MTX_SSILU
    pVec W(N);
    pdouble d;
    pdouble* Wp = &W[0];
  #else 
    tVec& W = w;
    tdouble* Wp = &W[0];
    tdouble d;
  #endif
  const tdouble* vp = &v[0];
  const tdouble* wp = &w[0];
  
  register tnlong i,j;
  for (i = 0; i < N; ++i) {
    d=vp[i];
    for (j=ija[i];j<ija[i+1];++j) {
      d-=sa[j]*wp[ija[j]];
    }
    Wp[i] = d/sa[i];
  }
  //calculate L^T*x=y
  for (i = N; i > 0; --i) {
    Wp[i-1] = Wp[i-1]/sa[i-1];
    for (j=ija[i-1];j<ija[i];++j) {
      Wp[ija[j]]-=Wp[i-1]*sa[j];
    }
  }
  #if FLX_KAHAN_MTX_SSILU
    tVec_assemble(w,W);
  #endif
}

void FlxMtxSparsSymILU::MultMv(const pVec& v, pVec& w) const
{
  // calculate L*y=f
  const tnlong N = nrows();
  tnlong i,j;
  pdouble d;
  const pdouble* vp = &v[0];
  pdouble* wp = &w[0];
  for (i = 0; i < N; ++i) {
    d=vp[i];
    for (j=ija[i];j<ija[i+1];++j) {
      d-=wp[ija[j]]*sa[j];
    }
    wp[i] = d/sa[i];
  }
  //calculate L^T*x=y
  for (i = N; i > 0; --i) {
    wp[i-1] = wp[i-1]/sa[i-1];
    for (j=ija[i-1];j<ija[i];++j) {
      wp[ija[j]]-=wp[i-1]*sa[j];
    }
  }
}

FlxMtxSparsSymILU::~FlxMtxSparsSymILU()
{
  delete [] sa;
  delete [] ija;
}

const tdouble FlxMtxSparsSymILU::operator()(const tnlong& i, const tnlong& j) const
{
  #if FLX_DEBUG
    if (i < 0 || j < 0 || i >= nrows() || j >= nrows()) {
      std::ostringstream ssV;
      ssV << "Index not within matrix.";
      throw FlxException("FlxMtxSparsSym::operator()_2", ssV.str() );
    }
  #endif
  if (i == j) return sa[i];
  if (j > i) return 0.0;
  for (register tnlong k = ija[i];k<ija[i+1];k++) {
    if (ija[k]==j) {
      return sa[k];
    }
    else if (ija[k]>j) {
      return 0.0;
    }
  }
  return 0.0;
}

FlxMtx_base& FlxMtxSparsSymILU::operator*=(const tdouble& s)
{
  throw FlxException_NotImplemented("FlxMtxSparsSymILU::operator*=");
}

void FlxMtxSparsSymILU::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  throw FlxException_NotImplemented("FlxMtxSparsSymILU::add_value");
}

const tdouble FlxMtxSparsSFEMSym::operator()(const tnlong& i, const tnlong& j) const
{
  tnlong ii = i - (i%Kdim);
  tnlong in = ii/Kdim;
  ii = i%Kdim;
  tnlong jj = j - (j%Kdim);
  tnlong jn = jj/Kdim;
  jj = j%Kdim;
  
  if (in == jn) {
    return sa[in]*(sb[in]->operator()(ii,jj));
  }
  if (jn > in) {
    std::swap(in, jn);
    std::swap(ii, jj);
  }
  for (tnlong k = ija[in];k<ija[in+1];k++) {
    if (ija[k]==jn) {
      if (sa[k]==0.0) {
	return 0.0;
      } else {
	return sa[k]*(sb[k]->operator()(ii,jj));
      }
    }
    else if (ija[k]>jn) {
      return 0.0;
    }
  }
  return 0.0;
}

const bool FlxMtxSparsSFEMSym::isPositiveDefinite(tuint& r, const bool fixIt)
{
  return sb[0]->isPositiveDefinite(r,fixIt);
}


void FlxMtxSparsSFEMSym::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  throw FlxException_NotImplemented("FlxMtxSparsSFEMSym::add_value");
}

const bool FlxMtxSparsSFEMSym::solve_CG(tVec& x, tVec& f, tdouble& eps, tnlong& iter, tuint pcn, bool startZero)
{
  if (!startZero) {
    tVec xx(Kdim);
    tVec ff = f[std::slice(0,Kdim,1)];
    tnlong iterV = iter;
    tdouble epsV = eps;
    sb[0]->solve_CG(xx,ff,epsV,iterV,1,true);
    x[std::slice(0,Kdim,1)] = xx;
  }
  return FlxMtx_baseS::solve_CG(x, f, eps, iter, pcn, startZero);
}

// FlxMtxDiag* FlxMtxSparsSFEMSym::get_diag() const
// {
//   FlxMtxDiag* d = new FlxMtxDiag(nrows());
//   FlxMtxDiag* di = sb[0]->get_diag();
//   
//   tnlong N = (ija[0]-1);
//   for (tnlong i = 0; i < N; ++i) {
//     for (tnlong j = 0; j < Kdim; ++j) {
//       d[i*Kdim+j] = di[j];
//     }
//   }
//   delete di;
//   return d;
// }

void FlxMtxSparsSFEMSym::output_Mtx(std::ostream& sout) const
{
  tnlong N = ija[0]-1;
  
  for (tnlong i =0;i<N;++i) {
    tnlong k=0;
    for (tnlong j = ija[i]; j<ija[i+1];++j) {
      while (k<ija[j]) {
	sout << "0" << "\t";
	k++;
      }
      sout << GlobalVar.Double2String(sa[j]) << "*K"<< (box.find(sb[j]))->second << "\t";
      k++;
    }
    while (k<i) {
      sout << "0" << "\t";
      k++;
    }
    sout << GlobalVar.Double2String(sa[i])  << "*K"<< (box.find(sb[i]))->second;
    sout << std::endl;
  }
  
  for (std::map<FlxMtx_baseS*, tuint>::const_iterator it = box.begin(); it != box.end(); ++it) {
    sout << "K" << it->second << "=";
    it->first->output_Mtx(sout);
  }
}

void FlxMtxSparsSFEMSym::assembleMinv(int i)
{
  if (i == 3) {
    tnlong N = (ija[0]-1);
    tVec c0ii(N);
    for (tnlong j = 0; j < N; ++j) {
      c0ii[j] = sa[j];
    }
    FlxMtxSparsSym* S = dynamic_cast<FlxMtxSparsSym*>(sb[0]);
    if (S == NULL) {
      std::ostringstream ssV;
      ssV << "Matrix is not a symmetric sparse matrix.";
      throw FlxException("FlxMtxSparsSFEMSym::assembleMinv_1", ssV.str() );
    }
    Minv = new FlxMtxPrecnInvSFEMSym(*S,c0ii);
  } else if (i == 4 || i==5 || i==6) {
    tnlong N = (ija[0]-1);
    tVec c0ii(N);
    for (tnlong j = 0; j < N; ++j) {
      c0ii[j] = sa[j];
    }
    FlxMtxSparsSym* S = dynamic_cast<FlxMtxSparsSym*>(sb[0]);
    if (S == NULL) {
      std::ostringstream ssV;
      ssV << "Matrix is not a symmetric sparse matrix.";
      throw FlxException("FlxMtxSparsSFEMSym::assembleMinv_1", ssV.str() );
    }
    if (i == 4 ) {
      Minv = new FlxMtxPrecnILUSFEMSym(*S,c0ii,true);
    } else if (i == 5) {
      Minv = new FlxMtxPrecnILUSFEMSym(*S,c0ii);
    } else {
      Minv = new FlxMtxPrecnILUSFEMSym(*S,c0ii,false,true);
    }
  } else {
    FlxMtx_baseS::assembleMinv(i);
  }
}

FlxMtxSparsSFEMSym::~FlxMtxSparsSFEMSym()
{
  delete[] sa;
  delete[] sb;
  delete[] ija;
}

FlxMtxSparsSFEMSym::FlxMtxSparsSFEMSym(std::valarray< tdouble >& fM, std::valarray< int >& KM, FlxMtx_baseS** StfMtxV, tulong P)
:Kdim(StfMtxV[0]->ncols())
{
  // assign box
    for ( register int i = 0; i <= KM.max(); ++i) {
      std::pair<FlxMtx_baseS*, tuint> Element(StfMtxV[i],i);
      box.insert(Element);
    }
  
  tulong nmax = P+1;
  for (tulong i = 1; i<P;++i) {
    for (tulong j = 0; j<i;++j) {
      if (fM[(i*i+i)/2+j]!=0.0) nmax++;
    }
  }  
  sa = new tdouble[nmax];
  sb = new FlxMtx_baseS*[nmax];
  ija = new tnlong[nmax];
  for (tnlong i = 0;i<P;i++) {
    sa[i] = fM[(i*i+i)/2+i];
    sb[i] = StfMtxV[KM[(i*i+i)/2+i]];
  }
  ija[0] = P+1;
  tnlong k=P;
  for (tnlong i=0;i<P;i++) {
    for (tnlong j=0;j<i;j++) {
      if (fM[(i*i+i)/2+j]!=0.0) {
	k++;
	sa[k]=fM[(i*i+i)/2+j];
	sb[k] = StfMtxV[KM[(i*i+i)/2+j]];
	ija[k]=j;
      }
    }
    ija[i+1]=k+1;
  }
}

FlxMtx_base& FlxMtxSparsSFEMSym::operator*=(const tdouble& s)
{
  const tnlong N=ija[ija[0]-1];
  for (register tnlong i=0;i<N;i++) sa[i]*=s;
  return *this;
}

void FlxMtxSparsSFEMSym::MultMv(const tVec& v, tVec& w) const
{
  tnlong N = ija[0]-1;
  #if FLX_KAHAN_MTX_SSFEM
    pVec W(N);
    tVec t(N);
  #else
    tVec& W = w;
  #endif
  for (tnlong i=0;i<N;++i) {
    #if FLX_KAHAN_MTX_SSFEM
      t = ((*sb[i])*v[std::slice(i*Kdim,Kdim,1)])*sa[i];
      for (tnlong j=0;j<Kdim;++j) {
	W[i*Kdim+j] = t[j];
      }
    #else
      W[std::slice(i*Kdim,Kdim,1)] = ((*sb[i])*v[std::slice(i*Kdim,Kdim,1)])*sa[i];
    #endif
  }
  for (tnlong i=0;i<N;++i) {
    for (tnlong k=ija[i];k<ija[i+1];++k) {
      #if FLX_KAHAN_MTX_SSFEM
	t = ((*sb[k])*v[std::slice(ija[k]*Kdim,Kdim,1)])*sa[k];
	for (tnlong j=0;j<Kdim;++j) {
	  W[i*Kdim+j] += t[j];
	}
	t = ((*sb[k])*v[std::slice(i*Kdim,Kdim,1)])*sa[k];
	for (tnlong j=0;j<Kdim;++j) {
	  W[ija[k]*Kdim+j] += t[j];
	}
      #else
	W[std::slice(i*Kdim,Kdim,1)] += ((*sb[k])*v[std::slice(ija[k]*Kdim,Kdim,1)])*sa[k];
	W[std::slice(ija[k]*Kdim,Kdim,1)]+=((*sb[k])*v[std::slice(i*Kdim,Kdim,1)])*sa[k];
      #endif
    }
  }
  #if FLX_KAHAN_MTX_SSFEM
    tVec_assemble(w,W);
  #endif
}

void FlxMtxSparsSFEMSym::MultMv(const pVec& v, pVec& w) const
{
  const tnlong N = ija[0]-1;
  for (register tnlong i=0;i<N;++i) {
    w[std::slice(i*Kdim,Kdim,1)] = ((*sb[i]).operator*(v[std::slice(i*Kdim,Kdim,1)]))*sa[i];
  }
  for (register tnlong i=0;i<N;++i) {
    for (register tnlong k=ija[i];k<ija[i+1];++k) {
      w[std::slice(i*Kdim,Kdim,1)] += ((*sb[k])*v[std::slice(ija[k]*Kdim,Kdim,1)])*sa[k];
      w[std::slice(ija[k]*Kdim,Kdim,1)]+=((*sb[k])*v[std::slice(i*Kdim,Kdim,1)])*sa[k];
    }
  }
}

void FlxMtxPrecnInvSFEMSym::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  throw FlxException_NotImplemented("FlxMtxPrecnInvSFEMSym::add_value");
}

const tdouble FlxMtxPrecnInvSFEMSym::operator()(const tnlong& i, const tnlong& j) const
{
  throw FlxException_NotImplemented("FlxMtxPrecnInvSFEMSym::operator()");
}

FlxMtx_base& FlxMtxPrecnInvSFEMSym::operator*=(const tdouble& s)
{
  throw FlxException_NotImplemented("FlxMtxPrecnInvSFEMSym::operator*=");
}

FlxMtxPrecnInvSFEMSym::FlxMtxPrecnInvSFEMSym(FlxMtxSparsSym& K, const tVec& c0ii): c0iiInv(c0ii), Kdim(K.nrows())
{
  const tnlong N=c0ii.size();
  tdouble* c0iiInvp = &c0iiInv[0];
  for (register tnlong i = 0; i < N; ++i) {
    c0iiInvp[i] = 1/c0iiInvp[i];
  }
  
  FlxMtxLTri L(Kdim);
  
//   FlxMtxSym S(K);
//   L.CholeskyDec(S);
  
  L.CholeskyDec(K);
  L.Invert();
  Kinv = new FlxMtxSym(L);
}

FlxMtxPrecnInvSFEMSym::~FlxMtxPrecnInvSFEMSym()
{
  delete Kinv;
}

void FlxMtxPrecnInvSFEMSym::MultMv(const tVec& v, tVec& w) const
{
  const tnlong N = c0iiInv.size();
  for (register tnlong i=0;i<N;i++) {
    w[std::slice(i*Kdim,Kdim,1)] = ((*Kinv)*v[std::slice(i*Kdim,Kdim,1)])*c0iiInv[i];
  }
}

void FlxMtxPrecnILUSFEMSym::add_value(const tnlong& i, const tnlong& j, const tdouble& v)
{
  throw FlxException_NotImplemented("FlxMtxPrecnILUSFEMSym::add_value");
}

FlxMtx_base& FlxMtxPrecnILUSFEMSym::operator*=(const tdouble& s)
{
  throw FlxException_NotImplemented("FlxMtxPrecnILUSFEMSym::operator*=");
}

const tdouble FlxMtxPrecnILUSFEMSym::operator()(const tnlong& i, const tnlong& j) const
{
  throw FlxException_NotImplemented("FlxMtxPrecnILUSFEMSym::operator()");
}

FlxMtxPrecnILUSFEMSym::~FlxMtxPrecnILUSFEMSym()
{
  delete ILU;
}

void FlxMtxPrecnILUSFEMSym::MultMv(const tVec& v, tVec& w) const
{
  const tnlong N = c0iiInv.size();
  for (register tnlong i=0;i<N;i++) {
    w[std::slice(i*Kdim,Kdim,1)] = ((*ILU)*v[std::slice(i*Kdim,Kdim,1)])*c0iiInv[i];
  }
}

void FlxMtxPrecnILUSFEMSym::MultMv(const pVec& v, pVec& w) const
{
  const tnlong N = c0iiInv.size();
  for (register tnlong i=0;i<N;i++) {
    w[std::slice(i*Kdim,Kdim,1)] = ((*ILU)*v[std::slice(i*Kdim,Kdim,1)])*c0iiInv[i];
  }
}

FlxMtxPrecnILUSFEMSym::FlxMtxPrecnILUSFEMSym(FlxMtxSparsSym& K, const tVec& c0ii, bool FullDecomp, bool mod0diagentry)
: c0iiInv(c0ii), Kdim(K.nrows())
{	
  const tnlong N=c0ii.size();
  tdouble* c0iiInvp = &c0iiInv[0];
  for (register tnlong i = 0; i < N; ++i) {
    c0iiInvp[i] = 1/c0iiInvp[i];
  }
  
  if (FullDecomp) {
    ILU = new FlxMtxSparsSymLU(K);
  } else {
    ILU = new FlxMtxSparsSymILU(K,mod0diagentry);
  }
}

