/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxMtx_Eigen.h"



void MtxEigenValue(FlxMtxSym& Amtx, const tuint M, tVec& EigenValues, std::vector< tVec >& Eigenvectors, const int Mode)
{
  const tuint N = Amtx.nrows();
  if (N < M) {
    std::ostringstream ssV;
    ssV << "Cannot compute more Eigenvalues (" << M << ") than number of DOFs in the system(" <<  N << ").";
    throw FlxException("MtxEigenValue_G_0.0", ssV.str());
  }
  
  #if FLX_DEBUG
    if (EigenValues.size() != M) {
      std::ostringstream ssV;
      ssV << "Problem with size of vector. (" << EigenValues.size() << "; " << M << ")";
      throw FlxException("MtxEigenValue_G_0.1", ssV.str() );
    }
    for (tuint i=0;i<M;i++) {
      if (Eigenvectors[i].size()!=N) {
	std::ostringstream ssV;
	ssV << "Problem with size of vector.";
	throw FlxException("MtxEigenValue_G_0.2", ssV.str() );
      }
    }
  #endif
  
  if (Mode == 1) {
    MtxEigenValue_ARP(Amtx,M,EigenValues,Eigenvectors);
  } else if (Mode == 2) {
    MtxEigenValue_GSL(Amtx,M,EigenValues,Eigenvectors);
  } else if (Mode == 3) {
    MtxEigenValue_GSLstab(Amtx,M,EigenValues,Eigenvectors);
  } else {
    throw FlxException_Crude("MtxEigenValue_G_0.3");
  }
  
  // define orientation of eigenvectors
    for (tuint i=0;i<M;++i) {
      tdouble* t = &(Eigenvectors[i][0]);
      tdouble m = fabs(t[0]);
      bool p = (t[0]>=0.0);
      for (tuint j=1;j<N;++j) {
	if (fabs(t[j])>m) {
	  m = fabs(t[j]);
	  p = (t[j]>=0.0);
	}
      }
      if (!p) {
	for (tuint j=0;j<N;++j) t[j] *= -1.0;
      }
    }
}

void MtxEigenValue(FlxMtxSym& Amtx, FlxMtxSym& Bmtx, const tuint M, tVec& EigenValues, std::vector<tVec>& Eigenvectors, const int Mode)
{
  const tuint N = Amtx.nrows();
  if (N < M) {
    std::ostringstream ssV;
    ssV << "Cannot compute more Eigenvalues (" << M << ") than number of DOFs in the system (" <<  N << ").";
    throw FlxException("MtxEigenValue_G_1.0", ssV.str());
  }
  
  #if FLX_DEBUG
    if (EigenValues.size() != M) {
      std::ostringstream ssV;
      ssV << "Problem with size of vector. (" << EigenValues.size() << "; " << M << ")";
      throw FlxException("MtxEigenValue_G_1.1", ssV.str() );
    }
    for (tuint i=0;i<M;i++) {
      if (Eigenvectors[i].size()!=Amtx.ncols()) {
	std::ostringstream ssV;
	ssV << "Problem with size of vector.";
	throw FlxException("MtxEigenValue_G_1.2", ssV.str() );
      }
    }
    if (Amtx.ncols()!=Bmtx.ncols()) {
      std::ostringstream ssV;
      ssV << "Matrix sizes do not match";
      throw FlxException("MtxEigenValue_G_1.3", ssV.str() );
    }
  #endif
  
  if (Mode == 1) {
    MtxEigenValue_ARP(Amtx,Bmtx,M,EigenValues,Eigenvectors);
  } else if (Mode == 2) {
    MtxEigenValue_GSL(Amtx,Bmtx,M,EigenValues,Eigenvectors);
  } else if (Mode == 3) {
    MtxEigenValue_GSLstab(Amtx,Bmtx,M,EigenValues,Eigenvectors);
  } else {
    throw FlxException_Crude("MtxEigenValue_G_1.4");
  }
  
  // define orientation of eigenvectors
    for (tuint i=0;i<M;++i) {
      tdouble* t = &(Eigenvectors[i][0]);
      const tdouble a = Norm2(t,N)/sqrt(tdouble(N))/2.;
      bool p = false;
      for (tuint j=0;j<N;++j) {
	if (fabs(t[j])>=a) {
	  p = (t[j]>=0.0);
	  break;
	}
      }
      if (!p) {
	for (tuint j=0;j<N;++j) t[j] *= -1.0;
      }
    }
}


