/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_FlxMtx_Eigen_H
#define fesslix_FlxMtx_Eigen_H

#include "flxMtx.h"


// -------------------------- General Interface ----------------------------------------------------------------------

// Mode:
//   1 -> ARP
//   2 -> GSL


/**
* @brief Calculates Eigenvalues of Amtx
*/
void MtxEigenValue(FlxMtxSym& Amtx, const tuint M, tVec& EigenValues, std::vector<tVec>& Eigenvectors, const int Mode);


/**
* @brief Calculates Eigenvalues of a generalized matrix eigenvalue problem Ax = lBx
*/
void MtxEigenValue(FlxMtxSym& Amtx, FlxMtxSym& Bmtx, const tuint M, tVec& EigenValues, std::vector<tVec>& Eigenvectors, const int Mode);




// -------------------------- Arnoldi - ARpack ----------------------------------------------------------------------

/**
* @brief Calculates Eigenvalues of Amtx
*/
void MtxEigenValue_ARP(FlxMtxSym& Amtx, const int M, tVec& EigenValues, std::vector<tVec>& Eigenvectors);

/**
* @brief Calculates Eigenvalues of a generalized matrix eigenvalue problem Ax = lBx
*/
void MtxEigenValue_ARP(FlxMtxSym& Amtx, FlxMtxSym& Bmtx, const int M, tVec& EigenValues, std::vector<tVec>& Eigenvectors);

// -------------------------- GSL ----------------------------------------------------------------------

/**
* @brief Calculates Eigenvalues of Amtx
*/
void MtxEigenValue_GSL(FlxMtxSym& Amtx, const int M, tVec& EigenValues, std::vector<tVec>& Eigenvectors);


/**
* @brief Calculates Eigenvalues of a generalized matrix eigenvalue problem Ax = lBx
*/
void MtxEigenValue_GSL(FlxMtxSym& Amtx, FlxMtxSym& Bmtx, const int M, tVec& EigenValues, std::vector<tVec>& Eigenvectors);


// -------------------------- GSL stab----------------------------------------------------------------------

/**
* @brief Calculates Eigenvalues of Amtx
*/
void MtxEigenValue_GSLstab(FlxMtxSym& Amtx, const int M, tVec& EigenValues, std::vector<tVec>& Eigenvectors);



/**
* @brief Calculates Eigenvalues of a generalized matrix eigenvalue problem Ax = lBx
*/
void MtxEigenValue_GSLstab(FlxMtxSym& Amtx, FlxMtxSym& Bmtx, const int M, tVec& EigenValues, std::vector<tVec>& Eigenvectors);




#endif // fesslix_FlxMtx_Eigen_H

