/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef flxexception_flxexception_H
#define flxexception_flxexception_H

#include <string>
#include <sstream>
#include <ostream>

/**
* @brief A class for exception handling.
*/
class FlxException {
  private:
    std::string errNumber;
    std::string titel;
    std::string msg;
  public:
    /**
    * @brief Create a FlxException class.
    * @param errnumber Where did the error occur? (Format: Class::Function_InternNumber)
    * @param Titel A short message describing the error.
    * @param Msg More error information.
    * @return FlxException
    */
    FlxException ( std::string errnumber, std::string Titel = "", std::string Msg = "" ) 
      : errNumber(errnumber), titel(Titel), msg(Msg) { }
    /**
    * @brief Get error information (usefull for catch-blocks).
    * @return string describing the error.
    */
    std::string what();
};

/**
* @brief A class for exception handling.
*/
class FlxException_NeglectInInteractive : public FlxException {
  public:
    /**
    * @brief Create a FlxException_NeglectInInteractive class.
    * @param errnumber Where did the error occur? (Format: Class::Function_InternNumber)
    * @param Titel A short message describing the error.
    * @param Msg More error information.
    * @return FlxException
    */
    FlxException_NeglectInInteractive ( std::string errnumber, std::string Titel = "", std::string Msg = "" ) 
      : FlxException(errnumber, Titel, Msg) { }
};

/**
* @brief A class for exception handling.
*/
class FlxException_NotImplemented : public FlxException {
  public:
    /**
    * @brief Create a FlxException_NotImplemented class.
    * @param errnumber Where did the error occur? (Format: Class::Function_InternNumber).
    * @return FlxException
    */
    FlxException_NotImplemented ( std::string errnumber ) 
      : FlxException(errnumber, "Feature not implemented", "The requested feature has not yet been implemented.") { }
};

/**
* @brief A class for exception handling.
*/
class FlxException_Crude : public FlxException {
  public:
    /**
    * @brief Create a FlxException_Crude class.
    * @param errnumber Where did the error occur? (Format: Class::Function_InternNumber).
    * @return FlxException
    */
    FlxException_Crude ( std::string errnumber ) 
      : FlxException(errnumber, "ERROR", "Actually, this error should have never occured ...") { }
};

/**
* @brief Used to terminate the program
*/
class FlxEndE {
  public:
    FlxEndE ( ) { }
};

/**
* @brief Used to exit from a certain read level
*/
class FlxExitE {
  public:
    FlxExitE ( ) { }
};

/**
* @brief A class for printing alerts.
*/
class FlxAlert {
  private:
    static std::ostream* cerr;
  public:
    /**
    * @brief Create a FlxAlert class.
    * @param &cerrV output_error_stream (Format: ostream).
    */
    FlxAlert () {}
    static void set_err_stream(std::ostream &cerrV) { cerr = &cerrV; }
    /**
    * @brief Output an alert.
    * @param alertID Where did the error occur? (Format: Class::Function_InternNumber)
    * @param alertStr A short message describing the alert.
    */
    void alert(std::string alertID, std::string alertStr);
};



#endif // flxexception_flxexception_H
