/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxfunction.h"
#include "flxfunction_fun.h"
#include "flxfunction_ope.h"
#include "flxfunction_ope_calc.h"
#include "flxmtxfun.h"

using namespace std;

FunReadBase* FunReadBase::startLink = NULL;
FlxConstantBox* FlxBoxBase::ConstantBox = NULL;
FlxVarBox* FlxBoxBase::VarBox = NULL;
FlxFunctionBox* FlxBoxBase::funBox = NULL;
FunReadSTART* FunReadFunBase_0::FunctionList = NULL;
tVec* FunPara::ParaList = NULL;
#if FLX_DEBUG
  int FlxVarBox::Cinst = 0;
  int FlxFunctionBox::Cinst = 0;
  int FlxFunctionReader::Cinst = 0;
#endif
int FunReadPara::numbofpara = 0;
tuint FlxFunDeg::default_deg = 0;

void FunBase::child_optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  FunBase* child_t = optf;
  optf = NULL;
  while (child_t->optimize(optf,foi)) {
    if (optf!=NULL) {
      delete child_t;
      child_t = optf;
      optf = NULL;
    }
  }
  optf = child_t;
}

void FunBase::calc_me(FunBasePtr& numref)
{
  numref = new FunNumber(calc());
}

const bool FunBase::is_number(FunBase* ftc)
{
  return (dynamic_cast<FunNumber*>(ftc) != NULL);
}

const tdouble FunBase::expectation()
{
  const tuint DIM = RndBox->get_NumbRnd();
  // determine dependence 
    bVec bv(false,DIM);
    dependOn_RV(bv);
  if (check_for_randomness(bv)) {
    return RndBox->calc_expectation_numerical(this);
  } else {
    return calc();
  }
}

bool check_dependence(const bVec& b1, const bVec& b2, const FlxRndBox& rndbox)
{
  #if FLX_DEBUG
    if (b1.size()!=b2.size()) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("check_dependence_1", ssV.str() );
    }
    if (b1.size()!=rndbox.get_NumbRnd()) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("check_dependence_1", ssV.str() );
    }
  #endif
  size_t DIM = b1.size();
  for (size_t i = 0; i < DIM; ++i) {
    if (b1[i]) {
      for (size_t j = 0; j < DIM; ++j) {
	if (b2[j]) {
	  if (rndbox.corr_get(i,j)!=0.0) return true;
	}
      }
    }
  }
  return false;
}

bool check_for_randomness(const bVec& bv)
{
  size_t DIM = bv.size();
  for (size_t i = 0; i < DIM; ++i) {
    if (bv[i]) return true;
  }
  return false;
}

bool check_for_single_dependence(tuint& index, const bVec& bv)
{
  size_t DIM = bv.size();
  bool b=false;
  for (size_t i = 0; i < DIM; ++i) {
    if (bv[i]) {
      if (b) {
	return false;
      } else {
	b=true;
	index=i;
      }
    }
  }
  if (b) {
    return true;
  } else {
    index = DIM;
    return false;
  }
}

FlxVarBox::FlxVarBox()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxVarBox' created ...";
      throw FlxException("FlxVarBox::FlxVarBox", ssV.str() );
    }
  #endif
  FlxBoxBase::set_varBox(this);
  FunBaseFun_MtxConst::set_dataVarConsts(this);
}

FlxVarBox::~FlxVarBox() {
  for (map<string, FlxFunction*>::iterator pos = box.begin(); pos != box.end(); ++pos) delete pos->second;
}

void FlxVarBox::insert(std::string name, FlxFunction* value) {
  #if FLX_DEBUG
    if (value==NULL) {
      throw FlxException("FlxVarBox::insert", "ERROR" );
    }
  #endif
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pair<std::string, FlxFunction*>Element(name, value);
  if ( ! box.insert(Element).second ) {
    map<std::string, FlxFunction*>::iterator pos;
    pos = box.find(name);;
    pos->second->assign(value);
  }
}

string FlxVarBox::get(FlxFunction* fun)
{
  for (map<string, FlxFunction*>::iterator pos = box.begin(); pos != box.end(); ++pos) {
    if (pos->second == fun) {
      return pos->first;
    }
  }
  std::ostringstream ssV;
  ssV << "Function not part of the list.";
  throw FlxException("FlxVarBox::get", ssV.str() );
}

FlxFunction* FlxVarBox::get(std::string name) {
  map<std::string, FlxFunction*>::iterator pos;
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pos = box.find(name);
  if ( pos != box.end() ) {
    return pos->second;
  } else {
    return NULL;
  }
}

void FlxVarBox::declareV(std::string name) {
  if ( get(name) == NULL ) {
    insert( name, new FlxFunction ( new FunNumber(0) ) );
  }
}

void FlxBoxBase::set_funBox(FlxFunctionBox* funBoxV)
{
  funBox = funBoxV;
}

void FlxBoxBase::set_constBox(FlxConstantBox* ConstantBoxV)
{
  ConstantBox = ConstantBoxV;
}

void FlxBoxBase::set_varBox(FlxVarBox* VarBoxV)
{
  VarBox = VarBoxV;
}

FlxFunction::~FlxFunction() {
  if (instances == NULL ) return;
  if ( (*instances) == 0)  {
    if ( fun ) delete fun;
    if ( instances ) delete instances;
  } else {
    (*instances)--;
  }
}

void FlxFunction::check_FlxFunction(const FlxFunction* funtocheck)
{
  const FlxFunction_Combine_Add* fa = dynamic_cast<const FlxFunction_Combine_Add*>(funtocheck);
  if (fa!=NULL) {
    throw FlxException_NotImplemented("FlxFunction::check_FlxFunction");
  }
}

const bool FlxFunction::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  child_optimize(fun,foi);
  return false;
}

FlxFunction::FlxFunction(const FlxFunction& FlxF)
{
  check_FlxFunction(&FlxF);
  instances = FlxF.instances;
  fun = FlxF.fun;
  (*instances)++;
}

FlxFunction& FlxFunction::operator=(const FlxFunction& FlxF)
{
  check_FlxFunction(&FlxF);
  if ( this != &FlxF ) {
    instances = FlxF.instances;
    fun = FlxF.fun;
    (*instances)++;
  }
  return *this;
}

void FlxFunction::assign(FlxFunction* funA) {
  check_FlxFunction(funA);
  if (this == funA) return;
  if (this->fun == funA->fun) return;
  if ((*instances)==0) {
    delete fun; 
    delete instances;
  } else {
    (*instances)--;
  }
  fun = funA->fun; 
  funA->fun = NULL;
  instances = funA->instances;
  funA->instances = NULL;
  delete funA;
}

const bool FlxFunction::dependOn_RV(const tuint& RVindex)
{
  #if FLX_DEBUG
  if (RVindex>=RndBox->get_NumbRnd()) {
    throw FlxException_Crude("FlxFunction::dependOn_RV");
  }
  #endif
  bVec bv(RndBox->get_NumbRnd());
  dependOn_RV(bv);
  return bv[RVindex];
}

const bool FlxFunction::dependOn_RV()
{
  bVec bv(false,RndBox->get_NumbRnd());
  dependOn_RV(bv);
  for (tuint i=0;i<bv.size();++i) {
    if (bv[i]) return true;
  }
  return false;
}

const tdouble FlxFunction::expectation()
{
  const tuint DIM = RndBox->get_NumbRnd();
  if (DIM==0) { 
    return fun->calc();
  } else {
    return fun->expectation();
  }
}

const bool FlxFunction::is_zero()
{
  FunNumber* S = dynamic_cast<FunNumber*>(fun);
  if (S==NULL) return false;
  if (fabs(S->get_thenumber()) <= GlobalVar.TOL()) {
    return true;
  } else {
    return false;
  }
}

const tdouble FlxFunction::cast2positive(const bool errSerious)
{
  const tdouble cv = fun->calc();
  if (cv<=0.0) {
    std::ostringstream ssV; 
    ssV << "Number must not be negative or zero (" << cv << ")."; 
    FlxError(errSerious,"FlxFunction::cast2positive", "Expected positive double!", ssV.str() );
  }
  return cv;
}

const tdouble FlxFunction::cast2positive_or0(const bool errSerious)
{
  const tdouble cv = fun->calc();
  if (cv<0.0) {
    std::ostringstream ssV; 
    ssV << "Number must not be negative (" << cv << ")."; 
    FlxError(errSerious,"FlxFunction::cast2positive_or0", "Expected positive double or zero!", ssV.str() );
  }
  return cv;
}

const tnlong FlxFunction::cast2tnlong(const bool errSerious)
{
  const tdouble cv = fun->calc();
  if ( cv <= 0.0) { 
    std::ostringstream ssV; 
    ssV << "Number must not be negative or zero (" << cv << ")."; 
    FlxError(errSerious,"FlxFunction::cast2tnlong", "Expected unsigned integer!", ssV.str() );     
  }
  return static_cast<tnlong>(cv);
}

const tuint FlxFunction::cast2tuint(const bool errSerious)
{
  const tdouble cv = fun->calc();
  if ( cv <= 0.0) { 
    std::ostringstream ssV; ssV << "Number must not be negative or zero (" << cv << ")."; FlxError(errSerious,"FlxFunction::cast2tuint", "Expected unsigned integer!", ssV.str() );
  }
  return static_cast<tuint>(cv);
}

const tuint FlxFunction::cast2tuintW0(const bool errSerious)
{
  const tdouble cv = fun->calc();
  if ( cv < 0.0) { std::ostringstream ssV; ssV << "Number must not be negative (" << cv << ")."; FlxError(errSerious,"FlxFunction::cast2tuintW0", "Expected unsigned integer!", ssV.str() );     }
  return static_cast<tuint>(cv);
}

const tnlong FlxFunction::cast2tnlongW0(const bool errSerious)
{
  const tdouble cv = fun->calc();
  if ( cv < 0.0) { std::ostringstream ssV; ssV << "Number must not be negative (" << cv << ")."; FlxError(errSerious,"FlxFunction::cast2tnlongW0", "Expected unsigned integer!", ssV.str() );     }
  return static_cast<tnlong>(cv);
}

FlxFunction_Combine_Add::FlxFunction_Combine_Add(FlxFunction* f1V, FlxFunction* f2V)
:f1(f1V), f2(f2V)
{
  instances = new tuint(0);
  fun = new FunAdd(f1->fun,f2->fun);
}

FlxFunction_Combine_Add::~FlxFunction_Combine_Add()
{
  FunAdd* fa = dynamic_cast<FunAdd*>(fun);
  fa->set_childs_zero();
  delete fa;
  fun = NULL;
  
  delete f1; f1=NULL;
  delete f2; f2=NULL;
}

void FlxFunction_Combine_Add::assign(FlxFunction* funA)
{
  throw FlxException_NotImplemented("FlxFunction_Combine_Add::assign");
}

FlxFunctionBox::FlxFunctionBox() {
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxFunctionBox' created ...";
      throw FlxException("FlxFunctionBox::FlxFunctionBox", ssV.str() );
    }
  #endif
  FlxBoxBase::set_funBox(this);
  // initialize random seed
    srand ( tuint(time(NULL)) );
    rv_initialize(true);
  // all objects of predefined functions have to be listed there
    flxfunction_fun_insert(*this);
}

FlxFunctionBox::~FlxFunctionBox() {
  for (map<std::string, FunReadFunBase*>::iterator pos = box.begin(); pos != box.end(); ++pos) delete pos->second;
}

void FlxFunctionBox::insert(std::string name, FunReadFunBase* fR) {
  pair<string, FunReadFunBase*>Element(name, fR);
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  if ( ! box.insert(Element).second ) {
    map<string, FunReadFunBase*>::iterator pos;
    pos = box.find(name);
    delete pos->second;
    pos->second = fR;
  }
}

void FlxFunctionBox::declareF(const std::string& name) {
  if ( get(name) == NULL ) {
    insert( name, new FunReadFunDummy() );
  }
}

FunReadFunBase* FlxFunctionBox::get(string name) {
  map<string, FunReadFunBase*>::iterator pos;
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pos = box.find(name);
  if ( pos != box.end() ) {
    return pos->second;
  } else {
    return NULL;
  }
}


const bool FunBaseOperat1::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  child_optimize(child_1, foi);
  if ( is_number(child_1) ) {
    calc_me(optf);
    return true;
  } else {
    return false;
  }
}

const bool FunBaseOperat2::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  child_optimize(child_1, foi);
  child_optimize(child_2, foi);
  if ( is_number(child_1) && is_number(child_2) ) {
    calc_me(optf);
    return true;
  } else {
    return false;
  }
}

const bool FunBaseOperat3::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  child_optimize(child_1, foi);
  child_optimize(child_2, foi);
  child_optimize(child_3, foi);
  if ( is_number(child_1) && is_number(child_2) && is_number(child_3) ) {
    calc_me(optf);
    return true;
  } else {
    return false;
  }
}

FunBaseFun_onePara::FunBaseFun_onePara(std::vector<FunBase*>* ParaListV)
:child_1(ParaListV->operator[](0))
{
  for ( size_t i = 1; i < ParaListV->size(); ++i ) 
    delete ParaListV->operator[](i); 
  delete ParaListV;
}

FunBaseFun_onePara::~FunBaseFun_onePara()
{
  delete child_1;
}

const std::string FunBaseFun_onePara::write()
{
  std::string str1 = write_v() + "(";
  str1 += child_1->write();
  str1 += ")";
  return str1;
}

const bool FunBaseFun_onePara::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  child_optimize(child_1, foi);
  if ( is_number(child_1) ) {
    calc_me(optf);
    return true;
  } else {
    return false;
  }
}

FunBaseFun_multPara::~FunBaseFun_multPara() {
  for ( size_t i = 0; i < ParaList->size(); ++i ) 
    delete ParaList->operator[](i); 
  delete ParaList;
}

const bool FunBaseFun_multPara::search_circref(FlxFunction* fcr) {
  bool b1 = false;
  for ( size_t i = 0; i < ParaList->size(); ++i ) {
    b1 = b1 || ParaList->operator[](i)->search_circref(fcr); 
  }
  return b1;
}

const std::string FunBaseFun_multPara::write()
{
  std::string str1 = write_v() + "(";
  for (size_t i = 0; i < ParaList->size(); ++i) {
    if (i > 0) str1 += ",";
    str1 += ParaList->operator[](i)->write();
  }
  str1 += ")";
  return str1;
}

void FunBaseFun_multPara::dependOn_RV(bVec& bv)
{
  for ( size_t i = 0; i < ParaList->size(); ++i ) 
    ParaList->operator[](i)->dependOn_RV(bv) ;
}

const bool FunBaseFun_multPara::dependOn_Const(const tdouble*const thenumber)
{
  for ( size_t i = 0; i < ParaList->size(); ++i ) {
    if ( ParaList->operator[](i)->dependOn_Const(thenumber) ) return true;
  }
  return false;
}

const bool FunBaseFun_multPara::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  bool b1=true;
  for ( size_t i = 0; i < ParaList->size(); ++i ) {
    child_optimize( ParaList->operator[](i), foi );
    b1 = b1 && is_number(ParaList->operator[](i));
  }
  if ( b1 ) {
    calc_me(optf);
    return true;
  } else {
    return false;
  }
}

void FunUser::dependOn_RV(bVec& bv)
{
  fun->dependOn_RV(bv);
  FunBaseFun_multPara::dependOn_RV(bv);
}

const bool FunUser::dependOn_Const(const tdouble*const thenumber)
{
  if (fun->dependOn_Const(thenumber)) return true;
  return FunBaseFun_multPara::dependOn_Const(thenumber);
}

const bool FunUser::search_circref(FlxFunction* fcr) {
  return FunBaseFun_multPara::search_circref(fcr) || fun->search_circref(fcr);
}

const tdouble FunUser::calc()
{
  tVec t(ParaList->size());
  for (size_t i = 0; i < t.size(); ++i) {
    t[i] = ParaListP[i]->calc();
  }
  tVec *tT = FunPara::ParaList;
  FunPara::ParaList = &t;
  tdouble res = fun->calc();
  FunPara::ParaList = tT;
  return res;
}

const tdouble FunPara::calc()
{
  if (ParaList == NULL) {
    ostringstream ssV_2;
    ssV_2 << "Paralist is not defined.";
    throw FlxException("FunPara::calc_1", ssV_2.str(), "This error should not have been occured.");  
  }
  if ( ParaList->size() >= theindex && theindex >= 1 ) {
    return ParaList->operator[](theindex - 1);
  } else {
    ostringstream ssV_2;
    ssV_2 << "A function uses a parameter which is out of range.";
    throw FlxException("FunPara::calc_2", ssV_2.str(), "This error is based on faulty function definition.");  
  }
}

const std::string FunPara::write()
{
  return "$" + theindex;
}

const std::string FunDummyFun::write_v()
{
  ostringstream ssV_2;
  ssV_2 << "ERROR";
  throw FlxException("FunDummyFun::write_v", ssV_2.str());  
}

const bool FunDummyFun::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  calc_me(optf);
  return true;
}

const std::string FunDummy::write()
{
  ostringstream ssV_2;
  ssV_2 << "ERROR";
  throw FlxException("FunDummy::write", ssV_2.str());  
}

const bool FunDummy::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  calc_me(optf);
  return true;
}

FunRV_fast::FunRV_fast(tuint RVindex) : RVindex(RVindex), RVptr(RndBox->getRV_UserIndex(RVindex))
{

}

const tdouble FunRV_fast::moments(const int& nth)
{
  if (nth==0) { 
    return 1.0;
  } else if (nth<0) {
    ostringstream ssV_2;
    ssV_2 << "Negative moments cannot be calculated with this algorithm.";
    throw FlxException("FunRV_fast::moments_2", ssV_2.str() );  
  } else if (nth == 1) {
    return RVptr->mean();
  } else {
    return RVptr->moment(nth);
  }
}

const tdouble FunRV_fast::calc()
{
  return RVptr->get_x_smp();
}

const std::string FunRV_fast::write()
{
  std::ostringstream ssV;
  ssV << "rv(" << RVindex << ")";
  return ssV.str();
}

void FunRV_fast::dependOn_RV(bVec& bv)
{
  tuint iv = RndBox->get_TrueIndex_from_User(RVindex);
  bv[iv] = true;
}

void FunRV::dependOn_RV(bVec& bv)
{
  const tdouble cv = child_1->calc();
  if (cv < 0) {
    ostringstream ssV_2;
    ssV_2 << "Index of user defined random variable must not be smaller than zero.";
    throw FlxException("FunRV::dependOn_RV", ssV_2.str() );  
  }
  const tuint iv = RndBox->get_TrueIndex_from_User(tuint(cv));
  bv[iv] = true;
}

const tdouble FunRV::calc()
{
  const tdouble res = child_1->calc();
  if (res < 0) {
    ostringstream ssV_2;
    ssV_2 << "Index of user defined random variable must not be smaller than zero.";
    throw FlxException("FunRV::calc", ssV_2.str() );  
  }
  return (RndBox->getRV_UserIndex(tuint(res)))->get_x_smp();
}

const bool FunRV::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  if ( FunBaseFun_onePara::optimize(optf, foi) && optf != NULL) {
    delete optf;
    optf = NULL;
  }
  if (is_number(child_1)) {
    const tdouble t = child_1->calc();
    if (t < 0) {
      ostringstream ssV_2;
      ssV_2 << "Index of user defined random variable must not be smaller than zero.";
      throw FlxException("FunRV::optimize", ssV_2.str() );  
    }
    optf = new FunRV_fast(tuint(t));
    return true;
  } else {
    return false;
  }
}

const tdouble FunRV::moments(const int& nth)
{
  tdouble res = child_1->calc();
  if (res < 0) {
    ostringstream ssV_2;
    ssV_2 << "Index of user defined random variable must not be smaller than zero.";
    throw FlxException("FunRV::moments_1", ssV_2.str() );  
  }
  if (nth==0) { 
    return 1.0;
  } else if (nth<0) {
    ostringstream ssV_2;
    ssV_2 << "Negative moments cannot be calculated with this algorithm.";
    throw FlxException("FunRV::moments_2", ssV_2.str() );  
  } else if (nth == 1) {
    return RndBox->getRV_UserIndex(tuint(res))->mean();
  } else {
    return RndBox->getRV_UserIndex(tuint(res))->moment(nth);
  }
}

FunReadBase::FunReadBase (const int Priority, bool isEndNode) : Next(NULL), priority(Priority) {
  if ( (!( isEndNode && priority == -1)) && priority < 0 ) {
    ostringstream ssV_2;
    ssV_2 << "Priority (" << priority << ") not allowed - value has to be greater than '0'.";
    throw FlxException("FunReadBase::FunReadBase_1", ssV_2.str(), "This error is based on faulty source code.");   
  }
}

FunReadBase* FunReadBase::insert ( FunReadBase *TheFunRead ) {
  if ( priority == -1 || TheFunRead->priority < priority ) {
    if ( TheFunRead->priority > 0 ) {
      TheFunRead->Next = this;
      return TheFunRead;
    } else {
      ostringstream ssV_2;
      ssV_2 << "Priority (" << priority << ") not allowed - value has to be greater than '0'.";
      throw FlxException("FunReadBase::insert_1", ssV_2.str(), "This error is based on faulty source code.");
    }
  } else {
    Next = Next->insert(TheFunRead);
    return this;
  }
}

vector< FunBase*>* FunReadFunBase::read_parameters(const int NumbOfPara, const bool errSerious) {
  vector< FunBase*>* paraL = new vector< FunBase*>(0);
  bool bnf = false;
  try {
    if ( reader->whatIsNextChar() != ')' ) {
      do {
	if ( bnf ) { 
	  reader->getChar(',',errSerious); 
	} else {
	  bnf = true;
	}
	paraL->push_back( FunctionList->read(errSerious) );
      } while ( reader->whatIsNextChar() == ',' );
    }

    if ( NumbOfPara >= 0 && paraL->size() != tuint(NumbOfPara) ) {
      ostringstream ssV_2;
      ssV_2 << "Function expects " << NumbOfPara << " parameters, and not " << paraL->size() << ".";
      FlxError(errSerious,"FunReadFunBase::read_parameters", ssV_2.str(), reader->getCurrentPos());
    }
    
    if (NumbOfPara==0) {
      paraL->push_back(new FunDummy());
    }
    
  } catch (FlxException& e) {
    for (size_t i = 0; i<paraL->size();++i) {
      delete paraL->operator[](i);
    }
    delete paraL;
    throw;
  }

  return paraL;
}

tdoublePtr FunReadFunBase::read_const_var(const bool errSerious,const bool define)
{
  const string strV = reader->getWord(false,errSerious);
  tdouble* d1 = ConstantBox->get(strV,define);
  if ( d1 == NULL ) {
    ostringstream ssV_2;
    ssV_2 << "Const-variable '" << strV << "' does not exist.";
    FlxError(errSerious,"FunReadFunBase::read_const_var", ssV_2.str(), reader->getCurrentPos());
  }
  return d1;
}

FunReadFunUser::~FunReadFunUser() {
  delete fun; 
}



FlxFunctionReader::FlxFunctionReader (  ) {
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxFunctionReader' created ...";
      throw FlxException("FlxFunctionReader::FlxFunctionReader", ssV.str() );
    }
  #endif
  
  FunctionList = new FunReadSTART();

  // thereafter all other FunReadXXXXXX Objects have to be listed (except 'FunReadSTART' and 'FunReadEND')
  FunctionList->insert( new FunReadNumber() );		// 1000
  FunctionList->insert( new FunReadPara() );		// 1020
  
  flxfunction_ope_insert(FunctionList);
}

FunBase* FlxFunctionReader::read (bool errSerious) {
  FunBase* rtn = NULL;
  try {
    rtn = FunctionList->read( errSerious );
    FunBasePtr optf = NULL;
    while (rtn->optimize(optf,Fun_OptimizeInfo())) {
      if (optf) {
	delete rtn;
	rtn = optf;
	optf = NULL;
      }
    }  
  } catch ( FlxException &e) { 
    if (rtn) delete rtn;
    throw;
  }
  return rtn;
}

FunBase* FunReadEND::read (bool errSerious) {
  ostringstream ssV_2;
  char c = reader->getChar(false);
  ssV_2 << "Character (" << c << ")[" << int(c) << "] not expected at this point.";
  FlxError(errSerious,"FunReadEND::read_1", ssV_2.str(), reader->getCurrentPos()); return NULL; // dummy-return
}

FunReadSTART::FunReadSTART() : FunReadBase(0) {
  startLink = this;
  FunReadFunBase_0::set_data(this);
  this->Next = new FunReadEND();
}

FunBase* FunReadSTART::read (bool errSerious) {
  return Next->read( errSerious );
}

FunBase* FunReadNumber::read (bool errSerious) {
  if ( reader->nextCanBeNumber() ) {
    bool is_minus = false;
    if (reader->whatIsNextChar()=='-') {
      reader->getChar();
      is_minus = true;
      if ( reader->nextCanBeNumber() == false ) {
	return new FunMult(new FunNumber(-1.0), Next->read(errSerious));
      }
    }
    tdouble d1 = reader->getNumber<tdouble>(errSerious);
    if (is_minus) d1*=-1;
    return new FunNumber(d1);
  }
  else return Next->read( errSerious );
}

FunBase* FunReadPara::read(bool errSerious) {
  if ( reader->whatIsNextChar() == '$' ) {
    reader->getChar();
    // parameter of functions
      if ( reader->nextCanBeNumber() ) {
	int i1 = reader->getNumber<int>(errSerious);
	if ( numbofpara == 0 ) {
	  ostringstream ssV;
	  ssV << "Parameters of type '$" << i1 << "' are only valid within functions.";
	  FlxError(errSerious,"FunReadPara::read_1", ssV.str(), reader->getCurrentPos());
	}
	if ( i1 < 1 || numbofpara < i1 ) {
	  ostringstream ssV;
	  ssV << "Index '" << i1 << "' out of range [0; " << numbofpara << "].";
	  FlxError(errSerious,"FunReadPara::read_2", ssV.str(), reader->getCurrentPos());
	}
	return new FunPara(static_cast<tuint>(i1));
      }
    // not allowed
      ostringstream ssV_2;
      ssV_2 << "Character not expeted ('" << reader->whatIsNextChar() << "').";
      FlxError(errSerious,"FunReadPara::read_3", ssV_2.str(), reader->getCurrentPos()); return NULL; // dummy-return
  } else return Next->read( errSerious );
}

void FunReadPara::set_NumbOfPara(tuint NumbOfPara, bool errSerious) {
  numbofpara = NumbOfPara; 
}




const tnlong tnlong_from(const tdouble& d, const std::string Descr, const bool errSerious, const bool zero_is_allowed)
{
  if ( (zero_is_allowed && d<0.0) || (!zero_is_allowed && d<= 0.0) ) {
    std::ostringstream ssV;
    ssV << Descr << " must not be a negative number (" << d << ").";
    FlxError(errSerious,"tnlong_from", ssV.str() );
  }
  return tnlong(d);
}

const tuint tuint_from(const tdouble& d, const std::string Descr, const bool errSerious, const bool zero_is_allowed)
{
  if ( (zero_is_allowed && d<0.0) || (!zero_is_allowed && d<= 0.0) ) {
    std::ostringstream ssV;
    ssV << Descr << " must not be a negative number (" << d << ").";
    FlxError(errSerious,"tuint_from", ssV.str() );
  }
  return tuint(d);
}

FlxFunDeg::FlxFunDeg(ReadStream* reader, FlxFunctionReader* funReader, bool errSerious)
:deg(0),fun(NULL)
{
  FlxFunction* tf = NULL;
  try {
    if (reader->whatIsNextChar()=='[') {
      reader->getChar('[',errSerious);
      fun = new FlxFunction(funReader,errSerious);
      reader->getChar(',',errSerious);
      tf = new FlxFunction(funReader,errSerious);
      deg = tf->cast2tuintW0(errSerious);
      delete tf; tf = NULL;
      reader->getChar(']',errSerious);
    } else {
      deg = default_deg;
      fun = new FlxFunction(funReader,errSerious);
    }
  } catch (FlxException& e) {
    if (fun) delete fun;
    if (tf) delete tf;
    throw;
  }
}

FlxFunDeg::FlxFunDeg(FlxFunction* funF)
:deg(default_deg),fun(funF)
{
  
}

FlxFunDeg::FlxFunDeg(const FlxFunDeg& ref)
{
  deg = ref.deg;
  fun = new FlxFunction(*ref.fun);
}

const std::string FlxFunDeg::write()
{
  std::ostringstream ssV;
  ssV << "[" << fun->write() << "," << deg << "]";
  return ssV.str();
}

void FlxFunDeg::assign(FlxFunDeg* funA)
{
  if (this == funA) return;
  deg = funA->deg;
  if (this->fun != funA->fun) {
    delete fun;
    fun = funA->fun;
    funA->fun = NULL;
  }
  delete funA;
}

