/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxfunction_fun_read.h"
#include "flxfunction_fun.h"

using namespace std;

void flxfunction_fun_insert(FlxFunctionBox& FunBox)
{
  FunBox.insert("sin", new FunReadFunSin() );
  FunBox.insert("arcsin", new FunReadFunArcsin() );
  FunBox.insert("cos", new FunReadFunCos() );
  FunBox.insert("arccos", new FunReadFunArccos() );
  FunBox.insert("tan", new FunReadFunTan() );
  FunBox.insert("arctan", new FunReadFunArctan() );
  FunBox.insert("ln", new FunReadFunLn() );
  FunBox.insert("log", new FunReadFunLog() );
  FunBox.insert("exp", new FunReadFunExp() );
  FunBox.insert("cot", new FunReadFunCot() );
  FunBox.insert("arccot", new FunReadFunArccot() );
  FunBox.insert("sinh", new FunReadFunSinh() );
  FunBox.insert("cosh", new FunReadFunCosh() );
  FunBox.insert("tanh", new FunReadFunTanh() );
  FunBox.insert("arctanh", new FunReadFunArctanh() );
  FunBox.insert("sqrt", new FunReadFunSqrt() );
  FunBox.insert("frac", new FunReadFunFrac() );
  FunBox.insert("abs", new FunReadFunAbs() );
  FunBox.insert("sig", new FunReadFunSig() );
  FunBox.insert("factorial", new FunReadFunFactorial() );
  FunBox.insert("pdfn", new FunReadFunPDFn() );
  FunBox.insert("cdfn", new FunReadFunCDFn() );
  FunBox.insert("invcdfn", new FunReadFunInvCDFn() );
  FunBox.insert("relidx", new FunReadFunRelIdx() );
  FunBox.insert("faillsf", new FunReadFunFailLSF() );
  FunBox.insert("hvi", new FunReadFunHvi() );
  FunBox.insert("rvs", new FunReadFunRVS() );
  FunBox.insert("time0", new FunReadFunTime0() );
  FunBox.insert("expectation", new FunReadFunExpectation() );
  FunBox.insert("integ", new FunReadFunInteg() );
  FunBox.insert("sum", new FunReadFunSum() );
  FunBox.insert("root", new FunReadFunRoot() );
  FunBox.insert("optimize1d", new FunReadFunOptimize1D() );
  FunBox.insert("deg2gauss", new FunReadFundeg2gauss() );
  FunBox.insert("binomialcoeff", new FunReadFunBinomialCoeff() );
  FunBox.insert("pdfn2", new FunReadFunPDFn2() );
  FunBox.insert("evalr", new FunReadFunEvalR() );
  FunBox.insert("evalc", new FunReadFunEvalC() );
  FunBox.insert("autocorr_exp", new FunReadFunAutoCorrExp() );
  FunBox.insert("autocorr_exp2", new FunReadFunAutoCorrExp2() );
  FunBox.insert("tgamma", new FunReadFunGamma() );
  FunBox.insert("igamma", new FunReadFunIGamma() );
  FunBox.insert("ibeta", new FunReadFunIBeta() );
  FunBox.insert("ibeta_inv", new FunReadFunIBeta_inv() );
  FunBox.insert("betafun", new FunReadFunBeta() );
  
  FunBox.insert("rv", new FunReadFunRV() );
}



// -------------------------- Random Variables (Distributions) ------ START ----------------------------------------------------------------------


FunRVS::FunRVS(string rvn, std::vector< FunBase* >* ParaListV, bool errSerious): FunBaseFun_multPara(ParaListV)
{ 
  transform(rvn.begin(), rvn.end(), rvn.begin(), (int(*)(int)) std::tolower);
  if (rvn=="unid") {
    rvf = &FunRVS::uniD;
  } else if (rvn=="unib") {
    rvf = &FunRVS::uniB;
  } else if (rvn=="stdb") {
    rvf = &FunRVS::stdB;
  } else if (rvn=="normalb") {
    rvf = &FunRVS::normalB;
  } else if (rvn=="logb") {
    rvf = &FunRVS::logB;
  } else {
    // TODO Add more distributions
    ostringstream ssV_2;
    ssV_2 << "Type '" << rvn << "' not known for a random variables.";
    FlxError(errSerious,"FunRVS::FunRVS", ssV_2.str());
  }
}


const int FunRVS::get_numb_of_para(string rvn)
{
  transform(rvn.begin(), rvn.end(), rvn.begin(), (int(*)(int)) std::tolower);
  if (rvn=="unid" || rvn=="unib" || rvn=="normalb") {
    return 2;
  } else if (rvn=="stdb") {
    return 0;
  } else if ( rvn=="logb" ) {
    return 3;
  } else {
    return -1;
  }
}

const tdouble FunRVS::uniD() const
{
  const tdouble a = ParaList->operator[](0)->calc();
  const tdouble b = ParaList->operator[](1)->calc();
  return (tdouble(rand())/RAND_MAX)*(b-a)+a;
}

const tdouble FunRVS::uniB() const
{
  const tdouble a = ParaList->operator[](0)->calc();
  const tdouble b = ParaList->operator[](1)->calc();
  return rv_uniform()*(b-a)+a;
}

const tdouble FunRVS::stdB() const
{
  return rv_normal();
}

const tdouble FunRVS::normalB() const
{
  const tdouble mv = ParaList->operator[](0)->calc();
  const tdouble sd = ParaList->operator[](1)->calc();
  return sd*rv_normal()+mv;
}

const tdouble FunRVS::logB() const
{
  const tdouble mean = ParaList->operator[](0)->calc();
  const tdouble sigma = ParaList->operator[](1)->calc();
  const tdouble epsilon = ParaList->operator[](2)->calc();
  const tdouble temp=std::log(pow2(sigma)/pow2(mean-epsilon) +1 );
  const tdouble lambda = std::log(mean-epsilon)-0.5*temp;
  const tdouble xi = std::sqrt(temp);
  return exp(rv_normal()*xi+lambda)+epsilon;
}

void FunRVS::dependOn_RV(bVec& bv) 
{
  ostringstream ssV;
  ssV << "It is not possible to check 'rv' for dependence on a random variable.";
  ostringstream ssV_2;
  ssV_2 << "Please use managed random variables.";
  throw FlxException("FunRVS::dependOn_RV", ssV.str(), ssV_2.str());
}


// -------------------------- Random Variables (Distributions) ------ END ----------------------------------------------------------------------


void FunTime0::dependOn_RV(bVec& bv) {
  ostringstream ssV;
  ssV << "It is not possible to check 'time0' for dependence on a random variable.";
  throw FlxException("FunTime0::dependOn_RV", ssV.str());
}

FunBase* FunReadFunRVS::read( bool errSerious )
{
  if ( reader->getNextType() != ReadStream::STRING ) {
    ostringstream ssV_2;
    ssV_2 << "Type of the random variable expected.";
    FlxError(errSerious,"FunReadFunRVS::read_1", ssV_2.str(), reader->getCurrentPos());
  }
  string rvn = reader->getWord(false,errSerious);
  
  int np = FunRVS::get_numb_of_para(rvn);
  if (np < 0) {
    ostringstream ssV_2;
    ssV_2 << "Type '" << rvn << "' not known for a random variables.";
    FlxError(errSerious,"FunReadFunRVS::read_2", ssV_2.str(), reader->getCurrentPos());
  }
  
  if (np>0) {
    reader->getChar(',',errSerious);
  }
  
  return new FunRVS(rvn, read_parameters(np,errSerious), errSerious ); 
}

FunInteg::~FunInteg()
{
  delete funI;
  delete startF;
  delete endF;
  delete gaussF;
  if (intF) delete intF;
}

const bool FunInteg::search_circref(FlxFunction* fcr)
{
  return funI->search_circref(fcr) || startF->search_circref(fcr) || endF->search_circref(fcr) || gaussF->search_circref(fcr) || intF->search_circref(fcr);
}

void FunInteg::dependOn_RV(bVec& bv)
{
  funI->dependOn_RV(bv);
  startF->dependOn_RV(bv);
  endF->dependOn_RV(bv);
  gaussF->dependOn_RV(bv);
  intF->dependOn_RV(bv);
}

const bool FunInteg::dependOn_Const(const tdouble*const thenumber)
{
  if (startF->dependOn_Const(thenumber)) return true;
  if (endF->dependOn_Const(thenumber)) return true;
  if (gaussF->dependOn_Const(thenumber)) return true;
  if (intF->dependOn_Const(thenumber)) return true;
  if (theconst==thenumber) { 
    return false;
  } else {
    return funI->dependOn_Const(thenumber);
  }
}

const bool FunInteg::optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) {
  child_optimize(funI, foi);
  child_optimize(startF, foi);
  child_optimize(endF, foi);
  child_optimize(gaussF, foi);
  if (intF) child_optimize(intF, foi);
  return false;
}

const std::string FunInteg::write()
{
  string str1 = "integ(";
  str1+=ConstantBox->get(theconst);
  str1+="=[";
  str1+=startF->write();
  str1+=",";
  str1+=endF->write();
  str1+="],";
  str1+=funI->write();
  str1+=",gp=";
  str1+=gaussF->write();
  if (intF!=NULL) {
    str1+=",int=";
    str1+=intF->write();
  }
  str1+=")";
  return str1;
}

const tdouble FunInteg::calc()
{
  // get number of integration points
    const tuint N = tnlong_from(gaussF->calc(),"number of Gauss-points",true,false);
  // determine start and end
    const tdouble start = startF->calc();
    const tdouble end = endF->calc();
    if (end<=start) {
      if (fabs(start-end)<=GlobalVar.TOL()) return 0.0;
      std::ostringstream ssV;
      ssV << "Error with integration boundaries.";
      throw FlxException("FunInteg::calc", ssV.str() );
    }
  if (intF==NULL) {
    return FlxFun_GaussIntegration(funI,theconst,start,end,N,*GI);
  } else {
    // get the number of intervals
    const tuint I = tnlong_from(intF->calc(),"the number of intervals",true,false);
    const tdouble dx = (end - start)/I;
    tdouble res = 0.0;
    tdouble t = start;
    for (register tuint i=0;i<I;++i) {
      res+=FlxFun_GaussIntegration(funI,theconst,t,t+dx,N,*GI);
      t+=dx;
    }
    return res;
  }
}


FunBase* FunReadFunInteg::read(bool errSerious)
{
  FunBase* funI = NULL;
  FunBase *startF=NULL,*endF=NULL,*gaussF=NULL,*intF=NULL;
  try {
    // get the constant
      tdouble* d1 = read_const_var(errSerious,true);
    // integration bounds
      reader->getChar('=',errSerious);
      reader->getChar('[',errSerious);
      startF = FunctionList->read(errSerious);
      reader->getChar(',',errSerious);
      endF = FunctionList->read(errSerious);
      reader->getChar(']',errSerious);
      reader->getChar(',',errSerious);
    // function to integrate
      funI = FunctionList->read(errSerious);
    reader->getChar(',',errSerious);
    reader->getWord("gp",errSerious);
    reader->getChar('=',errSerious);
    gaussF = FunctionList->read(errSerious);
    intF = NULL;
    if (reader->whatIsNextChar()==',') {
      reader->getChar(',',errSerious);
      reader->getWord("int",errSerious);
      reader->getChar('=',errSerious);
      intF = FunctionList->read(errSerious);
    }
    return new FunInteg(funI,d1,startF,endF,gaussF,intF);  
  } catch (FlxException &e) {
    if (funI) delete funI;
    if (startF!=NULL) delete startF;
    if (endF!=NULL) delete endF;
    if (gaussF!=NULL) delete gaussF;
    if (intF!=NULL) delete intF;
    throw;
  }
}

FunSum::~FunSum()
{
  delete funS;
  delete startF;
  delete cond;
  delete step;
}

const tdouble FunSum::calc()
{
  const tdouble STORE_VAL = *theconst;
  tdouble res = 0.0;
  *theconst = startF->calc();
  while (cond->calc()>0.0) {
    res += funS->calc();
    *theconst = step->calc();
  }
  *theconst = STORE_VAL;
  return res;
}

const bool FunSum::search_circref(FlxFunction* fcr)
{
  return funS->search_circref(fcr) || startF->search_circref(fcr) || cond->search_circref(fcr) || step->search_circref(fcr);
}

void FunSum::dependOn_RV(bVec& bv)
{
  funS->dependOn_RV(bv);
  startF->dependOn_RV(bv);
  cond->dependOn_RV(bv);
  step->dependOn_RV(bv);
}

const bool FunSum::dependOn_Const(const tdouble*const thenumber)
{
  if (startF->dependOn_Const(thenumber)) return true;
  if (cond->dependOn_Const(thenumber)) return true;
  if (step->dependOn_Const(thenumber)) return true;
  if (theconst==thenumber) { 
    return false;
  } else {
    return funS->dependOn_Const(thenumber);
  }
}

const bool FunSum::optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) {
  child_optimize(funS, foi);
  child_optimize(startF, foi);
  child_optimize(cond, foi);
  child_optimize(step, foi);
  return false;
}

const std::string FunSum::write()
{
  string str1 = "sum(";
  str1+=ConstantBox->get(theconst);
  str1+="=";
  str1+=startF->write();
  str1+=",";
  str1+=cond->write();
  str1+=",";
  str1+=step->write();
  str1+=",";
  str1+=funS->write();
  str1+=")";
  return str1;
}

FunBase* FunReadFunSum::read(bool errSerious)
{
  FunBase* funS = NULL;
  FunBase *startF=NULL,*cond=NULL,*step=NULL;
  try {
    // get the constant
      tdouble* d1 = read_const_var(errSerious,true);
    // start-value
      reader->getChar('=',errSerious);
      startF = FunctionList->read(errSerious);
      reader->getChar(',',errSerious);
    // condition
      cond = FunctionList->read(errSerious);
      reader->getChar(',',errSerious);
    // step-expression
      step = FunctionList->read(errSerious);
      reader->getChar(',',errSerious);
    // function to integrate
      funS = FunctionList->read(errSerious);
    return new FunSum(funS,d1,startF,cond,step);  
  } catch (FlxException &e) {
    if (startF) delete startF;
    if (cond) delete cond;
    if (step) delete step;
    if (funS) delete funS;
    throw;
  }
}

FunRoot::~FunRoot()
{
  delete funR;
  delete startF;
  delete endF;
  delete dx;
  delete dy;
}

const bool FunRoot::search_circref(FlxFunction* fcr)
{
  return funR->search_circref(fcr) || startF->search_circref(fcr) || endF->search_circref(fcr) || dx->search_circref(fcr) || dy->search_circref(fcr);
}

void FunRoot::dependOn_RV(bVec& bv)
{
  funR->dependOn_RV(bv);
  startF->dependOn_RV(bv);
  endF->dependOn_RV(bv);
  dx->dependOn_RV(bv);
  dy->dependOn_RV(bv);
}

const bool FunRoot::dependOn_Const(const tdouble*const thenumber)
{
  if (startF->dependOn_Const(thenumber)) return true;
  if (endF->dependOn_Const(thenumber)) return true;
  if (dx->dependOn_Const(thenumber)) return true;
  if (dy->dependOn_Const(thenumber)) return true;
  if (theconst==thenumber) { 
    return false;
  } else {
    return funR->dependOn_Const(thenumber);
  }
}

const bool FunRoot::optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) {
  child_optimize(funR, foi);
  child_optimize(startF, foi);
  child_optimize(endF, foi);
  child_optimize(dx, foi);
  child_optimize(dy, foi);
  return false;
}

const std::string FunRoot::write()
{
  string str1 = "root(";
  str1+=ConstantBox->get(theconst);
  str1+="=[";
  str1+=startF->write();
  str1+=",";
  str1+=endF->write();
  str1+="],";
  str1+=funR->write();
  str1+=",";
  switch (ID) {
    case 0:
      str1+="bisec";
      break;
    case 1:
      str1+="rgfsi";
      break;
    default:
      throw FlxException_Crude("FunRoot::write");
  }
  str1+=",dy=";
  str1+=dy->write();
  str1+=",dx=";
  str1+=dx->write();
  str1+=")";
  return str1;
}

const tdouble FunRoot::calc()
{
  switch (ID) {
    case 0:
      return FlxFun_RootSearch_Bisec(funR,theconst,startF->calc(),endF->calc(),dx->calc(),dy->calc());
    case 1:
      return FlxFun_RootSearch_RegulaFalsi(funR,theconst,startF->calc(),endF->calc(),dx->calc(),dy->calc());
    default:
      throw FlxException_Crude("FunRoot::calc");
  }
}

FunBase* FunReadFunRoot::read(bool errSerious)
{
  FunBase* funR = NULL;
  FunBase *startF=NULL,*endF=NULL,*dy=NULL,*dx=NULL;
  try {
    // get the constant
      tdouble* d1 = read_const_var(errSerious,true);
    reader->getChar(',',errSerious);
    reader->getChar('[',errSerious);
    startF = FunctionList->read(errSerious);
    reader->getChar(',',errSerious);
    endF = FunctionList->read(errSerious);
    reader->getChar(']',errSerious);
    reader->getChar(',',errSerious);
    funR = FunctionList->read(errSerious);
    reader->getChar(',',errSerious);
    const std::string method = reader->getWord(true,errSerious);
    tuint ID;
    // which method do we have to call?
      if (method == "bisec") {
	ID = 0;
      } else if (method == "rgfsi") {
	ID = 1;
      } else {
	std::ostringstream ssV;
	ssV << "Unknown method '" << method << "' for root-serach.";
	throw FlxException("FunReadFunRoot::read", ssV.str() );
      }
    if (reader->whatIsNextChar()==',') {
      reader->getChar(',',errSerious);
      reader->getWord("dy",errSerious);
      reader->getChar('=',errSerious);
      dy = FunctionList->read(errSerious);
      if (reader->whatIsNextChar()==',') {
	reader->getChar(',',errSerious);
	reader->getWord("dx",errSerious);
	reader->getChar('=',errSerious);
	dx = FunctionList->read(errSerious);
      }
    }
    if (dx==NULL) {
      dx = new FunNumber(1e-6);
    }
    if (dy==NULL) {
      dy = new FunNumber(1e-8);
    }
    return new FunRoot(ID,funR,d1,startF,endF,dy,dx);  
  } catch (FlxException &e) {
    if (funR) delete funR;
    if (startF) delete startF;
    if (endF) delete endF;
    if (dx) delete dx;
    if (dy) delete dy;
    throw;
  }
}

FunOptimize1D::~FunOptimize1D()
{
  delete funR;
  delete startF;
  delete endF;
  delete tau;
  delete nex;
}

const bool FunOptimize1D::search_circref(FlxFunction* fcr)
{
  return funR->search_circref(fcr) || startF->search_circref(fcr) || endF->search_circref(fcr) || tau->search_circref(fcr) || nex->search_circref(fcr);
}

void FunOptimize1D::dependOn_RV(bVec& bv)
{
  funR->dependOn_RV(bv);
  startF->dependOn_RV(bv);
  endF->dependOn_RV(bv);
  tau->dependOn_RV(bv);
  nex->dependOn_RV(bv);
}

const bool FunOptimize1D::dependOn_Const(const tdouble*const thenumber)
{
  if (startF->dependOn_Const(thenumber)) return true;
  if (endF->dependOn_Const(thenumber)) return true;
  if (tau->dependOn_Const(thenumber)) return true;
  if (nex->dependOn_Const(thenumber)) return true;
  if (theconst==thenumber) { 
    return false;
  } else {
    return funR->dependOn_Const(thenumber);
  }
}

const bool FunOptimize1D::optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) {
  child_optimize(funR, foi);
  child_optimize(startF, foi);
  child_optimize(endF, foi);
  child_optimize(tau, foi);
  child_optimize(nex, foi);
  return false;
}

const std::string FunOptimize1D::write()
{
  string str1 = "optimize1D(";
  str1+=ConstantBox->get(theconst);
  str1+=",[";
  str1+=startF->write();
  str1+=",";
  str1+=endF->write();
  str1+="],";
  str1+=funR->write();
  str1+=",";
  str1+=tau->write();
  str1+=",";
  str1+=nex->write();
  str1+=")";
  return str1;
}

const tdouble FunOptimize1D::calc()
{
  return FlxFun_Optimize1D(funR,theconst,startF->calc(),endF->calc(),tau->calc(),tuint_from(nex->calc(),"maximum number of iterations",false));
}

FunBase* FunReadFunOptimize1D::read(bool errSerious)
{
  FunBase *startF=NULL,*endF=NULL,*tau=NULL,*nex=NULL;
  // get the constant
    tdouble* d1 = read_const_var(errSerious,true);
  try {
      reader->getChar(',',errSerious);
	reader->getChar('[',errSerious);
	  startF = FunctionList->read(errSerious);
	  reader->getChar(',',errSerious);
	  endF = FunctionList->read(errSerious);
	reader->getChar(']',errSerious);
      reader->getChar(',',errSerious);
	FunBase* funR = FunctionList->read(errSerious);
      if (reader->whatIsNextChar()==',') {
	reader->getChar(',',errSerious);
	tau = FunctionList->read(errSerious);
	if (reader->whatIsNextChar()==',') {
	  reader->getChar(',',errSerious);
	  nex = FunctionList->read(errSerious);
	}
      }
      if (tau==NULL) {
	tau = new FunNumber(1e-6);
      }
      if (nex==NULL) {
	nex = new FunNumber(1e3);
      }
      return new FunOptimize1D(funR,d1,startF,endF,tau,nex);  
  } catch (FlxException &e) {
    if (startF) delete startF;
    if (endF) delete endF;
    if (tau) delete tau;
    if (nex) delete nex;
    throw;
  }
}

FunBase* FunReadFunEvalR::read(bool errSerious)
{
  FunBase* funI = FunctionList->read(errSerious);
  const tdouble cv = funI->calc();
  delete funI;
  return new FunNumber(cv);
}

FunBase* FunReadFunEvalC::read(bool errSerious)
{
  FunBase* funI = FunctionList->read(errSerious);
  Fun_OptimizeInfo foi(true);
  FunBase::child_optimize(funI,foi);
  return funI;
}

FunBase* FunReadFunAutoCorrExp::read(bool errSerious)
{
  tdouble* d1 = read_const_var(errSerious);
  reader->getChar(',');
  tdouble* d2 = read_const_var(errSerious);
  return new FunAutoCorrExp(d1,d2);
}

FunBase* FunReadFunAutoCorrExp2::read(bool errSerious)
{
  tdouble* d1 = read_const_var(errSerious);
  reader->getChar(',');
  tdouble* d2 = read_const_var(errSerious);
  return new FunAutoCorrExp2(d1,d2);
}

const tdouble FunDeg2Gauss::calc()
{
  return GI->pDegree2GPs(tuint_from(child_1->calc(),"Degree of a polynomial"));
}

const tdouble FunBinomialCoeff::calc()
{
  const int i1 = int(ParaListP[0]->calc());
  const int i2 = int(ParaListP[1]->calc());
  return BinomCoeff(i1,i2);
}

const tdouble FunPDFn2::calc()
{
  const tdouble u1 = ParaListP[0]->calc();
  const tdouble u2 = ParaListP[1]->calc();
  const tdouble r = ParaListP[2]->calc();
  const tdouble t = 1-pow2(r);
  return exp(-0.5*(pow2(u1)+pow2(u2)-2.0*r*u1*u2)/t)/(2.0*PI*sqrt(t));
}

const tdouble FunAutoCorrExp::calc()
{
  return exp(-1.0*(*d)/(*l));
}

const std::string FunAutoCorrExp::write()
{
  string str1 = "autocov_exp(";
  str1+=ConstantBox->get(d);
  str1+=",";
  str1+=ConstantBox->get(l);
  str1+=")";
  return str1;
}

const bool FunAutoCorrExp::optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) {
  if (foi.computeConstants) {
    calc_me(optf);
    return true;
  } else {
    return false;
  }
}

const tdouble FunAutoCorrExp2::calc()
{
  return exp(-1.0*pow2((*d)/(*l)));
}

const tdouble FunGamma::calc()
{
  const tdouble c1 = child_1->calc();
  try {
    return flxgamma( c1 );
  } catch (...) {
    std::ostringstream ssV;
    ssV << "Error evaluating gamma-function at '" << c1 << "'.";
    throw FlxException("FunGamma::calc", ssV.str() );
  }
}

const tdouble FunIGamma::calc()
{
  const tdouble c1 = ParaListP[0]->calc();
  const tdouble c2 = ParaListP[1]->calc();
  try {
    return flxgamma( c1, c2 );
  } catch (...) {
    std::ostringstream ssV;
    ssV << "Error evaluating incomplete gamma-function at '(" << c1 << "," << c2 << ")'.";
    throw FlxException("FunIGamma::calc", ssV.str() );
  }
}

const tdouble FunBeta::calc()
{
  const tdouble alpha = ParaListP[0]->calc();
  const tdouble beta = ParaListP[1]->calc();
  return BetaFun( alpha,beta);
}

const tdouble FunIBeta::calc()
{
  const tdouble alpha = ParaListP[0]->calc();
  const tdouble beta = ParaListP[1]->calc();
  const tdouble x = ParaListP[2]->calc();
  try {
    return iBeta_reg( alpha,beta,x );
  } catch (...) {
    std::ostringstream ssV;
    ssV << "Error evaluating regularized incomplete beta-function at '(" << alpha << "," << beta << "," << x << ")'.";
    throw FlxException("FunIBeta::calc", ssV.str() );
  }
}

const tdouble FunIBeta_inv::calc()
{
  const tdouble alpha = ParaListP[0]->calc();
  const tdouble beta = ParaListP[1]->calc();
  const tdouble p = ParaListP[2]->calc();
  try {
    return iBeta_reg_inv( alpha,beta,p );
  } catch (...) {
    std::ostringstream ssV;
    ssV << "Error evaluating inverse of the regularized incomplete beta-function at '(" << alpha << "," << beta << "," << p << ")'.";
    throw FlxException("FunIBeta::calc", ssV.str() );
  }
}
