/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxfunction_fun_calc_H
#define fesslix_flxfunction_fun_calc_H

#include "flxfunction_ext.h"


/**
* @brief arithmetic class: sinus
*/
class FunSin : public FunBaseFun_onePara {
  public:
    FunSin (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::sin( child_1->calc() ); }
    const std::string write_v() { return "sin";}
};

/**
* @brief arithmetic class: arcsin
*/
class FunArcsin : public FunBaseFun_onePara {
  public:
    FunArcsin (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::asin( child_1->calc() ); }
    const std::string write_v() { return "arcsin";}
};

/**
* @brief arithmetic class: cosinus
*/
class FunCos : public FunBaseFun_onePara {
  public:
    FunCos (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::cos( child_1->calc() ); }
    const std::string write_v() { return "cos";}
};

/**
* @brief arithmetic class: arccos
*/
class FunArccos : public FunBaseFun_onePara {
  public:
    FunArccos (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::acos( child_1->calc() ); }
    const std::string write_v() { return "arccos";}
};

/**
* @brief arithmetic class: tangens
*/
class FunTan : public FunBaseFun_onePara {
  public:
    FunTan (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::tan( child_1->calc() ); }
    const std::string write_v() { return "tan";}
};

/**
* @brief arithmetic class: arctan
*/
class FunArctan : public FunBaseFun_onePara {
  public:
    FunArctan (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::atan( child_1->calc() ); }
    const std::string write_v() { return "arctan";}
};

/**
* @brief arithmetic class: ln
*/
class FunLn : public FunBaseFun_onePara {
  public:
    FunLn (FunBase *child_1) : FunBaseFun_onePara(child_1) {};
    FunLn (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::log( child_1->calc() ); }
    const std::string write_v() { return "ln";}
};

/**
* @brief arithmetic class: log
*/
class FunLog : public FunBaseFun_onePara {
  public:
    FunLog (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::log10( child_1->calc() ); }
    const std::string write_v() { return "log";}
};

/**
* @brief arithmetic class: exp = e^x
*/
class FunExp : public FunBaseFun_onePara {
  public:
    FunExp (FunBase *child_1) : FunBaseFun_onePara(child_1) {};
    FunExp (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::exp( child_1->calc() ); }
    const std::string write_v() { return "exp";}
};

/**
* @brief arithmetic class: cot
*/
class FunCot : public FunBaseFun_onePara {
  public:
    FunCot (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return 1.0 / std::tan( child_1->calc() ); }
    const std::string write_v() { return "cot";}
};

/**
* @brief arithmetic class: arccot
*/
class FunArccot : public FunBaseFun_onePara {
  public:
    FunArccot (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return arccot( child_1->calc() ); }
    const std::string write_v() { return "arccot";}
};

/**
* @brief arithmetic class: sinh
*/
class FunSinh : public FunBaseFun_onePara {
  public:
    FunSinh (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::sinh( child_1->calc() ); }
    const std::string write_v() { return "sinh";}
};

/**
* @brief arithmetic class: cosh
*/
class FunCosh : public FunBaseFun_onePara {
  public:
    FunCosh (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::cosh( child_1->calc() ); }
    const std::string write_v() { return "cosh";}
};

/**
* @brief arithmetic class: tanh
*/
class FunTanh : public FunBaseFun_onePara {
  public:
    FunTanh (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::tanh( child_1->calc() ); }
    const std::string write_v() { return "tanh";}
};

/**
* @brief arithmetic class: arctanh
*/
class FunArctanh : public FunBaseFun_onePara {
  public:
    FunArctanh (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return arctanh( child_1->calc() ); }
    const std::string write_v() { return "arctanh";}
};

/**
* @brief arithmetic class: sqrt
*/
class FunSqrt : public FunBaseFun_onePara {
  public:
    FunSqrt (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::sqrt( child_1->calc() ); }
    const std::string write_v() { return "sqrt";}
};

/**
* @brief arithmetic class: frac
*/
class FunFrac : public FunBaseFun_onePara {
  public:
    FunFrac (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return frac( child_1->calc() ); }
    const std::string write_v() { return "frac";}
};

/**
* @brief arithmetic class: abs
*/
class FunAbs : public FunBaseFun_onePara {
  public:
    FunAbs (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return std::fabs( child_1->calc() ); }
    const std::string write_v() { return "abs";}
};

/**
* @brief arithmetic class: sig
*/
class FunSig : public FunBaseFun_onePara {
  public:
    FunSig (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { 
      const tdouble res = child_1->calc();
      if (std::fabs(res)<=GlobalVar.TOL()) return 0.0;
      else if ( res > 0 ) return 1.0;
      else return -1.0;
    }
    const std::string write_v() { return "sig";}
};

/**
* @brief arithmetic class: factorial
*/
// TODO: tread overflow
class FunFactorial : public FunBaseFun_onePara {
  public:
    FunFactorial (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { 
      const int n=int(floor(fabs(child_1->calc())+0.5));
      return Factorial(n);
    }
    const std::string write_v() { return "factorial";}
};

/**
* @brief arithmetic class: pdfn
*/
class FunPDFn : public FunBaseFun_onePara {
  public:
    FunPDFn (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return  rv_phi( child_1->calc() ); }
    const std::string write_v() { return "pdfn";}
};

/**
* @brief arithmetic class: cdfn
*/
class FunCDFn : public FunBaseFun_onePara {
  public:
    FunCDFn (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return rv_Phi( child_1->calc() ); }
    const std::string write_v() { return "cdfn";}
};

/**
* @brief arithmetic class: invcdfn
*/
class FunInvCDFn : public FunBaseFun_onePara {
  public:
    FunInvCDFn (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return rv_InvPhi( child_1->calc() ); }
    const std::string write_v() { return "invcdfn";}
};

/**
* @brief arithmetic class: relidx
*/
class FunRelIdx : public FunBaseFun_onePara {
  public:
    FunRelIdx (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return -rv_InvPhi( child_1->calc() ); }
    const std::string write_v() { return "relidx";}
};

/**
* @brief arithmetic class: faillsf
*/
class FunFailLSF : public FunBaseFun_onePara {
  public:
    FunFailLSF (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return (child_1->calc() <= 0.0); }
    const std::string write_v() { return "faillsf";}
};

/**
* @brief arithmetic class: hvi
*/
class FunHvi : public FunBaseFun_onePara {
  public:
    FunHvi (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return ( !(child_1->calc() < 0.0));  }
    const std::string write_v() { return "hvi";}
};

/**
* @brief arithmetic class: rvs (simple random variable)
*/
class FunRVS : public FunBaseFun_multPara {
  public:
    /**
    * @brief uniform distirbution (using c++ 'rand')
    */
    const tdouble uniD() const;
    const tdouble uniB() const;
    const tdouble stdB() const;
    const tdouble normalB() const;
    const tdouble logB() const;

    typedef const tdouble (FunRVS::*rvfT)() const;
    /**
    * @brief points to a probability distribution
    */
    rvfT rvf;
  public:
    FunRVS (std::string rvn, std::vector<FunBase*> *ParaListV, bool errSerious=true);
    const tdouble calc() { return (this->*rvf)(); }
    static const int get_numb_of_para(std::string rvn);
    const std::string write_v() { return "rvs";}
    void dependOn_RV(bVec& bv);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

/**
* @brief arithmetic class: time0
*/
class FunTime0 : public FunBaseFun_multPara {
  public:
    FunTime0 (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc() { return tdouble(std::time(0)); }
    const std::string write() { return "time0()";}
    const std::string write_v() { return "time0";}
    void dependOn_RV(bVec& bv);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi) { return false; }
};

/**
* @brief arithmetic class: expectation of an expression
*/
class FunExpectation : public FunBaseFun_onePara {
  public:
    FunExpectation (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc() { return child_1->expectation(); }
    const std::string write_v() { return "expectation";}
    void dependOn_RV(bVec& bv) { }
    const tdouble expectation() { return calc(); }
};

/**
* @brief arithmetic class: numerical integration of a function
*/
class FunInteg : public FunBase, public FlxBoxBase {
  private:
    /**
    * @brief the function to integrate
    */
    FunBase* funI;
    /**
    * @brief the integration variable
    */
    tdouble* theconst;
    /**
    * @brief start of the integration
    */
    FunBase* startF;
    /**
    * @brief end of the integration
    */
    FunBase* endF;
    /**
    * @brief the number of Gauss-points to use
    */
    FunBase* gaussF;
    /**
    * @brief the number of intervals to use
    */
    FunBase* intF;
  public:
    FunInteg(FunBase* funI, tdouble* theconst, FunBase* startF, FunBase* endF, FunBase* gaussF, FunBase* intF) : funI(funI),theconst(theconst),startF(startF),endF(endF),gaussF(gaussF),intF(intF) {}
    ~FunInteg();
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: numerical integration of a function
*/
class FunSum : public FunBase, public FlxBoxBase {
  private:
    /**
    * @brief the function to sum
    */
    FunBase* funS;
    /**
    * @brief the integration variable
    */
    tdouble* theconst;
    /**
    * @brief start of the integration
    */
    FunBase* startF;
    /**
    * @brief condition
    */
    FunBase* cond;
    /**
    * @brief step-expression
    */
    FunBase* step;
  public:
    FunSum(FunBase* funS, tdouble* theconst, FunBase* startF, FunBase* cond, FunBase* step) : funS(funS),theconst(theconst),startF(startF),cond(cond),step(step) {}
    ~FunSum();
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: root search (a single one) of a function
*/
class FunRoot : public FunBase, public FlxBoxBase {
  protected:
    /**
    * @brief the ID of the method
    * 0: bisection
    * 1: regula falsi
    */
    const int ID;
    /**
    * @brief the function to investigate
    */
    FunBase* funR;
    /**
    * @brief the root-search variable
    */
    tdouble* theconst;
    /**
    * @brief start of the serach-interval
    */
    FunBase* startF;
    /**
    * @brief end of the search-interval
    */
    FunBase* endF;
    /**
    * @brief stop criteria
    */
    FunBase* dx;
    FunBase* dy;
  public:
    FunRoot(const int ID, FunBase* funR, tdouble* theconst, FunBase* startF, FunBase* endF, FunBase* dx, FunBase* dy) : ID(ID), funR(funR),theconst(theconst),startF(startF),endF(endF),dx(dx),dy(dy) {}
    ~FunRoot();
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr);
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
    const std::string write();
};

/**
* @brief arithmetic class: root search (a single one) of a function
*/
class FunOptimize1D : public FunBase, public FlxBoxBase {
  protected:
    /**
    * @brief the function to investigate
    */
    FunBase* funR;
    /**
    * @brief the root-search variable
    */
    tdouble* theconst;
    /**
    * @brief start of the serach-interval
    */
    FunBase* startF;
    /**
    * @brief end of the search-interval
    */
    FunBase* endF;
    /**
    * @brief stop criteria
    */
    FunBase* tau;
    FunBase* nex;
  public:
    FunOptimize1D(FunBase* funR, tdouble* theconst, FunBase* startF, FunBase* endF, FunBase* tau, FunBase* nex) : funR(funR),theconst(theconst),startF(startF),endF(endF),tau(tau),nex(nex) {}
    ~FunOptimize1D();
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr);
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv);
    const bool dependOn_Const(const tdouble* const thenumber);
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
    const std::string write();
};

/**
* @brief arithmetic class: sinus
*/
class FunDeg2Gauss : public FunBaseFun_onePara {
  public:
    FunDeg2Gauss (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "sin";}
};

/**
* @brief arithmetic class: binomial coefficient
*/
class FunBinomialCoeff : public FunBaseFun_multPara {
  public:
    FunBinomialCoeff (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "binomialcoeff";}
    void dependOn_RV(bVec& bv) { }
    const tdouble expectation() { return calc(); }
};

/**
* @brief arithmetic class: pdfn2
*/
class FunPDFn2 : public FunBaseFun_multPara {
  public:
    FunPDFn2 (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "pdfn2";}
};

/**
* @brief arithmetic class: numerical integration of a function
*/
class FunAutoCorrExp : public FunBase, public FlxBoxBase {
  protected:
    const tdouble* d;
    const tdouble* l;
  public:
    FunAutoCorrExp(const tdouble* d, const tdouble* l) : d(d), l(l) {}
    const tdouble calc();
    const bool search_circref(FlxFunction* fcr) { return false; }
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { }
    const bool dependOn_Const(const tdouble* const thenumber) { return thenumber==d || thenumber==l;} 
    const bool optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi);
};

/**
* @brief arithmetic class: numerical integration of a function
*/
class FunAutoCorrExp2 : public FunAutoCorrExp {
  public:
    FunAutoCorrExp2(const tdouble* d, const tdouble* l) : FunAutoCorrExp(d,l) {}
    const tdouble calc();
};

/**
* @brief arithmetic class: Gamma-Function
*/
class FunGamma : public FunBaseFun_onePara {
  public:
    FunGamma (std::vector<FunBase*> *ParaListV) : FunBaseFun_onePara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "tgamma";}
};

/**
* @brief arithmetic class: incomplete Gamma-Function
*/
class FunIGamma : public FunBaseFun_multPara {
  public:
    FunIGamma (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "igamma";}
};

/**
* @brief arithmetic class: regularized incomplete Beta-Function
*/
class FunBeta : public FunBaseFun_multPara {
  public:
    FunBeta (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "beta";}
};

/**
* @brief arithmetic class: regularized incomplete Beta-Function
*/
class FunIBeta : public FunBaseFun_multPara {
  public:
    FunIBeta (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "ibeta";}
};

/**
* @brief arithmetic class: regularized incomplete Beta-Function
*/
class FunIBeta_inv : public FunBaseFun_multPara {
  public:
    FunIBeta_inv (std::vector<FunBase*> *ParaListV) : FunBaseFun_multPara(ParaListV) {};
    const tdouble calc();
    const std::string write_v() { return "ibeta_inv";}
};


#endif // fesslix_flxfunction_fun_calc_H
