/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#define fesslix_flxglobal_CPP

#include "flxmath.h"
#include "flxdefault.h"

#include <string>
#include <sstream>
#include <fstream>
#include <limits>


void print_FLX_GNU(std::ostream& sout)
{
  // TODO print a cool Fesslix Banner
  
  print_GNU(sout);
}


void print_GNU(std::ostream& sout)
{
  sout << "Fesslix Copyright (C) 2010-2012 Wolfgang Betz" << std::endl;
  sout << "This is free software; see the source code for copying conditions." << std::endl;
  sout << "There is ABSOLUTELY NO WARRANTY; not even for MERCHANTABILITY or" << std::endl;
  sout << "FITNESS FOR A PARTICULAR PURPOSE.  For details, type `warranty;'." << std::endl << std::endl;
  sout << "Additional information about Fesslix is available at http://www.fesslix.org." << std::endl << std::endl;
}

size_t bVec_NumbTrue(const bVec& bv)
{
  size_t N=0;
  for (size_t i=0;i<bv.size();++i) {
    if (bv[i]) ++N;
  }
  return N;
}

void FlxError(bool errSerious, std::string errnumber, std::string Titel, std::string Msg)
{
  if (errSerious) {
    throw FlxException(errnumber,Titel,Msg);
  } else {
    throw FlxException_NeglectInInteractive(errnumber,Titel,Msg);
  }
}

std::string strGlobalVar::Double2String(tdouble dv, bool checkTOL)
{
  if (checkTOL) {
    if (fabs(dv) <= GlobalVar.TOL()) dv = 0.0;
  }
  std::ostringstream ssV;
  // perform some settings (output-type)
    if (D2S_Type == 0) {
      // do nothing
    } else if (D2S_Type == 1) {
      ssV << std::scientific;
    } else if (D2S_Type == 2) {
      if ( fabs(dv)>=D2S_BValU || fabs(dv)<D2S_BValL ) {
	ssV << std::scientific;
      }
    } else if (D2S_Type == 3) {
      ssV << std::fixed;
    }
    ssV.precision(D2S_Prec);
  // create the number
    ssV << dv;
    std::string strV = ssV.str();
  // edit the number
    // delete 0s after the e
    if (D2S_Del0) {
      size_t iI = strV.rfind("e");
      if (iI != std::string::npos) {
	iI+=2;
	tuint n = 0;
	while ( strV[iI+n] == '0' ) {
	  ++n;
	}
	strV.erase(iI, n);
      }
    }
    // delete + after e
    if (D2S_DelP) {
      size_t iI = strV.rfind("+");
      if (iI != std::string::npos) {
	strV.erase(iI,1);
      }
    }
    // delete 0s after .XYZ
    if (D2S_Del0 && strV.rfind(".") != std::string::npos) {
      size_t iI = strV.rfind("e");
      tuint n;
      if (iI != std::string::npos) {
	n = iI-1;
      } else {
	n = strV.length()-1;
      }
      tuint m = 0;
      while ( strV[n-m] == '0' ) {
	++m;
      }
      strV.erase(n-m+1,m);
      iI = strV.rfind("e");
      if (iI != std::string::npos) {
	n = iI-1;
      } else {
	n = strV.length()-1;
      }
      if (strV[n] == '.') {
	strV.erase(n,1);
      }
    }
    // delete e if it is at the end of the string
    if (D2S_Del0) {
      size_t iI = strV.rfind("e");
      if (iI != std::string::npos) {
	if (iI == strV.length()-1) {
	  strV.erase(iI,1);
	}
      }
    }
  return strV;
}

void strGlobalVar::Double2String_setPrec(tuint prec)
{
  D2S_Prec = prec;
}

void strGlobalVar::Double2String_setType(tuint type)
{
  D2S_Type = type;
}

void strGlobalVar::Double2String_setBValU(tdouble bv)
{
  D2S_BValU = fabs(bv);
}

void strGlobalVar::Double2String_setBValL(tdouble bv)
{
  D2S_BValL = fabs(bv);
}

void strGlobalVar::Double2String_setDel0(bool d0)
{
  D2S_Del0 = d0;
}

void strGlobalVar::Double2String_setDelP(bool dP)
{
  D2S_DelP = dP;
}

void strGlobalVar::Double2String_log()
{
  slog(4) << "  floating point conversion:" << std::endl;
  slog(4) << "        Precision:              " << D2S_Prec << std::endl;
  slog(4) << "        Type:                   " << D2S_Type;
  if (D2S_Type==2) {
    slog(4) << " [" << D2S_BValL << "; " << D2S_BValU << "]";
  }
  slog(4) << std::endl;
  slog(4) << "        Del0:                   ";
  if (D2S_Del0) {
    slog(4) << "yes";
  } else {
    slog(4) << "no";
  }
  slog(4) << std::endl;
  slog(4) << "        DelP:                   ";
  if (D2S_DelP) { 
    slog(4) << "yes";
  } else {
    slog(4) << "no";
  }
  slog(4) << std::endl;
}

std::ostream& strGlobalVar::slog(int logLevel_)
{
  if (slogP == NULL || logLevel_ > logLevel) return *sdummy;
  else return *slogP; 
}

void strGlobalVar::set_TOL(tdouble tolV)
{
  TOLval = fabs(tolV);
  if (TOLval > 1e-6) {
    slog(2) << std::endl << "ALERT: TOL (" << TOLval << ") is very large. This leads to wrong results!" << std::endl << std::endl;
  }
}

strGlobalVar::strGlobalVar() 
: sdummy(NULL),slogP(NULL),stdcout(NULL), 
  // NOTE this are not the default values - the default values are set in the settings of the configuration file
  D2S_Prec(DEFAULT_FPC_PREC), D2S_Type(DEFAULT_FPC_TYPE), D2S_BValU(DEFAULT_FPC_BVALU), D2S_BValL(DEFAULT_FPC_BVALL), D2S_Del0(DEFAULT_FPC_DEL0), D2S_DelP(DEFAULT_FPC_DELP),
  TOLval(DEFAULT_TOL),prelog_active(true),prelog_stream(""),
  logLevel_strong(DEFAULT_LOG_LEVEL),exe_dir(""),leak_check(false),
  logLevel(DEFAULT_LOG_LEVEL),
  MT19937_init_calls(DEFAULT_MT19937_INIT_CALLS), MT19937_init_RAND(DEFAULT_MT19937_INIT_RAND), MT19937_init_seed(DEFAULT_MT19937_INIT_SEED), MT19937_init_seedvalue(DEFAULT_MT19937_INIT_SEEDVALUE),
  Exp_Num_Intervals_1D(DEFAULT_EXP_NUM_INTERVALS_1D),Exp_Num_Samples_1D(DEFAULT_EXP_NUM_SAMPLES_1D),Exp_Num_Diff_Redistribute_1D(DEFAULT_EXP_NUM_DIFF_REDISTRIBUTE_1D),
  Exp_Num_Lbound_1D(DEFAULT_EXP_NUM_LBOUND_1D),Exp_Num_Ubound_1D(DEFAULT_EXP_NUM_UBOUND_1D),Exp_Num_Bound_MCI(DEFAULT_EXP_NUM_BOUND_MCI),
  Exp_Num_Intervals_MCI(DEFAULT_EXP_NUM_INTERVALS_MCI), Exp_Num_SmpPerInterval_MCI(DEFAULT_EXP_NUM_SMPPERINTERVAL_MCI), Exp_Num_Smp_Redistribute_MCI(DEFAULT_EXP_NUM_SMP_REDISTRIBUTE_MCI), Exp_Num_Diff_Redistribute_MCI(DEFAULT_EXP_NUM_DIFF_REDISTRIBUTE_MCI),
  Exp_Num_Rnd_Moments(DEFAULT_EXP_NUM_RND_MOMENTS),Exp_Num_Rnd_Moments_Central(DEFAULT_EXP_NUM_RND_MOMENTS_CENTRAL),
  LegendrePolyH_init_numb(DEFAULT_LEGENDRE_NUMB), allowInteractiveMode(true), prgBar(DEFAULT_FLX_PRGBAR), configDir(""),configDir_local(""), pwd(""),
  sqrtEps(std::sqrt(std::numeric_limits<tdouble>::epsilon()))
{
}

strGlobalVar::~strGlobalVar()
{
//   std::cout << "strGlobalVar::~strGlobalVar  " << slogP << std::endl;
  using namespace std;
  if (sdummy != NULL) delete sdummy;
  if (slogP != NULL) {
    ofstream* thestream = dynamic_cast<ofstream*> (slogP);
    if ( thestream != NULL ) {
      thestream->close();
      delete thestream;
    }
  }
}

void strGlobalVar::logLevel_strong_set(int logL)
{
  logLevel_strong = logL; 
  logLevel_strong_reset();
}

void strGlobalVar::prelog_activated(bool b)
{
  if (prelog_active==b) return;
  prelog_active=b;
  prelog_stream = "";
}

void strGlobalVar::prelog_append(char c)
{
  prelog_stream+=c;
}

void strGlobalVar::prelog_append(std::string str)
{
  prelog_stream+=str;
}

void strGlobalVar::exe_dir_set(std::string str)
{
  char c1 = 47;
  char c2 = 92;
  size_t i1 = str.rfind(c1);
    if (i1 > str.size() ) i1 = str.size();
  size_t i2 = str.rfind(c2);
    if (i2 > str.size() ) i2 = str.size();
  if ( (i2 > i1 && i2!=str.size()) || i1 == str.size() ) i1 = i2;
  if (i1 == str.size() ) {
    exe_dir = str;
  } else {
    exe_dir = str.substr(0,i1+1);
  }
  #ifdef __WINDOWS__ 
    if (exe_dir.substr(exe_dir.size()-4,4)==".exe") {
      exe_dir = "";
    }
  #endif
}

void strGlobalVar::prelog_prepare_write()
{
  bool b1=true;
  size_t pos; char c;
  do {
    pos = prelog_stream.size();
    if (prelog_stream == "" || pos <= 0 || prelog_stream.empty() ) return;
    --pos;
    c = prelog_stream[pos];
    if (c == 10 || c == 9 || c == 13 || c == ' ' ) {
      prelog_stream.erase(pos,1);
    } else {
      b1 = false;
    }
  } while (b1);
  b1 = true;
  do {
    if (prelog_stream == "" || prelog_stream.size() <= 0 || prelog_stream.empty() ) return;
    c = prelog_stream[0];
    if (c == 10 || c == 9 || c == 13 || c == ' ' ) {
      prelog_stream.erase(0,1);
    } else {
      b1 = false;
    }
  } while (b1);
}

void strGlobalVar::prelog_clear()
{
  prelog_stream = "";
}

std::string strGlobalVar::prelog_force_write()
{
  prelog_prepare_write();
  std::string strtmp = prelog_stream;
//   prelog_stream = "";
  return strtmp;
}

void strGlobalVar::prelog_write()
{
  if (prelog_active) {
    if ( prelog_isNOTempty() ) {
      prelog_prepare_write();
      slog(4) << prelog_stream << std::endl;
      prelog_stream = "";
    }
  } else {
    prelog_stream = "";
  }
}


