/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxio.h"

#include <fstream>
#include <cstring>
#include <cctype>


using namespace std;

ReadStream* FlxReaderBase::reader = NULL;


ofstream* open_stream(string filename)
{
  std::ofstream *theStream = new std::ofstream(filename.c_str());
  if ( ! theStream->is_open() || theStream == NULL ) {
    std::ostringstream ssV;
    ssV << "File (" << filename << ") could not be opened.";
    throw FlxException("flxio.h::open_stream", ssV.str() );
  }
  return theStream;
}

istream_warper::istream_warper(istream* theStream, const string filename): theStream(theStream),filename(filename)
{
  if (filename=="(String)") return;
  std::ifstream *filstr = dynamic_cast<std::ifstream*> (theStream);
  if ( ! filstr->is_open() || filstr == NULL ) {
    ostringstream ssV;
    ssV << "File (" << filename << ") could not be opened.";
    // free memory
      std::string str1 = filename;
      if (str1 != "(cin)") delete theStream; 
    throw FlxException("istream_warper::istream_warper_1", ssV.str() );
  }
}

istream_warper::~istream_warper()
{
  std::ifstream *filstr = dynamic_cast<std::ifstream*> (theStream);
  if ( filstr != NULL) {
    filstr->close();
  }
  const std::string str1 = filename;
  if (str1 != "(cin)") {
    if (theStream!=NULL) delete theStream; 
  }
}

const int istream_warper::get()
{
  if (!mystack.empty()) {
    char ch = mystack.top();
    mystack.pop();
    return ch;
  } else {
    return theStream->get();
  }
}

const string istream_warper::get_line(const char delim)
{
  std::string st;
  char ch;
  while ((ch=get())!=delim) {
    st += ch;
  }
  return st;
}

const int istream_warper::peek()
{
  if (!mystack.empty()) {
    return mystack.top();
  } else {
    return theStream->peek();
  }
}

const bool istream_warper::eof()
{
  if (!mystack.empty()) {
    return false;
  } else {
    return theStream->eof();
  }
}

void istream_warper::putback(const int ch)
{
  if (ch < 0) {
    ostringstream ssV;
    ssV << "ERROR (" << ch << ")" ;
    throw FlxException("istream_warper::putback", ssV.str());
  }
  mystack.push(ch);
}

ReadStream::ReadStream ( const char* FileName, bool do_log, int tabWidth ) 
: theStream(new istream_warper(new std::ifstream(FileName), FileName)), TabWidth(tabWidth), lineNumb(1), charNumb(0),do_log(do_log) { 
  setNext(); 
}

ReadStream::ReadStream ( string strV, bool do_log, int tabWidth ) 
: theStream(new istream_warper(new std::istringstream(strV+"\n            "),"(String)")), TabWidth(tabWidth), lineNumb(1), charNumb(0),do_log(do_log) { 
  setNext(); 
}

// ReadStream::ReadStream(istream& istr, int tabWidth)
// : theStream(&istr), filename("(cin)"), TabWidth(tabWidth),lineNumb(1),charNumb(0)
// {
//   setNext();
// }


ReadStream::~ReadStream ( ) { 
  delete theStream;
}

ReadStream::InpType ReadStream::getType(char ch)
{
  if ( int(ch) < 0 ) return ENDOFFILE;
  if ( ch >= '0' && ch <= '9' ) return NUMERAL;
  if ( (ch >= 'A' && ch <= 'Z') || (ch >= 'a' && ch <= 'z') || ch == '_' ) return STRING;
  if ( ch == '#' ) return COMMENT;
  if ( ch == '(' || ch == ')' || ch == '[' || ch == ']' || ch == '{' || ch == '}' ) return BRAKET;
  if ( ch == 34 || ch == 96 ) return QUOTE;
  if ( ch >= '!' && ch <= '~' ) return OTHER;
  if ( ch == 9 		// tab
    || ch == 10		// new line
    || ch == 13		// carriage return
    || ch == 32		// blank
    ) return NONE;

  return UNKNOWN;
}

const ReadStream::InpType ReadStream::getNextType () {
  if ( theStream->eof() ) return ENDOFFILE;
  return getType( theStream->peek() );
}

const std::string ReadStream::getCurrentPos () {
  ostringstream ssV;
  ssV << "Line: " << lineNumb << "; Column: " << charNumb << "; File: " << theStream->get_FileName();
  return ssV.str();
}

const char ReadStream::stream_getnext_log()
{
  char t = theStream->get();
  log_char(t);
  return t;
}

void ReadStream::setNext ( const bool DOlog ) {
  char ch;
  while ( getNextType() == NONE || getNextType() == COMMENT ) {
    if ( getNextType() == COMMENT ) {			// ... if there is a comment
      theStream->get_line();
//       while ( (ch=theStream->get()) != 10 ) {
// 	if ( theStream->eof() ) return;
// 	if (int(ch)<0) return;
// 	if ( ch == 9 ) { 			// tab
// 	  charNumb += TabWidth;
// 	} else {
// 	  charNumb++;
// 	}
//       }
      lineNumb++; 
      charNumb=0;
    } else { 						// ... if it is NOT a comment
      if ( (ch=theStream->get()) == 10 ) {  		// new line
	lineNumb++;
	charNumb=0;
      } else {
	if ( ch == 9 ) {			// tab
	  charNumb += TabWidth;
	} else if ( ch == 13 ) {
	} else {
	  charNumb++;
	}
      }
      if (DOlog) log_char(ch);
    }
  }
}

const char ReadStream::getChar ( const bool DOsetNext, const bool DOlog ) {
  charNumb++;
  char ch = (DOlog?stream_getnext_log():theStream->get());
  if ( DOsetNext ) setNext(DOlog);
  return ch;
}

const char ReadStream::getChar ( const char theChar, const bool errSerious, const bool DOsetNext ) {
  char ch = getChar(DOsetNext);
  if ( ch != theChar ) {
    ostringstream ssV;
    ssV << "Expected '" << theChar << "' (and NOT '" << ch << "' [" << int(ch) << "])";
    FlxError(errSerious,"ReadStream::getChar_1", ssV.str(), getCurrentPos());
  }
  return ch;
}

const char ReadStream::whatIsNextChar() {
  return theStream->peek();
}

const std::string ReadStream::whatIsNextString( const int length, const bool lowercase) {
  if ( length <= 0 ) {
    ostringstream ssV;
    ssV << "'length' has to be greater than zero.";
    throw FlxException("ReadStream::whatIsNextString_1", ssV.str() );
  }
  // get next string
    #ifndef FLX_CV_1
      int i1[length];
      char ch1[length+1];
    #else
      int *i1 = new int[length];
      char *ch1 = new char[length+1];
    #endif
    if (theStream->eof()) {
      return "";
    }
    for (int i = 0; i < length; ++i) {
      i1[i] = theStream->get();
      if (theStream->eof()) {
	ch1[i] = '\0';
	break;
      }
      ch1[i] = i1[i];
    }
    ch1[length] = '\0';
    string str1(ch1);
  // write it back
    for (int i = std::strlen(ch1)-1;i>=0;--i) {
      if (ch1[i] != 0) theStream->putback(i1[i]);
    }
  #ifdef FLX_CV_1
    delete [] i1;
    delete [] ch1;
  #endif
  if (lowercase) {
    std::transform(str1.begin(), str1.end(), str1.begin(), (int(*)(int)) std::tolower);
  }
  return str1;
}

void ReadStream::getExpr(const char* strWordC, bool errSerious)
{
  #ifndef FLX_CV_1
    char chS[std::strlen(strWordC)];
  #else
    char *chS = new char[std::strlen(strWordC)];
  #endif
  int i = 0;
  while (strWordC[i]!='\0') {
    chS[i] = theStream->get();
    switch (chS[i]) {
      case 9:
	charNumb+=TabWidth;
	break;
      case 10:
	lineNumb++;
	charNumb=0;
	break;
      case 13:
	break;
      default:
	charNumb++;
    }
    ++i;
  }
  chS[i]='\0';
  
  std::string str1(chS);
  std::string str2(strWordC);
  
  #ifdef FLX_CV_1
    delete [] chS;
  #endif
  if (str1!=str2) {
    ostringstream ssV_2;
    ssV_2 << "Expected '" << str2 << "' and not '" << str1 << "'.";
    FlxError(errSerious,"ReadStream::getExpr", ssV_2.str(), getCurrentPos());
  }
  
  if (do_log) GlobalVar.prelog_append(str1);
}

const std::string ReadStream::getWord(const char* strWordC, bool errSerious)
{
  std::string strWord(strWordC);
  std::transform(strWord.begin(), strWord.end(), strWord.begin(), (int(*)(int)) std::tolower);
  std::string s1 = getWord(true,errSerious);
  if (s1!=strWord) {
    ostringstream ssV_2;
    ssV_2 << "Expected '" << strWord << "' and not '" << s1 << "'.";
    FlxError(errSerious,"ReadStream::getWord_11", ssV_2.str(), getCurrentPos());
  }
  return s1;
}

const std::string ReadStream::getWord (bool lowercase, bool errSerious, const bool numbeg) {
  ostringstream ssV;

  // check if the first character is of type STRING
  if ( !numbeg && getNextType() != STRING ) {
    ostringstream ssV_2;
    char c = getChar(false);
    ssV_2 << "Character '" << c << "' (" << int(c) << ") not expected. (Only 'A'-'Z', 'a'-'z' and '_' is valid.)";
    FlxError(errSerious,"ReadStream::getWord_1", ssV_2.str(), getCurrentPos());
  } 
  
  // read the word
  while ( getNextType() == STRING || getNextType() == NUMERAL ) {
    ssV << getChar(false);
  }
  setNext();
  if (lowercase) {
    std::string s1 = ssV.str();
    std::transform(s1.begin(), s1.end(), s1.begin(), (int(*)(int)) std::tolower);
    return s1;
  } else {
    return ssV.str();
  }
}

const std::string ReadStream::getText(const bool MultiLine, const bool errSerious) {
  getChar('"',errSerious,false);
  string strText = "";
  while ( whatIsNextChar() != '"' ) {
    if ( ! MultiLine ) {
      if ( whatIsNextChar() == 10 || whatIsNextChar() == 9 ) {
	ostringstream ssV_2;
	ssV_2 << "Character (" << getChar(false) << ") not expected.";
	FlxError(errSerious,"ReadStream::getText_1", ssV_2.str(), getCurrentPos());
      } 
    } else {
      if (whatIsNextChar()==10) {
	++lineNumb;
      }
    }
    strText += getChar(false);
  }
  getChar('"',errSerious);
  setNext();
  return strText;
}

const bool ReadStream::getBool(const bool errSerious) {
  bool b1;
  if ( nextCanBeNumber() ) {
    b1 = (fabs(getNumber<tdouble>(errSerious))> GlobalVar.TOL());
  } else {
    if ( getNextType() != ReadStream::STRING ) {
      ostringstream ssV;
      ssV << "Expected 'true' or 'false'.";
      FlxError(errSerious,"ReadStream::getBool_1", ssV.str(), getCurrentPos()); b1 = false;
    } else {
      string str1 = getWord(false,errSerious);
      transform(str1.begin(), str1.end(), str1.begin(), (int(*)(int)) std::tolower);
      if ( str1 == "true" ) {
	b1 = true;
      } else if ( str1 == "false" ) {
	b1 = false;
      } else {
	ostringstream ssV;
	ssV << "Expected 'true' or 'false'.";
	FlxError(errSerious,"ReadStream::getBool_2", ssV.str(), getCurrentPos()); b1 = false;
      }
    }
  }
  return b1;
}


const bool ReadStream::nextCanBeNumber() {
  char ch = theStream->peek();
  // check if the first character is of type NUMERAL or if it is a leading sign
  if ( getNextType() != NUMERAL && ch != '+' && ch != '-' && ch != '.' ) return false;
  else return true;
}


FlxIstream::FlxIstream(string nameV, bool errEOFv)
:name(nameV),errEOF(errEOFv)
{

}

void FlxIstream::reachedEOF()
{
  if (errEOF) {
    ostringstream ssV;
    ssV << "No more numbers to input. Input stream '" << name << "' is empty.";
    throw FlxException("FlxIstream::reachedEOF_1", ssV.str() ); 
  } else {
    ostringstream ssV;
    ssV << "Warning: No more numbers to input. Input stream '" << name << "' is empty.";
    GlobalVar.alert.alert("FlxIstream::reachedEOF_2", ssV.str() );
  }
}

bool FlxIstream::get_vec(tVec& v, size_t& lastIndex, bool suppressErr)
{
  for (register size_t i = 0; i<v.size();++i) {
    if (!get_value(v[i],suppressErr)) {
      lastIndex = i;
      return false;
    }
  }
  lastIndex = v.size();
  return true;
}

void FlxIstream_file::copyStream(FlxIstream* rhsB, bool errSerious)
{
  FlxIstream_file* rhs = dynamic_cast<FlxIstream_file*> (rhsB);
  if (rhs==NULL) {
    std::ostringstream ssV;
    ssV << "It is not possible to replace a file-input-stream with a non-file-input-stream.";
    FlxError(errSerious,"FlxIstream_file::copyStream_1", ssV.str() );
  }
  #if FLX_DEBUG
    if (rhs == this) {
      std::ostringstream ssV;
      ssV << "ERROR";
      FlxError(errSerious,"FlxIstream_file::copyStream_2", ssV.str() );
    }
    if (rhs->iReader == this->iReader) {
      std::ostringstream ssV;
      ssV << "ERROR";
      FlxError(errSerious,"FlxIstream_file::copyStream_3", ssV.str() );
    }
  #endif
  name = rhs->name;
  if (iReader != NULL) delete iReader;
    iReader = rhs->iReader;
    rhs->iReader = NULL;
  errEOF = rhs->errEOF;
  blocksize = rhs->blocksize;
  if (SeqVec != NULL) delete SeqVec;
    SeqVec = rhs->SeqVec;
    rhs->SeqVec = NULL;
  index = rhs->index;
  lastindex = rhs->lastindex;
  Cnumb = rhs->Cnumb;
  Cvec = rhs->Cvec;
  curCol = rhs->curCol;
  curVecIdx = rhs->curVecIdx;
  delete rhs;
}

void FlxIstream_file::resetStream(ReadStream* iReaderV, bool errEOFv, tuint blocksizeV)
{
  #if FLX_DEBUG
    if (iReaderV == iReader) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxIstream_file::resetStream_1", ssV.str() );
    }
  #endif
  if (iReader != NULL) delete iReader;
  iReader = iReaderV;
  errEOF = errEOFv;
  blocksize = blocksizeV;
  lastindex = blocksize;
  if (SeqVec != NULL) delete SeqVec;
  SeqVec = new tVec(blocksize);
  read_block();
}

FlxIstream_file::FlxIstream_file(string nameV, ReadStream* iReaderV, bool errEOFv, tuint blocksizeV, const int Cnumb, const std::vector<int> &Cvec)
: FlxIstream(nameV,errEOFv)
, iReader(NULL), blocksize(blocksizeV), SeqVec(NULL), index(0), lastindex(0),
Cnumb(Cnumb),Cvec(Cvec),curCol(1),curVecIdx(0)
{
  resetStream(iReaderV,errEOFv,blocksizeV);
}

FlxIstream_file::~FlxIstream_file()
{
  if (iReader != NULL) {
    delete iReader;
  }
  delete SeqVec;
}

bool FlxIstream_file::check_eof()
{
  if (iReader->getNextType() == ReadStream::ENDOFFILE) return true;
  else return false;
}

void FlxIstream_file::set_next()
{
  if (iReader->whatIsNextChar() == ',') {
    iReader->getChar(',');
  } else if (iReader->whatIsNextChar() == ';') {
    iReader->getChar(';');
  }  
}

void FlxIstream_file::read_block()
{
  if (Cnumb==1 && Cvec.size()==1) {
    for (tuint i = 0; i < blocksize; ++i) {
      if (check_eof()) {
	lastindex = i;
	break;
      }
      (*SeqVec)[i] = iReader->getNumber<tdouble>();
      if (!check_eof()) {
	set_next();
      }
    }
  } else {
    register tuint i=0;
    tdouble d;
    while (i<blocksize) {
      if (check_eof()) {
	lastindex = i;
	break;
      }
      d = iReader->getNumber<tdouble>();
      if (!check_eof()) {
	set_next();
      }
      if (curCol==Cvec[curVecIdx]) {	
	// read number
	  (*SeqVec)[i] = d;
	// update Cvec-Position
	  ++curVecIdx;
	  ++i;
	  if (curVecIdx==((int)Cvec.size())) curVecIdx = 0;
      }
      ++curCol;
      if (curCol>Cnumb) curCol = 1;
    }
  }
  index = 0;
}

bool FlxIstream_file::get_value(tdouble& v, bool suppressErr)
{
  if (index == lastindex) {
    if (!suppressErr) reachedEOF();
    v = 0.0;
    return false;
  } else {
    v = (*SeqVec)[index];
    ++index;
    if (index == blocksize) read_block();
    return true;
  }
}

FlxIstream_vector::FlxIstream_vector(string nameV, FlxIstream* is, bool errEOFv, const tnlong Nreserve)
: FlxIstream(nameV, errEOFv), index(0), numbEl(0)
{
  iVec.reserve(Nreserve);
  if (is!=NULL) {
    tdouble d;
    while (is->get_value(d,true)) {
      iVec.push_back(d);
    }
    numbEl = iVec.size();
  }
}

void FlxIstream_vector::copyStream(FlxIstream* rhsB, bool errSerious)
{
  FlxIstream_vector* rhs = dynamic_cast<FlxIstream_vector*> (rhsB);
  if (rhs==NULL) {
    std::ostringstream ssV;
    ssV << "It is not possible to replace a vector-input-stream with a non-vector-input-stream.";
    FlxError(errSerious,"FlxIstream_vector::copyStream_1", ssV.str() );
  }
  #if FLX_DEBUG
    if (rhs == this) {
      std::ostringstream ssV;
      ssV << "ERROR";
      FlxError(errSerious,"FlxIstream_vector::copyStream_2", ssV.str() );
    }
  #endif
  name = rhs->name;
  errEOF = rhs->errEOF;
  index = rhs->index;
  numbEl = rhs->numbEl;
  iVec = rhs->iVec;
  delete rhs;
}

bool FlxIstream_vector::get_value(tdouble& v, bool suppressErr)
{
  if (index == numbEl) {
    if (!suppressErr) reachedEOF();
    v = 0.0;
    index = 0;
    return false;
  } else {
    v = iVec[index];
    ++index;
    return true;
  }
}

void FlxIstream_vector::appendNumber(const tdouble& d)
{
  iVec.push_back(d);
  numbEl = iVec.size();
}

void FlxIstream_vector::sortStream()
{
  std::sort(iVec.begin(),iVec.end());
  index = 0;
}



