/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxmath_rnd_H
#define fesslix_flxmath_rnd_H

#include "flxmath.h"





/**
* @brief initializes the random number generator
*/
void rv_initialize(bool startup, bool user_seed=false, tuint user_seed_int = 0, tuint user_init_calls = 0);

/**
* @brief realization of a random variable with a uniform distribution [0;1]
*/
const tdouble rv_uniform();

/**
* @brief realization of a random variable with a standard normal distribution
*/
const tdouble rv_normal();
void rv_normal(tVec& y);

/**
* @brief realization of a random variable with a lognormal distribution
*/
inline const tdouble rv_lognormal(tdouble &lambda, tdouble &xi, tdouble &epsilon) {
  return exp(rv_normal()*xi+lambda)+epsilon;
}

/**
* @brief probability density function of the standard normal distribution: phi
*/
inline const tdouble rv_phi(const tdouble &y) { return exp(-0.5*y*y) / sqrt(2*PI) ;  }

/**
* @brief cumulative distribution function of the standard normal distribution: Phi
*/
const tdouble rv_Phi(const tdouble& y);

/**
* @brief inverse of the cumulative distribution function of the standard normal distribution: Phi^-1
*/
const tdouble rv_InvPhi(const tdouble& p);

/**
* @brief cdf of the Chi Square distribution
*/
const tdouble rv_cdf_ChiSquare(const tuint& v, const tdouble& x);

/**
* @brief pdf of the Chi Square distribution
*/
const tdouble rv_pdf_ChiSquare(const tuint& v, const tdouble& x);

/**
* @brief inverse of the cdf of the Chi Square distribution
*/
const tdouble rv_InvCDF_ChiSquare(const tuint& v, const tdouble& p);


#endif // fesslix_flxmath_rnd_H
