/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxmtxfun.h"
#include "flxmtxfun_fun.h"
#include "flxmtxfun_ope.h"

using namespace std;

FunMtxReadBase* FunMtxReadBase::startLink = NULL;
FunMtxReadSTART* FunReadMtxFunBase::MtxFunList = NULL;
FlxConstMtxBox* FunBaseFun_MtxConst::mtxConsts = NULL;
FlxConstantBox* FunBaseFun_MtxConst::scalarConsts = NULL;
FlxVarBox* FunBaseFun_MtxConst::varConsts = NULL;
FunMtxReadSTART* FunReadFunBase_Mtx::MtxFunList = NULL;
FlxConstMtxBox* FlxMtxBoxBase::ConstMtxBox = NULL;
FlxMtxFunBox* FlxMtxBoxBase::MtxFunBox = NULL;
#if FLX_DEBUG
  int FlxMtxFunBox::Cinst = 0;
  int FlxMatrixReader::Cinst = 0;
#endif

FlxMtxFunBox::FlxMtxFunBox()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxConstMtxBox' created ...";
      throw FlxException("FlxConstMtxBox::FlxConstMtxBox", ssV.str() );
    }
  #endif
  FlxMtxBoxBase::set_MtxFunBox(this);
  flxmtxfun_fun_insert(*this, *funBox);
}

FlxMtxFunBox::~FlxMtxFunBox()
{
  for (map<std::string, FunReadMtxFunBase*>::iterator pos = box.begin(); pos != box.end(); ++pos) delete pos->second;
}

void FlxMtxFunBox::insert(string name, FunReadMtxFunBase* fR)
{
  pair<string, FunReadMtxFunBase*>Element(name, fR);
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  if ( ! box.insert(Element).second ) {
    map<string, FunReadMtxFunBase*>::iterator pos;
    pos = box.find(name);
    delete pos->second;
    pos->second = fR;
  }
}

FunReadMtxFunBase* FlxMtxFunBox::get(string name)
{
  map<string, FunReadMtxFunBase*>::iterator pos;
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pos = box.find(name);
  if ( pos != box.end() ) {
    return pos->second;
  } else {
    return NULL;
  }
}



FlxMtxFun::~FlxMtxFun()
{
  if (instances == NULL) return;
  if ( (*instances) == 0)  {
    if ( fun != NULL ) delete fun;
    if ( instances != NULL ) delete instances;
  } else {
    (*instances)--;
  }
}

const tVec FlxMtxFun::calcVec()
{
  SMtxBase* res = fun->calc();
  // extract Vector
    if (res->get_ncols()==1 || res->get_nrows()==1) {
      tVec x = res->get_tVec();
      delete res;
      return x;
    } else {
      std::ostringstream ssV;
      ssV << "Expected a vector ... and not a matrix." << std::endl << *res;
      delete res;
      throw FlxException("FlxMtxFun::calcVec", ssV.str() );
    }
}

const iVec FlxMtxFun::cast2tuintNo0()
{
  tVec t = calcVec();
  iVec r(t.size());
  tVec_cast2tuintNo0(t,r);
  return r;
}

void FlxMtxFun::assign(FlxMtxFun* funA)
{
  if (this == funA) return;
  if (this->fun == funA->fun) return;
  if ((*instances)==0) {
    delete fun; 
    delete instances;
  } else {
    (*instances)--;
  }
  fun = funA->fun; 
  funA->fun = NULL;
  instances = funA->instances;
  funA->instances = NULL;
  delete funA;
}

FlxMtxFun& FlxMtxFun::operator=(const FlxMtxFun& FlxF)
{
  if ( this != &FlxF ) {
    instances = FlxF.instances;
    fun = FlxF.fun;
    (*instances)++;
  }
  return *this;
}

FlxMtxFun::FlxMtxFun(const FlxMtxFun& FlxF)
{
  instances = FlxF.instances;
  fun = FlxF.fun;
  (*instances)++;
}


const std::string FunMtxDummy::write()
{
  ostringstream ssV_2;
  ssV_2 << "ERROR";
  throw FlxException("FunMtxDummy::write", ssV_2.str());  
}

SMtxValue::~SMtxValue()
{
  for (tuint i = 0; i < vecV->size(); ++i) {
    delete vecV->operator[](i);
  }
  delete vecV;
}

const std::string SMtxValue::write() 
{
//   std::cout << nrows << ", " << ncols << "...";
  std::string strV = "{";
  tuint i,j;
  for (i = 0; i < nrows; ++i) {
    if (i > 0) strV += ";";
    for (j = 0; j < ncols; ++j) {
      if (j>0) strV += ",";
      strV += vecV->operator[](i*ncols+j)->write();
    }
  }  
  strV += "}";
  return strV;
}

SMtxBase* SMtxValue::calc()
{
  // Scalar
  if (nrows == 1 && ncols == 1) {
    return new SMtxScalar(vecV->operator[](0)->calc());
  // Vector
  } else if (nrows == 1 || ncols == 1 ) {
    const tuint n = (ncols==1)?nrows:ncols;
    tVec* d = new tVec(n);
    for (tuint i = 0; i < n; i++) {
      d->operator[](i) = vecV->operator[](i)->calc();
    }
    return new SMtxVector(d,n,nrows==1);
  // Matrix
  } else { 
    FlxMtx* d = new FlxMtx(nrows,ncols);
    tuint i,j;
    for (i = 0; i < nrows; ++i) {
      for (j = 0; j < ncols; ++j) {
	d->operator()(i,j) = vecV->operator[](i*ncols+j)->calc();
      }
    }
    return new SMtxMatrix(d);
  }
}

void SMtxValue::dependOn_RV(bVec& bv)
{
  for (size_t i = 0; i < vecV->size(); ++i) {
    vecV->operator[](i)->dependOn_RV(bv);
  }
}

const bool SMtxValue::dependOn_Const(const tdouble*const thenumber)
{
  for (size_t i = 0; i < vecV->size(); ++i) {
    if (vecV->operator[](i)->dependOn_Const(thenumber)) return true;
  }
  return false;
}

const bool SMtxValue::search_circref(FlxFunction* fcr)
{
  for (size_t i = 0; i < vecV->size(); ++i) {
    if ( vecV->operator[](i)->search_circref(fcr) ) return true;
  }
  return false;
}

FunMtxReadBase::FunMtxReadBase(const int Priority, bool isEndNode): Next(NULL),priority(Priority)
{
  if ( (!( isEndNode && priority == -1)) && priority < 0 ) {
    ostringstream ssV_2;
    ssV_2 << "Priority (" << priority << ") not allowed - value has to be greater than '0'.";
    throw FlxException("FunMtxReadBase::FunMtxReadBase_1", ssV_2.str(), "This error is based on faulty source code.");   
  }
}

FunMtxReadBase* FunMtxReadBase::insert(FunMtxReadBase* TheFunRead)
{
  if ( priority == -1 || TheFunRead->priority < priority ) {
    if ( TheFunRead->priority > 0 ) {
      TheFunRead->Next = this;
      return TheFunRead;
    } else {
      ostringstream ssV_2;
      ssV_2 << "Priority (" << priority << ") not allowed - value has to be greater than '0'.";
      throw FlxException("FunMtxReadBase::insert_1", ssV_2.str(), "This error is based on faulty source code.");
    }
  } else {
    Next = Next->insert(TheFunRead);
    return this;
  }
}



FlxMatrixReader::FlxMatrixReader()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxConstantBox' created ...";
      throw FlxException("FlxConstantBox::FlxConstantBox", ssV.str() );
    }
  #endif
  FunctionList = new FunMtxReadSTART();
  flxmtxfun_ope_insert(*FunctionList);  
}

FunMtxBase* FlxMatrixReader::read(bool errSerious)
{
  FunMtxBase* rtn = NULL;
  try {
    rtn = FunctionList->read(  errSerious );
  } catch ( FlxException &e) { 
    if (rtn) delete rtn;
    throw;
  }
  return rtn;
}




FunMtxBase* FunMtxReadEND::read(bool errSerious)
{
  ostringstream ssV_2;
  char c = reader->getChar(false);
  ssV_2 << "Character (" << c << ")[" << int(c) << "] not expected at this point.";
  FlxError(errSerious,"FunMtxReadEND::read_1", ssV_2.str(), reader->getCurrentPos()); return NULL;	// DUMMY-RETURN
}

FunMtxReadSTART::FunMtxReadSTART() : FunMtxReadBase(0)
{
  startLink = this;
  FunReadMtxFunBase::set_dataMtx(this);
  this->Next = new FunMtxReadEND();
}

FunMtxBase* FunMtxReadSTART::read(bool errSerious)
{
  return Next->read( errSerious );
}

std::list< FunMtxBase* >* FunReadFunBase_Mtx::read_para(tuint Pnmb, bool errSerious)
{
  std::list< FunMtxBase* >* plst = new std::list< FunMtxBase* >();
  try {
    do {
      plst->push_back(MtxFunList->read(errSerious));
      if ( reader->whatIsNextChar() == ',' ) {
	reader->getChar(',',errSerious);
      } else {
	break;
      }
    } while ( true );
    if (Pnmb > 0 && plst->size() != Pnmb) {
      ostringstream ssV;
      ssV << "Expected '" << Pnmb << "' parameters and not '" << plst->size() << "' parameters.";
      FlxError(errSerious,"FunReadFunBase_Mtx::read_para", ssV.str(), reader->getCurrentPos());
    };
  } catch (FlxException &e) {
    for (list< FunMtxBase* >::const_iterator i = plst->begin(); i!=plst->end(); ++i) {
      delete (*i);
    }
    delete plst;
  }
  return plst;
}

std::list< FlxString* >* FunReadFunBase_MtxConst::read_para(tuint Pnmb, bool errSerious)
{
  std::list< FlxString* >* plst = new std::list< FlxString* >();
  try {
    do {
      plst->push_back(new FlxString(false,false));
      if ( reader->whatIsNextChar() == ',' ) {
	reader->getChar(',',errSerious);
      } else {
	break;
      }
    } while ( true );
    if (Pnmb > 0 && plst->size() != Pnmb) {
      ostringstream ssV;
      ssV << "Expected '" << Pnmb << "' parameters and not '" << plst->size() << "' parameters.";
      FlxError(errSerious,"FunReadFunBase_MtxConst::read_para", ssV.str(), reader->getCurrentPos());
    }
  } catch (FlxException &e) {
    for (list< FlxString* >::const_iterator i = plst->begin(); i!=plst->end(); ++i) {
      delete (*i);
    }
    delete plst;
  }
  return plst;
}

FunBaseFun_Mtx::~FunBaseFun_Mtx()
{
  for (list< FunMtxBase* >::const_iterator i = ParaList_Mtx->begin(); i!=ParaList_Mtx->end(); ++i) {
    delete (*i);
  }
  delete ParaList_Mtx;
}

void FunBaseFun_Mtx::dependOn_RV(bVec& bv)
{
  FunBaseFun_multPara::dependOn_RV(bv);
  for (list< FunMtxBase* >::const_iterator i = ParaList_Mtx->begin(); i!=ParaList_Mtx->end(); ++i) {
    (*i)->dependOn_RV(bv);
  }
}

const bool FunBaseFun_Mtx::dependOn_Const(const tdouble* const thenumber)
{
  if (FunBaseFun_multPara::dependOn_Const(thenumber)) return true;
  for (list< FunMtxBase* >::const_iterator i = ParaList_Mtx->begin(); i!=ParaList_Mtx->end(); ++i) {
    if ((*i)->dependOn_Const(thenumber)) return true;
  }
  return false;
}

const bool FunBaseFun_Mtx::search_circref(FlxFunction* fcr)
{
  if ( FunBaseFun_multPara::search_circref(fcr) ) return true;
  for (list< FunMtxBase* >::const_iterator i = ParaList_Mtx->begin(); i!=ParaList_Mtx->end(); ++i) {
    if ( (*i)->search_circref(fcr) ) return true;
  }
  return false;
}

const std::string FunBaseFun_Mtx::write()
{
  std::string str1 = write_v() + "(";
  bool b1 = false;
  for (list< FunMtxBase* >::const_iterator i = ParaList_Mtx->begin(); i!=ParaList_Mtx->end(); ++i) {
    if (b1) str1 += ",";
    else b1 = true;
    str1 += (*i)->write();
  } 
  str1 += ")";
  return str1;
}

FunBaseFun_MtxConst::~FunBaseFun_MtxConst()
{
  for (list< FlxString* >::const_iterator i = ParaList_MtxName->begin(); i!=ParaList_MtxName->end(); ++i) {
    delete (*i);
  }
  delete ParaList_MtxName;
}

const std::string FunBaseFun_MtxConst::write()
{
  std::string str1 = write_v() + "(";
  bool b1 = false;
  for (list< FlxString* >::const_iterator i = ParaList_MtxName->begin(); i!=ParaList_MtxName->end(); ++i) {
    if (b1) str1 += ",";
    else b1 = true;
    str1 += (*i)->write();
  } 
  str1 += ")";
  return str1;
}

const bool FunBaseFun_MtxConst::search_circref(FlxFunction* fcr)
{
  if ( FunBaseFun_multPara::search_circref(fcr) ) return true;
  for (list< FlxString* >::const_iterator i = ParaList_MtxName->begin(); i!=ParaList_MtxName->end(); ++i) {
    if ( (*i)->search_circref(fcr) ) return true;
  }
  return false;
}

void FunBaseFun_MtxConst::dependOn_RV(bVec& bv)
{
  throw FlxException_NotImplemented("FunBaseFun_MtxConst::dependOn_RV");
}

const bool FunBaseFun_MtxConst::dependOn_Const(const tdouble*const thenumber)
{
  throw FlxException_NotImplemented("FunBaseFun_MtxConst::dependOn_Const");
}



