/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * flxfunction.h
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxmtxfun_ope_read.h"
#include "flxmtxfun_ope.h"

using namespace std;


void flxmtxfun_ope_insert(FunMtxReadSTART& FunList)
{
  FunList.insert( new FunMtxReadAdd() );		// 40
  FunList.insert( new FunMtxReadMult() );		// 50
  FunList.insert( new FunMtxReadValue() );		// 1000
  FunList.insert( new FunMtxReadWord() );		// 1010
  FunList.insert( new FunMtxReadBracket() );		// 1100
}

const std::string FunMtxConst::write()
{
  return strV;
}

SMtxBase* FunMtxConst::calc()
{
  SMtxBase* d1 = ConstMtxBox->get(strV);
  #if FLX_DEBUG
  if ( d1 != NULL ) {
  #endif
    return d1->copy();
  #if FLX_DEBUG
  } else {
    std::ostringstream ssV;
    ssV << "Matrix-constant '" << strV << "' has not yet been defined.";
    throw FlxException("FunMtxConst::calc", ssV.str() );
  }
  #endif
}

const std::string FunMtxAdd::write()
{ 
  std::string strV = "";
  if (child_1->precedence() > precedence()) {
    strV += "(" + child_1->write() + ")";
  } else {
    strV += child_1->write();
  }
  strV += "+";
  if (child_2->precedence() > precedence()) {
    strV += "(" + child_2->write() + ")";
  } else if ( !isPlus && child_2->precedence() == precedence() ) {
    strV += "(" + child_2->write() + ")";
  } else {
    strV += child_2->write();
  }
  return strV;
}

const std::string FunMtxMult::write()
{ 
  std::string strV = "";
  if (child_1->precedence() > precedence()) {
    strV += "(" + child_1->write() + ")";
  } else {
    strV += child_1->write();
  }
  strV += "*";
  if (child_2->precedence() > precedence()) {
    strV += "(" + child_2->write() + ")";
  } else {
    strV += child_2->write();
  }
  return strV;
}

SMtxBase* FunMtxMult::calc()
{ 
  SMtxBase* v1 = child_1->calc();
  SMtxBase* v2 = child_2->calc();
  if (v2->get_ncols() == 1 && v2->get_nrows() == 1 ) {
    *v1*=v2->operator()(0,0); 
    delete v2;
    return v1;
  } else if (v1->get_ncols() == 1 && v1->get_nrows() == 1 ) {
    *v2*=v1->operator()(0,0); 
    delete v1;
    return v2;
  } else if (v1->get_nrows() == 1 && v2->get_ncols() == 1 && v1->get_ncols() == v2->get_nrows() ) {
    SMtxVector* vec1 = dynamic_cast<SMtxVector*> (v1);
    SMtxVector* vec2 = dynamic_cast<SMtxVector*> (v2);
    if (vec1 == NULL || vec2 == NULL) {
      ostringstream ssV_2;
      ssV_2 << "Error.";
      throw FlxException("FunMtxMult::calc_1", ssV_2.str() );
    }
    tdouble t = dot(vec1->get_Vector(),vec2->get_Vector());
    delete v1;
    delete v2;
    return new SMtxScalar(t);
  } else if ( v1->get_ncols() == v2->get_nrows() ) {
    // Get Matrix 1
      FlxMtx_base* mtx1;
      SMtxMatrix* m1 = dynamic_cast<SMtxMatrix*> (v1);
      if (m1 == NULL) {
	SMtxVector* vec1 = dynamic_cast<SMtxVector*> (v1);
	if (vec1 == NULL) {
	  ostringstream ssV_2;
	  ssV_2 << "Error.";
	  throw FlxException("FunMtxMult::calc_3", ssV_2.str() );
	} else {
	  mtx1 = vec1->get_Matrix();
	  delete v1;
	  v1 = NULL;
	}
      } else {
	mtx1 = &(m1->get_Matrix());
      };
    // Get Matrix 2
      FlxMtx_base* mtx2;
      SMtxMatrix* m2 = dynamic_cast<SMtxMatrix*> (v2);
      if (m2 == NULL) {
	SMtxVector* vec2 = dynamic_cast<SMtxVector*> (v2);
	if (vec2 == NULL) {
	  ostringstream ssV_2;
	  ssV_2 << "Error.";
	  throw FlxException("FunMtxMult::calc_4", ssV_2.str() );
	} else {
	  mtx2 = vec2->get_Matrix();
	  delete v2;
	  v2 = NULL;
	}
      } else {
	mtx2 = &(m2->get_Matrix());
      };
    // Multiply
      FlxMtx mtxE = (*mtx1)*(*mtx2);
      FlxMtx* mtxEp = new FlxMtx(mtxE.nrows(),mtxE.ncols());
      *mtxEp = mtxE;
      if (v1 != NULL) delete v1; else delete mtx1;
      if (v2 != NULL) delete v2; else delete mtx2;
      return new SMtxMatrix(mtxEp);
  } else {
    ostringstream ssV_2;
    ssV_2 << "Matrices can not be multiplied.";
    throw FlxException("FunMtxMult::calc_2", ssV_2.str() );
  }
}

FunMtxBase* FunMtxReadMult::read(bool errSerious)
{
  FunMtxBase *theNode = Next->read( errSerious );
  try {
    while ( reader->whatIsNextChar() == '*' ) {
      reader->getChar();
      theNode = new FunMtxMult(theNode, Next->read( errSerious ) );
    }
  } catch (FlxException& e) {
    if ( theNode!=NULL) { delete theNode; theNode=NULL; }
    throw;
  }
  return theNode;
}

FunMtxBase* FunMtxReadWord::read(bool errSerious)
{
  if ( reader->getNextType() == ReadStream::STRING ) {
    string strV = reader->getWord(false,errSerious);
    
    // predefined functions
      FunReadMtxFunBase* fr = MtxFunBox->get(strV);
      if ( fr != NULL ) {
	reader->getChar('(',errSerious);
	FunMtxBase* fb = fr->read( errSerious );
	reader->getChar(')',errSerious);
	return fb;
      }

    // user-defined constants
      SMtxBase* d1 = ConstMtxBox->get(strV);
      if ( d1 != NULL ) {
	return new FunMtxConst( strV );
      }

    // Word is unknown ...
      ostringstream ssV_2;
      ssV_2 << "'" << strV << "' has not been defined yet.";
      FlxError(errSerious,"FunMtxReadWord::read_1", ssV_2.str(), reader->getCurrentPos()); return NULL; // DUMMY-RETURN
  }
  else return Next->read( errSerious );
}

FunMtxBase* FunMtxReadBracket::read( bool errSerious )
{
  if ( reader->whatIsNextChar() == '(' ) {
    reader->getChar();
    FunMtxBase *theNode = startLink->read( errSerious );
    if ( reader->getChar() == ')' ) {
      return theNode;
    } else {
      ostringstream ssV_2;
      ssV_2 << "Right parenthesis '(' expected.";
      FlxError(errSerious,"FunMtxReadBracket::read_1", ssV_2.str(), reader->getCurrentPos()); return NULL; // DUMMY-RETURN
    }
  } else {
    return Next->read( errSerious );
  }
}

FunMtxBase* FunMtxReadValue::read(bool errSerious)
{
  if ( reader->whatIsNextChar() == '{' ) {
    tuint nrows = 1, ncols = 1;
    reader->getChar('{',errSerious);
    std::vector<FunBase*>* vecV = new vector<FunBase*>();
    try {
      // obtain number of columns
	FunBase* fp = FunctionList->read(errSerious);
	vecV->push_back(fp);
	while ( reader->whatIsNextChar() == ',' ) {
	  reader->getChar(',',errSerious);
	  ncols++;
	  fp = FunctionList->read(errSerious);
	  vecV->push_back(fp);
	};
      // obtain number of rows
	while (reader->whatIsNextChar() == ';' ) {
	  reader->getChar(';',errSerious);
	  nrows++;
	  fp = FunctionList->read(errSerious);
	  vecV->push_back(fp);
	  for (tuint i = 1; i < ncols; ++i) {
	    reader->getChar(',',errSerious);
	    fp = FunctionList->read(errSerious);
	    vecV->push_back(fp);
	  };
	}
      reader->getChar('}',errSerious);
      return new SMtxValue(vecV,nrows,ncols);
    } catch ( FlxException &e) { 
      for (size_t i=0; i<vecV->size();++i) {
	delete vecV->operator[](i);
      }
      delete vecV;
      throw;
    }
  } else if (reader->nextCanBeNumber()) {
    std::vector<FunBase*>* vecV = new vector<FunBase*>();
    FunBase* fp = FunctionList->read(errSerious);
    vecV->push_back(fp);
    return new SMtxValue(vecV,1,1);
  } else {
    return Next->read( errSerious );
  }
}

FunMtxBase* FunMtxReadAdd::read(bool errSerious)
{
  FunMtxBase *theNode = Next->read( errSerious );
  bool IsPlus;
  try {
    while ( reader->whatIsNextChar() == '+' || reader->whatIsNextChar() == '-' ) {
      if ( reader->getChar() == '+' ) {
	IsPlus = true;
      }
      else {
	IsPlus = false;
      }
      theNode = new FunMtxAdd(theNode, Next->read( errSerious ), IsPlus);
    }
  } catch (FlxException& e) {
    if ( theNode!=NULL) { delete theNode; theNode=NULL; }
    throw;
  }
  return theNode;
}
