/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxmtxfun_ope_calc_H
#define fesslix_flxmtxfun_ope_calc_H

#include "flxmtxfun.h"

/**
* @brief arithmetic class: stores a constant from the ConstantBox
*/
class FunMtxConst : public FunMtxBase, public FlxMtxBoxBase {
  private:
    std::string strV;
  public:
    FunMtxConst (std::string strV) : strV(strV) {};
    SMtxBase* calc();
    const bool search_circref(FlxFunction* fcr) { return false;};
    const std::string write();
    const tuint precedence() { return 0; }
    void dependOn_RV(bVec& bv) { }
    const bool dependOn_Const(const tdouble* const thenumber) { return false; }
};

/**
* @brief arithmetic class: add (subtract) two values
*/
class FunMtxAdd : public FunMtxBaseOperat2 {
  private:
    const bool isPlus;
  public:
    FunMtxAdd (FunMtxBase *child1, FunMtxBase *child2, const bool IsPlus) : FunMtxBaseOperat2(child1, child2), isPlus(IsPlus) {};
    SMtxBase* calc() { 
      SMtxBase* v1 = child_1->calc();
      SMtxBase* v2 = child_2->calc();
      if ( isPlus ) *v1 += *v2;
      else *v1 += ((*v2).operator*=(-1));
      delete v2;
      return v1;
    }
    const std::string write();
    const tuint precedence() { return 12; }
};

/**
* @brief arithmetic class: multiply (divide) two values
*/
class FunMtxMult : public FunMtxBaseOperat2 {
  public:
    FunMtxMult (FunMtxBase *child1, FunMtxBase *child2) : FunMtxBaseOperat2(child1, child2) {};
    SMtxBase* calc ();
    const std::string write();
    const tuint precedence() { return 11; }
};


#endif // fesslix_flxmtxfun_ope_calc_H

