/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxobjcommon.h"

#include <fstream>
#include <iostream>

using namespace std;


void FlxCreateObjReaders_Common::createObjReaders(FlxObjectReadBox* objReadBox) {
  objReadBox->insert("var", new FlxObjReadVar());
  objReadBox->insert("fun", new FlxObjReadFun());
  objReadBox->insert("mtxconst", new FlxObjReadMtxConst());
  objReadBox->insert("calc", new FlxObjReadCalc());
  objReadBox->insert("print_hlegendrepoly", new FlxObjReadPrintHLegendrePoly());
  objReadBox->insert("print_hgegenbauerpoly", new FlxObjReadPrintHGegenbauerPoly());
  objReadBox->insert("echo", new FlxObjReadEcho());
  objReadBox->insert("mtxcalc", new FlxObjReadMtxCalc());
  objReadBox->insert("transform_mtx2octave", new FlxObjReadTransformMtx2Octave());
  objReadBox->insert("mtxconst_free", new FlxObjReadMtxConst_free());
  objReadBox->insert("mtxcoeff", new FlxObjReadMtxCoeff());
  objReadBox->insert("filestream", new FlxObjReadFileStream());
  objReadBox->insert("ifstream", new FlxObjReadInputFileStream());
  objReadBox->insert("ivstream", new FlxObjReadInputVectorStream());
  objReadBox->insert("ivstream_append", new FlxObjReadivstream_append());
  objReadBox->insert("istream_write", new FlxObjReadistream_write());
  objReadBox->insert("dstream", new FlxObjReadDistributorStream());
  objReadBox->insert("default", new FlxObjReadDefault());
  objReadBox->insert("if", new FlxObjReadIf());
  objReadBox->insert("while", new FlxObjReadWhile());
  objReadBox->insert("for", new FlxObjReadFor());
  objReadBox->insert("sfor", new FlxObjReadSFor());
  objReadBox->insert("timer", new FlxObjReadTimer());
  objReadBox->insert("time", new FlxObjReadTime());
  objReadBox->insert("funplot", new FlxObjReadFunPlot());
  objReadBox->insert("rnd", new FlxObjReadRnd());
  objReadBox->insert("end", new FlxObjReadEnd());
  objReadBox->insert("exit", new FlxObjReadExit());
  objReadBox->insert("interactive", new FlxObjReadInteractive());
  objReadBox->insert("read", new FlxObjReadReadFile());
  objReadBox->insert("streamset", new FlxObjReadStreamSet());
  objReadBox->insert("intervalcount", new FlxObjReadIntervalCount());
  objReadBox->insert("filter", new FlxObjReadFilter());
  objReadBox->insert("warranty", new FlxObjReadWarranty());
}

void FlxCreateObjReaders_Common::createFunReaders(FlxData* dataBox)
{
  FlxDataBase::set_data(dataBox);
  dataBox->FunBox.insert("isread", new FunReadFunISread() );
  dataBox->FunBox.insert("objexec", new FunReadObjExec() );
}

void FlxObjMtxCoeff::task()
{
  SMtxBase& mtx = *(data->ConstMtxBox.get(cname,true));
  tuint i = i_fun->cast2tuintW0(false);
  tuint j = j_fun->cast2tuintW0(false);
  tdouble d = c_fun->calc();
  if (i>=mtx.get_nrows() || j>=mtx.get_ncols()) {
    std::ostringstream ssV;
    ssV << "Index of coefficient (" << i << "," << j << ") are not within the matrix.";
    throw FlxException_NeglectInInteractive("FlxObjMtxCoeff::task", ssV.str() );
  }
  mtx.insert(i,j,d);
}

void FlxObjFun::task()
{
  #if FLX_DEBUG
    if (fun == NULL) {
      ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxObjFun::task", ssV.str() );
    }
  #endif
  data->FunBox.insert(cname, fun);
  fun = NULL;
  GlobalVar.slog(4) << "fun: Function '" << cname << "' declared." << std::endl;
}

FlxObjFun::~FlxObjFun()
{
  if ( fun != NULL ) delete fun;
}

FlxObjBase* FlxObjReadVar::read() {   
  string cname = reader->getWord(false,false);
  isdefined(cname, 'V',false);
  reader->getChar('=',false);
  FlxFunction* fi = NULL;
  FlxFunction* fV = NULL;
  try {
    fi = new FlxFunction(funReader,false);
    fV = data->VarBox.get(cname);
    if ( fi->search_circref(fV) ) {
      ostringstream ssV;
      ssV << "Circular reference in '" << cname << "'.";
      throw FlxException("FlxObjReadVar::read_1", ssV.str(), reader->getCurrentPos() );
    }
  } catch ( FlxException &e) {
    if (fi!=NULL) delete fi;
    if (fV!=NULL) delete fV;
    throw;
  }
  read_optionalPara(false);
  FlxObjBase* obj1 = new FlxObjVar( get_doLog(), cname, fi );
  data->VarBox.declareV(cname);
  return obj1;
}

FlxObjBase* FlxObjReadMtxConst::read()
{
  const string cname = reader->getWord(false,false);
  return read(cname);
}

FlxObjBase* FlxObjReadMtxConst::read(const string cname)
{
  reader->getChar('=',false);
  FlxMtxFun* mf = new FlxMtxFun(mtxReader,false);
  try {
  read_optionalPara(false);
  isdefined(cname,'M',false);
  data->ConstMtxBox.declareC(cname);
  return new FlxObjMtxConst( get_doLog(), cname, mf );;
  } catch (FlxException &e) {
    delete mf;
    throw;
  }
}

FlxObjBase* FlxObjReadMtxCoeff::read()
{
  string cname = reader->getWord(false,false);
  reader->getChar('(',false);
  FlxFunction* i_fun = new FlxFunction(funReader,false);
  reader->getChar(',',false);
  FlxFunction* j_fun=NULL; FlxFunction* c_fun=NULL;
  try {
    j_fun = new FlxFunction(funReader,false);
    reader->getChar(')',false);
    reader->getChar('=',false);
    c_fun = new FlxFunction(funReader,false);
    read_optionalPara(false);
    return new FlxObjMtxCoeff(get_doLog(),cname,i_fun,j_fun,c_fun);
  } catch (FlxException &e) {
    delete i_fun;
    if (j_fun!=NULL) delete j_fun;
    if (c_fun!=NULL) delete c_fun;
    throw;
  }
}

FlxObjBase* FlxObjReadCalc::read() {
  FlxFunction* f1 = new FlxFunction(funReader,false);
  try {
    read_optionalPara(false);
    FlxObjBase* obj1 = new FlxObjCalc(get_doLog(), f1, get_stream(), get_checkTOL() );
    return obj1;
  } catch (FlxException &e) {
    delete f1;
    throw;
  }
}

FlxObjReadPrintHLegendrePoly::FlxObjReadPrintHLegendrePoly()
{
  // deriv
    AllDefParaBox->insert(new FlxOptionalParaBool(false,"hlegendre::deriv"));
    ParaBox.insert("deriv", "hlegendre::deriv" );
}

FlxObjBase* FlxObjReadPrintHLegendrePoly::read()
{
  reader->getChar('(');
  FlxFunction* f1 = new FlxFunction(funReader,false);
  try {
    reader->getChar(')');
    read_optionalPara(false);
    FlxObjBase* obj1 = new FlxObjPrintHLegendrePoly(get_doLog(), f1, get_stream(), get_verbose(), get_optPara_bool("deriv") );
    return obj1;
  } catch (FlxException &e) {
    delete f1;
    throw;
  }
}

FlxObjBase* FlxObjReadPrintHGegenbauerPoly::read()
{
  reader->getChar('(');
  FlxFunction* f1 = new FlxFunction(funReader,false);
  try {
    reader->getChar(')');
    read_optionalPara(false);
    FlxObjBase* obj1 = new FlxObjPrintHGegenbauerPoly(get_doLog(), f1, get_stream(), get_verbose() );
    return obj1;
  } catch (FlxException &e) {
    delete f1;
    throw;
  }
}

FlxObjBase* FlxObjReadEcho::read()
{
  FlxString* strV = NULL;
  try {
    strV = new FlxString(true,false);
    read_optionalPara(false);
  } catch (FlxException &e) {
    if (strV!=NULL) delete strV;
    throw;
  }
  return new FlxObjEcho(get_doLog(), strV, get_stream() );
}

FlxObjBase* FlxObjReadWarranty::read()
{	
  read_optionalPara(false);
  return new FlxObjWarranty(get_doLog(), get_stream() );
}

FlxObjBase* FlxObjReadMtxCalc::read()
{
  FlxMtxFun* f1 = NULL;
  try {
    f1 = new FlxMtxFun(mtxReader,false);
    read_optionalPara(false);
    FlxObjBase* obj1 = new FlxObjMtxCalc(get_doLog(), f1, get_stream() );
    return obj1;
  } catch ( FlxException &e) {
    if (f1!=NULL) delete f1;
    throw;
  }
}

FlxObjBase* FlxObjReadTransformMtx2Octave::read()
{
  FlxString* funStr = new FlxString(false,false);
  try {
    read_optionalPara(false);
    return new FlxObjTransformMtx2Octave(get_doLog(), funStr, get_stream() );
  } catch ( FlxException &e) {
    if (funStr) delete funStr;
    throw;
  }
}

FlxObjBase* FlxObjReadMtxConst_free::read()
{
  FlxString* funStr = new FlxString(false,false);
  try {
    read_optionalPara(false);
    return new FlxObjMtxConst_free(get_doLog(), funStr, get_stream() );
  } catch ( FlxException &e) {
    if (funStr) delete funStr;
    throw;
  }
}

FlxObjBase* FlxObjReadStreamSet::read()
{
  FlxFunction* f1=NULL;
  try {
    std::string str1 = reader->getWord(true,false);
    reader->getChar('=',false);
    if (str1 == "prec") {
      f1 = new FlxFunction(funReader,false);
    } else {
      ostringstream ssV;
      ssV << "Unknown property '" << str1 << "'.";
      throw FlxException_NeglectInInteractive("FlxObjReadStreamSet::read_1", ssV.str(), reader->getCurrentPos() );
    }
    read_optionalPara(false);
    if (str1 == "prec") {
      return new FlxObjStreamSet_Prec(get_doLog(),f1,get_stream());
    } else {
      ostringstream ssV;
      ssV << "Unknown property '" << str1 << "'.";
      throw FlxException_NeglectInInteractive("FlxObjReadStreamSet::read_11", ssV.str(), reader->getCurrentPos() );
    }
  } catch (FlxException &e) {
    if (f1!=NULL) delete f1;
    throw;
  }
}


FlxObjBase* FlxObjReadInteractive::read()
{
  read_optionalPara(false);
  return new FlxObjInteractive(get_doLog(), get_stream());
}

FlxObjBase* FlxObjReadTimer::read() {
  string task = reader->getWord(false,false);
  string timer = reader->getWord(false,false);
  transform(task.begin(), task.end(), task.begin(), (int(*)(int)) std::tolower);
  if ( task == "start" ) {
    read_optionalPara(false);
    return new FlxObjTimerStart(get_doLog(),timer);
  } else if ( task == "stop" ) {
    read_optionalPara(false);
    return new FlxObjTimerStop(get_doLog(),timer);
  } else if ( task == "print" ) {
    read_optionalPara(false);
    return new FlxObjTimerPrint(get_doLog(),timer, get_stream());
  } else if ( task == "define" ) {
    read_optionalPara(false);
    return new FlxObjTimerDefine(get_doLog(),timer);
  } else if ( task == "delete" ) {
    read_optionalPara(false);
    return new FlxObjTimerDelete(get_doLog(),timer);
  } else {
    ostringstream ssV;
    ssV << "Unknown action '" << task << "'.";
    throw FlxException_NeglectInInteractive("FlxObjReadTimer::read_1", ssV.str(), reader->getCurrentPos() );
  }
}

FlxObjBase* FlxObjReadFunPlot::read()
{
  std::vector<FlxFunction*> V;
  std::vector<FlxMtxFun*> M;
  std::vector<bool> B;
  bool b=false;
  try {
    do {
      if (b) reader->getChar(',',false);
      else b=true;
      if (reader->whatIsNextChar()=='{') {
	reader->getChar('{',false);
	FlxMtxFun* f = new FlxMtxFun(mtxReader,false);
	M.push_back(f);
	B.push_back(false);
	reader->getChar('}',false);
      } else {
	FlxFunction* f = new FlxFunction(funReader,false);
	V.push_back(f);
	B.push_back(true);
      }
    } while (reader->whatIsNextChar() == ',');
    read_optionalPara(false);
    return new FlxObjFunPlot(get_doLog(),B,V,M,get_stream(),get_checkTOL());
  } catch ( FlxException &e) {
    for (size_t i = 0; i < V.size(); ++i) {
      delete V[i];
    }
    for (size_t i = 0; i < M.size(); ++i) {
      delete M[i];
    }
    throw;
  }
}

FlxObjReadRnd::FlxObjReadRnd()
{
  // ppdim
    std::vector<FunBase*>* vecV = new std::vector<FunBase*>;
    vecV->push_back(new FunNumber(10));
    vecV->push_back(new FunNumber(10));
    vecV->push_back(new FunNumber(10));
    FlxMtxFun* mtxvalue = new FlxMtxFun( new SMtxValue(vecV,3,1) );
    AllDefParaBox->insert(new FlxOptionalParaMtxFun(mtxvalue,"flxfem::elplot::pointsperdim"));
    ParaBox.insert("ppdim", "flxfem::elplot::pointsperdim" );
  // errmeasure
    AllDefParaBox->insert(new FlxOptionalParaFun(1,"rfplot::errmeasure"));
    ParaBox.insert("errmeasure", "rfplot::errmeasure" );  
}

FlxObjBase* FlxObjReadRnd::read()
{
  std::string keyS = reader->getWord(true,false);
  
  // generate samples
  if (keyS == "smp") {
    read_optionalPara(false);
    return new FlxObjRndSmp(get_doLog(),get_stream(),get_verbose());
    // generate samples
  } else if (keyS == "print") {
    read_optionalPara(false);
    return new FlxObjRndSmp(get_doLog(),get_stream(),get_verbose(),false);
  // define new user defined managed random varaible
  } else if (keyS == "new") {
    FlxFunction* index = new FlxFunction(funReader,false);
    try {
      std::string rvStr = reader->getWord(true,false);
      FlxRnd_Generator_base* rndG = FlxRnd_Generator_base::createRV(FlxRnd_Generator_base::get_rvt(rvStr,false));
      read_optionalPara(false);
      return new FlxObjRndCreate(get_doLog(),index, rndG);
    } catch ( FlxException &e) {
      delete index;
      throw;
    }
  // influence random variables (by set)
  } else if (keyS == "set") {
    std::string strV = reader->getWord(true,false);
    bool origSpace;
    if ( strV == "x") {
      origSpace = true;
    } else if ( strV == "y") {
      origSpace = false;
    } else {
      ostringstream ssV;
      ssV << "Unknown keyword '" << keyS << "'.";
      throw FlxException_NeglectInInteractive("FlxObjReadRnd::read_1", ssV.str(), reader->getCurrentPos() );
    }
    FlxMtxFun* vec = new FlxMtxFun(mtxReader,false);
    read_optionalPara(false);
    return new FlxObjRndSet(get_doLog(),vec, origSpace);
  // influence random variables (by track)
  } else if (keyS == "track") {
    std::string keyS2 = reader->getWord(true,false);
    if (keyS2 == "record") {
      FlxFunction* numb = new FlxFunction(funReader,false);
      try {
	read_optionalPara(false);
	return new FlxObjRndTrackRecord(get_doLog(),numb,get_stream(), get_verbose());
      } catch (FlxException &e) {
	delete numb;
	throw;
      }
    } else if ( keyS2 == "replay" ) {
      FlxString* isname = new FlxString(false,false);
      read_optionalPara(false);
      return new FlxObjRndTrackReplay(get_doLog(),isname);
    } else if ( keyS2 == "stop" ) {
      read_optionalPara(false);
      return new FlxObjRndTrackStop(get_doLog());
    } else {
      ostringstream ssV;
      ssV << "Unknown keyword '" << keyS2 << "'.";
      throw FlxException_NeglectInInteractive("FlxObjReadRnd::read_2", ssV.str(), reader->getCurrentPos() );
    }
  // influence random variables (by seed)
  } else if (keyS == "seed") {
    FlxFunction* seedv = NULL; FlxFunction* icv = NULL;
    try {
      seedv = new FlxFunction(funReader,false);
      if (reader->whatIsNextChar() == ';') {
	icv = new FlxFunction(new FunNumber(0.0));
      } else {
	icv = new FlxFunction(funReader,false);
      }
      read_optionalPara(false);
      return new FlxObjRndSeed(get_doLog(),seedv, icv);
    } catch (FlxException &e) {
      if (seedv!=NULL) delete seedv;
      if (icv!=NULL) delete icv;
      throw;
    }
  } else if (keyS == "corr") {
    if (reader->whatIsNextString(6,true)=="create") {
      reader->getWord("create",false);
      read_optionalPara(false);
      return new FlxObjRndCorr_Create(get_doLog());
    } else {
      FlxFunction* rv1=NULL; FlxFunction* rv2=NULL; FlxFunction* rho=NULL;
      try {
	reader->getWord("rv1",false);
	rv1 = new FlxFunction(funReader,false);
	reader->getWord("rv2",false);
	rv2 = new FlxFunction(funReader,false);
	reader->getWord("rho",false);
	rho = new FlxFunction(funReader,false);
	read_optionalPara(false);
	return new FlxObjRndCorr_Set(get_doLog(),rv1,rv2,rho);
      } catch ( FlxException &e) {
	if (rv1!=NULL) delete rv1;
	if (rv2!=NULL) delete rv2;
	if (rho!=NULL) delete rho;
	throw;
      }
    }
  // deal with random fields
  } else if (keyS == "rf") {
    std::string keyS2 = reader->getWord(true,false);
    // define new user defined managed random field
    if (keyS2 == "new") {
      FlxFunction* index = new FlxFunction(funReader,false);
      try {
	std::string rfStr = reader->getWord(true,false);
	FlxRF_Generator_base* rndG = FlxRF_Generator_base::createRF(FlxRF_Generator_base::get_rft(rfStr,false));
	read_optionalPara(false);
	return new FlxObjRndRfCreate(get_doLog(),index, rndG);
      } catch ( FlxException &e) {
	delete index;
	throw;
      }
    } else if (keyS2 == "elset") {
      FlxFunction* index = new FlxFunction(funReader,false);
      FlxFunction* elindex = NULL;
      FlxKLelementProp_base* KLelProp = NULL;
      try {
	const flxDomain_base::geoel_type gelt = flxDomain_base::get_geoel_type(reader->getWord(false,false),false);
	elindex = new FlxFunction(funReader,false);
	switch (gelt) {
	  case flxDomain_base::geoel_edge:
	    KLelProp = new FlxKLelementProp_1D(false);
	    break;
	  case flxDomain_base::geoel_face:
	    KLelProp = new FlxKLelementProp_2D(false);
	    break;
	  case flxDomain_base::geoel_box:
	    KLelProp = new FlxKLelementProp_3D(false);
	    break;
	  default:
	    throw FlxException_NotImplemented("FlxObjReadRnd::read_3");
	}
	read_optionalPara(false);
	return new FlxObjRndRfElSet(get_doLog(),index, gelt,elindex, KLelProp);
      } catch ( FlxException &e) {
	delete index;
	if (elindex!=NULL) delete elindex;
	if (KLelProp!=NULL) delete KLelProp;
	throw;
      }
    } else if (keyS2 == "solve") {
      FlxFunction* index = new FlxFunction(funReader,false);
      FlxRFsolveProp* RFsolveProp = NULL;
      try {
	RFsolveProp = new FlxRFsolveProp(false);
	read_optionalPara(false);
	return new FlxObjRndRfSolve(get_doLog(),index,RFsolveProp,get_stream(),get_verbose());
      } catch ( FlxException &e) {
	delete index;
	if (RFsolveProp!=NULL) delete RFsolveProp;
	throw;
      }
    } else if (keyS2 == "err") {
      FlxFunction* index = new FlxFunction(funReader,false);
      try {
	read_optionalPara(false);
	return new FlxObjRndRfErr(get_doLog(),index,get_optPara_FlxFunction("errmeasure"),get_stream(),get_verbose());
      } catch (FlxException &e) {
	delete index;
	throw;
      }
    } else if (keyS2 == "plot") { 
      FlxFunction* index = new FlxFunction(funReader,false);
      try {
	read_optionalPara(false);
	return new FlxObjRndRfPlot(get_doLog(),index,get_optPara_FlxMtxFun("ppdim"),get_optPara_FlxFunction("errmeasure"),get_stream(),get_verbose());
      } catch (FlxException &e) {
	delete index;
	throw;
      }
    } else if (keyS2 == "plotcov") {
      FlxFunction* index = new FlxFunction(funReader,false);
      FlxFunPoint* p = NULL;
      try {
	p = new FlxFunPoint(reader, funReader,false);
	read_optionalPara(false);
	return new FlxObjRndRfPlotCov(get_doLog(),index,p,get_optPara_FlxMtxFun("ppdim"),get_stream(),get_verbose());
      } catch (FlxException &e) {
	delete index;
	if (p!=NULL) delete p;
	throw;
      }
    } else {
      ostringstream ssV;
      ssV << "Unknown keyword '" << keyS2 << "'.";
      throw FlxException_NeglectInInteractive("FlxObjReadRnd::read_4", ssV.str(), reader->getCurrentPos() );
    }
  } else {
    ostringstream ssV;
    ssV << "Unknown keyword '" << keyS << "'.";
    throw FlxException_NeglectInInteractive("FlxObjReadRnd::read_5", ssV.str(), reader->getCurrentPos() );
  }
}

FlxObjBase* FlxObjReadIntervalCount::read()
{
  FlxObjBase* Ivblock = NULL;
  reader->getChar('(',false);
  FlxFunction* Inumb = new FlxFunction(funReader,false);
  try {
    reader->getChar(')',false);
    Ivblock = read_block(true,false);
    read_optionalPara(false);
    return new FlxObjIntervalCount(get_doLog(),Inumb, Ivblock);
  } catch (FlxException &e) {
    delete Inumb;
    if (Ivblock!=NULL) delete Ivblock;
    throw;
  }
}

FlxObjBase* FlxObjReadFilter::read()
{
  reader->getChar('(',false);
  
  if (reader->getNextType() != ReadStream::STRING) {
    ostringstream ssV_2;
    ssV_2 << "Name of the 'const' variable to use expected.";
    throw FlxException_NeglectInInteractive("FunReadMtxFunFilter::read_1", ssV_2.str(), reader->getCurrentPos());
  }
  tdouble* cv = data->ConstantBox.get(reader->getWord(true),true);
  reader->getChar(';',false);
  FlxMtxFun* seqMtx = new FlxMtxFun(mtxReader,false);
  FlxObjBase* block = NULL;
  try {
    reader->getChar(')',false);
    block = read_block(true,false);
    read_optionalPara(false);
    return new FlxObjFilter(get_doLog(),cv, seqMtx, block);
  } catch (FlxException &e) {
    delete seqMtx;
    if (block!=NULL) delete block;
    throw;
  }
}
  
void FlxObjTimerPrint::task()
{
  if ( data->TimerBox.get(tname)->IsRunning() ) {
    ostringstream ssV;
    ssV << "Timer '" << tname << "' is running.";
    throw FlxException("FlxObjTimerPrint::task_1", ssV.str(), "To output information out of a timer you have to stop it first.");
  }
  clock_t ct = data->TimerBox.get(tname)->get_ticks();
  std::string tstr = GlobalVar.Double2String(tdouble(ct)/CLOCKS_PER_SEC);
  sout() << "timer(" << tname << ") = " << tstr << "sec. " << endl; //[" << ct << "ticks]" << endl;
  GlobalVar.slog(4) << "timer: timer '" << tname << "' has a value of t=" << tstr << "s." << std::endl;
}

void FlxObjTime::task()
{
  FlxTimer t;
  t.start();
  InternListTime->exec();
  t.stop();
  sout() << "time = " << GlobalVar.Double2String(tdouble(t.get_ticks())/CLOCKS_PER_SEC) << "sec. " << std::endl;
  if (GlobalVar.check_logNOTcout()) {
    GlobalVar.slog(4) << "time: measured time: " << tdouble(t.get_ticks())/CLOCKS_PER_SEC << "s." << std::endl;
  }
}

FlxObjTime::~FlxObjTime()
{
  delete InternListTime;
}

void FlxObjFunPlot::task()
{
  std::vector<FlxFunction*>::iterator j = V.begin();
  std::vector<FlxMtxFun*>::iterator k = M.begin();
  
  for (std::vector<bool>::iterator i = B.begin(); i != B.end(); ++i) {
    if (*i == true) {
      sout() << GlobalVar.Double2String((*j)->calc(),checkTOL) << "\t";
      ++j;
    } else {
      SMtxBase* mb = (*k)->calc();
      tuint l,m;
      for (l = 0; l < mb->get_nrows(); ++l) {
	for (m = 0; m < mb->get_ncols(); ++m) {
	  sout() << GlobalVar.Double2String(mb->operator()(l,m),checkTOL) << "\t";
	}
      }
      ++k;
      delete mb;
    }
  }
  sout() << std::endl;
}

FlxObjFunPlot::~FlxObjFunPlot() {
  for (std::vector<FlxFunction*>::iterator i = V.begin(); i != V.end(); ++i) {
    delete (*i);
  }
  for (std::vector<FlxMtxFun*>::iterator i = M.begin(); i != M.end(); ++i) {
    delete (*i);
  }
}

FlxObjRndCorr_Set::~FlxObjRndCorr_Set()
{
  delete rv1;
  delete rv2;
  delete rho;
}

void FlxObjRndCorr_Set::task()
{
  tuint i1 = rv1->cast2tuintW0(false);
  tuint i2 = rv2->cast2tuintW0(false);
  tdouble d = rho->calc();
  data->RndBox.corr_set(i1,i2,d);
  GlobalVar.slog(4) << "Random variable: set correlation of rv(" << i1 << ") and rv(" << i2 << ") to " << GlobalVar.Double2String(d) << "." << std::endl;
}

void FlxObjRndCorr_Create::task()
{
  data->RndBox.corr_create();
  GlobalVar.slog(4) << "Random variable: created Nataf transformation." << std::endl;
}

void FlxObjRndSmp::task() {
  if (genSMP) {
    data->RndBox.gen_smp();
    GlobalVar.slog(4) << "Random variable: samples generated." << std::endl;
  } else {
    GlobalVar.slog(4) << "Random variable: output current realization" << std::endl;
  }
  ostringstream ssV;
  ssV << " Number of Random Variables: " << data->RndBox.get_NumbRnd() << ";" << std::endl;
  const tVec& xR = data->RndBox.get_x_Vec();
  const tVec& yR = data->RndBox.get_y_Vec();
  for (size_t i = 0; i < data->RndBox.get_NumbRnd(); ++i) {
    ssV << "  RV" << i+1 << ": x=" << xR[i] << "; ";
    ssV << " y=" << yR[i] << "; " << std::endl;
  }
  if (is_verbose()||!genSMP) {
    sout() << ssV.str();
  }
  if ((is_verbose()||!genSMP) && (GlobalVar.check_logNOTcout() )) {
    GlobalVar.slog(4) << ssV.str();
  }
}

void FlxObjRndTrackRecord::task()
{
  tuint nmb = numb->cast2int();
  if ( numb == 0) return;
  sout() << rv_normal();
  for (tuint i = 1; i < nmb; ++i) {
    sout() << std::endl << rv_normal();
  }
  sout() << std::endl;
}

void FlxObjRndTrackReplay::task()
{
  data->RndBox.RndCreator.replay_start( &data->IstreamBox.get(isname->eval_word()) );
}

FlxObjCalc::FlxObjCalc( const bool dolog, FlxFunction* funV, std::string ostreamV, const bool checkTOL )
: FlxObjOutputBase(dolog,ostreamV,false,checkTOL), fun(funV), ansptr(data->ConstantBox.get("ans",true))
{

}

void FlxObjPrintHLegendrePoly::task()
{
  tuint i = fun->cast2tuint(false);
  flxLegendrePolyH LPH;
  LPH.resize(i);
  
  if (is_verbose()) {
    if (deriv) {
      LPH.code_dPn(i,sout());
    } else {
      LPH.code_Pn(i,sout());
    }
  } else {
    sout() << "  " << i;
    if (i==1) sout() << "st";
    else if (i==2) sout() << "nd";
    else if (i==3) sout() << "rd";
    else sout() << "th";
    sout() << ": ";
    LPH.print_Pn(i-1,sout());
    sout() << std::endl;
  }
}

void FlxObjPrintHGegenbauerPoly::task()
{
  const tuint i = fun->cast2tuint(false);
  if (i<2) return;
  flxGegenbauerPolyH GPH(i-2);
  GPH.code_hGBP(i,sout());
}

void FlxObjWarranty::task()
{
  sout() << std::endl;
  sout() << "Fesslix Copyright (C) 2010-2011 Wolfgang Betz" << std::endl << std::endl;
  sout() << "This program is free software; you can redistribute it and/or modify" << std::endl;
  sout() << "it under the terms of the GNU General Public License as published by" << std::endl;
  sout() << "the Free Software Foundation; either version 3 of the License, or" << std::endl;
  sout() << "(at your option) any later version." << std::endl << std::endl;
  sout() << "This program is distributed in the hope that it will be useful," << std::endl;
  sout() << "but WITHOUT ANY WARRANTY; without even the implied warranty of" << std::endl;
  sout() << "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the" << std::endl;
  sout() << "GNU General Public License for more details." << std::endl  << std::endl;
  sout() << "You should have received a copy of the GNU General Public License" << std::endl;
  sout() << "along with this program.  If not, see <http://www.gnu.org/licenses/>." << std::endl  << std::endl; 
}

void FlxObjMtxCalc::task()
{
  sout() << fun->write() << " = " << std::endl;
  SMtxBase* mp = fun->calc();
  sout() << *mp << "(" << mp->get_nrows() << "," << mp->get_ncols() << ")" << std::endl;
  delete mp;
}

void FlxObjTransformMtx2Octave::task()
{
  const std::string mtxName = funStr->eval_word(true);
  SMtxBase* mp = data->ConstMtxBox.get(mtxName);
  sout() << "[";
  const tuint Nrows = mp->get_nrows();
  const tuint Ncols = mp->get_ncols();
  for (tuint i=0;i<Nrows;++i) {
    if (i>0) sout() << ";";
    for (tuint j=0;j<Ncols;++j) {
      sout() << " " << GlobalVar.Double2String(mp->operator()(i,j));
    }
  }
  sout() << "]" << std::endl;
}

void FlxObjMtxConst_free::task()
{
  const std::string mtxName = funStr->eval_word(true);
  data->ConstMtxBox.freeC(mtxName);
}

void FlxObjStreamSet_Prec::task()
{
  int i = fun->cast2tuintW0(false);
  sout().precision(i);
}

void FlxObjRndCreate::task() {
  tuint ui = index->cast2tuintW0(false);
  data->RndBox.insert(ui,rG->generate_RV(false));
}

void FlxObjRndRfCreate::task()
{
  tuint ui = index->cast2tuintW0(false);
  data->RFbox.insert(ui,rG->generate_RF(false));
}

void FlxObjRndRfElSet::task()
{
  const tuint rfindex = index->cast2tuintW0(false);
  const tuint elindexI = elindex->cast2tuintW0(false);
  flxDomain_base* db = NULL;
  flxElement_KL_base* kb = NULL;
  std::ostringstream ssV;
  switch (typeEL) {
    case flxDomain_base::geoel_edge:
    {
      flxEdge* et = data->FEM->EdgeList.get(elindexI,true);
      FlxKLelementProp_1D* KLelProp_1D = dynamic_cast<FlxKLelementProp_1D*>(KLelProp);
      #if FLX_DEBUG
	if (KLelProp_1D==NULL) throw FlxException_Crude("FlxObjRndRfElSet::task_1");
      #endif
      kb = et->create_KL_element(KLelProp_1D);
      db = et;
      ssV << "edge";
      break;
    }
    case flxDomain_base::geoel_face:
    {
      flxFace* et = data->FEM->FaceList.get(elindexI,true);
      FlxKLelementProp_2D* KLelProp_2D = dynamic_cast<FlxKLelementProp_2D*>(KLelProp);
      #if FLX_DEBUG
	if (KLelProp_2D==NULL) throw FlxException_Crude("FlxObjRndRfElSet::task_2");
      #endif
      kb = et->create_KL_element(KLelProp_2D);
      db = et;
      ssV << "face";
      break;
    }
    case flxDomain_base::geoel_box:
    {
      flxBox* et = data->FEM->BoxList.get(elindexI,true);
      FlxKLelementProp_3D* KLelProp_3D = dynamic_cast<FlxKLelementProp_3D*>(KLelProp);
      #if FLX_DEBUG
	if (KLelProp_3D==NULL) throw FlxException_Crude("FlxObjRndRfElSet::task_3");
      #endif
      kb = et->create_KL_element(KLelProp_3D);
      db = et;
      ssV << "box";
      break;
    }
    default:
      throw FlxException_Crude("FlxObjRndRfElSet::task_4");
  }
  try {
    ssV << " " << elindexI;
    FlxRandomField_base* rfb = data->RFbox.get(rfindex,false);
    rfb->set_element(db,kb,true,rfindex,ssV.str());
  } catch (FlxException& e) {
    if (kb) delete kb;
    throw;
  }
}

void FlxObjRndRfSolve::task()
{
  const tuint idx = index->cast2tuintW0(false);
  FlxRandomField_base* rf = data->RFbox.get(idx,false);
  const tuint DOF = rf->solve(RFsolveProp,sout());
  GlobalVar.slog(3) << "Random field: assembled random field with index '" << idx << "'. (" << DOF << " DOFs)" << std::endl;
  data->ConstantBox.insert("g_solver_ndof",tdouble(DOF));
  if (RFsolveProp->get_calcerr()) GlobalVar.slog(3) << "  mean-error(1)=" << GlobalVar.Double2String(rf->get_domain_error(1)) << std::endl;
}

void FlxObjRndRfErr::task()
{
  const tuint idx = index->cast2tuintW0(false);
  const int erM = errMeasure->cast2int();
  FlxRandomField_base* rf = data->RFbox.get(idx,false);
  sout() << "RF-error(field=" << GlobalVar.Double2String(idx) << ",measure=" << GlobalVar.Double2String(erM) << ") = " ;
  sout() << GlobalVar.Double2String(rf->get_domain_error(erM)) << std::endl;
}

void FlxObjRndRfPlot::task()
{
  const tuint idx = index->cast2tuintW0(false);
  FlxRandomField_base* rf = data->RFbox.get(idx,false);
  const int erM = errMeasure->cast2int();
  rf->plot(sout(),ppdim,erM);
}

void FlxObjRndRfPlotCov::task()
{
  const tuint idx = index->cast2tuintW0(false);
  FlxRandomField_base* rf = data->RFbox.get(idx,false);
  flxPoint pp =  p->calc();
  rf->plotcov(sout(),pp,ppdim);
}

void FlxObjRndSet::task()
{
  tVec v = vec->calcVec();
  if (v.size() != data->RndBox.get_NumbRnd()) {
    std::ostringstream ssV;
    ssV << "Vector sizes do not match.";
    throw FlxException("FlxObjRndSet::task", ssV.str() );
  }
  GlobalVar.slog(3) << "Random variable: set value of RVs in ";
  if (origSpace) {
    data->RndBox.set_x_Vec(v);
    GlobalVar.slog(3) << "original space: ";
  } else {
    data->RndBox.set_y_Vec(v);
    GlobalVar.slog(3) << "standard normal space: ";
  }
  GlobalVar.slog(3) << v << ";" << std::endl;
}

void FlxObjRndSeed::task()
{
  rv_initialize(false, true, seedv->cast2tuint(false), icv->cast2tuintW0(false));
}

void FlxObjIntervalCount::task()
{
  if (funNp != NULL) {
    Np = funNp->cast2tuint(false);
    delete funNp;
    funNp = NULL;
  }
  countI++;
  if (countI >= Np) {
    InternList->exec();
    countI = 0;
  }
}

FlxObjIntervalCount::~FlxObjIntervalCount()
{
  if (funNp != NULL) delete funNp;
  delete InternList;
}

FlxObjFilter::~FlxObjFilter()
{
  delete seqMtx;
  delete block;
}

void FlxObjFilter::task()
{
  SMtxBase* mtx = seqMtx->calc();
  tuint nr = mtx->get_nrows();
  tuint nc = mtx->get_ncols();
  tuint j,i;
  for (j = 0; j < nc; ++j) {
    for (i = 0; i < nr; ++i) {
      *cv = mtx->operator()(i,j);
      block->exec();
    }
  }
  delete mtx;
}

void FlxObjFileStream::task() {
  std::ofstream *theStream;
  std::string fn = filename->eval();
  if ( trunc ) {
    theStream = new std::ofstream(fn.c_str());
  } else {
    theStream = new std::ofstream(fn.c_str(), ios_base::out | ios_base::app); 
  }
  if ( ! theStream->is_open() || theStream == NULL ) {
    std::ostringstream ssV;
    ssV << "File (" << fn << ") could not be opened.";
    delete theStream;
    throw FlxException("FlxObjFileStream::task_1", ssV.str() );
  }
  data->OstreamBox.insert(streamname, theStream);
  GlobalVar.slog(4) << "filestream: stream '" << streamname << "' is directed into file '" << fn << "'." << std::endl;
}

void FlxObjInputFileStream::task()
{
  std::string fns = filename->eval();
  const char* test = fns.c_str();
  ReadStream* rs=NULL;
  FlxIstream_file* istrm=NULL;
  try {
    rs = new ReadStream(test);
    istrm = new FlxIstream_file(streamname,rs,erreof,bs->cast2tuint(false),Cnumb,Cvec);
    rs = NULL;
    data->IstreamBox.insert(streamname, istrm, false);
    GlobalVar.slog(3) << "ifstream: file '" << fns << "' directed to input stream '" << streamname << "'." << std::endl;
  } catch (FlxException &e) {
    if (rs!=NULL) delete rs;
    if (istrm!=NULL) delete istrm;
    throw;
  }
}

void FlxObjInputVectorStream::task()
{
  FlxIstream *istrm_input = NULL;
  std::string istrm_input_str = "";
  if (inputStreamName!=NULL) {
    istrm_input_str = inputStreamName->eval_word(true);
  }
  if (istrm_input_str!="") {
    istrm_input = &(data->IstreamBox.get(istrm_input_str));
  }
  FlxIstream_vector* istrm=NULL;
  try {
    istrm = new FlxIstream_vector(streamname,istrm_input,erreof,Nreserve->cast2tnlong(false));
    data->IstreamBox.insert(streamname, istrm);
    GlobalVar.slog(3) << "ivstream: created vector stream." << std::endl;
  } catch (FlxException &e) {
    if (istrm!=NULL) delete istrm;
    throw;
  }
}

FlxObjInputVectorStream::~FlxObjInputVectorStream()
{
  if (inputStreamName!=NULL) delete inputStreamName; 
  delete Nreserve;
}

void FlxObjistream_write::task()
{
  FlxIstream &is = data->IstreamBox.get(isname->eval_word(true));
  tdouble d;
  while (is.get_value(d,true)) {
    sout() << GlobalVar.Double2String(d) << std::endl;
  };
}

void FlxObjInteractive::print_prompt() {
  //YES, that is right:  cout !!! NOT sout !!!
  std::cout << GlobalVar.get_flxPrompt();
}

void FlxObjInteractive::task() {
  if (!GlobalVar.allowInteractiveMode) {
    std::ostringstream ssV_2;
    ssV_2 << "Interactive input mode is not allowed in this environment.";
    throw FlxException_NeglectInInteractive("FlxObjInteractive::task", ssV_2.str() );
  }
  FlxObjBase* ob = NULL;
  ReadStream *Ireader = NULL;
  std::string strV;
  if (!GlobalVar.check_logNOTcout()) {
    if (GlobalVar.prelog_isActive()) {
      resetInputLog=true;
      GlobalVar.prelog_activated(false);
    }
  }
  try {
    GlobalVar.slog(3) << "interactive: start of the interactive input mode"<< std::endl;
    print_GNU(std::cout); // cout and NOT sout !!!
    for (;;) {
      int wert;
      bool b_ml = false;
      ostringstream ssV;
      print_prompt();
      while ( (wert = cin.get()) != EOF && (wert != 10 || b_ml) ) {
	if ( wert == 92 && !b_ml) {
	  b_ml = true;
	} else if ( wert == 10) {
	  b_ml = false;
	} else {
	  if (b_ml) {
	    if ( (char)wert != ' ' && (char)wert != '\t' ) {
	      ssV << '\\';
	      b_ml = false;
	    }
	  }
	  ssV << (char)wert;
	}
      }
      Ireader = new ReadStream(ssV.str(),true);
      data->ReadManager.push(Ireader);
      try {
	while ( Ireader->getNextType() != ReadStream::ENDOFFILE ) {
	  // print the prompt for output of input to log
	    GlobalVar.prelog_append(GlobalVar.get_flxPrompt());
	  ob = NULL;
	  ob = EvaluateCmd->evaluateCmd();
	  ob->exec();
	  GlobalVar.logLevel_strong_reset();
	  GlobalVar.slog_flush();
	  delete ob;
	}
	data->ReadManager.pop();
	delete Ireader;
      } catch ( FlxException_NeglectInInteractive &e) {
	std::cerr << std::endl << e.what() << std::endl;
	GlobalVar.logLevel_strong_reset();
	if (GlobalVar.check_logNOTcout()) {
	  GlobalVar.slog(1) << std::endl << e.what() << std::endl;
	}
	if (ob) delete ob;
	data->ReadManager.pop();
	delete Ireader;
      } catch ( FlxException &e) {
	if (ob) delete ob;
	data->ReadManager.pop();
	delete Ireader;
	throw;
      }
    }
    GlobalVar.slog(3) << "interactive: end of the interactive input mode"<< std::endl;
  } catch ( FlxExitE &e) {
    GlobalVar.logLevel_strong_reset();
    GlobalVar.slog(4) << "exit: 'exit;' was called - interactive mode will be closed." << std::endl;
    delete ob;
    data->ReadManager.pop();
    delete Ireader;
    do_resetInputLog();
    GlobalVar.slog(3) << "interactive: end of the interactive input mode"<< std::endl;
  } catch ( FlxEndE &e) {
    GlobalVar.logLevel_strong_reset();
    delete ob;
    data->ReadManager.pop();
    delete Ireader;
    do_resetInputLog();
    GlobalVar.slog(3) << "interactive: end of the interactive input mode"<< std::endl;
    throw;
  } catch ( FlxException &e) {
    GlobalVar.logLevel_strong_reset();
    do_resetInputLog();
    throw;
  };
  do_resetInputLog();
}

void FlxObjInteractive::do_resetInputLog()
{
  if (!resetInputLog) return;
  resetInputLog = false;
  GlobalVar.prelog_activated(true);
}

void FlxObjReadFile::task() {
  // open the file
    std::string fn = filename->eval();
    ReadStream *Freader = new ReadStream(fn.c_str(),true); 
    data->ReadManager.push(Freader);
    GlobalVar.slog(3) << "read: start parsing parameter file '" << fn << "'." << std::endl;
  // read from the file
    FlxObjBase* ob;
    try {
      while ( Freader->getNextType() != ReadStream::ENDOFFILE ) {
	// print the prompt for output of input to log
	  GlobalVar.prelog_append(GlobalVar.get_flxPrompt());
	ob = EvaluateCmd->evaluateCmd();
	ob->exec();
	delete ob;
      }
    } catch ( FlxExitE &e) {
      GlobalVar.slog(4) << "exit: 'exit;' was called - file '" << fn << "' will be closed." << std::endl;
      delete ob;
    } catch ( FlxEndE &e) {
      delete ob;
      data->ReadManager.pop();
      delete Freader;
      GlobalVar.slog(3) << "read: stop parsing parameter file '" << fn << "'." << std::endl;
      throw;
    };
  // close the file
    data->ReadManager.pop();
    delete Freader;
    GlobalVar.slog(3) << "read: stop parsing parameter file '" << fn << "'." << std::endl;
}

void FlxObjDistributorStream::task() {
  ostream *theStream;
  ostream& stream_1 = data->OstreamBox.get(stream1);
  ostream& stream_2 = data->OstreamBox.get(stream2);
  theStream = new flxStreamAlloc(stream_1, stream_2);
  data->OstreamBox.insert(streamname, theStream);
  GlobalVar.slog(4) << "dstream: stream '" << streamname << "' is directed into streams '" << stream1 << "' and '" << stream2 << "'." << std::endl;
}

void FlxObjDefault::task() {
  para->set_default(value);
}

FlxObjDefault::~FlxObjDefault() {
  para->free_value(value);
}


FlxObjReadFileStream::FlxObjReadFileStream() {
  AllDefParaBox->insert(new FlxOptionalParaBool(true,"filestream::truncate"));
  ParaBox.insert("truncate", "filestream::truncate" );
}

FlxObjBase* FlxObjReadFileStream::read() {
  FlxString* fn = NULL;
  try {
    string sn = reader->getWord(false,false);
    reader->getChar('(',false);
    fn = new FlxString(false,false);
    reader->getChar(')',false);
    read_optionalPara(false);
    bool bt = *(static_cast<bool*>( ParaBox.get("truncate")->get() ));
    FlxObjBase* obj1 = new FlxObjFileStream(get_doLog(),sn, fn, bt);
    return obj1;
  } catch (FlxException &e) {
    if (fn!=NULL) delete fn;
    throw;
  }
}

FlxObjReadInputFileStream::FlxObjReadInputFileStream()
{
  AllDefParaBox->insert(new FlxOptionalParaFun(1e3,"ifstream::blocksize"));
  ParaBox.insert("blocksize", "ifstream::blocksize" );
  
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"istream::erreof"));
  ParaBox.insert("erreof", "istream::erreof" );
  
  AllDefParaBox->insert(new FlxOptionalParaFun(1,"ifstream::colnumb"));
  ParaBox.insert("colnumb", "ifstream::colnumb" );
  
  AllDefParaBox->insert(new FlxOptionalParaFlxString("","ifstream::pcol"));
  ParaBox.insert("pcol", "ifstream::pcol" );
}

const tuint FlxObjReadInputFileStream::get_Columns(std::vector< int >& Cvec, const bool errSerious)
{
  FlxFunction* fun = NULL;
  FlxString* funS = NULL;
  try {
    // get the number of columns
      fun = static_cast<FlxFunction*>( ParaBox.get("colnumb")->get() ); 
      tuint Cnumb = fun->cast2tuint();
      delete fun;
    // get the columns to read
      funS = static_cast<FlxString*>( ParaBox.get("pcol")->get() ); 
      std::string strT = funS->eval();
      delete funS;
      if (strT!="") {
	ReadStream rs(strT);
	while (rs.getNextType()!=ReadStream::ENDOFFILE) {
	  Cvec.push_back(rs.getNumber<int>(false));
	  if (rs.getChar()!=',') break;
	}
      }
    // check the columns to read
      if (Cvec.empty() && Cnumb==1) {
	Cvec.push_back(1);
      } else if (Cvec.empty() && Cnumb > 1) {
	ostringstream ssV;
	ssV << "Error reading parameter 'pcol'";
	FlxError(errSerious,"FlxObjReadInputFileStream::get_Columns_1", ssV.str(), reader->getCurrentPos());
      }
      int istart=0;
      for (size_t i=0;i<Cvec.size();++i) {
	if (Cvec[i]<=istart && Cvec[i]>=(int)Cnumb) {
	  ostringstream ssV;
	  ssV << "Error reading parameter 'pcol'";
	  FlxError(errSerious,"FlxObjReadInputFileStream::get_Columns_2", ssV.str(), reader->getCurrentPos());
	} else {
	  istart = Cvec[i];
	}
      }
      if (Cvec.size() == Cnumb) {
	Cvec.clear();
	Cvec.push_back(1);
	Cnumb = 1;
      }
    return Cnumb;
  } catch (FlxException &e) {
    if (fun!=NULL) delete fun;
    if (funS!=NULL) delete funS;
    throw;
  }
}

FlxObjBase* FlxObjReadInputFileStream::read()
{
  FlxString* fn = NULL;
  try {
  string sn = reader->getWord(false,false);
  reader->getChar('(',false);
  fn = new FlxString(false,false);
  reader->getChar(')',false);
  read_optionalPara(false);
  // read column-specification
    std::vector<int> Cvec;
    int Cnumb = get_Columns(Cvec,false);
  return new FlxObjInputFileStream(get_doLog(),sn, fn, get_optPara_FlxFunction("blocksize"), Cnumb, Cvec, get_optPara_bool("erreof"));
  } catch (FlxException &e) {
    if (fn!=NULL) delete fn;
    throw;
  }
}

FlxObjReadInputVectorStream::FlxObjReadInputVectorStream()
{
  AllDefParaBox->insert(new FlxOptionalParaBool(false,"istream::erreof"));
  ParaBox.insert("erreof", "istream::erreof" );
  
  AllDefParaBox->insert(new FlxOptionalParaFun(1e5,"ivstream::nreserve"));
  ParaBox.insert("nreserve", "ivstream::nreserve" );
}

FlxObjBase* FlxObjReadInputVectorStream::read()
{
  FlxString* fn = NULL;
  try {
  string sn = reader->getWord(false,false);
  reader->getChar('(',false);
  if (reader->whatIsNextChar()!=')') { 
    fn = new FlxString(false,false);
  }
  reader->getChar(')',false);
  read_optionalPara(false);
  return new FlxObjInputVectorStream(get_doLog(),sn, fn, get_optPara_FlxFunction("nreserve"), get_optPara_bool("erreof"));
  } catch (FlxException &e) {
    if (fn!=NULL) delete fn;
    throw;
  }
}

FlxObjBase* FlxObjReadivstream_append::read()
{
  string sn = reader->getWord(false,false);
  reader->getChar('=');
  FlxFunction *fun = NULL;
  try {
    FlxIstream_vector* isv = dynamic_cast<FlxIstream_vector*>(&(data->IstreamBox.get(sn)));
    if (isv==NULL) {
      std::ostringstream ssV;
      ssV << "A vector-input-stream '" << sn << "' does not exist.";
      throw FlxException_NeglectInInteractive("FlxObjReadivstream_append::read", ssV.str(), reader->getCurrentPos() );
    }
    fun = new FlxFunction(funReader,false);
    return new FlxObjivstream_append(get_doLog(),*isv,fun);
  } catch (FlxException &e) {
    if (fun!=NULL) delete fun;
    throw;
  }
}

FlxObjBase* FlxObjReadistream_write::read()
{
  FlxString *isname = new FlxString(false,false);
  try {
    read_optionalPara(false);
    return new FlxObjistream_write(get_doLog(),isname,get_stream());
  } catch (FlxException &e) {
    delete isname;
    throw;
  }
}

FlxObjBase* FlxObjReadReadFile::read() {
  FlxString* fn = new FlxString(false,false);
  read_optionalPara(false);
  return new FlxObjReadFile(get_doLog(),fn);
}

FlxObjBase* FlxObjReadDistributorStream::read() {
  string dstream = reader->getWord(false,false);
  reader->getChar('(',false);
  string stream1 = reader->getWord(false,false);
  reader->getChar(',',false);
  string stream2 = reader->getWord(false,false);
  reader->getChar(')',false);
  read_optionalPara(false);
  return new FlxObjDistributorStream(get_doLog(),dstream, stream1, stream2);
}

void FlxObjFloatingPointConversion::task()
{
  switch (id) {
    case 0:
      GlobalVar.Double2String_setPrec(fun->cast2tuintW0(false));
      break;
    case 1:
      GlobalVar.Double2String_setType(fun->cast2tuintW0(false));
      break;
    case 2:
      GlobalVar.Double2String_setBValU(fun->calc());
      break;
    case 3:
      GlobalVar.Double2String_setBValL(fun->calc());
      break;
    case 4:
      if ( fun->calc() == 0.0 ) {
	GlobalVar.Double2String_setDel0(false);
      } else {
	GlobalVar.Double2String_setDel0(true);
      }
      break;
    case 5:
      if ( fun->calc() == 0.0 ) {
	GlobalVar.Double2String_setDelP(false);
      } else {
	GlobalVar.Double2String_setDelP(true);
      }
      break;
  }
}

void FlxObjSetVariousDefault::task()
{
  switch (ID) {
    case 1:
      GlobalVar.logLevel_strong_set(fun->cast2int());
      break;
    case 2:
      FlxFunDeg::set_deg(fun->cast2tuintW0());
      break;
    default:
      #if FLX_DEBUG
	throw FlxException_Crude("FlxObjSetVariousDefault::task");
      #endif
      return;
  }
}


FlxObjBase* FlxObjReadDefault::read_special(string& key)
{
  // floating point conversion
    if (key == "flxoutput::float::prec") {
      FlxFunction* f1 = new FlxFunction(funReader,false);
      read_optionalPara(false);
      return new FlxObjFloatingPointConversion(get_doLog(),f1,0);
    } else if (key == "flxoutput::float::type") {
      FlxFunction* f1 = new FlxFunction(funReader,false);
      read_optionalPara(false);
      return new FlxObjFloatingPointConversion(get_doLog(),f1,1);
    } else if (key == "flxoutput::float::bvalu") {
      FlxFunction* f1 = new FlxFunction(funReader,false);
      read_optionalPara(false);
      return new FlxObjFloatingPointConversion(get_doLog(),f1,2);
    } else if (key == "flxoutput::float::bvall") {
      FlxFunction* f1 = new FlxFunction(funReader,false);
      read_optionalPara(false);
      return new FlxObjFloatingPointConversion(get_doLog(),f1,3);
    } else if (key == "flxoutput::float::del0") {
      FlxFunction* f1 = new FlxFunction(funReader,false);
      read_optionalPara(false);
      return new FlxObjFloatingPointConversion(get_doLog(),f1,4);
    } else if (key == "flxoutput::float::delp") {
      FlxFunction* f1 = new FlxFunction(funReader,false);
      read_optionalPara(false);
      return new FlxObjFloatingPointConversion(get_doLog(),f1,5);
    }
  // loglevel
    if (key == "log::level") {
      FlxFunction* f1 = new FlxFunction(funReader);
      read_optionalPara(false);
      return new FlxObjSetVariousDefault(get_doLog(),1,f1);
    }
  // default polynomial degree of FlxFunDeg
    if (key == "flxfundeg::degree") {
      FlxFunction* f1 = new FlxFunction(funReader);
      read_optionalPara(false);
      return new FlxObjSetVariousDefault(get_doLog(),2,f1);
    }
    
    return NULL;
}

FlxObjBase* FlxObjReadDefault::read() {
  // extract the DefSetName
    string sn = reader->getWord(false,false);
    while ( reader->whatIsNextChar() == ':' ) {
      reader->getChar(':',false);
      reader->getChar(':',false);
      sn += "::";
      sn += reader->getWord(false,false);
    }
    transform(sn.begin(), sn.end(), sn.begin(), (int(*)(int)) std::tolower);
    reader->getChar('=',false);
  // check for special treatment
    FlxObjBase* obj1 = read_special(sn);
    if (obj1) return obj1;
  // get the Parameter-Class
    FlxOptionalParaBase* pb = AllDefParaBox->get(sn);
    if ( pb == NULL ) {
      ostringstream ssV;
      ssV << "Unknown parameter '" << sn << "'.";
      throw FlxException_NeglectInInteractive("FlxObjReadDefault::read_1", ssV.str(), reader->getCurrentPos());
    }
    void *vp = pb->read_value(false);
    read_optionalPara(false);
    obj1 = new FlxObjDefault(get_doLog(),vp, pb);
    return obj1;
}

FlxObjIf::~FlxObjIf() {
  delete funIf;
  delete InternListThen;
  delete InternListElse;
}

FlxObjWhile::~FlxObjWhile() {
  delete funWhile;
  delete InternListLoop;
}

FlxObjFor::~FlxObjFor() {
  delete funCond;
  delete funConst;
  delete InternListLoop;
  delete ConstDef;
}

FlxObjBase* FlxObjReadWhile::read() {
  FlxFunction* f1 = NULL;
  FlxObjBase* loopblock = NULL;
  try {
    reader->getChar('(',false);
    f1 = new FlxFunction(funReader,false);
    reader->getChar(')',false);
    loopblock = read_block(true,false);
    read_optionalPara(false);
    return new FlxObjWhile(get_doLog(),f1, loopblock, get_maxpasses());
  } catch ( FlxException& i) {
    if (f1!=NULL) delete f1;
    if (loopblock!=NULL) delete loopblock;
    throw;
  }
}

FlxObjBase* FlxObjReadFor::read() {
  FlxObjReadConst* objRC = NULL;
  FlxObjConst* obj1 = NULL;
  FlxFunction* funCond = NULL; FlxFunction* funConst = NULL;
  FlxObjBase* loopblock = NULL;
  try {
    reader->getChar('(',false);
    string cname = reader->getWord(false,false);
    objRC = new FlxObjReadConst();
    obj1 = dynamic_cast<FlxObjConst*> ( objRC->read(cname) );
    delete objRC; objRC=NULL;
    tdouble* d1 = data->ConstantBox.get(cname);
    reader->getChar(';',false);
    funCond = new FlxFunction(funReader,false);
    reader->getChar(';',false);
    funConst = new FlxFunction(funReader,false);
    reader->getChar(')',false);
    loopblock = read_block(true,false);
    read_optionalPara(false);
    return new FlxObjFor(get_doLog(),d1, obj1, funCond, funConst, loopblock, get_maxpasses());
  } catch (FlxException &e) {
    if (objRC) delete objRC;
    if (obj1) delete obj1;
    if (funCond) delete funCond;
    if (funConst) delete funConst;
    if (loopblock) delete loopblock;
    throw;
  }
}

FlxObjBase* FlxObjReadSFor::read() {
  FlxFunction* funTo = NULL;
  FlxObjBase* loopblock = NULL;
  try {
    reader->getChar('(',false);
    string cname = reader->getWord(false,false);
    reader->getChar(';',false);
    funTo = new FlxFunction(funReader,false);
    bool start0 = false;
    if (reader->whatIsNextChar()==';') {
      reader->getChar(';',false);
      start0 = reader->getBool(false);
    }
    reader->getChar(')',false);
    data->ConstantBox.declareC(cname);
    tdouble* d1 = data->ConstantBox.get(cname);
    loopblock = read_block(true,false);
    return new FlxObjSFor(get_doLog(),d1, funTo, start0, loopblock);
  } catch (FlxException &e) {
    if (funTo!=NULL) delete funTo;
    if (loopblock!=NULL) delete loopblock;
    throw;
  }
}

FlxObjReadTime::FlxObjReadTime()
{
  // Default stream
    AllDefParaBox->insert(new FlxOptionalParaStream("cout","flxoutput::stream"));
    ParaBox.insert("stream", "flxoutput::stream" );
}

string FlxObjReadTime::get_stream()
{
  return *(static_cast<string*>( ParaBox.get("stream")->get() ));
}

FlxObjBase* FlxObjReadTime::read()
{
  FlxObjBase* timeblock = read_block(false,false);
  read_optionalPara(false);
  return new FlxObjTime(get_doLog(),timeblock,get_stream());
}

FlxObjBase* FlxObjReadIf::read() {
  FlxFunction* f1 = NULL;
  FlxObjBase* thenblock = NULL; FlxObjBase* elseblock = NULL;
  try {
    reader->getChar('(',false);
    f1 = new FlxFunction(funReader,false);
    reader->getChar(')',false);
    thenblock = read_block(false,false);
    if ( reader->getNextType() == ReadStream::STRING ) {
      string strV = reader->getWord(false,false);
      transform(strV.begin(), strV.end(), strV.begin(), (int(*)(int)) std::tolower);
      if ( strV != "else" ) {
	ostringstream ssV;
	ssV << "Expected 'else' or ';' (not '" << strV << "').";
	throw FlxException_NeglectInInteractive("FlxObjReadIf::read_1", ssV.str(), reader->getCurrentPos());
      }
      elseblock = read_block(false,false);
    }
    read_optionalPara(false);
    return new FlxObjIf(get_doLog(),f1, thenblock, elseblock);
  } catch ( FlxException &e) {
    if (f1!=NULL) delete f1;
    if (thenblock!=NULL) delete thenblock;
    if (elseblock!=NULL) delete elseblock;
    throw;
  }
}

FlxObjBase* FlxObjReadFun::read() {
  string cname = reader->getWord(true,false);
  // check if function is already defined
    if ( data->FunBox.get(cname) != NULL ) {
      ostringstream ssV_2;
      ssV_2 << "Function '" << cname << "' was already defined - redefinition of functions is not allowed.";
      throw FlxException_NeglectInInteractive("FlxObjReadFun::read_1", ssV_2.str(), reader->getCurrentPos() );
    }
  // check if a variable with that name exists already
    isdefined(cname, 'F',false);
  // number of parameters
    reader->getChar('(',false);
    
    FlxFunction* i1f = NULL;
    tuint i1=0;
    try {
      i1f = new FlxFunction(funReader,false);
      i1 = i1f->cast2tuintW0(false);
      delete i1f;
    } catch (FlxException &e) {
      if (i1f!=NULL) delete i1f; 
      throw;
    }
    reader->getChar(')',false);
  reader->getChar('=',false);
  
  FunReadPara::set_NumbOfPara(i1,false);
  FlxFunction *fp = new FlxFunction(funReader,false);
  try {
    read_optionalPara(false);
    FlxObjBase* obj1 = new FlxObjFun( get_doLog(), cname, new FunReadFunUser(cname, fp, i1) );
    FunReadPara::set_NumbOfPara(0,false);
    data->FunBox.declareF(cname);
    return obj1;
  } catch (FlxException &e) {
    delete fp;
    throw;
  }
}

FlxObjBase* FlxObjReadEnd::read()
{
  read_optionalPara(false);
  return new FlxObjEnd(get_doLog());
}

FlxObjBase* FlxObjReadExit::read()
{
  read_optionalPara(false);
  return new FlxObjExit(get_doLog());
}

FunBase* FunReadFunISread::read(bool errSerious)
{
  FlxString* strV = new FlxString(false,errSerious);
  return new FunISread( strV );
}

FunISread::~FunISread()
{
  if (strV != NULL) delete strV;
}

void FunISread::dependOn_RV(bVec& bv)
{
  std::ostringstream ssV_2;
  ssV_2 << "It is not allowed to check an input stream for dependence on a random variable.";
  throw FlxException("FunISread::dependOn_RV", ssV_2.str() ); 
}

const bool FunISread::search_circref(FlxFunction* fcr)
{
  if (strV != NULL) {
    return strV->search_circref(fcr);
  } else {
    return false;
  }
}

void FunISread::set_istrm()
{
  strS = strV->eval_word(true);
  istrm = &(data->IstreamBox.get(strS));
  delete strV;
  strV = NULL;
}

const std::string FunISread::write()
{
  if (istrm==NULL) {
    set_istrm();
  }
  return "isread(" + strS + ")";
}

const tdouble FunISread::calc()
{
  if (istrm==NULL) {
    set_istrm();
  }
  tdouble cv;
  istrm->get_value(cv);
  return cv;
}

const tdouble FunObjExec::calc()
{
  InternListLoop->exec();
  return fun->calc();
}

const bool FunObjExec::search_circref(FlxFunction* fcr)
{
  return fun->search_circref(fcr);
}

const std::string FunObjExec::write()
{
  return "objexec(" + fun->write() + ",{...})";
}

const bool FunObjExec::dependOn_Const(const tdouble* const thenumber)
{
  throw FlxException_NotImplemented("FunObjExec::dependOn_Const");
}

void FunObjExec::dependOn_RV(bVec& bv)
{
  throw FlxException_NotImplemented("FunObjExec::dependOn_RV");
}

const bool FunObjExec::optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi)
{
  child_optimize(fun,foi);
  return false;
}

FunBase* FunReadObjExec::read(bool errSerious)
{
  FunBase* fun = FunctionList->read(errSerious); 
  FlxObjBase* loopblock = NULL;
  try {
    reader->getChar(',',errSerious);
    loopblock = read_block(true,errSerious);
    return new FunObjExec(fun,loopblock);
  } catch (FlxException &e) {
    delete fun;
    if (loopblock!=NULL) delete loopblock;
    throw;
  }
}

FlxObjBase* FunReadObjExec::read()
{
  throw FlxException_Crude("FunReadObjExec::read_1");
}















