/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxobjects.h"

using namespace std;

FlxDefParaBox* FlxDefParaBoxBase::AllDefParaBox = NULL;
FlxEvaluateCmd* FlxEvaluateCmdBase::EvaluateCmd = NULL;
#if FLX_DEBUG
  int FlxDefParaBox::Cinst = 0;
  int FlxEvaluateCmd::Cinst = 0;
#endif

void FlxOptionalParaBase::read(bool errSerious) {
  set( read_value(errSerious) );
  is_set = true;
}

void FlxOptionalParaFun::set_default(void* defV)
{
  FlxFunction* valueV = static_cast<FlxFunction*>(defV);
  if (defv!=NULL) {
    delete defv;
  }
  defv = new FlxFunction(*valueV); 
  GlobalVar.slog(3) << "default: set '" << pName << "' to '" << defv->write() << "'." << std::endl;
}

FlxOptionalParaFun::FlxOptionalParaFun(const tdouble defV, const string pName)
: FlxOptionalParaBase(pName), defv(NULL), value(NULL)
{
  defv = new FlxFunction(new FunNumber(defV));
}

FlxOptionalParaFun::~FlxOptionalParaFun()
{
  if ( value != NULL ) {
    delete value; 
  }
  if ( defv != NULL ) {
    delete defv;
  }
}

void FlxOptionalParaFun::free_value(void* valueP)
{
  FlxFunction* valueV = static_cast<FlxFunction*>(valueP);
  if ( valueP != NULL) { delete valueV; }
}

void FlxOptionalParaFun::set(void* valueP)
{
  FlxFunction* valueV = static_cast<FlxFunction*>(valueP);
  if ( value==NULL) {
    value = valueV;
  } else {
    value->assign(valueV);
  }
}

void* FlxOptionalParaFun::get()
{
  if ( value == NULL || is_set == false ) {
    is_set = false;
    return new FlxFunction(*defv);
  } else {
    is_set = false;
    return new FlxFunction(*value);
  }
}

FlxOptionalParaFunDeg::FlxOptionalParaFunDeg(tdouble defV, string pName)
: FlxOptionalParaBase(pName), defv(NULL), value(NULL)
{
  defv = new FlxFunDeg(new FlxFunction(new FunNumber(defV)));
}

FlxOptionalParaFunDeg::~FlxOptionalParaFunDeg()
{
  if ( value != NULL ) {
    delete value; 
  }
  if ( defv != NULL ) {
    delete defv;
  }
}

void FlxOptionalParaFunDeg::set(void* valueP)
{
  FlxFunDeg* valueV = static_cast<FlxFunDeg*>(valueP);
  if ( value==NULL) {
    value = valueV;
  } else {
    value->assign(valueV);
  }
}

void* FlxOptionalParaFunDeg::get()
{
  if ( value == NULL || is_set == false ) {
    is_set = false;
    return new FlxFunDeg(*defv);
  } else {
    is_set = false;
    return new FlxFunDeg(*value);
  }
}

void FlxOptionalParaFunDeg::set_default(void* defV)
{
  FlxFunDeg* valueV = static_cast<FlxFunDeg*>(defV);
  if (defv!=NULL) {
    delete defv;
  }
  defv = new FlxFunDeg(*valueV); 
  GlobalVar.slog(3) << "default: set '" << pName << "' to '" << defv->write() << "'." << std::endl;
}

void FlxOptionalParaFunDeg::free_value(void* valueP)
{
  FlxFunDeg* valueV = static_cast<FlxFunDeg*>(valueP);
  if ( valueP != NULL) { delete valueV; }
}

void FlxOptionalParaFunDegDetect::free_value(void* valueP)
{
  flxFun_DegreeDetect* valueV = static_cast<flxFun_DegreeDetect*>(valueP);
  if ( valueP != NULL) { delete valueV; }
}

void FlxOptionalParaFunDegDetect::set_default(void* defV)
{
  flxFun_DegreeDetect* valueV = static_cast<flxFun_DegreeDetect*>(defV);
  if (defv!=NULL) {
    delete defv;
  }
  defv = new flxFun_DegreeDetect(*valueV); 
  GlobalVar.slog(3) << "default: set '" << pName << "' to '" << defv->write() << "'." << std::endl;
}

void* FlxOptionalParaFunDegDetect::get()
{
  if ( value == NULL || is_set == false ) {
    is_set = false;
    return new flxFun_DegreeDetect(*defv);
  } else {
    is_set = false;
    return new flxFun_DegreeDetect(*value);
  }
}

FlxOptionalParaFunDegDetect::~FlxOptionalParaFunDegDetect()
{
  if ( value != NULL ) {
    delete value; 
  }
  if ( defv != NULL ) {
    delete defv;
  }
}

void FlxOptionalParaFunDegDetect::set(void* valueP)
{
  flxFun_DegreeDetect* valueV = static_cast<flxFun_DegreeDetect*>(valueP);
  if ( value==NULL) {
    value = valueV;
  } else {
    *value = *valueV;
    delete valueV;
  }
}

FlxOptionalParaMeshRefine_1D::~FlxOptionalParaMeshRefine_1D()
{
  if ( value != NULL ) {
    delete value; 
  }
  if ( defv != NULL ) {
    delete defv;
  }
}

void FlxOptionalParaMeshRefine_1D::set_default(void* defV)
{
  flxFEMrefine_1D* valueV = static_cast<flxFEMrefine_1D*>(defV);
  if (defv!=NULL) {
    delete defv;
  }
  defv = new flxFEMrefine_1D(*valueV); 
  GlobalVar.slog(3) << "default: set '" << pName << "' to '" << defv->write() << "'." << std::endl;
}

void FlxOptionalParaMeshRefine_1D::set(void* valueP)
{
  flxFEMrefine_1D* valueV = static_cast<flxFEMrefine_1D*>(valueP);
  if ( value==NULL) {
    value = valueV;
  } else {
    *value = *valueV;
    delete valueV;
  }
}

void* FlxOptionalParaMeshRefine_1D::get()
{
  if ( value == NULL || is_set == false ) {
    is_set = false;
    if (defv==NULL) return NULL;
    else return new flxFEMrefine_1D(*defv);
  } else {
    is_set = false;
    return new flxFEMrefine_1D(*value);
  }
}

void FlxOptionalParaMeshRefine_1D::free_value(void* valueP)
{
  if (valueP==NULL) return;
  flxFEMrefine_1D* valueV = static_cast<flxFEMrefine_1D*>(valueP);
  if ( valueV != NULL) { delete valueV; }
}

void FlxOptionalParaFlxString::set(void* valueP)
{
  FlxString* valueV = static_cast<FlxString*>(valueP);
  if ( value==NULL) {
    value = valueV;
  } else {
    value->assign(valueV);
  }
}

void* FlxOptionalParaFlxString::get()
{
  if ( value == NULL || is_set == false ) {
    is_set = false;
    return new FlxString(*defv);
  } else {
    is_set = false;
    return new FlxString(*value);
  }
}

void FlxOptionalParaFlxString::set_default(void* defV)
{
  FlxString* valueV = static_cast<FlxString*>(defV);
  if (defv!=NULL) {
    delete defv;
  }
  defv = new FlxString(*valueV); 
  GlobalVar.slog(3) << "default: set '" << pName << "' to '" << defv->write() << "'." << std::endl;
}

FlxOptionalParaFlxString::FlxOptionalParaFlxString(string defV, string pName)
: FlxOptionalParaBase(pName), defv(new FlxString(new FlxString_String(defV,true))), value(NULL)
{

}

FlxOptionalParaFlxString::~FlxOptionalParaFlxString()
{
  if ( value != NULL ) {
    delete value; 
  }
  if ( defv != NULL ) {
    delete defv;
  }
}

void FlxOptionalParaFlxString::free_value(void* valueP)
{
  FlxString* valueV = static_cast<FlxString*>(valueP);
  if ( valueP != NULL) { delete valueV; }
}

void FlxOptionalParaMtxFun::set_default(void* defV)
{
  FlxMtxFun* valueV = static_cast<FlxMtxFun*>(defV);
  if (defv!=NULL) {
    delete defv;
  }
  defv = new FlxMtxFun(*valueV);
  GlobalVar.slog(3) << "default: set '" << pName << "' to '" << defv->write() << "'." << std::endl;
}

void* FlxOptionalParaMtxFun::get()
{
  if ( value == NULL || is_set == false ) {
    is_set = false;
    return new FlxMtxFun(*defv);
  } else {
    is_set = false;
    return new FlxMtxFun(*value);
  }
}

void FlxOptionalParaMtxFun::free_value(void* valueP)
{
  FlxMtxFun* valueV = static_cast<FlxMtxFun*>(valueP);
  if ( valueP != NULL) { delete valueV; }
}

FlxOptionalParaMtxFun::~FlxOptionalParaMtxFun()
{
  if ( value != NULL ) {
    delete value; 
  }
  if ( defv != NULL ) {
    delete defv;
  }
}

void FlxOptionalParaMtxFun::set(void* valueP)
{
  FlxMtxFun* valueV = static_cast<FlxMtxFun*>(valueP);
  if ( value==NULL) {
    value = valueV;
  } else {
    value->assign(valueV);
  }
}




FlxOptionalParaStream::~FlxOptionalParaStream() {
  if ( value != NULL ) 
    delete value;
}

void* FlxOptionalParaStream::get() {
  if ( value == NULL || is_set == false ) {
    is_set = false;
    return &defv;
  } else {
    is_set = false;
    return value;
  }
}

void* FlxOptionalParaStream::read_value(bool errSerious) {
  return ( new string(reader->getWord(false,errSerious)) );
}

void FlxOptionalParaStream::set(void* valueP) {
  string* valueV = static_cast<string*>(valueP);
  if ( value == NULL ) {
    value = new string( *valueV );
  } else {
    *value = *valueV;
  }
  delete valueV;
}

void FlxOptionalParaStream::free_value(void* valueP) {
  string* valueV = static_cast<string*>(valueP);
  if ( valueP != NULL) { delete valueV; }
}


void FlxOptionalParaStream::set_default(void* defV) {
  defv = *(static_cast<string*>(defV));
  GlobalVar.slog(3) << "default: set '" << pName << "' to '" << defv << "'." << std::endl;
}

FlxOptionalParaBool::~FlxOptionalParaBool() {
  if ( value != NULL ) {
    delete value;
  }
}

void* FlxOptionalParaBool::read_value(bool errSerious) {
  FlxFunction* t = new FlxFunction(funReader,errSerious);
  const bool bt = (t->cast2int()!=0);
  delete t;
  return ( new bool(bt) );
}

void FlxOptionalParaBool::set(void* valueP) {
  bool* valueV = static_cast<bool*>(valueP);
  if ( value == NULL ) {
    value = new bool( *valueV );
  } else {
    *value = *valueV;
  }
  delete valueV;
}

void* FlxOptionalParaBool::get() {
  if ( value == NULL || is_set == false ) {
    is_set = false;
    return &defv;
  } else {
    is_set = false;
    return value;
  }
}

void FlxOptionalParaBool::free_value(void* value) {
  bool* valueV = static_cast<bool*>(value);
  if ( value != NULL) { delete valueV; }
}





void FlxOptionalParaBox::insert(string name, string defname) {
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  transform(defname.begin(), defname.end(), defname.begin(), (int(*)(int)) std::tolower);
  FlxOptionalParaBase* para = AllDefParaBox->get(defname);
  pair<std::string, FlxOptionalParaBase*>Element(name, para);
  if ( ! box.insert(Element).second ) {
    ostringstream ssV;
    ssV << "Optional parameter '" << name << "' could not be inserted into ParaBox.";
    throw FlxException("FlxOptionalParaBox::insert_1", ssV.str() );
  }
}

FlxOptionalParaBase* FlxOptionalParaBox::get(std::string name) {
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  map<std::string, FlxOptionalParaBase*>::iterator pos;
  pos = box.find(name);
  if ( pos != box.end() ) {
    return pos->second;
  } else {
    return NULL;
  }
}

void FlxOptionalParaBool::set_default(void* defV) {
  defv = *(static_cast<bool*>(defV));
  GlobalVar.slog(3) << "default: set '" << pName << "' to '" << (defv?"true":"false") << "'." << std::endl;
}

FlxDefParaBox::FlxDefParaBox()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxDefParaBox' created ...";
      throw FlxException("FlxDefParaBox::FlxDefParaBox", ssV.str() );
    }
  #endif
  FlxDefParaBoxBase::set_AllDefParaBox(this);
}

FlxDefParaBox::~FlxDefParaBox() {
  for (map<std::string, FlxOptionalParaBase*>::iterator pos = box.begin(); pos != box.end(); ++pos) delete pos->second;
}

void FlxDefParaBox::insert(FlxOptionalParaBase* value) {
  pair<std::string, FlxOptionalParaBase*>Element(value->get_name(), value);
  if ( ! box.insert(Element).second ) {;
    delete value;
  }
}

FlxOptionalParaBase* FlxDefParaBox::get(std::string name) {
  transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  map<std::string, FlxOptionalParaBase*>::iterator pos = box.find(name);
  if ( pos != box.end() ) {
    return pos->second;
  } else {
    return NULL;
  }
}



void FlxEvaluateCmd::check_ending() {
  char ch = reader->getChar();
  if ( ch != ';' ) {
    if ( int(ch) != -1 ) { 	// cheap and nasty - this produces a bug. (';' is not necessary at the end of the file)
      ostringstream ssV;
      ssV << "Expected ';' (and NOT '" << ch << "' [" << int(ch) << "])";
      throw FlxException_NeglectInInteractive("FlxEvaluateCmd::check_ending_1", ssV.str(), reader->getCurrentPos());
    }
  }
}

FlxEvaluateCmd::FlxEvaluateCmd()
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxEvaluateCmd' created ...";
      throw FlxException("FlxEvaluateCmd::FlxEvaluateCmd", ssV.str() );
    }
  #endif
  FlxEvaluateCmdBase::set_EvaluateCmd(this);   
}

FlxObjBase* FlxEvaluateCmd::evaluateCmd( ) {
  // Detect multiple ;;;'s
  while (reader->whatIsNextChar() == ';') {
    reader->getChar(true,false);
  }
  string str1 = reader->getWord(false,false);
  FlxObjBase* obj1 = NULL;

  // Are there some objects which are not allowed ( for example "fun" in loops? ) 
    if ( data->IgnoreBox.isActive_iL_recur() ) {
      if ( data->IgnoreBox.isOnIgnoreList_recur(str1) ) {
	ostringstream ssV;
	ssV << "'" << str1 << "'not allowed at this point.";
	throw FlxException_NeglectInInteractive("FlxEvaluateCmd::evaluateCmd_1", ssV.str(), reader->getCurrentPos());
      }
    }

  // check if there is a keyword for an object
    FlxObjReadBase *objR = ObjReadBox.get(str1);
    if (objR != NULL) {
      obj1 = NULL;
      try {
	obj1 = objR->read();
	check_ending();
      } catch ( FlxException &e) {
	if (obj1!=NULL) delete obj1;
	throw;
      };
      GlobalVar.prelog_write();
      return obj1;
    } 

  // check if the keyword refers to a procedure
  obj1 = data->SubBox.get(str1);
  if ( obj1 != NULL ) {
    FlxObjReadBase *objR = ObjReadBox.get("procedure");
    FlxObjReadSubDo* objRSD = dynamic_cast<FlxObjReadSubDo*> ( objR );
    try {
    obj1 = objRSD->read(obj1);
      check_ending();
    } catch ( FlxException &e) {
      if (obj1!=NULL) delete obj1;
      throw;
    };
    GlobalVar.prelog_write();
    return obj1;
  }

  // if no valid keyword was found, check if it is an assignment of a 'const' variable
    if ( reader->whatIsNextChar() != '=' ) {
      ostringstream ssV;
      ssV << "Expected keyword ('" << str1 << "' is not valid).";
      throw FlxException_NeglectInInteractive("FlxEvaluateCmd::evaluateCmd_2", ssV.str(), reader->getCurrentPos());
    } else {	// assign 'const' variables
      FlxObjReadConst* objRC = dynamic_cast<FlxObjReadConst*> ( ObjReadBox.get("const") );
      obj1 = objRC->read(str1);
      try {
	check_ending();
      } catch ( FlxException &e) {
	delete obj1;
	throw;
      };
      GlobalVar.prelog_write();
      return obj1;
    }

}




FlxObjReadBase* FlxObjectReadBox::get(std::string name) {
  map<string, FlxObjReadBase*>::iterator pos;
  std::transform(name.begin(), name.end(), name.begin(), (int(*)(int)) std::tolower);
  pos = box.find(name);
  if ( pos != box.end() ) {
    return pos->second;
  } else {
    return NULL;
  }
}

void FlxObjectReadBox::insert(std::string name, FlxObjReadBase* value) {
  pair<string, FlxObjReadBase*> Element(name, value);
  if ( ! box.insert(Element).second ) {
    ostringstream ssV;
    ssV << "Error during inserting " << name << " in FlxObjectReadBox.";
    throw FlxException("FlxObjectReadBox::insert_1", ssV.str());
  }
}

FlxObjectReadBox::~FlxObjectReadBox() {
  for (map<string, FlxObjReadBase*>::iterator pos = box.begin(); pos != box.end(); ++pos) delete pos->second;
}

FlxObjectReadBox::FlxObjectReadBox() {
  insert("const", new FlxObjReadConst());
  insert("sub", new FlxObjReadSub());
  insert("procedure", new FlxObjReadSubDo());
  
}

FlxObjReadBase::FlxObjReadBase()
{
  // Default do-log-setting
    AllDefParaBox->insert(new FlxOptionalParaBool(true,"flxlog::dolog"));
    ParaBox.insert("dolog", "flxlog::dolog" );
}

const bool FlxObjReadBase::get_doLog()
{
  return *(static_cast<bool*>(ParaBox.get("dolog")->get() ));
}

void FlxObjReadBase::read_optionalPara(bool errSerious) {
  if ( reader->whatIsNextChar() != '{' ) { return; }
  string strV;
  reader->getChar('{',errSerious);
  while ( reader->getNextType() == ReadStream::STRING ) {
    strV = reader->getWord(false,errSerious);
    FlxOptionalParaBase* para = ParaBox.get(strV);
    if ( para == NULL ) {
      ostringstream ssV;
      ssV << "Unknown optional Parameter '" << strV << "'.";
      FlxError(errSerious,"FlxObjReadBase::read_optionalPara_1", ssV.str());
    }
    reader->getChar('=',errSerious);
    para->read(errSerious);
    reader->getChar(';',errSerious);
  }
  reader->getChar('}',errSerious);
}

FlxFunction* FlxObjReadBase::get_optPara_FlxFunction(string str)
{
  void *v1 = ParaBox.get(str)->get();
  FlxFunction* fun = static_cast<FlxFunction*>( v1 ); 
  if (fun == NULL) {
    throw FlxException_Crude("FlxObjReadBase::get_optPara_FlxFunction");
  }
  return fun;
}

FlxFunDeg* FlxObjReadBase::get_optPara_FlxFunDeg(string str)
{
  void *v1 = ParaBox.get(str)->get();
  FlxFunDeg* fun = static_cast<FlxFunDeg*>( v1 ); 
  if (fun == NULL) {
    throw FlxException_Crude("FlxObjReadBase::get_optPara_FlxFunDeg");
  }
  return fun;
}

FlxMtxFun* FlxObjReadBase::get_optPara_FlxMtxFun(string str)
{
  void *v1 = ParaBox.get(str)->get();
  FlxMtxFun* fun = static_cast<FlxMtxFun*>( v1 ); 
  if (fun == NULL) {
    throw FlxException_Crude("FlxObjReadBase::get_optPara_FlxMtxFun");
  }
  return fun;
}

flxFun_DegreeDetect* FlxObjReadBase::get_optPara_DegreeDetect(string str)
{
  void *v1 = ParaBox.get(str)->get();
  flxFun_DegreeDetect* fun = static_cast<flxFun_DegreeDetect*>( v1 ); 
  if (fun == NULL) {
    throw FlxException_Crude("FlxObjReadBase::get_optPara_DegreeDetect");
  }
  return fun;
}

flxFEMrefine_1D* FlxObjReadBase::get_optPara_Refine_1D(string str)
{
  void *v1 = ParaBox.get(str)->get();
  flxFEMrefine_1D* fun = static_cast<flxFEMrefine_1D*>( v1 ); 
  if (fun == NULL) {
    throw FlxException_Crude("FlxObjReadBase::get_optPara_Refine_1D");
  }
  return fun;
}

FlxString* FlxObjReadBase::get_optPara_FlxString(string str)
{
  void *v1 = ParaBox.get(str)->get();
  FlxString* fun = static_cast<FlxString*>( v1 ); 
  if (fun == NULL) {
    throw FlxException_Crude("FlxObjReadBase::get_optPara_FlxString");
  }
  return fun;
}

const bool FlxObjReadBase::get_optPara_bool(string str)
{
  bool* b1 = static_cast<bool*>( ParaBox.get(str)->get() ); 
  if (b1 == NULL) {
    throw FlxException_Crude("FlxObjReadBase::get_optPara_bool");
  }
  return *b1;
}

FlxObjReadLogBase::FlxObjReadLogBase()
{
  // Default verbose-setting
    AllDefParaBox->insert(new FlxOptionalParaBool(true,"flxlog::verbose"));
    ParaBox.insert("vlog", "flxlog::verbose" );
}

const bool FlxObjReadLogBase::get_verboseLog()
{
  return *(static_cast<bool*>(ParaBox.get("vlog")->get() ));
}

FlxObjReadOutputBase::FlxObjReadOutputBase() {
  // Default stream
    AllDefParaBox->insert(new FlxOptionalParaStream("cout","flxoutput::stream"));
    ParaBox.insert("stream", "flxoutput::stream" );
  // Default verbose-setting
    AllDefParaBox->insert(new FlxOptionalParaBool(true,"flxoutput::verbose"));
    ParaBox.insert("verbose", "flxoutput::verbose" );
  // Default check TOL for numerical values
    AllDefParaBox->insert(new FlxOptionalParaBool(false,"flxoutput::checktol"));
    ParaBox.insert("checktol", "flxoutput::checktol" );
}

const bool FlxObjReadOutputBase::get_verbose()
{
  return *(static_cast<bool*>(ParaBox.get("verbose")->get() ));
}

const bool FlxObjReadOutputBase::get_checkTOL()
{
  return *(static_cast<bool*>(ParaBox.get("checktol")->get() ));
}

const std::string FlxObjReadOutputBase::get_stream() {
  return *(static_cast<string*>( ParaBox.get("stream")->get() ));
}

void FlxObjConst::task()
{
  *cptr = fun->calc();
}

void FlxObjReadFCVbase::isdefined(const std::string thename, const char theFCVindex, const bool errSerious) const {
  if ( data->FunBox.get(thename) != NULL && theFCVindex != 'F' ) {
    ostringstream ssV;
    ssV << "A function with the name ('" << thename << "') is already defined.";
    FlxError(errSerious,"FlxObjReadFCVbase::isdefined_1", ssV.str(), reader->getCurrentPos());
  }
  if ( data->ConstantBox.get(thename) != NULL && theFCVindex != 'C' ) {
    ostringstream ssV;
    ssV << "A 'const' variable with the name ('" << thename << "') is already defined.";
    FlxError(errSerious,"FlxObjReadFCVbase::isdefined_2", ssV.str(), reader->getCurrentPos());
  }
  if ( data->VarBox.get(thename) != NULL && theFCVindex != 'V' ) {
    ostringstream ssV;
    ssV << "A 'var' variable with the name ('" << thename << "') is already defined.";
    FlxError(errSerious,"FlxObjReadFCVbase::isdefined_3", ssV.str(), reader->getCurrentPos());
  }
  if ( data->MtxFunBox.get(thename) != NULL && theFCVindex != 'G' ) {
    ostringstream ssV;
    ssV << "A matrix function with the name ('" << thename << "') is already defined.";
    FlxError(errSerious,"FlxObjReadFCVbase::isdefined_4", ssV.str(), reader->getCurrentPos());
  }
  if ( data->ConstMtxBox.get(thename) != NULL && theFCVindex != 'M' ) {
    ostringstream ssV;
    ssV << "A 'mtxconst' matrix-variable with the name ('" << thename << "') is already defined.";
    FlxError(errSerious,"FlxObjReadFCVbase::isdefined_5", ssV.str(), reader->getCurrentPos());
  }
}

FlxObjReadLoops::FlxObjReadLoops() {
  // optional parameters
  AllDefParaBox->insert(new FlxOptionalParaFun(0.0,"loops::maxpasses"));
  ParaBox.insert("maxpasses", "loops::maxpasses" );
}

tuint FlxObjReadLoops::get_maxpasses() {
  void *v1 = ParaBox.get("maxpasses")->get();
  FlxFunction* fun = static_cast<FlxFunction*>( v1 ); 
  tuint i1 = fun->cast2tuintW0();
  delete fun;
  return i1;
}

FlxObjBase* FlxObjReadConst::read() {   
  const string cname = reader->getWord(false,false);
  return read(cname);
}

FlxObjBase* FlxObjReadConst::read(std::string cname) {
  isdefined(cname, 'C', false);
  reader->getChar('=',false);
  read_optionalPara(false);
  return new FlxObjConst(get_doLog(), cname, new FlxFunction(funReader,false) );
}

void FlxObjSub::task()
{
  data->SubBox.insert(sub_name, InternListSub); InternListSub = NULL;
  GlobalVar.slog(4) << "sub: defined procedure '" << sub_name << "()'." << std::endl;
}

FlxObjBase* FlxObjReadCtrlStrct::read_block(bool UseIgnoreList, bool errSerious) {
  if (UseIgnoreList) {
    data->IgnoreBox.activate_iL_recur();	// activate IgnoreList
  }
  reader->getChar('{',errSerious);
  FlxObjBase* obR = NULL;
  FlxObjBase* ob;
  try {
    while ( reader->whatIsNextChar() != '}' ) {
      ob = EvaluateCmd->evaluateCmd( );
      if ( obR == NULL ) obR = ob;
      else obR->attach_obj(ob);
    }
    reader->getChar('}',errSerious);
  } catch ( FlxException &e ) {
    if (obR) delete obR;
    throw;
  }
  if (obR == NULL) {
    obR = new FlxObjDummy();
  }
  if (UseIgnoreList) {
    data->IgnoreBox.deactivate_iL_recur();	// deactivate IgnoreList
  }
  return obR;
}

FlxObjBase* FlxObjReadSub::read() {
  string sub_name = reader->getWord(false,false);
  reader->getChar('(',false);
  reader->getChar(')',false);
  FlxObjBase* loopblock = read_block(true,false);
  read_optionalPara(false);
  return new FlxObjSub( get_doLog(), sub_name, loopblock);
}

FlxObjBase* FlxObjReadSubDo::read()
{
  string str1 = reader->getWord(false,false);
  FlxObjBase* obj1 = data->SubBox.get(str1);
  if ( obj1 == NULL ) {
    ostringstream ssV;
    ssV << "A procedure with the name ('" << str1 << "') does not exist.";
    FlxException_NeglectInInteractive("FlxObjReadSubDo::read", ssV.str(), reader->getCurrentPos());
  }
  return read(obj1);
}

FlxObjBase* FlxObjReadSubDo::read(FlxObjBase* objProcedure)
{
  reader->getChar('(',false);
  reader->getChar(')',false);
  read_optionalPara(false);
  return new FlxObjSubDo(get_doLog(),objProcedure);
}


