/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxostools.h"

#include <iostream>

#include <stdio.h>  /* defines FILENAME_MAX */
#ifdef __WINDOWS__
  #include <direct.h>
  #define GetCurrentDir _getcwd
#else
  #include <unistd.h>
  #define GetCurrentDir getcwd
#endif

#ifdef __unix__
  #include <pwd.h>
#elif defined(__WINDOWS__)
  #include <windows.h>
  #include "Userenv.h"
  #pragma comment(lib, "userenv.lib")
#else
  #error Operating system not handled.
#endif

#ifdef __unix__
  #include <termios.h>
#elif defined(__WINDOWS__)
  #include <stdlib.h>
#else
  #error Operating system not handled.
#endif

char cCurrentPath[FILENAME_MAX];

const bool get_config_dir_global(std::string& configDir)
{
  #ifdef __unix__
    configDir = "/etc/fesslix/";
  #elif defined(__WINDOWS__)
    configDir = GlobalVar.exe_dir_get();
  #endif
  return true;
}

const bool get_config_dir_local(std::string& configDir)
{
  #ifdef __unix__
    configDir = getpwuid(getuid())->pw_dir;
    configDir+="/.config/fesslix/";
  #elif defined(__WINDOWS__)
    configDir = "";
  #endif
  return true;
}

const bool get_pwd(std::string& pwdStr) 
{
  if (!GetCurrentDir(cCurrentPath, sizeof(cCurrentPath) / sizeof(char)))
  {
    return false;
  }
  cCurrentPath[sizeof(cCurrentPath) - 1] = '\0'; /* not really required */
  pwdStr = cCurrentPath;
  return true;
}

#ifdef __unix__
int mygetch ( void ) 
{
  int ch;
  struct termios oldt, newt;
  
  tcgetattr ( STDIN_FILENO, &oldt );
  newt = oldt;
  newt.c_lflag &= ~( ICANON | ECHO );
  tcsetattr ( STDIN_FILENO, TCSANOW, &newt );
  ch = getchar();
  tcsetattr ( STDIN_FILENO, TCSANOW, &oldt );
  
  return ch;
}
#endif

void take_a_break(std::ostream& sout)
{
  #ifdef __WINDOWS__
    system ( "PAUSE" );
  #elif defined(__unix__)
    std::cout<<"Press any key to continue . . ." ;
    mygetch();
    std::cout << std::endl;
  #else
    std::cout<<"Press [Enter] to continue . . .";
    std::cin.get();
  #endif
}

