/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxrandomFEM.h"


FlxRndBox* FlxBoxBaseR::RndBox = NULL;
GaussIntegration* FlxBoxBaseR::GI = NULL;
FlxRFBox* FlxRndBox::RFbox = NULL;

void FlxBoxBaseR::set_Boxes(FlxRndBox* RndBoxV)
{
  RndBox = RndBoxV;
}

tdouble calc_expectation_numerical_base::calc_expectation(const tnlong& N_Interv, const tnlong& N, const tdouble& pRedistribute, const FlxRnd_base& rnd)
{
  const tdouble LB=GlobalVar.Exp_Num_Lbound_1D, UB=GlobalVar.Exp_Num_Ubound_1D;
  
  if ( N<=1 || N_Interv<=1 ) {
    return calc_Interval(UB,LB,(N<=1)?N_Interv:N,rnd).cast2double();
  }
  
  const tdouble step=(UB-LB)/N_Interv;
  register tnlong i; 
  pVec I_vec(0.0,N_Interv);
  std::valarray<tnlong> N_Vec(1,N_Interv);
  pdouble Integ=0.0, t;
  tdouble x_u = LB, x_o = x_u+step;
  
  // first crude sampling
    for (i=0;i<N_Interv;++i) {
      t = calc_Interval(x_o,x_u,1,rnd);
      I_vec[i] = t;
      Integ += sqrt(fabs(t.cast2double()));
      x_u=x_o;
      x_o+=step;
    } 
  //redistribution
    // adaptive 
      tnlong max = 0;
      tnlong maxI = 0;
      tnlong count = 0;
      tdouble tt = Integ.cast2double();
      for (i=0;i<N_Interv;++i) {
	N_Vec[i] = tnlong(std::floor((sqrt(fabs((I_vec[i]).cast2double()))/tt)*N));
	if (N_Vec[i]>max) {
	  maxI = i;
	  max = N_Vec[i];
	}
	count+=N_Vec[i];
      }
      N_Vec[maxI] += N - count;
    // redistribute (diffusion)
      count = 0;
      tnlong rd = tnlong(N_Vec[0]*pRedistribute/2.0);
      N_Vec[0]-=rd;
      count +=N_Vec[0];
      for (i=1;i<N_Interv-1;++i) {
	N_Vec[i]+=rd;
	rd = tnlong((N_Vec[i]-rd)*pRedistribute/2.0);
	N_Vec[i]-=2*rd;
	N_Vec[i-1]+=rd;
	count += N_Vec[i]+rd;
      }
      N_Vec[N_Interv-1]+=rd;
      rd = tnlong((N_Vec[N_Interv-1]-rd)*pRedistribute/2.0);
      N_Vec[N_Interv-1]-=rd;
      N_Vec[N_Interv-2]+=rd;
      count += N_Vec[N_Interv-1]+rd;
      N_Vec[maxI] += N - count;
  // resampling
    Integ=0.0;
    x_u = LB;
    x_o = x_u+step;
    for (i=0;i<N_Interv;++i) {
      if (N_Vec[i]<=1) {
	Integ+=I_vec[i];
      } else {
	Integ += calc_Interval(x_o,x_u,N_Vec[i],rnd);
      }
      x_u=x_o;
      x_o+=step;
    } 
  
  return Integ.cast2double();
}

pdouble calc_expectation_numerical_base::calc_Interval(const tdouble& x_o, const tdouble& x_u, const tnlong& N, const FlxRnd_base& rnd)
{
  const tdouble step=(x_o-x_u)/N;
  const tdouble steph=step/2.0;
  
  tnlong NpI;
  if (N>=1e4) {
    NpI = tnlong(floor(sqrt(tdouble(N))));
  } else {
    NpI = N;
  }
  
  tdouble y = x_u, x;
  tdouble CDF_new, CDF_old;
  tdouble f_new, f_old;
  pdouble Integ=0.0, IntegI;
  // assign values
    x = rnd.y2x(y);
    f_old = calc(x,y);
    bool b_comp;
    if (y>0.0) {
      b_comp = true;
      CDF_old = rv_Phi(-y);
    } else {
      b_comp = false;
      CDF_old = rv_Phi(y);
    }
    y+=step;
  register tnlong i,j,j_start=1;
  #if FLX_EXPNUM_1D_QUAD_APPROX
    tdouble p,a,b,xm,x_new;
  #endif
  for (int k=(NpI==N)?2:1; k<=2; ++k) {
    for (j = (NpI==N)?NpI:j_start; j <= NpI; ++j) {
      IntegI=0.0;
      for (i = 0; i < NpI; ++i) {
	// at x_mitte
	  #if FLX_EXPNUM_1D_QUAD_APPROX
	    xm = rnd.y2x(y-steph);
	    x_new = rnd.y2x(y);
	    p = ((x_new-x)==0.0)?(0.0):(xm-x)/(x_new-x);
	    x = x_new;
	  #else
	    x = rnd.y2x(y);
	  #endif
	// at x_oben
	  CDF_new = (b_comp)?rv_Phi(-y):rv_Phi(y);
	  f_new = calc(x,y);
	// at x_mitte
	  #if FLX_EXPNUM_1D_QUAD_APPROX
	    if (p==0.0||p==1.0) {
	      a = 0.0;
	      b = 0.0;
	    } else {
	      a=(((b_comp)?rv_Phi(-y+steph):rv_Phi(y-steph))-CDF_new*p-CDF_old*(1-p))/(p*((b_comp)?(p-1):(1-p)));
	      b=(calc(xm,y-steph)-f_new*p-f_old*(1-p))/(p*(1-p));
	    }
	  #endif
	IntegI+= 0.5*((b_comp)?(CDF_old-CDF_new):(CDF_new-CDF_old))*(f_new + f_old);
	#if FLX_EXPNUM_1D_QUAD_APPROX
	  IntegI+=b/6*((b_comp)?(CDF_old-CDF_new):(CDF_new-CDF_old));
	  IntegI+=a/6*(f_old-f_new);
	#endif
	y += step;
	f_old = f_new;
	CDF_old = CDF_new;
      }
      Integ+=IntegI;
    }
    if (N!=NpI) {
      NpI = N-pow2(NpI);
      j_start = NpI;
    }
  }
  return Integ;
}

tdouble calc_expectation_numerical_Fun::calc_expectation(const tnlong& N_Interv, const tnlong& N, const tdouble& pRedistribute, const FlxRnd_base& rnd)
{
  rndB = &rnd;
  // remember original setting
    const tdouble orig_y = rnd.get_y_smp();
    const tdouble orig_x = rnd.get_y_smp();
    tuint index = RndBox->get(rnd);
    orig_x_Ptr = &((*(RndBox->x_smp))[index]);
    orig_y_Ptr = &((*(RndBox->y_smp))[index]);
//     std::cout << "fun: " << fun->write() << std::endl;
    
  tdouble returnv = calc_expectation_numerical_base::calc_expectation(N_Interv,N,pRedistribute,rnd);
  
  // restore original setting
    *orig_x_Ptr = orig_x;
    *orig_y_Ptr = orig_y;
    
  return returnv;
}

tdouble calc_expectation_numerical_Fun::calc(const tdouble& x, const tdouble& y)
{
  *orig_x_Ptr = x;
  *orig_y_Ptr = y;
  return fun->calc();
}

tdouble calc_expectation_numerical_MCI::calc_expectation(FunBase* fun, const bVec& bv, const tnlong& N_Interv, const tnlong& N_mciI_1, const tnlong& N_mciI_2, const tdouble& pRedistribute)
{
  funR = fun;
  // remember x and y
    const tVec orig_X = RndBox->get_x_Vec();
    const tVec orig_Y = RndBox->get_y_Vec();
    
  pdouble Integ(0), IntegI, IntegI2;
  pVec Ivec(N_Interv);
  register tnlong i,j,m;
  
  // get dimension of the problem
    DIM = bVec_NumbTrue(bv);
    tVec tempVec(DIM);
    Index_V = new iVec(tuint(0),DIM);
    iVec& IndexV = *Index_V;
    for (i=0,j=0;i<bv.size();++i) {
      if (bv[i]) {
	IndexV[j++]=i;
      }
    }
  // create correlation information
    L = RndBox->create_small_L(IndexV);
  // create reference vectors
    y_smp = new tdouble*[DIM];
    x_smp = new tdouble*[DIM];
    for (i=0;i<DIM;++i) {
      y_smp[i] = &(RndBox->y_smp->operator[](IndexV[i]));
      x_smp[i] = &(RndBox->x_smp->operator[](IndexV[i]));
    }
    
  const tdouble Bound=GlobalVar.Exp_Num_Bound_MCI;
  const tdouble step=Bound/N_Interv;
  tdouble h_inv = Bound;
  
  tdouble x_u, x_o;
  tnlong N_mciI_T, N_mciI_T_I;
  std::valarray<tnlong> N_Vec(N_mciI_1,N_Interv);
  uD = new FlxRnd_Uniform(-1.0,1.0);
  for (int k=1;k<=2;++k) {
    x_u=0.0; x_o = step;
    for (i=0;i<N_Interv;++i) {
      // change properties of uniform distribution
	uD->reset_bound(x_u,x_o);
      IntegI = 0.0;
      N_mciI_T = N_Vec[i];
      if (N_mciI_T>=1e4) {
	N_mciI_T_I = tnlong(floor(sqrt(tdouble(N_mciI_T))));
	for (int l=1;l<=2;++l) {
	  for (j=(l==2)?N_mciI_T_I-1:1;j<N_mciI_T_I;++j) {
	    IntegI2 = 0.0;
	    for (m=0;m<N_mciI_T_I;++m) {
	      IntegI2+=Integrationstep(tempVec);
	    }
	    IntegI+=IntegI2;
	  }
	  N_mciI_T_I = N_mciI_T - pow2(N_mciI_T_I) + N_mciI_T_I;
	}
      } else {
	for (j=0;j<N_mciI_T;++j) {
	  IntegI+=Integrationstep(tempVec);
	}
      }
      Ivec[i]+=IntegI;
      Integ+=sqrt(fabs(IntegI.cast2double()));	// only relevant for k=1
      x_u = x_o;
      x_o += step;
    }
    if (k==1 && N_Interv > 1) {	//redistribution
      // adaptive 
	tnlong max = 0;
	tnlong maxI = 0;
	tnlong count = 0;
	tdouble tt = Integ.cast2double();
	for (i=0;i<N_Interv;++i) {
	  N_Vec[i] = tnlong(std::floor((sqrt(fabs((Ivec[i]).cast2double()))/tt)*N_mciI_2));
	  if (N_Vec[i]>max) {
	    maxI = i;
	    max = N_Vec[i];
	  }
	  count+=N_Vec[i];
	}
	N_Vec[maxI] += N_mciI_2 - count;
      // redistribute (diffusion)
	count = 0;
	tnlong rd = tnlong(N_Vec[0]*pRedistribute/2.0);
	N_Vec[0]-=rd;
	count +=N_Vec[0];
	for (i=1;i<N_Interv-1;++i) {
	  N_Vec[i]+=rd;
	  rd = tnlong((N_Vec[i]-rd)*pRedistribute/2.0);
	  N_Vec[i]-=2*rd;
	  N_Vec[i-1]+=rd;
	  count += N_Vec[i]+rd;
	}
	N_Vec[N_Interv-1]+=rd;
	rd = tnlong((N_Vec[N_Interv-1]-rd)*pRedistribute/2.0);
	N_Vec[N_Interv-1]-=rd;
	N_Vec[N_Interv-2]+=rd;
	count += N_Vec[N_Interv-1]+rd;
	N_Vec[maxI] += N_mciI_2 - count;
    } else if (k==1 && N_Interv == 1) {
      N_Vec[0] = N_mciI_2;
    }
  }
  delete uD;
  
  // sum of everything
    Integ = 0.0;
    j = N_Interv*N_mciI_1+N_mciI_2;
    tdouble jm = tdouble(j)/tdouble(N_Interv);
    for (i=0;i<N_Interv;++i) {
      Integ += Ivec[i] * (jm/(N_mciI_1+N_Vec[i]));
    }
    tdouble result = Integ.cast2double() * h_inv / j;
  
  // restore x and y
    *(RndBox->x_smp) = orig_X;
    *(RndBox->y_smp) = orig_Y;
  // free memory
    funR = NULL;
    delete [] y_smp;
    delete [] x_smp;
    if (L!=NULL) delete L;
    delete Index_V;
  return result;
}

void calc_expectation_numerical_MCI::create_y(tVec& y)
{
  RndBox->RndCreator.gen_smp(y);
}

void calc_expectation_numerical_MCI::set_y(const tVec& y)
{
  for (register size_t i = 0; i < DIM; ++i) *(y_smp[i]) = y[i];
}

void calc_expectation_numerical_MCI::set_x(const tVec& x)
{
  for (register size_t i = 0; i < DIM; ++i) *(x_smp[i]) = x[i];
}

void calc_expectation_numerical_MCI::update_x(const tVec& y, tVec& tempVec)
{
  // transform uncorrelated y to correlated x
    tVec &x = tempVec;
    y2w(y,x);
  // transform standard normal x to some other distribution
    for (size_t i = 0; i < DIM; ++i) x[i] = RndBox->box[Index_V->operator[](i)]->y2x(x[i]);
  // set the vector
    set_x(x);
}

tdouble calc_expectation_numerical_MCI::Integrationstep(tVec& tempVec)
{
  // generate y-vector
    tVec &y = tempVec;
    create_y(y);

  // transform to z-space
    const tdouble ly2 = Norm2_NOroot(y);
    const tdouble lz2 = uD->Inv_cdf_x(rv_cdf_ChiSquare(DIM,ly2));
    tVec &z = tempVec;
    z*= sqrt(lz2/ly2);
  
  // calculate f
    tdouble f = rv_pdf_ChiSquare(DIM,lz2);
    
  // update standard normal space
    set_y(z);
    
  // update original space
    update_x(z, tempVec);
    
  // return function * foverh;
    return funR->calc()*f;	// h has to be multiplied afterwards
}

void calc_expectation_numerical_MCI::y2w(const tVec& y, tVec& w) const
{
  if (L==NULL) {
    w = y;
  } else {
    L->MultMv(y,w);  
  }
}

FlxRnd_base::FlxRnd_base() 
: y_smp(NULL),x_smp(NULL),automaticV(true),isRandomField(false),isRandomField_automatic(false),
num_moment_array_size(0), num_moment_central_array_size(0)
, num_moment_array(new tdouble[GlobalVar.Exp_Num_Rnd_Moments-1]), num_moment_central_array(new tdouble[GlobalVar.Exp_Num_Rnd_Moments_Central-2])
{

}

FlxRnd_base::~FlxRnd_base()
{
  delete[] num_moment_array;
  delete[] num_moment_central_array;
}

const tdouble FlxRnd_base::moment(const int nth)
{
  if (nth==0) return 1.0;
  else if (nth==1) return mean();
  else if (nth>1 && nth <= (int)GlobalVar.Exp_Num_Rnd_Moments) {
    if (nth-1 <= (int)num_moment_array_size) {
      return num_moment_array[nth-2];
    } else {
      for (tuint i=num_moment_array_size+1;i<(tuint)nth;++i) {
	calc_expectation_numerical_nth_moment tmp(i+1);
	num_moment_array[i-1] = tmp.calc_expectation(GlobalVar.Exp_Num_Intervals_1D,GlobalVar.Exp_Num_Samples_1D,GlobalVar.Exp_Num_Diff_Redistribute_1D,*this);
      }
      num_moment_array_size=nth-1;
      return num_moment_array[nth-2];
    }
  } else {
    calc_expectation_numerical_nth_moment tmp(nth);
    return tmp.calc_expectation(GlobalVar.Exp_Num_Intervals_1D,GlobalVar.Exp_Num_Samples_1D,GlobalVar.Exp_Num_Diff_Redistribute_1D,*this);
  }
}

const tdouble FlxRnd_base::moment_central(const int nth)
{
  if (nth==0) return 1.0;
  else if (nth==1) return 0.0;
  else if (nth==2) return pow2(sd());
  else if (nth>2 && nth <= (int)GlobalVar.Exp_Num_Rnd_Moments) {
    if (nth-2 <= (int)num_moment_central_array_size) {
      return num_moment_central_array[nth-3];
    } else {
      for (tuint i=num_moment_central_array_size+2;i<(tuint)nth;++i) {
	calc_expectation_numerical_nth_moment tmp(i+1,mean());
	num_moment_central_array[i-2] = tmp.calc_expectation(GlobalVar.Exp_Num_Intervals_1D,GlobalVar.Exp_Num_Samples_1D,GlobalVar.Exp_Num_Diff_Redistribute_1D,*this);
      }
      num_moment_central_array_size=nth-2;
      return num_moment_central_array[nth-3];
    }
  } else {
    calc_expectation_numerical_nth_moment tmp(nth,mean());
    return tmp.calc_expectation(GlobalVar.Exp_Num_Intervals_1D,GlobalVar.Exp_Num_Samples_1D,GlobalVar.Exp_Num_Diff_Redistribute_1D,*this);
  }
}

const tdouble FlxRnd_StdN::moment(const int nth)
{
  #if FLX_DEBUG	
    if (nth < 0) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxRnd_StdN::moment", ssV.str() );
    }
  #endif
  if (nth%2 == 1) return 0.0;
  else return DoubleFactorial(nth-1);
}

const tdouble FlxRnd_Normal::moment(const int nth)
{
  #if FLX_DEBUG	
    if (nth < 0) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxRnd_Normal::moment", ssV.str() );
    }
  #endif
  switch (nth) {
    case 0:
      return 1.0;
    case 1:
      return mu;
    case 2:
      return pow2(mu)+pow2(sigma);
    case 3:
      return pow(mu,3)+3*mu*pow2(sigma);
    case 4:
      return pow(mu,4)+6*pow2(mu)*pow2(sigma)+3*pow(sigma,4);
    case 5:
      return pow(mu,5)+10*pow(mu,3)*pow2(sigma)+15*mu*pow(sigma,4);  
    case 6:
      return pow(mu,6)+15*pow(mu,4)*pow2(sigma)+45*pow2(mu)*pow(sigma,4)+15*pow(sigma,6); 
    case 7:
      return pow(mu,7)+21*pow(mu,5)*pow2(sigma)+105*pow(mu,3)*pow(sigma,4)+105*mu*pow(sigma,6);   
    case 8:
      return pow(mu,8)+28*pow(mu,6)*pow2(sigma)+210*pow(mu,4)*pow(sigma,4)+420*pow2(mu)*pow(sigma,6)+105*pow(sigma,8);   
  }
  throw FlxException_NotImplemented("FlxRnd_Normal::moment");
}

const tdouble FlxRnd_Normal::moment_central(const int nth)
{
  #if FLX_DEBUG	
    if (nth < 0) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxRnd_Normal::moment", ssV.str() );
    }
  #endif
  if (nth%2 == 1) return 0.0;
  else return pow(sigma,nth)*DoubleFactorial(nth-1);
}

const tdouble FlxRnd_LogNormal::x2y(const tdouble x) const
{
  #if FLX_DEBUG
    if ( x <= epsilon ) {
      std::ostringstream ssV;
      ssV << "x (" << x << ") out of range.";
      throw FlxException("FlxRnd_LogNormal::x2y", ssV.str() );
    }
  #endif
  return (std::log(x-epsilon)-lambda)/zeta;
}

const tdouble FlxRnd_LogNormal::pdf_x(const tdouble x) const
{
  #if FLX_DEBUG
    if ( x <= epsilon) {
      std::ostringstream ssV;
      ssV << "x (" << x << ") out of range.";
      throw FlxException("FlxRnd_LogNormal::pdf_x", ssV.str() );
    }
  #endif
  return exp(-0.5*pow2((std::log(x-epsilon)-lambda)/zeta))/((x-epsilon)*zeta*sqrt(2*PI)); 
}

const tdouble FlxRnd_LogNormal::moment(const int nth)
{
  if (epsilon==0.0) {
    return exp(nth*lambda+0.5*pow2(nth*zeta));
  } else {
    return FlxRnd_base::moment(nth);
  }
}

void FlxRnd_LogNormal::print_para(std::ostream& sout) const
{
  sout << "log-normal: lambda=" << lambda << "; zeta=" << zeta;
  sout << "; epsilon=" << epsilon << ";";
  sout << " (mean=" << mean() << "; sd=" << sd() << ";)";
}

const bool FlxRnd_LogNormal::check_value(const tdouble& xval) const
{
  return (xval > epsilon);
}

FlxRnd_LogNormal::FlxRnd_LogNormal(const FlxRnd_LogNormal& Distri)
: lambda(Distri.lambda), zeta(Distri.zeta), epsilon(Distri.epsilon)
{
}

void FlxRnd_Gumbel::print_para(std::ostream& sout) const
{
  sout << "gumbel: u=" << u << "; alpha=" << alpha << ";";
  sout << " (mean=" << mean() << "; sd=" << sd() << ";)";
}

FlxRnd_Gumbel::FlxRnd_Gumbel(const FlxRnd_Gumbel& Distri)
: u(Distri.u), alpha(Distri.alpha)
{
}

void FlxRnd_Uniform::print_para(std::ostream& sout) const
{
  sout << "uniform: a=" << a << "; b=" << b << ";";
  sout << " (mean=" << mean() << "; sd=" << sd() << ";)";
}

const bool FlxRnd_Uniform::check_value(const tdouble& xval) const
{
  return ((xval>=a && xval<=b)?true:false);
}

const tdouble FlxRnd_Uniform::pdf_x(const tdouble x) const
{
  #if FLX_DEBUG
    if ( x < a || x > b) {
      std::ostringstream ssV;
      ssV << "x (" << x << ") out of range.";
      throw FlxException("FlxRnd_Uniform::pdf_x", ssV.str() );
    }
  #endif
  return 1/(b-a);
}

const tdouble FlxRnd_Uniform::cdf_x(const tdouble x) const
{
  #if FLX_DEBUG
    if ( x < a || x > b) {
      std::ostringstream ssV;
      ssV << "x (" << x << ") out of range.";
      throw FlxException("FlxRnd_Uniform::cdf_x", ssV.str() );
    }
  #endif
  return (x-a)/(b-a);
}

const tdouble FlxRnd_Uniform::Inv_cdf_x(const tdouble p) const
{
  #if FLX_DEBUG
    if ( p < 0.0 || p > 1.0) {
      std::ostringstream ssV;
      ssV << "p (" << p << ") out of range.";
      throw FlxException("FlxRnd_Uniform::Inv_cdf_x", ssV.str() );
    }
  #endif
  return a+p*(b-a);
}

FlxRnd_Uniform::FlxRnd_Uniform(const FlxRnd_Uniform& Distri)
: a(Distri.a), b(Distri.b),mean_v((a+b)/2)
{
}

const tdouble FlxRnd_Uniform::moment(const int nth)
{
  #if FLX_DEBUG	
    if (nth < 0) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxRnd_Uniform::moment", ssV.str() );
    }
  #endif
  return (pow(b,nth+1)-pow(a,nth+1))/((nth+1)*(b-a));
}

const tdouble FlxRnd_Uniform::moment_central(const int nth)
{
  #if FLX_DEBUG	
    if (nth < 0) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxRnd_Uniform::moment_central", ssV.str() );
    }
  #endif
  return (pow(b-mean_v,nth+1)-pow(a-mean_v,nth+1))/((nth+1)*(b-a));
}

FlxRnd_Exponential::FlxRnd_Exponential(const FlxRnd_Exponential& Distri)
:lambda(Distri.lambda),lambda_inv(Distri.lambda_inv)
{
  
}

const tdouble FlxRnd_Exponential::pdf_x(const tdouble x) const
{
  #if FLX_DEBUG
    if ( x < 0.0) {
      std::ostringstream ssV;
      ssV << "x (" << x << ") out of range.";
      throw FlxException("FlxRnd_Exponential::pdf_x", ssV.str() );
    }
  #endif
  return lambda*std::exp(-lambda*x);
}

const tdouble FlxRnd_Exponential::cdf_x(const tdouble x) const
{
  #if FLX_DEBUG
    if ( x < 0.0) {
      std::ostringstream ssV;
      ssV << "x (" << x << ") out of range.";
      throw FlxException("FlxRnd_Exponential::cdf_x", ssV.str() );
    }
  #endif
  return 1.-std::exp(-lambda*x);
}

void FlxRnd_Exponential::print_para(std::ostream& sout) const
{
  sout << "exponential: lambda=" << lambda << ";";
  sout << " (mean=" << mean() << "; sd=" << sd() << ";)";
}

const bool FlxRnd_Exponential::check_value(const tdouble& xval) const
{
  return ((xval>=0.)?true:false);
}

void FlxRndCreator::replay_start(FlxIstream* rndReaderV)
{
  replay_stop();
  rndReader = rndReaderV;
  GlobalVar.slog(3) << "rnd track: started replay of semi random values." << std::endl;
}

FlxRnd_Beta::FlxRnd_Beta(const tdouble alpha, const tdouble beta, const tdouble a, const tdouble b)
:alpha(alpha),beta(beta),a(a),b(b),
mu((alpha*b+beta*a)/(alpha+beta)),
lnbetafun(GammaLn(alpha)+GammaLn(beta)-GammaLn(alpha+beta))
{
  if (b<=a) {
    throw FlxException_Crude("FlxRnd_Beta::FlxRnd_Beta");
  }
}

FlxRnd_Beta::FlxRnd_Beta(const FlxRnd_Beta& Distri)
:alpha(Distri.alpha),beta(Distri.beta),a(Distri.a),b(Distri.b),mu(Distri.mu),lnbetafun(Distri.lnbetafun)
{

}

const tdouble FlxRnd_Beta::y2x(const tdouble y) const
{
  return (iBeta_reg_inv(alpha,beta,rv_Phi(y))*(b-a)+a);
}

const tdouble FlxRnd_Beta::x2y(const tdouble x) const
{
  const tdouble xx = (x-a)/(b-a);	// scale x to [0;1]
  return rv_InvPhi(iBeta_reg(alpha,beta,xx));
}

const tdouble FlxRnd_Beta::pdf_x(const tdouble x) const
{
  #if FLX_DEBUG
    if ( x < a || x > b) {
      std::ostringstream ssV;
      ssV << "x (" << x << ") out of range.";
      throw FlxException("FlxRnd_Beta::pdf_x", ssV.str() );
    }
  #endif
  const tdouble xx = (x-a)/(b-a);	// scale x to [0;1]
  return exp(
    (alpha-1.)*log(xx)+(beta-1.)*log(1.-xx)
    -lnbetafun
  );
}

const tdouble FlxRnd_Beta::cdf_x(const tdouble x) const
{
  #if FLX_DEBUG
    if ( x < a || x > b) {
      std::ostringstream ssV;
      ssV << "x (" << x << ") out of range.";
      throw FlxException("FlxRnd_Beta::cdf_x", ssV.str() );
    }
  #endif
  const tdouble xx = (x-a)/(b-a);	// scale x to [0;1]
  return iBeta_reg(alpha,beta,xx);
}

void FlxRnd_Beta::print_para(std::ostream& sout) const
{
  sout << "beta: alpha=" << GlobalVar.Double2String(alpha) << ", beta=" << GlobalVar.Double2String(beta);
  sout << "; a=" << GlobalVar.Double2String(a) << ", b=" << GlobalVar.Double2String(b);
}

const bool FlxRnd_Beta::check_value(const tdouble& xval) const
{
  return ((xval>=a && xval<=b)?true:false);
}

const tdouble FlxRnd_Beta::Inv_cdf_x(const tdouble p) const
{
  #if FLX_DEBUG
    if ( p < 0.0 || p > 1.0) {
      std::ostringstream ssV;
      ssV << "p (" << p << ") out of range.";
      throw FlxException("FlxRnd_Beta::Inv_cdf_x", ssV.str() );
    }
  #endif
  const tdouble x = iBeta_reg_inv(alpha,beta,p);
  return x*(b-a)+a;	// scale x to [a;b]
}

const tdouble FlxRndCreator::gen_smp()
{
  if (rndReader == NULL) {
    return rv_normal();
  } else {
    tdouble t;
    if (!rndReader->get_value(t)) {
      replay_stop();
      GlobalVar.alert.alert("FlxRndCreator::gen_smp","Replay of semi random values stopped.");
      return gen_smp();
    }
    return t;
  }
}

const tdouble FlxRndCreator::gen_smp_uniform()
{
  if (rndReader == NULL) {
    return rv_uniform();
  } else {
    return rv_Phi(gen_smp());
  }
}

void FlxRndCreator::gen_smp(tVec& y)
{
  if (rndReader == NULL) {
    rv_normal(y);
  } else {
    size_t L;
    if (!rndReader->get_vec(y,L)) {
      replay_stop();
      GlobalVar.alert.alert("FlxRndCreator::gen_smp","Replay of semi random values stopped.");
      for (register size_t i = L; i<y.size(); ++i) {
	y[i] = gen_smp();
      }
    }
  }
}

void FlxRndCreator::replay_stop(bool doAlert)
{
  if (rndReader==NULL) {
    if (doAlert) GlobalVar.alert.alert("FlxRndCreator::replay_stop","Replay of semi random values already stopped.");
  }
  if (rndReader!=NULL) {
    rndReader = NULL;
    GlobalVar.slog(3) << "rnd track: stopped replay of semi random values." << std::endl;
  }
}

const tdouble FlxHLRVstd::get_Realization()
{
  return RndBox->RndCreator.gen_smp();
}

void FlxCorrelation::y2w(const tVec& y, tVec& w) const
{
  if (L) {
    L->MultMv(y,w);
  } else {
    w = y;
  }
}

void FlxCorrelation::w2y(tVec& y, const tVec& w) const
{
  if (L) {
    L->MultInv(w,y);
  } else {
    if (&y==&w) return;
    y = w;
  }
}

void FlxCorrelation::set_corr(tuint i, tuint j, tdouble corrV)
{
  #if FLX_DEBUG
    if (rhoCorr==NULL) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxCorrelation::set_corr_1", ssV.str() );
    }
    if (i>=rhoCorr->nrows() || j>=rhoCorr->ncols() ) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxCorrelation::set_corr_2", ssV.str() );
    }
  #endif
  if (i==j) {
    std::ostringstream ssV;
    ssV << "The correlation of a random variable with itself is per definition '1'. It is not possible to change this value.";
    throw FlxException("FlxCorrelation::set_corr_3", ssV.str() );
  }
  rhoCorr->operator()(i,j) = corrV;
  deallocate_L();
  CorrSet = true;
}

const tdouble FlxCorrelation::get_corr(tuint i, tuint j) const
{
  #if FLX_DEBUG
    if (i>=rhoCorr->nrows() || j>=rhoCorr->ncols() ) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxCorrelation::get_corr_1", ssV.str() );
    }
  #endif
  if (i==j) {
    return 1.0;
  }
  return rhoCorr->operator()(i,j);
}

const tdouble FlxCorrelation::calc_R_for_rhoPrime(FlxRnd_base* rv1, FlxRnd_base* rv2, const tdouble& rho)
{
  FlxRnd_base::rvtype t1 = rv1->get_Type();
  FlxRnd_base::rvtype t2 = rv2->get_Type();
  if (t1>t2) {
    std::swap(t1,t2);
    std::swap(rv1,rv2);
  }
  if ( t1==FlxRnd_base::rvStdN || t1==FlxRnd_base::rvNormal ) {
    if (t2 ==FlxRnd_base::rvStdN || t2==FlxRnd_base::rvNormal ) {
      return 1.0;
    } else if ( t2==FlxRnd_base::rvUnifrom ) {
      return 1.023;
    } else if ( t2==FlxRnd_base::rvGumbel ) {
      return 1.031;
    } else if ( t2==FlxRnd_base::rvLogN ) {
      FlxRnd_LogNormal* rv2_ln = dynamic_cast<FlxRnd_LogNormal*>(rv2);
      tdouble V_j = rv2_ln->get_CoeffOfVar_withoutEpsilon();
      return V_j/sqrt(log(1+pow2(V_j)));
    }
  } else if (t1==FlxRnd_base::rvLogN) {
    if (t2==FlxRnd_base::rvLogN) {
      FlxRnd_LogNormal* rv_ln = dynamic_cast<FlxRnd_LogNormal*>(rv1);
      tdouble V_i = rv_ln->get_CoeffOfVar_withoutEpsilon();
      rv_ln = dynamic_cast<FlxRnd_LogNormal*>(rv2);
      tdouble V_j = rv_ln->get_CoeffOfVar_withoutEpsilon();
      return log(1+rho*V_i*V_j)/rho/sqrt(log(1+pow2(V_i))*log(1+pow2(V_j)));
    } else if ( t2==FlxRnd_base::rvUnifrom ) {
      FlxRnd_LogNormal* rv_ln = dynamic_cast<FlxRnd_LogNormal*>(rv1);
      tdouble V_i = rv_ln->get_CoeffOfVar_withoutEpsilon();
      return 1.019+0.010*pow2(rho)+0.014*V_i+0.249*pow2(V_i);
    } else if ( t2==FlxRnd_base::rvGumbel ) {
      FlxRnd_LogNormal* rv_ln = dynamic_cast<FlxRnd_LogNormal*>(rv1);
      tdouble V_i = rv_ln->get_CoeffOfVar_withoutEpsilon();
      return 1.029+0.001*rho+0.004*pow2(rho)+0.014*V_i+0.233*pow2(V_i)-0.197*V_i*rho;
    }
  } else if (t1==FlxRnd_base::rvGumbel) {
    if ( t2==FlxRnd_base::rvGumbel ) {
      return 1.064-0.069*rho+0.005*pow2(rho);
    } else if ( t2==FlxRnd_base::rvUnifrom ) {
      return 1.055+0.015*pow2(rho);
    }
  } else if (t1==FlxRnd_base::rvUnifrom) {
    if ( t2==FlxRnd_base::rvUnifrom ) {
      return 1.047-0.047*pow2(rho);
    }
  }
  
  throw FlxException_NotImplemented("FlxCorrelation::calc_rhoPrime");
}

void FlxCorrelation::create_L(const bool onlyFirst)
{
  if (CorrSolved && onlyFirst) {
    return;
  } else {
    if (rhoCorr==NULL) {
      std::ostringstream ssV_2;
      ssV_2 << "No random variables exist!";
      throw FlxException("FlxCorrelation::create_L", ssV_2.str() ); 
    }
  }
  if (rhoCorr==NULL) return;
  deallocate_L(false);
  CorrSolved=true;
  // Nataf-Transformation
    tuint DIM = rhoCorr->nrows();
    FlxMtxSparsSym* rhoPrime = new FlxMtxSparsSym(*rhoCorr);
    tdouble rho, rhoPt;
    tuint i,j;
    tnlong count=0;
    for (i=0;i<DIM;++i) {
      for (j=0;j<i;++j) {
	rho = rhoPrime->operator()(i,j);
	if (rho!=0.0) {
	  rhoPt = calc_R_for_rhoPrime(RndBox->get(i),RndBox->get(j),rho)*rho;
	  rhoPrime->set_value(i,j, rhoPt );
	  ++count;
	}	
      }
    }
    // check for diagonal matrix
      if (count==0) {
	delete rhoPrime;
	return;
      }
  // CholeskyDecomposition
    FlxMtxLTri* Lt = new FlxMtxLTri(rhoPrime->nrows());
    try {
      Lt->CholeskyDec(*rhoPrime);
    } catch (FlxException& e) {
      delete Lt;
      delete rhoPrime;
      throw;
    }
    L = new FlxMtxSparsLTri(*Lt);
  // free memory
    delete Lt;
    delete rhoPrime;
}

FlxMtxSparsLTri* FlxCorrelation::create_small_L(const iVec& indexV)
{
  if (rhoCorr==NULL) {
    return NULL;
  }
  tuint i,j;
  // Nataf-Transformation
    tuint DIM = indexV.size();
    FlxMtxSym* rhoCorr_small = new FlxMtxSym(DIM);
    for (i=0;i<DIM;++i) {
      for (j=0;j<=i;++j) {
	rhoCorr_small->operator()(i,j) = rhoCorr->operator()(indexV[i],indexV[j]);
      }
    }
    FlxMtxSparsSym* rhoPrime = new FlxMtxSparsSym(*rhoCorr_small);

    tdouble rho, rhoPt;
    tnlong count=0;
    for (i=0;i<DIM;++i) {
      for (j=0;j<i;++j) {
	rho = rhoPrime->operator()(i,j);
	if (rho!=0.0) {
	  rhoPt = calc_R_for_rhoPrime(RndBox->get(indexV[i]),RndBox->get(indexV[j]),rho)*rho;
	  rhoPrime->set_value(i,j, rhoPt );
	  ++count;
	}
      }
    }
    // check for diagonal matrix
      if (count == 0) {
	delete rhoCorr_small;
	delete rhoPrime;
	return NULL;
      }
      
  // CholeskyDecomposition
    FlxMtxLTri* Lt = new FlxMtxLTri(rhoPrime->nrows());
    Lt->CholeskyDec(*rhoPrime);
    FlxMtxSparsLTri* L_return = new FlxMtxSparsLTri(*Lt);
  // free memory
    delete rhoCorr_small;
    delete Lt;
    delete rhoPrime;
  return L_return;
}

void FlxCorrelation::insertRV(const tuint& nRND)
{
  #if FLX_DEBUG
    if (rhoCorr!=NULL) {
      if (nRND<=rhoCorr->nrows()) {
	std::ostringstream ssV;
	ssV << "ERROR";
	throw FlxException("FlxCorrelation::insertRV_1", ssV.str() );
      }
    }
  #endif
  if (rhoCorr) {
    FlxMtxSym* rcT = rhoCorr;
    allocate_rhoCorr(nRND);
    // copy old matrix into new matrix
      for (tuint i=1; i<rcT->nrows();++i) {
	for (tuint j=0; j<i;++j) {
	  rhoCorr->operator()(i,j) = rcT->operator()(i,j);
	}
      }
    delete rcT;
  } else {
    allocate_rhoCorr(nRND);
  }
  deallocate_L();
}

void FlxCorrelation::deallocate_L(bool warning)
{
  if ( CorrSolved && warning ) GlobalVar.alert.alert("FlxCorrelation::deallocate_L","You have to run 'rnd corr create;' again !!!");
  if (L) {
    delete L; L=NULL;
    CorrSolved = false;
  }
}

void FlxCorrelation::allocate_rhoCorr(const tuint& nRND)
{
  rhoCorr = new FlxMtxSym(nRND);
  rhoCorr->set_zeroMtx();
  for (tuint i=0; i<nRND;++i) {
    rhoCorr->operator()(i,i) = 1.0;
  }
}

FlxCorrelation::~FlxCorrelation()
{
  if (CorrSet && !CorrSolved) {
    GlobalVar.alert.alert("FlxCorrelation::~FlxCorrelation", "'rnd corr create;' has not been executed !!!");
  }
  if (rhoCorr) delete rhoCorr;
  if (L) delete L;
}

FlxRndBox::FlxRndBox() : y_smp(NULL), x_smp(NULL), isLocked(false), boxM_size_inclRF(0)
{
  #if FLX_DEBUG
    ++Cinst;
    if (Cinst > 1) {
      std::ostringstream ssV;
      ssV << "More than one instance of 'FlxRndBox' created ...";
      throw FlxException("FlxRndBox::FlxRndBox", ssV.str() );
    }
  #endif
  FlxBoxBaseR::set_Boxes( this );
  
  insert("std", new FlxHLRVstd() ); 	// create the High Level Random Variable 'std'
}

FlxRndBox::~FlxRndBox()
{
  for (size_t i = 0; i < box.size(); i++) delete box[i];
  
  for ( std::map<std::string, FlxHLRVbase*>::iterator iter = boxH.begin(); iter != boxH.end(); ++iter ) {
    delete iter->second;
  }
  
  if (x_smp!=NULL) delete x_smp;
  if (y_smp!=NULL) delete y_smp;
}

FlxRnd_base* FlxRndBox::get(tuint index)
{
  #if FLX_DEBUG
    if ( index < 0 || box.size() <= index ) {
      std::ostringstream ssV;
      ssV << "Random variable with index '" << index << "' does not exist.";
      throw FlxException("FlxRndBox::get_1", ssV.str() );
    }
  #endif
  return box[index];
}

FlxHLRVbase* FlxRndBox::get(std::string index)
{
  std::transform(index.begin(), index.end(), index.begin(), (int(*)(int)) std::tolower);
  std::map<std::string, FlxHLRVbase*>::iterator pos = boxH.find(index);
  if ( pos != boxH.end() ) {
    return (pos->second);
  } else {
    std::ostringstream ssV_2;
    ssV_2 << "A 'High Level Random Variable' with the name  '" << index << "' has not yet been defined.";
    throw FlxException("FlxRndBox::get_2", ssV_2.str()); 
  }
}

const tuint FlxRndBox::get(const FlxRnd_base& rndV)
{
  for (size_t i = 0; i < box.size(); i++) {
    if (box[i] == &rndV) {
      return i;
    }
  }
  std::ostringstream ssV_2;
  ssV_2 << "The random variable is not a managed random variable.";
  throw FlxException("FlxRndBox::get_3", ssV_2.str()); 
}

const bool FlxRndBox::has_userDefined()
{
  if (boxM.size() > 0) return true;
  else return false;
}

const tuint FlxRndBox::get_TrueIndex_from_User(tuint index)
{
  std::map<tuint, tuint>::iterator pos = boxM.find(index);
  if ( pos != boxM.end() ) {
    return pos->second;
  } else {
    std::ostringstream ssV;
    ssV << "Random variable '" << index << "' has not been defined.";
    throw FlxException("FlxRndBox::get_UserIndex", ssV.str() );
  }
}

void FlxRndBox::gen_smp()
{
  for (size_t i = 0; i < box.size(); ++i) (*y_smp)[i] = RndCreator.gen_smp();
  y2x(*y_smp,*x_smp);
}

void FlxRndBox::get_x_Vec(tVec& x) const
{
  #if FLX_DEBUG
    if (x.size() != x_smp->size()) {
      std::ostringstream ssV;
      ssV << "Vector has wrong size.";
      throw FlxException("FlxRndBox::get_x_Vec", ssV.str() );
    }
  #endif
  for (size_t i = 0; i < x_smp->size(); ++i) {
    x[i] = x_smp->operator[](i);
  }
}

const tVec FlxRndBox::get_x_Vec_UD() const
{
  tVec t(boxM_size_inclRF);
  size_t tc = 0;
  for (size_t i = 0; i < box.size(); ++i) {
    if (box[i]->is_automatic()==false || (box[i]->is_automatic() && box[i]->is_RandomField() && !(box[i]->is_RandomField_automatic())) ) {
      t[tc++] = x_smp->operator[](i);
    }
  }
  return t;
}

const tVec FlxRndBox::get_y_Vec_UD() const
{
  tVec t(boxM_size_inclRF);
  size_t tc = 0;
  for (size_t i = 0; i < box.size(); ++i) {
    if (box[i]->is_automatic()==false || (box[i]->is_automatic() && box[i]->is_RandomField() && !(box[i]->is_RandomField_automatic())) ) {
      t[tc++] = y_smp->operator[](i);
    }
  }
  return t;
}

const tVec FlxRndBox::get_mean_Vec(bool output_just_user)
{
  if (output_just_user) {
    tVec X(boxM.size());
    size_t Xc = 0;
    for (size_t i = 0; i < box.size(); ++i) {
      if (box[i]->is_automatic()==false) {
	X[Xc++] = box[i]->mean();
      }
    }
    return X;
  } else {
    tVec X(box.size());
    for (size_t i = 0; i < box.size(); ++i) {
      X[i] = box[i]->mean();
    }
    return X;
  }
}

const tVec FlxRndBox::get_sd_Vec(bool output_just_user)
{
  if (output_just_user) {
    tVec X(boxM.size());
    size_t Xc = 0;
    for (size_t i = 0; i < box.size(); ++i) {
      if (box[i]->is_automatic()==false) {
	X[Xc++] = box[i]->sd();
      }
    }
    return X;
  } else {
    tVec X(box.size());
    for (size_t i = 0; i < box.size(); ++i) {
      X[i] = box[i]->sd();
    }
    return X;
  }
}

void FlxRndBox::corr_set(tuint i_udRV, tuint j_udRV, tdouble corrV)
{
  const tuint i = get_TrueIndex_from_User(i_udRV);
  tuint j = get_TrueIndex_from_User(j_udRV);
  if (box[i]->is_RandomField()) {
    std::ostringstream ssV;
    ssV << "It is not possible to set the correlation of a variable belonging to a random field.";
    throw FlxException("FlxRndBox::corr_set", ssV.str() );
  }
  corr.set_corr(i,j,corrV);
}

const tdouble FlxRndBox::corr_get(const tuint i, const tuint j) const
{
  return corr.get_corr(i,j);
}

void FlxRndBox::corr_create(bool onlyFirst)
{
  corr.create_L(onlyFirst);
}

const tdouble FlxRndBox::calc_expectation_numerical(FunBase* fun, bool silent)
{ 
  tuint DIM = get_NumbRnd();
  // determine dependence 
    bVec bv(false,DIM);
      fun->dependOn_RV(bv);
      tuint index;
  // check if expression depends on just one value
    if (check_for_single_dependence(index,bv)) {
	if (!silent) GlobalVar.slog(4) << "  numerical integration (1D) of E[" << fun->write() << "]" << std::endl;
	calc_expectation_numerical_Fun eF(fun);
	return eF.calc_expectation(GlobalVar.Exp_Num_Intervals_1D,GlobalVar.Exp_Num_Samples_1D,GlobalVar.Exp_Num_Diff_Redistribute_1D,*box[index]);
    } else {
      if (index==DIM) {	// no random variables !!!
	return fun->calc();
      }
    }
      
  // do a Monte carlo estimation ...
    if (!silent) {
      GlobalVar.slog(4) << "  numerical integration (MCI) of E[" << fun->write() << "]"; 
      GlobalVar.slog(4) << " (intv=" << GlobalVar.Double2String(GlobalVar.Exp_Num_Intervals_MCI) << "; spi=" << GlobalVar.Double2String(GlobalVar.Exp_Num_SmpPerInterval_MCI) << "; srd=" << GlobalVar.Double2String(GlobalVar.Exp_Num_Smp_Redistribute_MCI) << "; drd=" << GlobalVar.Double2String(GlobalVar.Exp_Num_Diff_Redistribute_MCI) << ")" << std::endl;
    }
    calc_expectation_numerical_MCI eF;
    return eF.calc_expectation(fun,bv,GlobalVar.Exp_Num_Intervals_MCI,GlobalVar.Exp_Num_SmpPerInterval_MCI,GlobalVar.Exp_Num_Smp_Redistribute_MCI,GlobalVar.Exp_Num_Diff_Redistribute_MCI);
}

FlxMtxSparsLTri* FlxRndBox::calc_Jinv()
{
  tuint DIM=get_NumbRnd();
  tVec w(DIM);
  const tVec &x=get_x_Vec();
  corr.y2w(get_y_Vec(),w);
//   std::cout << "y=" << get_y_Vec() << std::endl << "w=" << w << std::endl;
  // calc dx/dw (diagonal matrix!)
    FlxMtxDiag dxdw(DIM);
    for (tuint i=0;i<DIM;++i) {
      dxdw(i) = rv_phi(w[i])/(get(i)->pdf_x(x[i]));
    }
//     std::cout << "dxdw: " << dxdw << std::endl;
  // calc dw/dz (triangular matrix)
    const FlxMtxSparsLTri* Ji = corr.calc_Jinv();
    if (Ji==NULL) {
      return new FlxMtxSparsLTri(dxdw);
    } else {
      FlxMtxSparsLTri* Jinv = (*Ji).MultM_Mdiag(dxdw);
      return Jinv; 
    }
}

void FlxRndBox::x2y(tVec& y, const tVec& x) const
{
  size_t DIM = get_NumbRnd();
  #if FLX_DEBUG
    if (y.size()!= DIM || x.size()!=DIM ) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxRndBox::x2y", ssV.str() );
    }
  #endif
  // transform an correlated arbitrary distribution x to a correlated standard normal distribution w
    for (tuint i = 0; i < DIM; ++i) {
      y[i] = box[i]->x2y(x[i]);
    }
  // get rid of the correlation
    corr.w2y(y,y);
}

void FlxRndBox::y2x(const tVec& y, tVec& x) const
{
  size_t DIM = get_NumbRnd();
  #if FLX_DEBUG
    if (y.size()!= DIM || x.size()!=DIM ) {
      std::ostringstream ssV;
      ssV << "ERROR";
      throw FlxException("FlxRndBox::y2x", ssV.str() );
    }
  #endif
  // transform uncorrelated y to correlated x
    corr.y2w(y,x);
  // transform standard normal x to some other distribution
  for (tuint i = 0; i < DIM; ++i) {
    x[i] = box[i]->y2x(x[i]);
  }
}

void FlxRndBox::set_x_Vec(const tVec& x)
{
  #if FLX_DEBUG
    if (x.size() != get_NumbRnd()) {
      std::ostringstream ssV;
      ssV << "Vector sizes do not match.";
      throw FlxException("FlxRndBox::set_x_Vec_1", ssV.str() );
    }
  #endif
  *x_smp = x;
  x2y(*y_smp,*x_smp);
}

const void FlxRndBox::set_y_Vec(const tVec& y)
{
  #if FLX_DEBUG
    if (y.size() != get_NumbRnd()) {
      std::ostringstream ssV;
      ssV << "Vector sizes do not match.";
      throw FlxException("FlxRndBox::set_y_Vec_1", ssV.str() );
    }
  #endif
  if (y_smp!=&y) *y_smp = y;
  y2x(*y_smp,*x_smp);
}

const bool FlxRndBox::check_xVec(const tVec& xVec) const
{
  bool b1 = true;
  for (size_t i=0;i<box.size(); ++i) {
    b1 = b1 && box[i]->check_value(xVec[i]);
  }
  return b1;
}

const tuint FlxRndBox::insert(tuint index, FlxRnd_base* value)
{
  check_lock();
  value->set_UserVar();  // this variable has an user defined index ... and therefore is an user defined variable
  std::pair<tuint, tuint>Element(index, box.size());
  if ( ! boxM.insert(Element).second ) {;
    delete value;
    std::ostringstream ssV;
    ssV << "Random variable '" << index << "' has already been defined.";
    throw FlxException("FlxRndBox::insert_1", ssV.str() );
  }
  ++boxM_size_inclRF;
  GlobalVar.slog(3) << "Random variable: created new user defined RV with index '" << index << "'." << std::endl;
  GlobalVar.slog(4) << "  Type of distribution of RV: ";
  value->print_para(GlobalVar.slog(4));
  GlobalVar.slog(4) << std::endl;
  return insert(value);
}

void FlxRndBox::insert(std::string index, FlxHLRVbase* value)
{
  std::transform(index.begin(), index.end(), index.begin(), (int(*)(int)) std::tolower);
  std::pair<std::string, FlxHLRVbase*>Element(index, value);
  if ( ! boxH.insert(Element).second ) {;
    delete value;
    std::ostringstream ssV;
    ssV << "A 'High Level Random Variable' with the name '" << index << "' has already been defined.";
    throw FlxException("FlxRndBox::insert_2", ssV.str() );
  }
  GlobalVar.slog(3) << "High level RV: created new user HLRV with name '" << index << "'." << std::endl;
  GlobalVar.slog(4) << "  Distributed according to: " << value->print_info() << std::endl;
}

void FlxRndBox::set_lock(bool lockIt)
{
  if (lockIt) {
    ++isLocked;
  } else {
    --isLocked;
    #if FLX_DEBUG
      if (isLocked < 0) {
	std::ostringstream ssV;
	ssV << "ERROR";
	throw FlxException("FlxRndBox::set_lock", ssV.str() );
      }
    #endif
  }
}

void FlxRndBox::check_lock()
{
  if ( isLocked > 0 ) {
    std::ostringstream ssV;
    ssV << "At this point it is not possible to create managed random variables anymore.";
    throw FlxException("FlxRndBox::check_lock", ssV.str() );
  }
}

void FlxRndBox::set_RFbox(FlxRFBox* RFboxV)
{
  RFbox = RFboxV;
}

const tuint FlxRndBox::insert(FlxRnd_base* value)
{
  check_lock();
  box.push_back(value);
  // correct pointers to samples (x_smp, y_smp)
    const size_t DIM = box.size();
    tVec* x_smpT = new tVec(DIM);
    tVec* y_smpT = new tVec(DIM);
    for (size_t i=0;i<DIM-1;++i) {
      x_smpT->operator[](i) = x_smp->operator[](i);
      y_smpT->operator[](i) = y_smp->operator[](i);
    }
    x_smpT->operator[](DIM-1) = box[DIM-1]->mean();
    if (x_smp!=NULL) delete x_smp;
    if (y_smp!=NULL) delete y_smp;
    x_smp = x_smpT;
    y_smp = y_smpT;
    for (size_t i=0; i<DIM; ++i) {
      box[i]->x_smp = &(x_smp->operator[](i));
      box[i]->y_smp = &(y_smp->operator[](i));
    }
    RFbox->update_RV_index();
  // correct correlation matrix
    corr.insertRV(DIM);
  return DIM-1;
}

const tuint FlxRndBox::insert_RFv(std::vector< FlxRnd_base* >& boxRF)
{
  check_lock();
  const size_t index_start = box.size();
  for (size_t i=0; i<boxRF.size();++i) box.push_back(boxRF[i]);
  boxM_size_inclRF+=boxRF.size();
  // correct pointers to samples (x_smp, y_smp)
    const size_t DIM = box.size();
    tVec* x_smpT = new tVec(DIM);
    tVec* y_smpT = new tVec(DIM);
    for (size_t i=0;i<index_start;++i) {
      x_smpT->operator[](i) = x_smp->operator[](i);
      y_smpT->operator[](i) = y_smp->operator[](i);
    }
    for (size_t i=index_start; i<box.size(); ++i) {
      x_smpT->operator[](i) = box[i]->mean();
    }
    if (x_smp!=NULL) delete x_smp;
    if (y_smp!=NULL) delete y_smp;
    x_smp = x_smpT;
    y_smp = y_smpT;
    for (size_t i=0; i<DIM; ++i) {
      box[i]->x_smp = &(x_smp->operator[](i));
      box[i]->y_smp = &(y_smp->operator[](i));
    }
    RFbox->update_RV_index();
  // correct correlation matrix
    corr.insertRV(DIM);
  return index_start;
}




