/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxstring.h"

using namespace std;

FlxFunctionReader* FlxReaderBase2::funReader = NULL;
FlxStringFunBox* FlxString::StrFunBox = NULL;

FlxString::FlxString(FlxString_Base* strB) :
strLst(new std::list<FlxString_Base*>),instances(new tuint(0))
{
  strLst->push_back(strB);
}

FlxString::FlxString(const bool multiline, const bool errSerious) :
strLst(new std::list<FlxString_Base*>),instances(new tuint(0))
{
  bool b1 = false;
  try {
    do {
      if (b1) reader->getChar('&',errSerious);
      else b1 = true;
      FlxString_Base* strB;
      if (reader->whatIsNextChar() == '"' ) {
	strB = new FlxString_String(reader->getText(multiline,errSerious),false);
      } else if ( reader->whatIsNextChar() == '{' ) {
	reader->getChar('{',errSerious);
	strB = new FlxString_Fun(new FlxFunction(funReader,errSerious));
	reader->getChar('}',errSerious);
      } else if ( reader->whatIsNextChar() == '$' ) {
	reader->getChar('$',errSerious);
	strB = new FlxString_StrFun(StrFunBox->read(reader,errSerious));
      } else {
	strB = new FlxString_String(reader->getWord(false,errSerious),true);
      }
      strLst->push_back(strB);
    } while (reader->whatIsNextChar() == '&');
  } catch (FlxException &e) {
    std::list<FlxString_Base*>::iterator iter;
    for (iter = strLst->begin(); iter != strLst->end(); ++iter) {
      delete (*iter);
    }
    delete strLst;
    delete instances;
    throw;
  }
}

FlxString::FlxString(const FlxString& FlxStrF)
{
  instances = FlxStrF.instances;
  strLst = FlxStrF.strLst;
  (*instances)++;
}

FlxString& FlxString::operator=(const FlxString& FlxStrF)
{
  if ( this != &FlxStrF ) {
    instances = FlxStrF.instances;
    strLst = FlxStrF.strLst;
    (*instances)++;
  }
  return *this;
}

FlxString::~FlxString()
{
  if (instances == NULL) return;
  if ( (*instances) == 0)  {
    if (strLst!=NULL) {
      std::list<FlxString_Base*>::iterator iter;
      for (iter = strLst->begin(); iter != strLst->end(); ++iter) {
	delete (*iter);
      }
      delete strLst;
    }
    if ( instances != NULL ) delete instances;
  } else {
    (*instances)--;
  }
}

void FlxString::eval(ostream& os)
{
  std::list<FlxString_Base*>::iterator iter;
  for (iter = strLst->begin(); iter != strLst->end(); ++iter) {
    (*iter)->eval(os);
  }
}

const std::string FlxString::eval()
{
  std::ostringstream ssV;
  eval(ssV);
  return ssV.str();
}

const std::string FlxString::eval_word(const bool lowercase, const bool emptyAllow, const bool numbegallow)
{
  std::string strV = eval();
  if (strV.length() == 0) {
    if (emptyAllow) {
      return "";
    } else {
      std::ostringstream ssV_2;
      ssV_2 << "Word must not be empty.";
      throw FlxException("FlxString::eval_word_1", ssV_2.str()); 
    }
  }
  tuint i_start = 0;
  if (!numbegallow) {
    if ( ReadStream::getType( strV[0] ) != ReadStream::STRING ) {
      std::ostringstream ssV_2;
      ssV_2 << "Evaluated string '" << strV << "' is not of type 'word'.";
      throw FlxException("FlxString::eval_word_2", ssV_2.str()); 
    }
    i_start = 1;
  }
  for ( std::string::size_type i = i_start; i < strV.length(); ++i) {
    ReadStream::InpType iT = ReadStream::getType( strV[i] );
    if ( iT != ReadStream::STRING && iT != ReadStream::NUMERAL ) {
      std::ostringstream ssV_2;
      ssV_2 << "Evaluated string '" << strV << "' is not of type 'word'.";
      throw FlxException("FlxString::eval_word_3", ssV_2.str()); 
    }
  }
  if (lowercase) {
    std::transform(strV.begin(), strV.end(), strV.begin(), (int(*)(int)) std::tolower);
  }
  return strV;
}

const std::string FlxString::write()
{
  std::string strV = "";
  bool b1 = false;
  std::list<FlxString_Base*>::iterator iter;
  for (iter = strLst->begin(); iter != strLst->end(); ++iter) {
    if (b1) {
      strV+="&";
    } else {
      b1 = true;
    };
    strV+= (*iter)->write();
  }
  return strV;
}

void FlxString::assign(FlxString* strA)
{
  if (this == strA) return;
  if (this->strLst == strA->strLst) return;
  if ((*instances)==0) {
    std::list<FlxString_Base*>::iterator iter;
    for (iter = strLst->begin(); iter != strLst->end(); ++iter) {
      delete (*iter);
    }
    delete strLst; 
    delete instances;
  } else {
    (*instances)--;
  }
  strLst = strA->strLst; 
  strA->strLst = NULL;
  instances = strA->instances;
  strA->instances = NULL;
  delete strA;
}

const bool FlxString::search_circref(FlxFunction* fcr)
{
  std::list<FlxString_Base*>::iterator iter;
  for (iter = strLst->begin(); iter != strLst->end(); ++iter) {
    if ((*iter)->search_circref(fcr)) {
      return true;
    }
  }
  return false;
}

string FlxString_String::write()
{
  if (isWord) {
    return strV;
  } else {
    return "\"" + strV + "\"";
  }
}


