/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#ifndef fesslix_flxstring_H
#define fesslix_flxstring_H

#include "flxstringfun.h"

#include <list>

class FlxReaderBase2 : public FlxReaderBase {
  protected:
    static FlxFunctionReader *funReader;
  public:
    virtual ~FlxReaderBase2() {};
};


class FlxString_Base {
  public:
    virtual ~FlxString_Base() {};
    virtual void eval(std::ostream& os) = 0;
    virtual bool search_circref(FlxFunction* fcr) { return false; }
    virtual std::string write() = 0;
};

class FlxString_String : public FlxString_Base {
  private:
    std::string strV;
    bool isWord;
  public:
    FlxString_String(std::string strV, bool isWord) : strV(strV), isWord(isWord) {}
    void eval(std::ostream& os) { os << strV;}
    std::string write();
};

class FlxString_Fun : public FlxString_Base {
  private:
    FlxFunction* fun;
  public:
    FlxString_Fun(FlxFunction* fun) : fun(fun) {}
    ~FlxString_Fun() { delete fun; }
    void eval(std::ostream& os) { os << GlobalVar.Double2String(fun->calc());}
    bool search_circref(FlxFunction* fcr) { return fun->search_circref(fcr); }
    std::string write() { return "{" + fun->write() + "}"; }
};

class FlxString_StrFun : public FlxString_Base {
  private:
    StringFunBase* strFun;
  public:
    FlxString_StrFun(StringFunBase* strFun) : strFun(strFun) {}
    ~FlxString_StrFun() { delete strFun; }
    void eval(std::ostream& os) { os << strFun->eval();}
//     bool search_circref(FlxFunction* fcr) { return false; }
    std::string write() { return "$" + strFun->write(); }
};

class FlxString : public FlxReaderBase2 {
  private:
    std::list<FlxString_Base*>* strLst;
    /**
    * @brief instances of this FlxString (used to copy this FlxString)
    */
    tuint* instances;
    
    static FlxStringFunBox* StrFunBox;
  public:
    FlxString(FlxString_Base* strB);
    FlxString(const bool multiline = false, const bool errSerious = true);
    FlxString(const FlxString &FlxStrF);
    FlxString& operator=(const FlxString &FlxStrF);
    ~FlxString();
    const std::string eval();
    void eval(std::ostream& os);
    const std::string eval_word(const bool lowercase = false, const bool emptyAllow=false,const bool numbegallow=false);
    const bool search_circref(FlxFunction* fcr);
    const std::string write();
    /**
    * @brief set expression of strA to this class; and deletes strA;
    */
    void assign ( FlxString *strA );
    
    static void set_StrFunBox(FlxStringFunBox* StrFunBoxV) { StrFunBox = StrFunBoxV; }
};

#endif // fesslix_flxstring_H
