/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "flxdefault.h"
#include "fesslix.h"

#include <iostream>

#include "po_part0.h"



int main(int argc, char** argv)
{
  ReadStream *reader = NULL;
  GlobalVar.set_stdcout(&std::cout);
  GlobalVar.set_slog(&std::cout);
  
  // read options (command-line & ini-file)
    #include "po_part1.txt"
  
  try {
    #include "po_part2.txt"

    // check arguments
      if ( argc < 2 )  {
	std::cout << "Please specify some options !!!" << std::endl << std::endl;
	#if FLX_BOOST_PO
	  print_help(visible);
	#else
	  print_help();
	#endif
	#ifdef __WINDOWS__
	  take_a_break(std::cout);
	#else
	  if (pauseAtEnd) { take_a_break(std::cout); }
	#endif
	return RETURN_ERROR;
      }
      
  // Deal with the log-files
    create_log(logFile, logTrunc);
     
  // try to open the file with the Gauss points
    ReadStream* gaussRS = NULL;
    open_GaussFile(gaussFile, gaussRS);
    
  // create output-logFile
    if (!GlobalVar.check_logNOTcout()) logOutput=false;
    if (logOutput) {
      flxcout = new flxStreamAlloc(std::cout, GlobalVar.get_log());
    } else {
      flxcout = &std::cout;
    }
    
  // produce some logging-infos
    if (do_logInfo) {
      FesslixMain::logInfo(GlobalVar.slog(4),false);
    }
    
  // create main object
    FesslixMain fesslix(*flxcout, std::cerr, gaussRS);

    #if FLX_DEBUG
    // testing
      fesslix.test();
    #endif
  
      
  #include "po_part3.txt"
    
  // return success
    print_RETURN_SUCCESS();
    delete_FLXcout(logOutput,flxcout);
    if (pauseAtEnd) { take_a_break(std::cout); }
    return RETURN_SUCCESS;
  }
  catch ( FlxEndE const &e) {
    GlobalVar.logLevel_strong_reset();
    GlobalVar.slog(4) << "end: 'end;' was called" << std::endl;
    if (reader != NULL) delete reader;
    print_RETURN_SUCCESS();
    delete_FLXcout(logOutput,flxcout);
    if (pauseAtEnd) { take_a_break(std::cout); }
    return RETURN_SUCCESS;
  }
  #include "po_part4.txt"

}

