/* Fesslix - Stochastic Analysis
 * Copyright (C) 2010-2012 Wolfgang Betz
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>. 
 */

#include "sfem.h"

// using namespace std;
using namespace sfem;


void FlxCreateObjReaders_SFEM::createObjReaders(FlxObjectReadBox* objReadBox) {
   objReadBox->insert("ssfem", new FlxObjReadSFEM());
   objReadBox->insert("ssfemtest", new FlxObjReadSFEMtest()); 
}

void FlxCreateObjReaders_SFEM::createFunReaders(FlxData* dataBox)
{
  dataBox->FunBox.insert("ssfem_moment",new FunReadSFEM_moment() );
  dataBox->FunBox.insert("ssfem_covariance",new FunReadSFEM_covariance() );
}


void FlxObjSFEM_new::task()
{
  delete data->FEM;
  data->FEM = new flxSFEM(p,&data->ConstantBox,&data->RndBox);
  GlobalVar.slog(4) << "SFEM: created new SFEM calculation" << std::endl;
}

FlxObjReadSFEM::FlxObjReadSFEM() 
{
  // order
    AllDefParaBox->insert(new FlxOptionalParaFun(3,"flxrandom::order"));
    ParaBox.insert("order", "flxrandom::order");
}

int FlxObjReadSFEM::get_order()
{
  FlxOptionalParaBase* p1 = ParaBox.get("order"); 
  void *v1 = p1->get();
  FlxFunction* fun = static_cast<FlxFunction*>( v1 ); 
  int iv = fun->cast2int();
  delete fun;
  if (iv < 0) {
    std::ostringstream ssV_2;
    ssV_2 << "Order of the homogeneous chaos (" << iv << ") must not be a negativ value.";
    throw FlxException("FlxObjReadSFEM::get_order", ssV_2.str(), reader->getCurrentPos() );
  }
  return iv;
}

FlxObjBase* FlxObjReadSFEM::read_new(bool errSerious)
{
  read_optionalPara(errSerious);
  return new FlxObjSFEM_new(get_doLog(),get_order(),get_stream(),get_verbose());
}

FlxObjBase* FlxObjReadSFEMtest::read() {
  //FlxFunction* f1 = new FlxFunction(funReader);
  int M = reader->getNumber<int>();
  int p = reader->getNumber<int>();
  read_optionalPara(false);
  FlxObjBase* obj1 = new FlxObjSFEMtest(get_doLog(), M, p, get_stream() );
  return obj1;
  
//   if (pV<=0) {
//     ostringstream ssV_2;
//     ssV_2 << "Order of polynomail chaos (p=" << pV << ") must be greater or equal than one.";
//     throw FlxException("sfem::PolyChaos::GenerateSequence_1", ssV_2.str() );
//   }
//   if (MV<=0) {
//     ostringstream ssV_2;
//     ssV_2 << "Dimension of polynomail chaos (M=" << MV << ") must be greater or equal than one.";
//     throw FlxException("sfem::PolyChaos::GenerateSequence_2", ssV_2.str() );
//   }
}


void FlxObjSFEMtest::task()
{
  sout() << "Polynomial Chaos" << std::endl;
  
  PolyChaos* pb = new PolyChaos(M,p);

  tulong N = pb->get_SeqNumb();
//   sout() << "Number of Sequences: " << N << std::endl;
//   for (int i=0;i<=M;i++) {
//     for (tulong j=0;j<N;j++) {
//       for (tulong k=j;k<N;k++) {
// 	tdouble c = pb->calc_c_ijk(i,j,k);
// 	if (c!=0.0) {
// 	  sout() << "(" << i << ", " << j << ", " << k << "): " << c << std::endl;
// 	}
//       }
//     }
//   }
//   for (tulong i=0;i<N;i++) {
//     sout() << i << ": " << pb->get_ExpPB2(i) << "; " << endl;
//   }
  sout() << std::endl;
  sout() << N << ": " << pb->get_ExpPB2(N-1) << std::endl;
//   sout() << "(0,1,1): " << pb->calc_c_ijk(0,1,1) << endl;
  
  delete pb;
}

const std::string FunSFEM_moment::write()
{
  return "sfem_moment(" + write_dof() + "," + moment->write() + ")";
}

const bool FunSFEM_moment::optimize(FunBasePtr& optf, const Fun_OptimizeInfo &foi)
{
  child_optimize(moment,foi);
  return FunFEM_base::optimize(optf,foi);
}

const bool FunSFEM_moment::dependOn_Const(const tdouble*const thenumber)
{
  throw FlxException_NotImplemented("FunSFEM_moment::dependOn_Const");
}

const tdouble FunSFEM_moment::calc()
{
  flxSFEM* sfem = dynamic_cast<flxSFEM*> (data->FEM);
  if (sfem == NULL) {
    std::ostringstream ssV;
    ssV << "This is not a stochastic finite element analysis.";
    throw FlxException("FunSFEM_moment::calc_1", ssV.str() );
  }
  
  int nth=floor(fabs(moment->calc())+0.5);
  if (nth < 0) {
    std::ostringstream ssV;
    ssV << "Moment to calculate must not be negative.";
    throw FlxException("FunSFEM_moment::calc_1", ssV.str() );
  } else if ( nth == 0) {
    return 1.0;
  }
  
  flxDOFn* dofP = get_dof();
  const std::vector<FlxRnd_base*>& rndvec = sfem->RndBox->get_vector();
  FlxPeval* sp = sfem->get_Pol(dofP->get_DOFindex());
  tdouble mu = sp->calc_Expectation(rndvec);
  
  if (nth == 1) {
    delete sp;
    return mu;
  }
  
  mu*=-1;
  sp->add_term(FlxPeval_term(rndvec.size()),mu);
  
  FlxPeval* spU = sp;
  sp = new FlxPeval(*sp);
  FlxPeval* spT1;
  for ( int i = 2; i <= nth; ++i) {
    spT1 = sp;
    sp = new FlxPeval(*spT1,*spU);
    delete spT1;
  }
  delete spU;
  
  const tdouble cv = sp->calc_Expectation(rndvec);
  delete sp;
  return cv;
}

FunBase* FunReadSFEM_moment::read(bool errSerious)
{
  FunBase* node;
  FlxDof::dof dofV;
  read_node(node,dofV,errSerious);
  reader->getChar(',',errSerious); 
  FunBase* moment = FunctionList->read(errSerious);
  return new FunSFEM_moment(node,dofV,moment);
}

const std::string FunSFEM_covariance::write()
{
  return "sfem_covariance(" + write_dof() + "," + write_dof(node2,dof2) + ")";
}

const bool FunSFEM_covariance::optimize(FunBasePtr& optf, const Fun_OptimizeInfo& foi)
{
  child_optimize(node2,foi);
  child_optimize(joint2,foi);
  return FunFEM_base::optimize(optf,foi);
}

const bool FunSFEM_covariance::dependOn_Const(const tdouble*const thenumber)
{
  throw FlxException_NotImplemented("FunSFEM_covariance::dependOn_Const");
}

const tdouble FunSFEM_covariance::calc()
{
  flxSFEM* sfem = dynamic_cast<flxSFEM*> (data->FEM);
  if (sfem == NULL) {
    std::ostringstream ssV;
    ssV << "This is not a stochastic finite element analysis.";
    throw FlxException("FunSFEM_moment::calc_1", ssV.str() );
  }
  
  flxDOFn* dofP = get_dof();
  flxDOFn* dofP2 = get_dof(node2,dof2);
  const std::vector<FlxRnd_base*>& rndvec = sfem->RndBox->get_vector();
  FlxPeval* sp = sfem->get_Pol(dofP->get_DOFindex());
  tdouble mu = sp->calc_Expectation(rndvec);
  FlxPeval* sp2 = sfem->get_Pol(dofP2->get_DOFindex());
  tdouble mu2 = sp2->calc_Expectation(rndvec);

  mu*=mu2;
  
  FlxPeval* spT = sp;
  sp = new FlxPeval(*spT,*sp2);
  delete spT; delete sp2;
  mu2 = sp->calc_Expectation(rndvec);
  delete sp;
  return mu2 - mu;
}

FunBase* FunReadSFEM_covariance::read(bool errSerious)
{
  FunBase* node; FunBase* node2;
  FlxDof::dof dofV; FlxDof::dof dof2;
  read_node(node,dofV,errSerious);
  reader->getChar(',',errSerious);
  read_node(node2,dof2,errSerious);
  return new FunSFEM_covariance(node,dofV,node2,dof2);
}

FlxRSM_Fun_base* FlxRSM_Fun_pc::copy()
{
  return new FlxRSM_Fun_pc(DIM, pc.get_order(), new tVec(*paraVec));
}

void FlxRSM_Fun_pc::calc_termVec(const tVec& yVec, const tVec& xVec, tVec& termVec)
{
  const tulong k = pc.get_SeqNumb();
  #if FLX_DEBUG
    if (termVec.size() != k) {
      std::ostringstream ssV_2;
      ssV_2 << "ERROR. (" << termVec.size() << ", " << get_NumbOfTerms() << ")";
      throw FlxException("FlxRSM_Fun_pc::calc_termVec_0", ssV_2.str() );  
    }
  #endif
  if (yVec.size() != DIM) {
    std::ostringstream ssV_2;
    ssV_2 << "Number of random variables has changed.";
    throw FlxException("FlxRSM_Fun_pc::calc_termVec_1", ssV_2.str() );  
  }
  for (tulong i=0;i<k;++i) {
    termVec[i] = pc.calc_Poly(i,yVec);
  }  
}

void FlxRSM_Fun_pc::updateStart()
{
  FlxRSM_Fun_base::updateStart();
  int order = orderF->cast2int();
  pc.GenerateSequence(DIM,order);
}

const tdouble FunSFEM_pc_term::calc()
{
  return pc.calc_Poly(j,RndBox->get_y_Vec());
}

void FunSFEM_pc_term::dependOn_RV(bVec& bv)
{
  pc.dependOn_RV(j,bv);
}

const tdouble FunSFEM_pc_term::expectation()
{
 return (j==0);
}



